//=============================================================================
//
//   File : kvi_textencoding.cpp
//   Creation date : Wed 04 Dec 2002 03:33:21 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __LIBKVILIB__



#include "kvi_textencoding.h"
#include "kvi_string.h"
#include "kvi_env.h"

#include <qtextcodec.h>

typedef struct _EncodingDescriptor
{
	const char * name;
	unsigned int uCode;
} EncodingDescriptor;

static EncodingDescriptor encoding_descriptors[]=
{
	{ "utf8" , KVI_TEXT_ENCODING_UTF8 },

	{ "iso88591" , KVI_TEXT_ENCODING_ISO88591 },
	{ "latin1" , KVI_TEXT_ENCODING_ISO88591 },

	{ "iso88592" , KVI_TEXT_ENCODING_ISO88592 },
	{ "latin2" , KVI_TEXT_ENCODING_ISO88592 },

	{ "iso885915" , KVI_TEXT_ENCODING_ISO885915 },
	{ "latin9" , KVI_TEXT_ENCODING_ISO885915 },

	{ "iso88595" , KVI_TEXT_ENCODING_ISO88595 },

	{ "iso88597" , KVI_TEXT_ENCODING_ISO88597 },

	{ "iso88599" , KVI_TEXT_ENCODING_ISO88595 },
	{ "latin5" , KVI_TEXT_ENCODING_ISO88595 },

	{ "koi8r" , KVI_TEXT_ENCODING_KOI8R },

	{ "cp1251" , KVI_TEXT_ENCODING_CP1251 },
	{ "win1251" , KVI_TEXT_ENCODING_CP1251 },
	{ "windows1251" , KVI_TEXT_ENCODING_CP1251 },

#ifndef QT_NO_BIG_CODECS
	{ "big5", KVI_TEXT_ENCODING_BIG5 },

	{ "gb2312", KVI_TEXT_ENCODING_GB18030 },
	{ "gb18030", KVI_TEXT_ENCODING_GB18030 },

	{ "sjis", KVI_TEXT_ENCODING_SJIS },

	{ "euckr", KVI_TEXT_ENCODING_EUCKR },
#endif

	{ "tis620", KVI_TEXT_ENCODING_TIS620 },

	{ 0,0 }
};

static unsigned int internal_textEncodingFromName(const char * encName)
{
	KviStr szEncName = encName;
	szEncName.replaceAll('-',"");
	szEncName.replaceAll(' ',"");
	szEncName.toLower();
	for(int i=0;encoding_descriptors[i].name;i++)
	{
		if(kvi_strEqualCS(szEncName.ptr(),encoding_descriptors[i].name))return encoding_descriptors[i].uCode;
	}
	return KVI_TEXT_ENCODING_UNKNOWN;
}

unsigned int kvi_textEncodingFromName(const char * encName)
{
	unsigned int enc = internal_textEncodingFromName(encName);
	if(enc != KVI_TEXT_ENCODING_UNKNOWN)return enc;
	return KVI_TEXT_ENCODING_ISO88591; // Doh :D
}

typedef struct _EncodingName
{
	const char * szName;
	const char * szAlias;
} EncodingName;

/*
  \i Latin1
  \i Big5 -- Chinese
  \i Big5-HKSCS -- Chinese
  \i eucJP -- Japanese
  \i eucKR -- Korean
  \i GB2312 -- Chinese
  \i GBK -- Chinese
  \i GB18030 -- Chinese
  \i JIS7 -- Japanese
  \i Shift-JIS -- Japanese
  \i TSCII -- Tamil
  \i utf8 -- Unicode, 8-bit
  \i utf16 -- Unicode
  \i KOI8-R -- Russian
  \i KOI8-U -- Ukrainian
  \i ISO8859-1 -- Western
  \i ISO8859-2 -- Central European
  \i ISO8859-3 -- Central European
  \i ISO8859-4 -- Baltic
  \i ISO8859-5 -- Cyrillic
  \i ISO8859-6 -- Arabic
  \i ISO8859-7 -- Greek
  \i ISO8859-8 -- Hebrew, visually ordered
  \i ISO8859-8-i -- Hebrew, logically ordered
  \i ISO8859-9 -- Turkish
  \i ISO8859-10
  \i ISO8859-13
  \i ISO8859-14
  \i ISO8859-15 -- Western
  \i IBM 850
  \i IBM 866
  \i CP874
  \i CP1250 -- Central European
  \i CP1251 -- Cyrillic
  \i CP1252 -- Western
  \i CP1253 -- Greek
  \i CP1254 -- Turkish
  \i CP1255 -- Hebrew
  \i CP1256 -- Arabic
  \i CP1257 -- Baltic
  \i CP1258
  \i Apple Roman
  \i TIS-620 -- Thai
*/

static EncodingName encoding_names[KVI_TEXT_ENCODING_COUNT + 1]=
{
	{ "UTF-8", "Unicode" },
	{ "ISO 8859-1", "Latin-1 - Western European" },
	{ "ISO 8859-2", "Latin-2 - Central European" },
	{ "ISO 8859-5", "Cyrillic" },
	{ "ISO 8859-7", "Greek" },
	{ "ISO 8859-9", "Latin-5 - Turkish" },
	{ "ISO 8859-15", "Latin-9 - Western European" },
	{ "KOI8-R", "KOI - Cyrillic" },
	{ "CP-1251", "Windows-1251 - Cyrillic" },
#ifndef QT_NO_BIG_CODECS
	{ "Big5", "Chinese Traditional" },
	{ "GB18030", "Chinese Simplified" },
	{ "Shift-JIS", "Japanese" },
	{ "EUC-KR", "Korean" },
#endif
	{ "TIS-620", "Thai" },
	{ "Unknown" , "Unknown" }
};

const char * kvi_textEncodingName(unsigned int uEncoding)
{
	if(uEncoding >= KVI_TEXT_ENCODING_COUNT)return encoding_names[KVI_TEXT_ENCODING_COUNT].szName;
	return encoding_names[uEncoding].szName;
}

QTextCodec * kvi_textEncodingCodec(unsigned int uEncoding)
{
	if(uEncoding >= KVI_TEXT_ENCODING_COUNT)return QTextCodec::codecForLocale();
	QTextCodec * c = QTextCodec::codecForName(encoding_names[uEncoding].szName);
	if(!c)
	{
		debug("WARNING: Found no Qt codec for encoding %u (%s)",uEncoding,encoding_names[uEncoding].szName);
		return QTextCodec::codecForLocale();
	}
	return c;
}


const char * kvi_textEncodingNameAlias(unsigned int uEncoding)
{
	if(uEncoding >= KVI_TEXT_ENCODING_COUNT)return encoding_names[KVI_TEXT_ENCODING_COUNT].szAlias;
	return encoding_names[uEncoding].szAlias;
}

static const char * iso88591_locales[] =
{
	"c",            "en",         0
};

static const char * iso88592_locales[] =
{
	"croatian",     "cs",         "cs_CS",         "cs_CZ",
	"cz",           "cz_CZ",      "czech",         "hr",
	"hr_HR",        "hu",         "hu_HU",         "hungarian",
	"pl",           "pl_PL",      "polish",        "ro",
	"ro_RO",        "rumanian",   "serbocroatian", "sh",
	"sh_SP",        "sh_YU",      "sk",            "sk_SK",
	"sl",           "sl_CS",      "sl_SI",         "slovak",
	"slovene",      "sr_SP",      0
};

static const char * iso88595_locales[] =
{
	"mk",           "mk_MK",      "sp",            "sp_YU",
	0
};


static const char * iso88597_locales[] =
{
	"el",           "el_GR",      "greek",         0
};


static const char * const iso88599_locales[] =
{
	"tr",           "tr_TR",      "turkish",       0
};

static const char * iso885915_locales[] =
{
	"et",           "et_EE",      "br_FR",         "ca_ES",
	"de",           "de_AT",      "de_BE",         "de_DE",
	"de_LU",        "en_IE",      "es",            "es_ES",
	"eu_ES",        "fi",         "fi_FI",         "finnish",
	"fr",           "fr_FR",      "fr_BE",         "fr_LU",
	"french",       "ga_IE",      "gl_ES",         "it",
	"it_IT",        "oc_FR",      "nl",            "nl_BE",
	"nl_NL",        "pt",         "pt_PT",         "sv_FI",
	"wa_BE",        0
};

static const char * iso88595_or_koi8r_locales[] =
{
	"ru",           "ru_SU",      "ru_RU",         "russian",
	0
};

static const char * cp1251_locales[] =
{
	"be",           "be_BY",      "bg",            "bg_BG",
	"bulgarian",    0
};

static const char * tis_620locales[] = {
	"th",            "th_TH",     "thai", 0
};


static const char * check_locale_list(const char ** list,const char * locale)
{
	for(int i=0;list[i];i++)
		if(kvi_strEqualCI(list[i],locale))return list[i];
	return 0;
}

unsigned int kvi_textEncodingForLocale()
{
	KviStr szLocale = kvi_getenv("LC_ALL");
	if(szLocale.isEmpty())szLocale = kvi_getenv("LC_CTYPE");
	if(szLocale.isEmpty())szLocale = kvi_getenv("LANG");
	if(szLocale.isEmpty())szLocale = QTextCodec::locale();
	if(szLocale.isEmpty())
	{
		debug("Warning: All of $LC_ALL, $LC_CTYPE and $LANG are empty, no way to guess the system locale");
		return KVI_TEXT_ENCODING_ISO88591;
	}

	szLocale.stripWhiteSpace();

	int idx = szLocale.findFirstIdx('.');
	if(idx != -1)
	{
		// handle things like en_GB.UTF-8
		KviStr szEnc = szLocale;
		szEnc.cutToFirst('.');
		szEnc.cutFromFirst('@');
		unsigned int enc = internal_textEncodingFromName(szEnc.ptr());
		if(enc != KVI_TEXT_ENCODING_UNKNOWN)return enc;
	}

	szLocale.toLower();

	// I also saw a thingie like @euro...
	if(szLocale.findFirstIdx("@euro") != -1)return KVI_TEXT_ENCODING_ISO885915;

	szLocale.cutFromFirst('@');
	szLocale.cutFromFirst('.');

	if(check_locale_list(iso88591_locales,szLocale.ptr()))return KVI_TEXT_ENCODING_ISO88591;
	if(check_locale_list(iso88592_locales,szLocale.ptr()))return KVI_TEXT_ENCODING_ISO88592;
	if(check_locale_list(iso88595_locales,szLocale.ptr()))return KVI_TEXT_ENCODING_ISO88595;
	if(check_locale_list(iso88597_locales,szLocale.ptr()))return KVI_TEXT_ENCODING_ISO88597;
	if(check_locale_list(iso885915_locales,szLocale.ptr()))return KVI_TEXT_ENCODING_ISO885915;
	if(check_locale_list(cp1251_locales,szLocale.ptr()))return KVI_TEXT_ENCODING_CP1251;

	if(check_locale_list(iso88595_or_koi8r_locales,szLocale.ptr()))
	{
		// this is a mess : almost all russians use KOI8-R and incorrectly set 
		// $LANG to ru_RU : a check hack is needed
		// Qt does it better in fact...

		// unicode   koi8r   latin5   name
		// 0x044E    0xC0    0xEE     CYRILLIC SMALL LETTER YU
		// 0x042E    0xE0    0xCE     CYRILLIC CAPITAL LETTER YU
		int latin5 = tolower(0xce);
		int koi8r = tolower(0xe0);
		if((koi8r != 0xc0) && (latin5 == 0xee))return KVI_TEXT_ENCODING_ISO88595;
		return KVI_TEXT_ENCODING_KOI8R;
	}
	if(check_locale_list(tis_620locales,szLocale.ptr()))return KVI_TEXT_ENCODING_TIS620;

#ifndef QT_NO_BIG_CODECS
	if(szLocale.equalsCIN("zh",2))
	{
		if(szLocale.equalsCI("zh_TW")/* || szLocale.contains(".Big5",false)*/)return KVI_TEXT_ENCODING_BIG5;
		if(szLocale.equalsCI("zh_CN"))return KVI_TEXT_ENCODING_GB18030;
	}
	
	if(szLocale.equalsCIN("ja",2)/* && (szLocale.contains(".SJIS",false) || szLocale.contains(".Shift-JIS",false))*/)return KVI_TEXT_ENCODING_SJIS;

	if(szLocale.equalsCIN("ko",2)/* && szLocale.contains(".euc",false)*/)return KVI_TEXT_ENCODING_EUCKR;
#endif

	QTextCodec * c = QTextCodec::codecForLocale(); // try what Qt thinks is better
	if(c)
	{
		const char * cName = c->name();
		if(cName)
		{
			unsigned int enc = internal_textEncodingFromName(cName);
			if(enc != KVI_TEXT_ENCODING_UNKNOWN)return enc;
		}
	}

	return KVI_TEXT_ENCODING_ISO88591; // this is our default : can't do better for now
}

/*
static const char * const iso8859_3locales[] = {
    "eo", 0 };

static const char * const iso8859_4locales[] = {
    "ee", "ee_EE", 0 };



static const char * const pt_154locales[] = {
    "ba_RU", "ky", "ky_KG", "kk", "kk_KZ", 0 };

static const char * const iso8859_6locales[] = {
    "ar_AA", "ar_SA", "arabic", 0 };


static const char * const iso8859_8locales[] = {
    "hebrew", "he", "he_IL", "iw", "iw_IL", 0 };

static const char * const iso8859_13locales[] = {
    "lt", "lt_LT", "lv", "lv_LV", 0 };

static const char * const koi8_ulocales[] =
{
	"uk", "uk_UA", "ru_UA", "ukrainian", 0
};

static const char * const tcvnlocales[] = {
    "vi", "vi_VN", 0 };

*/

