//=============================================================================
//
//   File : kvi_wstring.cpp
//   Creation date : Wed 04 Dec 2002 20:42:00 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define _KVI_WSTRING_CPP_
#define __WSTRINGLIB__

#include "kvi_wstring.h"
#include "kvi_wstring_external.h"

#ifdef WSTRINGCONFIG_USE_QT
	#include <qglobal.h> // for QT_VERSION
	#ifdef WSTRINGCONFIG_COMPILE_FOR_KVIRC
		#include <qtextcodec.h>
		#include "kvi_textencoding.h"
	#endif // WSTRINGCONFIG_COMPILE_FOR_KVIRC
#endif // WSTRINGCONFIG_USE_QT


//=============================================================================================================
// constructors
//=============================================================================================================


#define ALLOC_EMPTY_STRING \
	m_len = 0; \
	m_ptr = _WSTRING_WMALLOC(1); \
	*m_ptr = KVI_WCHAR_NULL;

#define REALLOC_EMPTY_STRING \
		m_len = 0; \
		m_ptr = _WSTRING_WREALLOC(m_ptr,1); \
		*m_ptr = KVI_WCHAR_NULL;


KviWStr::KviWStr()
{
	ALLOC_EMPTY_STRING
}

KviWStr::KviWStr(const KviWStr &str)
{
	m_len = str.m_len;
	m_ptr = _WSTRING_WMALLOC(m_len + 1);
	_WSTRING_WMEMCPY(m_ptr,str.m_ptr,m_len + 1);
}


KviWStr::KviWStr(const kvi_wchar_t * str)
{
	if(!str)
	{
		ALLOC_EMPTY_STRING
	} else {
		m_len = kvi_wstrlen(str);
		m_ptr = _WSTRING_WMALLOC(m_len + 1);
		_WSTRING_WMEMCPY(m_ptr,str,m_len + 1);
	}
}

KviWStr::KviWStr(const kvi_wchar_t * str,kvi_wslen_t len)
{
	m_len = len;
	m_ptr = _WSTRING_WMALLOC(m_len + 1);
	_WSTRING_WMEMCPY(m_ptr,str,m_len);
	*(m_ptr + m_len) = KVI_WCHAR_NULL;
}

KviWStr::KviWStr(const kvi_wchar_t * begin,const kvi_wchar_t * end)
{
	m_len = end - begin;
	m_ptr = _WSTRING_WMALLOC(m_len + 1);
	_WSTRING_WMEMCPY(m_ptr,begin,m_len);
	*(m_ptr + m_len) = KVI_WCHAR_NULL;
}


#ifdef WSTRINGCONFIG_USE_QT
	KviWStr::KviWStr(const QString &str)
	{
		m_len = str.length();
		m_ptr = _WSTRING_WMALLOC(m_len + 1);

		if(m_len > 0)
		{
	#ifdef WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		_WSTRING_WMEMCPY(m_ptr,str.unicode(),m_len);
		*(m_ptr + m_len) = KVI_WCHAR_NULL;
	#else // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		#if QT_VERSION >= 0x030100
			_WSTRING_WMEMCPY(m_ptr,str.ucs2(),m_len);
			*(m_ptr + m_len) = KVI_WCHAR_NULL;
		#else // QT_VERSION < 0x030100
			kvi_wslen_t auxLen = m_len;
			kvi_wchar_t * auxPtr = m_ptr;
			QChar * auxStr = str.unicode();
			while(auxLen > 0)
			{
				*auxPtr++ = auxStr->unicode();
				auxStr++;
				auxLen--;
			}
			*auxPtr = KVI_WCHAR_NULL;
		#endif // QT_VERSION < 0x030100
	#endif // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		} else {
			*m_ptr = KVI_WCHAR_NULL;
		}
	}
#endif // WSTRINGCONFIG_USE_QT


//=============================================================================================================
// destructor
//=============================================================================================================


KviWStr::~KviWStr()
{
	_WSTRING_WFREE(m_ptr);
}


//=============================================================================================================
// buffer manipulation
//=============================================================================================================

void KviWStr::setLen(kvi_wslen_t len)
{
	m_len = len;
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);
	*(m_ptr + m_len) = KVI_WCHAR_NULL;
}

//=============================================================================================================
// buffer manipulation
//=============================================================================================================


void KviWStr::bufferDetach()
{
	ALLOC_EMPTY_STRING
}

void KviWStr::bufferAttach(kvi_wchar_t * str)
{
	_WSTRING_WFREE(m_ptr);
	m_len = kvi_wstrlen(str);
	m_ptr = str;
}


void KviWStr::bufferAttach(kvi_wchar_t * str,kvi_wslen_t len)
{
	_WSTRING_WFREE(m_ptr);
	m_len = len;
	m_ptr = str;
}


//=============================================================================================================
// local memory allocator
//=============================================================================================================


kvi_wchar_t * KviWStr::bufferAlloc(kvi_wslen_t nChars)
{
	return _WSTRING_WMALLOC(nChars);
}

kvi_wchar_t * KviWStr::bufferRealloc(kvi_wchar_t * str,kvi_wslen_t nChars)
{
	return _WSTRING_WREALLOC(str,nChars);
}

void KviWStr::bufferFree(kvi_wchar_t * str)
{
	_WSTRING_WFREE(str);
}

//======================================================================================================
// conversions
//======================================================================================================


#ifdef WSTRINGCONFIG_USE_QT
	QString KviWStr::qstring() const
	{
		QString ret;
		ret.setLength(m_len);
	#ifdef WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		_WSTRING_WMEMCPY(ret.unicode(),m_ptr,m_len);
	#else // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		QChar * c = ret.unicode();
		kvi_wchar_t * ptr = m_ptr;
		while(*ptr)
		{
			c->unicode() = *ptr;
			ptr++;
			c++;
		}
	#endif // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		return ret;
	}
#endif // WSTRINGCONFIG_USE_QT

//======================================================================================================
// strip stuff
//======================================================================================================

void KviWStr::stripWhiteSpace()
{
	// 0123456789
	//    abcd   0
	// ^        ^
	// left   right
	register kvi_wchar_t *left=m_ptr;
	register kvi_wchar_t *right=m_ptr+m_len-1;
	// skip initial spaces
	while(kvi_wisspace(*left))left++;
	if(*left){
		// valid string , left points to first non-space
		while((right >= left) && kvi_wisspace(*right))right--;
		// 0123456789
		//    abcd   0
		//    ^  ^
		// left   right	
		m_len = (right - left)+1;
		_WSTRING_WMEMMOVE(m_ptr,left,m_len);
		m_ptr = _WSTRING_WREALLOC(m_ptr,m_len+1);
		*(m_ptr+m_len)=KVI_WCHAR_NULL;
	} else {
		m_ptr = _WSTRING_WREALLOC(m_ptr,1);
		*m_ptr = KVI_WCHAR_NULL;
		m_len = 0;
	}
}

//======================================================================================================
// format
//======================================================================================================

void KviWStr::sprintf(const kvi_wchar_t * fmt,...)
{
	m_ptr=_WSTRING_WREALLOC(m_ptr,256);
	//First try
	kvi_va_list list;
	kvi_va_start(list,fmt);
	//print...with max 256 chars
	m_len=kvi_wvsnprintf(m_ptr,256,fmt,list);
	kvi_va_end(list);

	//check if we failed
	if(m_len < 0)
	{
		//yes , failed....
		int dummy=256;
		do{ //we failed , so retry with 256 more chars
			dummy+=256;
			//realloc
			m_ptr=_WSTRING_WREALLOC(m_ptr,dummy);
			//print...
			kvi_va_start(list,fmt);
			m_len=kvi_wvsnprintf(m_ptr,dummy,fmt,list);
			kvi_va_end(list);
		} while(m_len < 0);
	}
	//done...
	//now m_len is the length of the written string not including the terminator...
	//perfect! :)
	m_ptr=_WSTRING_WREALLOC(m_ptr,m_len+1);
}

void KviWStr::sprintf(const char * fmt,...)
{
	m_ptr=_WSTRING_WREALLOC(m_ptr,256);
	//First try
	kvi_va_list list;
	kvi_va_start(list,fmt);
	//print...with max 256 chars
	m_len=kvi_wvsnprintcf(m_ptr,256,fmt,list);
	kvi_va_end(list);

	//check if we failed
	if(m_len < 0)
	{
		//yes , failed....
		int dummy=256;
		do{ //we failed , so retry with 256 more chars
			dummy+=256;
			//realloc
			m_ptr=_WSTRING_WREALLOC(m_ptr,dummy);
			//print...
			kvi_va_start(list,fmt);
			m_len=kvi_wvsnprintcf(m_ptr,dummy,fmt,list);
			kvi_va_end(list);
		} while(m_len < 0);
	}
	//done...
	//now m_len is the length of the written string not including the terminator...
	//perfect! :)
	m_ptr=_WSTRING_WREALLOC(m_ptr,m_len+1);
}



//=============================================================================================================
// append
//=============================================================================================================

#ifdef WSTRINGCONFIG_USE_QT
	void KviWStr::append(const QString &str,int startIdx,int len)
	{
		if(startIdx >= (int)str.length())return;
		if(startIdx + len > (int)str.length())len = str.length() - startIdx;
		m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + len + 1);
	#ifdef WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		_WSTRING_WMEMCPY(m_ptr + m_len,str.unicode() + startIdx,len);
		m_len += len;
		*(m_ptr + m_len) = KVI_WCHAR_NULL;
	#else // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		#if QT_VERSION >= 0x030100
			_WSTRING_WMEMCPY(m_ptr + m_len,str.ucs2() + startIdx,len);
			m_len += len;
			*(m_ptr + m_len) = KVI_WCHAR_NULL;
		#else // QT_VERSION < 0x030100
			kvi_wchar_t * auxPtr = m_ptr + m_len;
			m_len += len;
			QChar * auxStr = str.unicode() + startIdx;
			while(len > 0)
			{
				*auxPtr++ = auxStr->unicode();
				auxStr++;
				len--;
			}
			*auxPtr = KVI_WCHAR_NULL;
		#endif // QT_VERSION < 0x030100
	#endif // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
	}
#endif // WSTRINGCONFIG_USE_QT
	
void KviWStr::append(kvi_wchar_t c)
{
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 2);
	*(m_ptr + m_len) = c;
	m_len++;
	*(m_ptr + m_len) = KVI_WCHAR_NULL;
}

void KviWStr::append(kvi_wchar_t * str)
{
	if(!str)return;
	int len = kvi_wstrlen(str);
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + len + 1);
	_WSTRING_WMEMCPY(m_ptr + m_len,str,len + 1);
	m_len += len;
}

void KviWStr::append(kvi_wchar_t * str,kvi_wslen_t len)
{
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + len + 1);
	_WSTRING_WMEMCPY(m_ptr + m_len,str,len);
	m_len += len;
	*(m_ptr + m_len) = KVI_WCHAR_NULL;
}

//=============================================================================================================
// prepend
//=============================================================================================================



void KviWStr::prepend(const char * str)
{
	int len = _WSTRING_STRLEN(str);
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + len + 1);
	_WSTRING_WMEMMOVE((m_ptr + len),m_ptr,m_len + 1);
	m_len += len;
	kvi_wchar_t * aux = m_ptr;
	while(*str)*aux++ = *str++;
}

void KviWStr::prepend(const char * str,kvi_wslen_t len)
{
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + len + 1);
	_WSTRING_WMEMMOVE((m_ptr + len),m_ptr,m_len + 1);
	m_len += len;
	kvi_wchar_t * aux = m_ptr;
	while(len > 0)
	{
		*aux++ = *str++;
		len--;
	}
}

void KviWStr::prepend(const KviWStr &str)
{
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + str.m_len + 1);
	_WSTRING_WMEMMOVE((m_ptr + str.m_len),m_ptr,m_len + 1);
	m_len += str.m_len;
	_WSTRING_WMEMCPY(m_ptr,str.m_ptr,str.m_len);
}

void KviWStr::prepend(kvi_wchar_t c)
{
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 2);
	_WSTRING_WMEMMOVE((m_ptr + 1),m_ptr,m_len + 1);
	*m_ptr = c;
	m_len += 1;
}


//=============================================================================================================
// conversion
//=============================================================================================================

void KviWStr::decodeLatin1(const char * str)
{
#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
	if(!str)
	{
		REALLOC_EMPTY_STRING
		return;
	}
	m_len = _WSTRING_STRLEN(str);
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);
	kvi_wchar_t * auxPtr = m_ptr;
	while(*str)
	{
		*auxPtr++ = *((unsigned char *)str);
		str++;
	}
	*(auxPtr) = KVI_WCHAR_NULL;
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC
}

void KviWStr::decodeLatin1(const char * str,kvi_wslen_t len)
{
#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
	m_len = len;
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);
	kvi_wchar_t * auxPtr = m_ptr;
	while(len)
	{
		*auxPtr++ = *((unsigned char *)str);
		str++;
		len--;
	}
	*(auxPtr) = KVI_WCHAR_NULL;
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC
}

#ifdef WSTRINGCONFIG_USE_KVISTR
	void KviWStr::encodeLatin1(KviStr &szBuffer)
	{
#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
		szBuffer.setLen(m_len);
		unsigned char * aux = (unsigned char *) szBuffer.ptr();
		kvi_wchar_t * ptr = m_ptr;
		while(*ptr)
		{
			*aux++ = (unsigned char) *ptr;
			ptr++;
		}
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC
	}
#endif // WSTRINGCONFIG_USE_KVISTR


#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
//for A in 0 1 2 3 4 5 6 7 8 9 a b c d e f; do for B in 0 1 2 3 4 5 6 7 8 9 a b c d e f ; do X=`zcat /usr/share/i18n/charmaps/ISO-8859-2.gz | grep "/x${A}${B}" | sed -e 's/<U/0x/g' | sed -e 's/>.*/,/g'`; echo -n $X ; done; done; echo ""

static kvi_wchar_t iso88592_2_unicode_table[0x100]=
{
	0x0000,0x0001,0x0002,0x0003,0x0004,0x0005,0x0006,0x0007,
	0x0008,0x0009,0x000A,0x000B,0x000C,0x000D,0x000E,0x000F,
	0x0010,0x0011,0x0012,0x0013,0x0014,0x0015,0x0016,0x0017,
	0x0018,0x0019,0x001A,0x001B,0x001C,0x001D,0x001E,0x001F,
	0x0020,0x0021,0x0022,0x0023,0x0024,0x0025,0x0026,0x0027,
	0x0028,0x0029,0x002A,0x002B,0x002C,0x002D,0x002E,0x002F,
	0x0030,0x0031,0x0032,0x0033,0x0034,0x0035,0x0036,0x0037,
	0x0038,0x0039,0x003A,0x003B,0x003C,0x003D,0x003E,0x003F,
	0x0040,0x0041,0x0042,0x0043,0x0044,0x0045,0x0046,0x0047,
	0x0048,0x0049,0x004A,0x004B,0x004C,0x004D,0x004E,0x004F,
	0x0050,0x0051,0x0052,0x0053,0x0054,0x0055,0x0056,0x0057,
	0x0058,0x0059,0x005A,0x005B,0x005C,0x005D,0x005E,0x005F,
	0x0060,0x0061,0x0062,0x0063,0x0064,0x0065,0x0066,0x0067,
	0x0068,0x0069,0x006A,0x006B,0x006C,0x006D,0x006E,0x006F,
	0x0070,0x0071,0x0072,0x0073,0x0074,0x0075,0x0076,0x0077,
	0x0078,0x0079,0x007A,0x007B,0x007C,0x007D,0x007E,0x007F,
	0x0080,0x0081,0x0082,0x0083,0x0084,0x0085,0x0086,0x0087,
	0x0088,0x0089,0x008A,0x008B,0x008C,0x008D,0x008E,0x008F,
	0x0090,0x0091,0x0092,0x0093,0x0094,0x0095,0x0096,0x0097,
	0x0098,0x0099,0x009A,0x009B,0x009C,0x009D,0x009E,0x009F,
	0x00A0,0x0104,0x02D8,0x0141,0x00A4,0x013D,0x015A,0x00A7,
	0x00A8,0x0160,0x015E,0x0164,0x0179,0x00AD,0x017D,0x017B,
	0x00B0,0x0105,0x02DB,0x0142,0x00B4,0x013E,0x015B,0x02C7,
	0x00B8,0x0161,0x015F,0x0165,0x017A,0x02DD,0x017E,0x017C,
	0x0154,0x00C1,0x00C2,0x0102,0x00C4,0x0139,0x0106,0x00C7,
	0x010C,0x00C9,0x0118,0x00CB,0x011A,0x00CD,0x00CE,0x010E,
	0x0110,0x0143,0x0147,0x00D3,0x00D4,0x0150,0x00D6,0x00D7,
	0x0158,0x016E,0x00DA,0x0170,0x00DC,0x00DD,0x0162,0x00DF,
	0x0155,0x00E1,0x00E2,0x0103,0x00E4,0x013A,0x0107,0x00E7,
	0x010D,0x00E9,0x0119,0x00EB,0x011B,0x00ED,0x00EE,0x010F,
	0x0111,0x0144,0x0148,0x00F3,0x00F4,0x0151,0x00F6,0x00F7,
	0x0159,0x016F,0x00FA,0x0171,0x00FC,0x00FD,0x0163,0x02D9
};


static unsigned char unicode_2_iso88592_table[0x2de]=
{
	0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,
	0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,
	0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,
	0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x3e,0x3f,
	0x40,0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4a,0x4b,0x4c,0x4d,0x4e,0x4f,
	0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5a,0x5b,0x5c,0x5d,0x5e,0x5f,
	0x60,0x61,0x62,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,0x6e,0x6f,
	0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x7b,0x7c,0x7d,0x7e,0x7f,
	0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8a,0x8b,0x8c,0x8d,0x8e,0x8f,
	0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0x9b,0x9c,0x9d,0x9e,0x9f,
	0xa0,0x3f,0x3f,0x3f,0xa4,0x3f,0x3f,0xa7,0xa8,0x3f,0x3f,0x3f,0x3f,0xad,0x3f,0x3f,
	0xb0,0x3f,0x3f,0x3f,0xb4,0x3f,0x3f,0x3f,0xb8,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0xc1,0xc2,0x3f,0xc4,0x3f,0x3f,0xc7,0x3f,0xc9,0x3f,0xcb,0x3f,0xcd,0xce,0x3f,
	0x3f,0x3f,0x3f,0xd3,0xd4,0x3f,0xd6,0xd7,0x3f,0x3f,0xda,0x3f,0xdc,0xdd,0x3f,0xdf,
	0x3f,0xe1,0xe2,0x3f,0xe4,0x3f,0x3f,0xe7,0x3f,0xe9,0x3f,0xeb,0x3f,0xed,0xee,0x3f,
	0x3f,0x3f,0x3f,0xf3,0xf4,0x3f,0xf6,0xf7,0x3f,0x3f,0xfa,0x3f,0xfc,0xfd,0x3f,0x3f,
	0x3f,0x3f,0xc3,0xe3,0xa1,0xb1,0xc6,0xe6,0x3f,0x3f,0x3f,0x3f,0xc8,0xe8,0xcf,0xef,
	0xd0,0xf0,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0xca,0xea,0xcc,0xec,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0xc5,0xe5,0x3f,0x3f,0xa5,0xb5,0x3f,
	0x3f,0xa3,0xb3,0xd1,0xf1,0x3f,0x3f,0xd2,0xf2,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0xd5,0xf5,0x3f,0x3f,0xc0,0xe0,0x3f,0x3f,0xd8,0xf8,0xa6,0xb6,0x3f,0x3f,0xaa,0xba,
	0xa9,0xb9,0xde,0xfe,0xab,0xbb,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0xd9,0xf9,
	0xdb,0xfb,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0xac,0xbc,0xaf,0xbf,0xae,0xbe,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0xb7,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,
	0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0xa2,0xff,0x3f,0xb2,0x3f,0xbd
};
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC


void KviWStr::decodeLatin2(const char * str)
{
#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
	if(!str)
	{
		REALLOC_EMPTY_STRING
		return;
	}
	m_len = _WSTRING_STRLEN(str);
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);
	kvi_wchar_t * auxPtr = m_ptr;
	while(*str)
	{
		*auxPtr++ = iso88592_2_unicode_table[*((unsigned char *)str)];
		str++;
	}
	*(auxPtr) = KVI_WCHAR_NULL;
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC
}

void KviWStr::decodeLatin2(const char * str,kvi_wslen_t len)
{
#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
	m_len = len;
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);
	kvi_wchar_t * auxPtr = m_ptr;
	while(len)
	{
		*auxPtr++ = iso88592_2_unicode_table[*((unsigned char *)str)];
		str++;
		len--;
	}
	*(auxPtr) = KVI_WCHAR_NULL;
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC
}

#ifdef WSTRINGCONFIG_USE_KVISTR
	void KviWStr::encodeLatin2(KviStr &szBuffer)
	{
#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
		szBuffer.setLen(m_len);
		unsigned char * aux = (unsigned char *) szBuffer.ptr();
		kvi_wchar_t * ptr = m_ptr;
		while(*ptr)
		{
			if(*ptr < 0x2de)*aux++ = unicode_2_iso88592_table[*ptr];
			else *aux++ = (unsigned char)'?'; // 0x3f
			ptr++;
		}
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC
	}
#endif // WSTRINGCONFIG_USE_KVISTR





void KviWStr::decodeLatin15(const char * str,kvi_wslen_t len)
{
#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
	m_len = len;
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);
	kvi_wchar_t * auxPtr = m_ptr;
	while(len)
	{
		switch(((unsigned char)*str))
		{
			case 0xa4:
				*auxPtr++ = 0x20ac;
			break;
			case 0xa6:
				*auxPtr++ = 0x0160;
			break;
			case 0xa8:
				*auxPtr++ = 0x0161;
			break;
			case 0xb4:
				*auxPtr++ = 0x017d;
			break;
			case 0xb8:
				*auxPtr++ = 0x017e;
			break;
			case 0xbc:
				*auxPtr++ = 0x0152;
			break;
			case 0xbd:
				*auxPtr++ = 0x0153;
			break;
			case 0xbe:
				*auxPtr++ = 0x0178;
			break;
			default:
				*auxPtr++ = *((unsigned char *)str);
			break;
		}
		str++;
		len--;
	}
	*(auxPtr) = KVI_WCHAR_NULL;
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC
}

#ifdef WSTRINGCONFIG_USE_KVISTR
	void KviWStr::encodeLatin15(KviStr &szBuffer)
	{
#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
		szBuffer.setLen(m_len);
		unsigned char * aux = (unsigned char *) szBuffer.ptr();
		kvi_wchar_t * ptr = m_ptr;
		while(*ptr)
		{
			if(*ptr < 0x100)
			{
				if(*ptr < 0xa4)
				{
					*aux++ = *((unsigned char *)ptr);
				} else {
					switch(*ptr)
					{
						case 0xa4:
						case 0xa6:
						case 0xa8:
						case 0xb4:
						case 0xb8:
						case 0xbc:
						case 0xbd:
						case 0xbe:
							*aux++ = '?'; // invalid characters
						break;
						default:
							*aux++ = *((unsigned char *)ptr);
						break;
					}
				}
			} else {
				switch(*ptr)
				{
					case 0x20ac:
						*aux++ = 0xa4;
					break;
					case 0x0160:
						*aux++ = 0xa6;
					break;
					case 0x0161:
						*aux++ = 0xa8;
					break;
					case 0x017d:
						*aux++ = 0xb4;
					break;
					case 0x017e:
						*aux++ = 0xb8;
					break;
					case 0x0152:
						*aux++ = 0xbc;
					break;
					case 0x0153:
						*aux++ = 0xbd;
					break;
					case 0x0178:
						*aux++ = 0xbe;
					break;
					default:
						*aux++ = '?';
					break;
				}
			}
			ptr++;
		}
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC
	}
#endif // WSTRINGCONFIG_USE_KVISTR







void KviWStr::decodeUtf8(const char * str,kvi_wslen_t len)
{
#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
	m_len = len; // worst case
	m_ptr = _WSTRING_WREALLOC(m_ptr,len + 1);

	int missing = 0;
	kvi_wchar_t * aux = m_ptr;
	while(len)
	{
		if(missing)
		{
			if((((unsigned char)(*str)) & 0xc0) == 0x80)
			{
				*aux = (*aux << 6) | (((unsigned char)(*str)) & 0x3f);
				if(!(--missing))aux++;
			} else {
				// invalid encoding : unicode.org says to use
				// the replacement and continue processing from next char
				*aux = KVI_WCHAR_REPLACEMENT;
				aux++;
				missing = 0;
			}
		} else {
			if(((unsigned char)(*str)) < 0x80)
			{
				*aux++ = *str;
			} else if( (((unsigned char)(*str)) & 0xe0) == 0xc0)
			{
				*aux = ((unsigned char)(*str)) & 0x1f;
				missing = 1;
			} else if( (((unsigned char)(*str)) & 0xf0) == 0xe0)
			{
				*aux = ((unsigned char)(*str)) & 0x0f;
				missing = 2;
			} // else we're out of our 16 bit range anyway
		}
		str++;
		len--;
	}

	// FIXME: what is missing > 0 now ? (the last char is invalid! (the encoding was invalid))

	m_len = aux - m_ptr;
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);

	*(aux) = KVI_WCHAR_NULL;
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC
}

void KviWStr::decodeUtf8(const char * str)
{
#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
	if(!str)
	{
		REALLOC_EMPTY_STRING
		return;
	}

	decodeUtf8(str,_WSTRING_STRLEN(str));
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC
}

#ifdef WSTRINGCONFIG_USE_KVISTR
	void KviWStr::encodeUtf8(KviStr &szBuffer)
	{
#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
		szBuffer.setLen(m_len * 3); // worst case

		unsigned char * aux = (unsigned char *) szBuffer.ptr();
		kvi_wchar_t * ptr = m_ptr;

		while(*ptr)
		{
			if(*ptr < 0x80)*aux++ = (unsigned char)*ptr;
			else {
				if(*ptr < 0x800)
				{
					*aux++ = 0xc0 | ((unsigned char) (*ptr >> 6));
				} else {
					*aux++ = 0xe0 | ((unsigned char) (*ptr >> 12));
					*aux++ = 0x80 | ((unsigned char) ((*ptr >> 6) & 0x3f));
				}
				*aux++ = 0x80| ((unsigned char) (*ptr & 0x3f));
			}
			ptr++;
		}

		szBuffer.setLen(((char *)aux) - szBuffer.ptr());
#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC
	}
#endif // WSTRINGCONFIG_USE_KVISTR

void KviWStr::decode(const char * str,unsigned int e)
{
	if(!str)
	{
		REALLOC_EMPTY_STRING
		return;
	}

	decode(str,_WSTRING_STRLEN(str),e);
}


#if !defined(WSTRINGCONFIG_USE_QT) || !defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
static unsigned char unicode_2_any_XX[0x100]=
{
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};


// koi8r character mapping tables

static kvi_wchar_t koi8r_2_unicode[0x100]=
{
	0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007,
	0x0008, 0x0009, 0x000A, 0x000B, 0x000C, 0x000D, 0x000E, 0x000F,
	0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017,
	0x0018, 0x0019, 0x001A, 0x001B, 0x001C, 0x001D, 0x001E, 0x001F,
	0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
	0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D, 0x002E, 0x002F,
	0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
	0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F,
	0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
	0x0048, 0x0049, 0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F,
	0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
	0x0058, 0x0059, 0x005A, 0x005B, 0x005C, 0x005D, 0x005E, 0x005F,
	0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
	0x0068, 0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F,
	0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
	0x0078, 0x0079, 0x007A, 0x007B, 0x007C, 0x007D, 0x007E, 0x007F,
	0x2500, 0x2502, 0x250C, 0x2510, 0x2514, 0x2518, 0x251C, 0x2524,
	0x252C, 0x2534, 0x253C, 0x2580, 0x2584, 0x2588, 0x258C, 0x2590,
	0x2591, 0x2592, 0x2593, 0x2320, 0x25A0, 0x2219, 0x221A, 0x2248,
	0x2264, 0x2265, 0x00A0, 0x2321, 0x00B0, 0x00B2, 0x00B7, 0x00F7,
	0x2550, 0x2551, 0x2552, 0x0451, 0x2553, 0x2554, 0x2555, 0x2556,
	0x2557, 0x2558, 0x2559, 0x255A, 0x255B, 0x255C, 0x255D, 0x255E,
	0x255F, 0x2560, 0x2561, 0x0401, 0x2562, 0x2563, 0x2564, 0x2565,
	0x2566, 0x2567, 0x2568, 0x2569, 0x256A, 0x256B, 0x256C, 0x00A9,
	0x044E, 0x0430, 0x0431, 0x0446, 0x0434, 0x0435, 0x0444, 0x0433,
	0x0445, 0x0438, 0x0439, 0x043A, 0x043B, 0x043C, 0x043D, 0x043E,
	0x043F, 0x044F, 0x0440, 0x0441, 0x0442, 0x0443, 0x0436, 0x0432,
	0x044C, 0x044B, 0x0437, 0x0448, 0x044D, 0x0449, 0x0447, 0x044A,
	0x042E, 0x0410, 0x0411, 0x0426, 0x0414, 0x0415, 0x0424, 0x0413,
	0x0425, 0x0418, 0x0419, 0x041A, 0x041B, 0x041C, 0x041D, 0x041E,
	0x041F, 0x042F, 0x0420, 0x0421, 0x0422, 0x0423, 0x0416, 0x0412,
	0x042C, 0x042B, 0x0417, 0x0428, 0x042D, 0x0429, 0x0427, 0x042A
};

static unsigned char unicode_2_koi8r_00[0x100]=
{
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
	0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x9a, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0xbf, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x9c, 0x3f, 0x9d, 0x3f, 0x3f, 0x3f, 0x3f, 0x9e,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x9f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char unicode_2_koi8r_04[0x100]=
{
	0x3f, 0xb3, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0xe1, 0xe2, 0xf7, 0xe7, 0xe4, 0xe5, 0xf6, 0xfa,
	0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 0xf0,
	0xf2, 0xf3, 0xf4, 0xf5, 0xe6, 0xe8, 0xe3, 0xfe,
	0xfb, 0xfd, 0xff, 0xf9, 0xf8, 0xfc, 0xe0, 0xf1,
	0xc1, 0xc2, 0xd7, 0xc7, 0xc4, 0xc5, 0xd6, 0xda,
	0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0,
	0xd2, 0xd3, 0xd4, 0xd5, 0xc6, 0xc8, 0xc3, 0xde,
	0xdb, 0xdd, 0xdf, 0xd9, 0xd8, 0xdc, 0xc0, 0xd1,
	0x3f, 0xa3, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char unicode_2_koi8r_22[0x100]=
{
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x95, 0x96, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x97, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x98, 0x99, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char unicode_2_koi8r_23[0x100]=
{
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x93, 0x9b, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char unicode_2_koi8r_25[0x100]=
{
	0x80, 0x3f, 0x81, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x82, 0x3f, 0x3f, 0x3f,
	0x83, 0x3f, 0x3f, 0x3f, 0x84, 0x3f, 0x3f, 0x3f,
	0x85, 0x3f, 0x3f, 0x3f, 0x86, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x87, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x88, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x89, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x8a, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0xa0, 0xa1, 0xa2, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8,
	0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0,
	0xb1, 0xb2, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
	0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x8b, 0x3f, 0x3f, 0x3f, 0x8c, 0x3f, 0x3f, 0x3f,
	0x8d, 0x3f, 0x3f, 0x3f, 0x8e, 0x3f, 0x3f, 0x3f,
	0x8f, 0x90, 0x91, 0x92, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x94, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char * unicode_2_koi8r[0x100]=
{
	unicode_2_koi8r_00, unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_koi8r_04, unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_koi8r_22, unicode_2_koi8r_23,
	unicode_2_any_XX  , unicode_2_koi8r_25, unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  ,
	unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  , unicode_2_any_XX  
};

// koi8r tables are 3584 bytes (-256 bytes of the shared table)


// iso88595 character mapping tables

static kvi_wchar_t iso88595_2_unicode[0x100]=
{
	0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007,
	0x0008, 0x0009, 0x000A, 0x000B, 0x000C, 0x000D, 0x000E, 0x000F,
	0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017,
	0x0018, 0x0019, 0x001A, 0x001B, 0x001C, 0x001D, 0x001E, 0x001F,
	0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
	0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D, 0x002E, 0x002F,
	0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
	0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F,
	0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
	0x0048, 0x0049, 0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F,
	0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
	0x0058, 0x0059, 0x005A, 0x005B, 0x005C, 0x005D, 0x005E, 0x005F,
	0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
	0x0068, 0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F,
	0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
	0x0078, 0x0079, 0x007A, 0x007B, 0x007C, 0x007D, 0x007E, 0x007F,
	0x0080, 0x0081, 0x0082, 0x0083, 0x0084, 0x0085, 0x0086, 0x0087,
	0x0088, 0x0089, 0x008A, 0x008B, 0x008C, 0x008D, 0x008E, 0x008F,
	0x0090, 0x0091, 0x0092, 0x0093, 0x0094, 0x0095, 0x0096, 0x0097,
	0x0098, 0x0099, 0x009A, 0x009B, 0x009C, 0x009D, 0x009E, 0x009F,
	0x00A0, 0x0401, 0x0402, 0x0403, 0x0404, 0x0405, 0x0406, 0x0407,
	0x0408, 0x0409, 0x040A, 0x040B, 0x040C, 0x00AD, 0x040E, 0x040F,
	0x0410, 0x0411, 0x0412, 0x0413, 0x0414, 0x0415, 0x0416, 0x0417,
	0x0418, 0x0419, 0x041A, 0x041B, 0x041C, 0x041D, 0x041E, 0x041F,
	0x0420, 0x0421, 0x0422, 0x0423, 0x0424, 0x0425, 0x0426, 0x0427,
	0x0428, 0x0429, 0x042A, 0x042B, 0x042C, 0x042D, 0x042E, 0x042F,
	0x0430, 0x0431, 0x0432, 0x0433, 0x0434, 0x0435, 0x0436, 0x0437,
	0x0438, 0x0439, 0x043A, 0x043B, 0x043C, 0x043D, 0x043E, 0x043F,
	0x0440, 0x0441, 0x0442, 0x0443, 0x0444, 0x0445, 0x0446, 0x0447,
	0x0448, 0x0449, 0x044A, 0x044B, 0x044C, 0x044D, 0x044E, 0x044F,
	0x2116, 0x0451, 0x0452, 0x0453, 0x0454, 0x0455, 0x0456, 0x0457,
	0x0458, 0x0459, 0x045A, 0x045B, 0x045C, 0x00A7, 0x045E, 0x045F
};

static unsigned char unicode_2_iso88595_00[0x100]=
{
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
	0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
	0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
	0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,
	0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
	0xa0, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0xfd,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0xad, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char unicode_2_iso88595_04[0x100]=
{
	0x3f, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
	0xa8, 0xa9, 0xaa, 0xab, 0xac, 0x3f, 0xae, 0xaf,
	0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7,
	0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
	0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
	0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
	0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
	0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
	0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
	0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
	0x3f, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7,
	0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0x3f, 0xfe, 0xff,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char unicode_2_iso88595_21[0x100]=
{
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0xf0, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char * unicode_2_iso88595[0x100]=
{
	unicode_2_iso88595_00, unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_iso88595_04, unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_iso88595_21, unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     
};

// iso88595 tables are 3072 bytes (-256 bytes of shared tables)


// iso88597 character mapping tables

static kvi_wchar_t iso88597_2_unicode[0x100]=
{
	0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007,
	0x0008, 0x0009, 0x000A, 0x000B, 0x000C, 0x000D, 0x000E, 0x000F,
	0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017,
	0x0018, 0x0019, 0x001A, 0x001B, 0x001C, 0x001D, 0x001E, 0x001F,
	0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
	0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D, 0x002E, 0x002F,
	0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
	0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F,
	0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
	0x0048, 0x0049, 0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F,
	0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
	0x0058, 0x0059, 0x005A, 0x005B, 0x005C, 0x005D, 0x005E, 0x005F,
	0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
	0x0068, 0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F,
	0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
	0x0078, 0x0079, 0x007A, 0x007B, 0x007C, 0x007D, 0x007E, 0x007F,
	0x0080, 0x0081, 0x0082, 0x0083, 0x0084, 0x0085, 0x0086, 0x0087,
	0x0088, 0x0089, 0x008A, 0x008B, 0x008C, 0x008D, 0x008E, 0x008F,
	0x0090, 0x0091, 0x0092, 0x0093, 0x0094, 0x0095, 0x0096, 0x0097,
	0x0098, 0x0099, 0x009A, 0x009B, 0x009C, 0x009D, 0x009E, 0x009F,
	0x00A0, 0x2018, 0x2019, 0x00A3, 0x00a4, 0x00a5, 0x00A6, 0x00A7,
	0x00A8, 0x00A9, 0x00aa, 0x00AB, 0x00AC, 0x00AD, 0x00ae, 0x2015,
	0x00B0, 0x00B1, 0x00B2, 0x00B3, 0x0384, 0x0385, 0x0386, 0x00B7,
	0x0388, 0x0389, 0x038A, 0x00BB, 0x038C, 0x00BD, 0x038E, 0x038F,
	0x0390, 0x0391, 0x0392, 0x0393, 0x0394, 0x0395, 0x0396, 0x0397,
	0x0398, 0x0399, 0x039A, 0x039B, 0x039C, 0x039D, 0x039E, 0x039F,
	0x03A0, 0x03A1, 0x00d2, 0x03A3, 0x03A4, 0x03A5, 0x03A6, 0x03A7,
	0x03A8, 0x03A9, 0x03AA, 0x03AB, 0x03AC, 0x03AD, 0x03AE, 0x03AF,
	0x03B0, 0x03B1, 0x03B2, 0x03B3, 0x03B4, 0x03B5, 0x03B6, 0x03B7,
	0x03B8, 0x03B9, 0x03BA, 0x03BB, 0x03BC, 0x03BD, 0x03BE, 0x03BF,
	0x03C0, 0x03C1, 0x03C2, 0x03C3, 0x03C4, 0x03C5, 0x03C6, 0x03C7,
	0x03C8, 0x03C9, 0x03CA, 0x03CB, 0x03CC, 0x03CD, 0x03CE, 0x00ff
};

static unsigned char unicode_2_iso88597_00[0x100]=
{
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
	0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
	0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
	0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,
	0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
	0xa0, 0x3f, 0x3f, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
	0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0x3f,
	0xb0, 0xb1, 0xb2, 0xb3, 0x3f, 0x3f, 0x3f, 0xb7,
	0x3f, 0x3f, 0x3f, 0xbb, 0x3f, 0xbd, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0xd2, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0xff
};
static unsigned char unicode_2_iso88597_03[0x100]=
{
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0xb4, 0xb5, 0xb6, 0x3f,
	0xb8, 0xb9, 0xba, 0x3f, 0xbc, 0x3f, 0xbe, 0xbf,
	0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
	0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
	0xd0, 0xd1, 0x3f, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
	0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
	0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
	0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
	0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7,
	0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char unicode_2_iso88597_20[0x100]=
{
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0xaf, 0x3f, 0x3f,
	0xa1, 0xa2, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char * unicode_2_iso88597[0x100]=
{
	unicode_2_iso88597_00, unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_iso88597_03,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_iso88597_20, unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     ,
	unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     , unicode_2_any_XX     
};

// iso88597 tables are 3072 bytes



// cp1251 character mapping tables

static kvi_wchar_t cp1251_2_unicode[0x100]=
{
	0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007,
	0x0008, 0x0009, 0x000A, 0x000B, 0x000C, 0x000D, 0x000E, 0x000F,
	0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017,
	0x0018, 0x0019, 0x001A, 0x001B, 0x001C, 0x001D, 0x001E, 0x001F,
	0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
	0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D, 0x002E, 0x002F,
	0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
	0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F,
	0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
	0x0048, 0x0049, 0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F,
	0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
	0x0058, 0x0059, 0x005A, 0x005B, 0x005C, 0x005D, 0x005E, 0x005F,
	0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
	0x0068, 0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F,
	0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
	0x0078, 0x0079, 0x007A, 0x007B, 0x007C, 0x007D, 0x007E, 0x007F,
	0x0402, 0x0403, 0x201A, 0x0453, 0x201E, 0x2026, 0x2020, 0x2021,
	0x20AC, 0x2030, 0x0409, 0x2039, 0x040A, 0x040C, 0x040B, 0x040F,
	0x0452, 0x2018, 0x2019, 0x201C, 0x201D, 0x2022, 0x2013, 0x2014,
	0x0098, 0x2122, 0x0459, 0x203A, 0x045A, 0x045C, 0x045B, 0x045F,
	0x00A0, 0x040E, 0x045E, 0x0408, 0x00A4, 0x0490, 0x00A6, 0x00A7,
	0x0401, 0x00A9, 0x0404, 0x00AB, 0x00AC, 0x00AD, 0x00AE, 0x0407,
	0x00B0, 0x00B1, 0x0406, 0x0456, 0x0491, 0x00B5, 0x00B6, 0x00B7,
	0x0451, 0x2116, 0x0454, 0x00BB, 0x0458, 0x0405, 0x0455, 0x0457,
	0x0410, 0x0411, 0x0412, 0x0413, 0x0414, 0x0415, 0x0416, 0x0417,
	0x0418, 0x0419, 0x041A, 0x041B, 0x041C, 0x041D, 0x041E, 0x041F,
	0x0420, 0x0421, 0x0422, 0x0423, 0x0424, 0x0425, 0x0426, 0x0427,
	0x0428, 0x0429, 0x042A, 0x042B, 0x042C, 0x042D, 0x042E, 0x042F,
	0x0430, 0x0431, 0x0432, 0x0433, 0x0434, 0x0435, 0x0436, 0x0437,
	0x0438, 0x0439, 0x043A, 0x043B, 0x043C, 0x043D, 0x043E, 0x043F,
	0x0440, 0x0441, 0x0442, 0x0443, 0x0444, 0x0445, 0x0446, 0x0447,
	0x0448, 0x0449, 0x044A, 0x044B, 0x044C, 0x044D, 0x044E, 0x044F
};

static unsigned char unicode_2_cp1251_00[0x100]=
{
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
	0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x98, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0xa0, 0x3f, 0x3f, 0x3f, 0xa4, 0x3f, 0xa6, 0xa7,
	0x3f, 0xa9, 0x3f, 0xab, 0xac, 0xad, 0xae, 0x3f,
	0xb0, 0xb1, 0x3f, 0x3f, 0x3f, 0xb5, 0xb6, 0xb7,
	0x3f, 0x3f, 0x3f, 0xbb, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};

static unsigned char unicode_2_cp1251_04[0x100]=
{
	0x3f, 0xa8, 0x80, 0x81, 0xaa, 0xbd, 0xb2, 0xaf,
	0xa3, 0x8a, 0x8c, 0x8e, 0x8d, 0x3f, 0xa1, 0x8f,
	0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
	0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
	0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
	0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
	0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
	0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
	0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7,
	0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff,
	0x3f, 0xb8, 0x90, 0x83, 0xba, 0xbe, 0xb3, 0xbf,
	0xbc, 0x9a, 0x9c, 0x9e, 0x9d, 0x3f, 0xa2, 0x9f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0xa5, 0xb4, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char unicode_2_cp1251_20[0x100]=
{
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x96, 0x97, 0x3f, 0x3f, 0x3f,
	0x91, 0x92, 0x82, 0x3f, 0x93, 0x94, 0x84, 0x3f,
	0x86, 0x87, 0x95, 0x3f, 0x3f, 0x3f, 0x85, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x89, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x8b, 0x9b, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x88, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char unicode_2_cp1251_21[0x100]=
{
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0xb9, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x99, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
	0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f
};
static unsigned char * unicode_2_cp1251[0x100]=
{
	unicode_2_cp1251_00, unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_cp1251_04, unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_cp1251_20, unicode_2_cp1251_21, unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   ,
	unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   , unicode_2_any_XX   
};

// cp1251 tables are 3328 bytes

#endif // !WSTRINGCONFIG_USE_QT || !WSTRINGCONFIG_COMPILE_FOR_KVIRC

void KviWStr::internalDecode(kvi_wchar_t * map,const char * str,unsigned int len)
{
	m_len = len;
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);
	kvi_wchar_t * aux = m_ptr;
	while(*str)
	{
		*aux++ = map[*((unsigned char *)str)];
		str++;
	}
	*aux = KVI_WCHAR_NULL;
}

void KviWStr::internalEncode(unsigned char ** map,KviStr &szBuffer)
{
	szBuffer.setLen(m_len);
	kvi_wchar_t * aux = m_ptr;
	unsigned char * ptr = (unsigned char *)szBuffer.ptr();
	while(*aux)
	{
		*ptr++ = map[*aux >> 8][*aux & 0xff];
		aux++;
	}
}


void KviWStr::decode(const char * str,kvi_wslen_t len,unsigned int e)
{
#if defined(WSTRINGCONFIG_USE_QT) && defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
	QTextCodec * c = kvi_textEncodingCodec(e);
	QString s = c->toUnicode(str,(int)len);
	if(s.isEmpty())
	{
		REALLOC_EMPTY_STRING
	
	} else {
		setStr(s,0,(int)s.length());
	}
#else
	switch(e)
	{
		case KVI_TEXT_ENCODING_ISO88591:
			decodeLatin1(str,len);
		break;
		case KVI_TEXT_ENCODING_ISO88592:
			decodeLatin2(str,len);
		break;
		case KVI_TEXT_ENCODING_ISO885915:
			decodeLatin15(str,len);
		break;
		case KVI_TEXT_ENCODING_UTF8:
			decodeUtf8(str,len);
		break;
		case KVI_TEXT_ENCODING_KOI8R:
			internalDecode(koi8r_2_unicode,str,len);
		break;
		case KVI_TEXT_ENCODING_ISO88595:
			internalDecode(iso88595_2_unicode,str,len);
		break;
		case KVI_TEXT_ENCODING_ISO88597:
			internalDecode(iso88597_2_unicode,str,len);
		break;
		case KVI_TEXT_ENCODING_CP1251:
			internalDecode(cp1251_2_unicode,str,len);
		break;
		default:
			decodeLatin1(str,len);
		break;
	}
#endif // WSTRINGCONFIG_USE_QT && WSTRINGCONFIG_COMPILE_FOR_KVIRC
}

#ifdef WSTRINGCONFIG_USE_KVISTR
	void KviWStr::encode(KviStr &szBuffer,unsigned int e)
	{
#if defined(WSTRINGCONFIG_USE_QT) && defined(WSTRINGCONFIG_COMPILE_FOR_KVIRC)
		QTextCodec * c = kvi_textEncodingCodec(e);
		int len = (int)m_len;
		szBuffer.setStr(c->fromUnicode(qstring(),len));
#else
		switch(e)
		{
			case KVI_TEXT_ENCODING_ISO88591:
				encodeLatin1(szBuffer);
			break;
			case KVI_TEXT_ENCODING_ISO88592:
				encodeLatin2(szBuffer);
			break;
			case KVI_TEXT_ENCODING_ISO885915:
				encodeLatin15(szBuffer);
			break;
			case KVI_TEXT_ENCODING_UTF8:
				encodeUtf8(szBuffer);
			break;
			case KVI_TEXT_ENCODING_KOI8R:
				internalEncode(unicode_2_koi8r,szBuffer);
			break;
			case KVI_TEXT_ENCODING_ISO88595:
				internalEncode(unicode_2_iso88595,szBuffer);
			break;
			case KVI_TEXT_ENCODING_ISO88597:
				internalEncode(unicode_2_iso88597,szBuffer);
			break;
			case KVI_TEXT_ENCODING_ISO88599:

				internalEncode(unicode_2_iso88599,szBuffer);
			break;
			case KVI_TEXT_ENCODING_CP1251:
				internalEncode(unicode_2_cp1251,szBuffer);
			break;
			default:
				encodeLatin1(szBuffer);
			break;
		}
#endif // WSTRINGCONFIG_USE_QT && WSTRINGCONFIG_COMPILE_FOR_KVIRC
	}
#endif // WSTRINGCONFIG_USE_KVISTR

//=============================================================================================================
// assignment operators
//=============================================================================================================

void KviWStr::operator=(const KviWStr &str)
{
	m_len = str.m_len;
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);
	_WSTRING_WMEMCPY(m_ptr,str.m_ptr,m_len + 1);
}

void KviWStr::operator=(const KviStr &str)
{
	m_len = str.len();
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);
	kvi_wchar_t * auxPtr = m_ptr;
	unsigned char * c = (unsigned char *)(str.ptr());
	while(*c)*auxPtr++ = *c++;
	*auxPtr = KVI_WCHAR_NULL;

}

void KviWStr::operator=(const kvi_wchar_t *str)
{
	if(!str)
	{
		REALLOC_EMPTY_STRING
		return;
	}
	m_len = kvi_wstrlen(str);
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);
	_WSTRING_WMEMCPY(m_ptr,str,m_len + 1);
}

void KviWStr::operator=(const char *str)
{
	if(!str)
	{
		REALLOC_EMPTY_STRING
		return;
	}

	m_len = _WSTRING_STRLEN(str);
	m_ptr = _WSTRING_WREALLOC(m_ptr,m_len + 1);
	kvi_wchar_t * auxPtr = m_ptr;
	while(*str)*auxPtr++ = *str++;
	*auxPtr = KVI_WCHAR_NULL;
}

//=============================================================================================================
// assignment operators
//=============================================================================================================

#ifdef WSTRINGCONFIG_USE_QT
	void KviWStr::setStr(const QString &str,int startIdx,int len)
	{
		if(startIdx >= (int)str.length())return;
		if(startIdx + len > (int)str.length())len = str.length() - startIdx;
		m_ptr = _WSTRING_WREALLOC(m_ptr,len + 1);
		m_len = len;
	#ifdef WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		_WSTRING_WMEMCPY(m_ptr,str.unicode() + startIdx,len);
		*(m_ptr + m_len) = KVI_WCHAR_NULL;
	#else // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		#if QT_VERSION >= 0x030100
			_WSTRING_WMEMCPY(m_ptr,str.ucs2() + startIdx,len);
			*(m_ptr + m_len) = KVI_WCHAR_NULL;
		#else // QT_VERSION < 0x030100
			kvi_wchar_t * auxPtr = m_ptr;
			QChar * auxStr = str.unicode() + startIdx;
			while(len > 0)
			{
				*auxPtr++ = auxStr->unicode();
				auxStr++;
				len--;
			}
			*auxPtr = KVI_WCHAR_NULL;
		#endif // QT_VERSION < 0x030100
	#endif // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
	}
#endif // WSTRINGCONFIG_USE_QT
