//=============================================================================
//
//   File : kvi_wstring_functions.cpp
//   Creation date : Wed 04 Dec 2002 19:58:00 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define _KVI_WSTRING_CPP_
#define __WSTRINGLIB__

#include "kvi_wstring_functions.h"
#include "kvi_wstring_external.h"

#ifdef WSTRINGCONFIG_USE_QT
	#include <qstring.h>
#endif // WSTRINGCONFIG_USE_QT

#ifdef WSTRINGCONFIG_USE_KVISTR
	#include "kvi_string.h"
#endif // WSTRINGCONFIG_USE_KVISTR

#include "kvi_wstring.h"

#include <ctype.h> // for hybrid functions

#ifndef WSTRINGCONFIG_USE_GCC_ix86_ASM
	kvi_wslen_t kvi_wstrlen(const kvi_wchar_t * str)
	{
		const kvi_wchar_t * ptr = str;
		while(*ptr)ptr++;
		return (ptr - str);
	}
#endif // !WSTRINGCONFIG_USE_GCC_ix86_ASM

#ifdef WSTRINGCONFIG_USE_QT
	#if QT_VERSION < 0x030100
		QString kvi_wchararray_to_qstring(const kvi_wchar_t * ptr)
		{
			if(!ptr)return QString::null;
			QString ret;
			kvi_wslen_t len = kvi_wstrlen(ptr);
			ret.setLength(len);
			#ifdef WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
				_WSTRING_WMEMCPY(ret.unicode(),ptr,len);
			#else // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
//#warning "HERE"
				QChar * c = ret.unicode();
				while(*ptr)
				{
					c->unicode() = *ptr;
					ptr++;
					c++
				}
			#endif // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
			return ret;
		}
	#endif // QT_VERSION < 0x030100

	QString kvi_wchararray_to_qstring_with_length(const kvi_wchar_t * ptr,kvi_wslen_t len)
	{
		QString ret;
		ret.setLength(len);
		#ifdef WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
			_WSTRING_WMEMCPY(ret.unicode(),ptr,len);
		#else // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
			QChar * c = ret.unicode();
			while(*ptr)
			{
				c->unicode() = *ptr;
				ptr++;
				c++;
			}
		#endif // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		return ret;
	}

	void kvi_append_wchararray_to_qstring_with_length(QString * qstrptr,const kvi_wchar_t * ptr,kvi_wslen_t len)
	{
		kvi_wslen_t oldLen = qstrptr->length();
		qstrptr->setLength(oldLen + len);
		#ifdef WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
			_WSTRING_WMEMCPY(qstrptr->unicode() + oldLen,ptr,len);
		#else // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
			QChar * c = (qstrptr->unicode() + oldLen);
			while(*ptr)
			{
				c->unicode() = *ptr;
				ptr++;
				c++;
			}
		#endif // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
	}


#endif // WSTRINGCONFIG_USE_QT


bool kvi_wstrEqualCIN(const kvi_wchar_t *str1,const kvi_wchar_t *str2,int len)
{
	while(len-- && *str1)if(kvi_wtolower(*str1++) != kvi_wtolower(*str2++))return false;
	return (len < 0);
}

bool kvi_wstrEqualCI(const kvi_wchar_t *str1,const kvi_wchar_t *str2)
{
	while(*str1)if(kvi_wtolower(*str1++) != kvi_wtolower(*str2++))return false;
	return (*str1 == *str2);
}


bool kvi_hstrEqualCIN(const kvi_wchar_t *str1,const char *str2,int len)
{
	while(len-- && *str1)if(kvi_wtolower(*str1++) != (kvi_wchar_t)tolower(*str2++))return false;
	return (len < 0);
}

bool kvi_hstrEqualCI(const kvi_wchar_t *str1,const char *str2)
{
	while(*str1)if(kvi_wtolower(*str1++) != (kvi_wchar_t)tolower(*str2++))return false;
	return (*str1 == *str2);
}

// %s = Latin1 char string (can't be null)
// %d = signed int (short,char)
// %u = unsigned int (short,char)
// %c = char value (kvi_wchar_t value)

// %f = double value

// %w = kvi_wchar_t string (can't be null)

// %S = Latin1 KviStr pointer (#ifdef WSTRINGCONFIG_USE_KVISTR) : can't be NULL!
// %W = KviWStr pointer : can't be NULL!
// %Q = QString pointer : can't be NULL!

#ifdef WSTRINGCONFIG_USE_KVISTR 
	#define WVSNPRINTF_KVISTR_BODY \
			case 'S': \
			{ \
				KviStr * pString = kvi_va_arg(list,KviStr *); \
				char * argString = pString->ptr(); \
				if(len <= pString->len())return (-1); \
				while(*argString)*p++ = *argString++; \
				len -= pString->len(); \
			} \
			break;
#else // !WSTRINGCONFIG_USE_KVISTR
	#define WVSNPRINTF_KVISTR_BODY
#endif // !WSTRINGCONFIG_USE_KVISTR


#ifdef WSTRINGCONFIG_USE_QT
	#ifdef WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		#define WVSNPRINTF_QSTRING_BODY \
			case 'Q': \
			{ \
				QString * pString = kvi_va_arg(list,QString *); \
				if(pString->length() > 0) \
				{ \
					if(len <= pString->length())return (-1); \
					_WSTRING_WMEMCPY(p,pString->unicode(),pString->length()); \
					p += pString->length(); \
					len -= pString->length(); \
				} \
			} \
			break;
	#else // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
		#if QT_VERSION >= 0x030100
			#define WVSNPRINTF_QSTRING_BODY \
				case 'Q': \
				{ \
					QString * pString = kvi_va_arg(list,QString *); \
					if(pString->length() > 0) \
					{ \
						if(len <= pString->length())return (-1); \
						_WSTRING_WMEMCPY(p,pString->ucs2(),pString->length()); \
						p += pString->length(); \
						len -= pString->length(); \
					} \
				} \
				break;
		#else // QT_VERSION < 0x030100
			#define WVSNPRINTF_QSTRING_BODY \
					case 'Q': \
					{ \
						QString * pString = kvi_va_arg(list,QString *); \
						if(pString->length() > 0) \
						{ \
							if(len <= pString->length())return (-1); \
							kvi_wslen_t ll = pString->length(); \
							QChar * argString = pString->unicode(); \
							while(ll > 0) \
							{ \
								*p++ = argString->unicode(); \
								argString++; \
								ll--; \
							} \
						} \
					} \
					break;
		#endif // QT_VERSION < 0x030100
	#endif // !WSTRINGCONFIG_SAFE_TO_MEMCPY_QCHAR
#else // !WSTRINGCONFIG_USE_QT
	#define WVSNPRINTF_QSTRING_BODY
#endif // !WSTRINGCONFIG_USE_QT



#define WVSNPRINTF_BODY \
\
	register kvi_wchar_t *p; \
	long int argValue; \
	unsigned long argUValue; \
\
	kvi_wchar_t numberBuffer[32]; \
	kvi_wchar_t *pNumBuf; \
	unsigned int tmp; \
\
	for(p=buffer ; *fmt ; ++fmt) \
	{ \
		if(len < 1)return (-1); \
\
		if(*fmt != '%') \
		{ \
			*p++ = *fmt; \
			--len; \
			continue; \
		} \
\
		++fmt; \
\
		switch(*fmt) \
		{ \
			case 's': \
			{ \
				char * argString = kvi_va_arg(list,char *); \
				argValue = _WSTRING_STRLEN(argString); \
				if(len <= argValue)return (-1); \
				while(*argString)*p++ = *argString++; \
				len -= argValue; \
			} \
			break; \
			case 'w': \
			{ \
				kvi_wchar_t * wArgString = kvi_va_arg(list,kvi_wchar_t *); \
				argValue = kvi_wstrlen(wArgString); \
				if(len <= argValue)return (-1); \
				_WSTRING_WMEMCPY(p,wArgString,argValue); \
				p += argValue; \
				len -= argValue; \
			} \
			break; \
			WVSNPRINTF_KVISTR_BODY \
			case 'W': \
			{ \
				KviWStr * pString = kvi_va_arg(list,KviWStr *); \
				if(len <= pString->len())return (-1); \
				/*if(pString->ptr())_WSTRING_WMEMCPY(p,pString->ptr(),pString->len());*/ \
				/*The above should NEVER happen: KviWStr is always non null*/ \
				/*if it happens, the cause must be somewhere else*/ \
				_WSTRING_WMEMCPY(p,pString->ptr(),pString->len()); \
				p += pString->len(); \
				len -= pString->len(); \
			} \
			break; \
			WVSNPRINTF_QSTRING_BODY \
			case 'd': \
				argValue = kvi_va_arg(list,int); \
				if(argValue < 0) \
				{ \
					*p++ = '-'; \
					if(--len == 0)return (-1); \
					argValue = -argValue; \
					if(argValue < 0)argValue = 0; \
				} \
				pNumBuf = numberBuffer; \
				do { \
					tmp = argValue / 10; \
					*pNumBuf++ = argValue - (tmp * 10) + '0'; \
				} while((argValue = tmp)); \
				argUValue = pNumBuf - numberBuffer; \
				if(((uint)len) <= argUValue)return (-1); \
				do { \
					*p++ = *--pNumBuf; \
				} while(pNumBuf != numberBuffer); \
				len -= argUValue; \
			break; \
			case 'u': \
				argUValue = kvi_va_arg(list,unsigned int); \
				pNumBuf = numberBuffer; \
				do { \
					tmp = argUValue / 10; \
					*pNumBuf++ = argUValue - (tmp * 10) + '0'; \
				} while((argUValue = tmp)); \
				argValue = pNumBuf - numberBuffer; \
				if(len <= argValue)return (-1); \
				do { \
					*p++ = *--pNumBuf; \
				} while(pNumBuf != numberBuffer); \
				len -= argValue; \
			break; \
			case 'f': \
			{ \
				double dVal = (double)kvi_va_arg(list,double); \
				char sprintfBuffer[32]; \
				argValue = sprintf(sprintfBuffer,"%f",dVal); \
				if(len <= argValue)return (-1); \
				char * pSprintfBuffer = sprintfBuffer; \
				while(*pSprintfBuffer)*p++ = *pSprintfBuffer++; \
				len -= argValue; \
			} \
			break; \
			case 'c': \
				*p++ = (kvi_wchar_t)kvi_va_arg(list,int); \
				--len; \
			break; \
			default: \
				*p++ = '%';  \
				if(--len == 0)return (-1); \
				if(*fmt){ \
					*p++ = *fmt; \
					--len; \
				} \
			break; \
		} \
		continue; \
	} \
	if(len < 1)return (-1); \
	*p = KVI_WCHAR_NULL; \
	return p-buffer;

int kvi_wvsnprintcf(kvi_wchar_t *buffer,kvi_wslen_t len,const char *fmt,kvi_va_list list)
{
	WVSNPRINTF_BODY
}

int kvi_wvsnprintf(kvi_wchar_t *buffer,kvi_wslen_t len,const kvi_wchar_t *fmt,kvi_va_list list)
{
	WVSNPRINTF_BODY
}


