//=============================================================================
//
//   File : kvi_ircserver.cpp
//   Creation date : Mon Jul 10 2000 03:42:59 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2000-2004 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __LIBKVILIB__
#include "kvi_ircserver.h"
#include "kvi_malloc.h"

// FIXME: This should be renamed to KviServer or sth like that

KviIrcServer::KviIrcServer()
{
	m_uFlags = 0;
	m_uPort  = 6667;
	m_pChannelList = 0;
	m_bAutoConnect = false;
}

KviIrcServer::KviIrcServer(const KviIrcServer &serv)
{
	m_szHostname         = serv.m_szHostname;
	m_szIp               = serv.m_szIp;
	m_szDescription      = serv.m_szDescription;
	m_szUser             = serv.m_szUser;
	m_szPass             = serv.m_szPass;
	m_uPort              = serv.m_uPort;
	m_szNick             = serv.m_szNick;
	m_szRealName         = serv.m_szRealName;
	m_uFlags             = serv.m_uFlags;
	m_szInitUMode = serv.m_szInitUMode;
	m_szOnConnectCommand = serv.m_szOnConnectCommand;
	m_szOnLoginCommand   = serv.m_szOnLoginCommand;
	m_szLinkFilter       = serv.m_szLinkFilter;
	if(serv.m_pChannelList)
		m_pChannelList = new QStringList(*(serv.m_pChannelList));
	else m_pChannelList = 0;
	m_bAutoConnect       = serv.m_bAutoConnect;
}

void KviIrcServer::operator=(const KviIrcServer &serv)
{
	m_szHostname         = serv.m_szHostname;
	m_szIp               = serv.m_szIp;
	m_szDescription      = serv.m_szDescription;
	m_szUser             = serv.m_szUser;
	m_szPass             = serv.m_szPass;
	m_uPort              = serv.m_uPort;
	m_szNick             = serv.m_szNick;
	m_szRealName         = serv.m_szRealName;
	m_uFlags             = serv.m_uFlags;
	m_szInitUMode = serv.m_szInitUMode;
	m_szOnConnectCommand = serv.m_szOnConnectCommand;
	m_szOnLoginCommand   = serv.m_szOnLoginCommand;
	m_szLinkFilter       = serv.m_szLinkFilter;
	if(m_pChannelList)delete m_pChannelList;
	if(serv.m_pChannelList)
		m_pChannelList = new QStringList(*(serv.m_pChannelList));
	else m_pChannelList = 0;
	m_bAutoConnect    = serv.m_bAutoConnect;
}


KviIrcServer::~KviIrcServer()
{
	if(m_pChannelList)delete m_pChannelList;
}


void KviIrcServer::setAutoJoinChannelList(QStringList * pNewChannelList)
{
	if(m_pChannelList)delete m_pChannelList;
	m_pChannelList = pNewChannelList;
}

KviIrcNetwork::KviIrcNetwork(const QString & name)
{
	m_pServerList = new KviPtrList<KviIrcServer>;
	m_pServerList->setAutoDelete(true);
	m_szName = name;
	m_pCurrentServer = 0;
}


KviIrcNetwork::~KviIrcNetwork()
{
	delete m_pServerList;
}

void KviIrcNetwork::insertServer(KviIrcServer *srv)
{
	m_pServerList->append(srv);
}

KviIrcServer * KviIrcNetwork::findServer(const QString &szHostname, unsigned short int port)
{
	for(KviIrcServer *s=m_pServerList->first();s;s=m_pServerList->next())
	{
		if(KviQString::equalCI(s->m_szHostname,szHostname) && (s->m_uPort == port))return s;
	}
	return 0;
}


void KviIrcNetwork::setCurrentServer(KviIrcServer *srv)
{
	if(m_pServerList->findRef(srv) != -1)m_pCurrentServer = srv;
}

