//=============================================================================
//
//   File : kvi_ircserverdb.cpp
//   Creation date : Mon Jul 10 2000 14:25:00 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2004 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __LIBKVILIB__

#include <qapplication.h>
#include <qlayout.h>
#include <qmessagebox.h>
#include <qcheckbox.h>

#include "kvi_ircserverdb.h"
#include "kvi_config.h"
#include "kvi_locale.h"
#include "kvi_netutils.h"






// NETWORK NAMES , SERVER NAMES & SERVER DESCRIPTIONS are in Latin1 charset

KviIrcServerDataBase::KviIrcServerDataBase()
{
	m_pNetworkList = new KviPtrList<KviIrcNetwork>;
	m_pNetworkList->setAutoDelete(true);
	m_pCurrentNetwork = 0;
	m_pAutoConnectOnStartupServers = 0;
}

KviIrcServerDataBase::~KviIrcServerDataBase()
{
	delete m_pNetworkList;
	if(m_pAutoConnectOnStartupServers)delete m_pAutoConnectOnStartupServers;
}

void KviIrcServerDataBase::clearAutoConnectOnStartupServers()
{
	if(!m_pAutoConnectOnStartupServers)return;
	delete m_pAutoConnectOnStartupServers;
	m_pAutoConnectOnStartupServers = 0;
}


void KviIrcServerDataBase::clear()
{
//	delete m_pNetworkList;
//	m_pNetworkList = new KviPtrList<KviIrcNetwork>;
//	m_pNetworkList->setAutoDelete(true);
	m_pNetworkList->clear();
	m_pCurrentNetwork = 0;
}

void KviIrcServerDataBase::insertNetwork(KviIrcNetwork *n)
{
	int i = 0;
	for(KviIrcNetwork * net = m_pNetworkList->first();net;net = m_pNetworkList->next())
	{
		if(KviQString::cmpCI(net->m_szName,n->m_szName) > 0)
		{
			m_pNetworkList->insert(i,n);
			return;
		}
		i++;
	}
	m_pNetworkList->append(n);
}

KviIrcNetwork * KviIrcServerDataBase::findNetwork(const QString &szName)
{
	for(KviIrcNetwork * net = m_pNetworkList->first();net;net = m_pNetworkList->next())
	{
		if(KviQString::equalCI(net->m_szName,szName))return net;
	}
	return 0;
}

void KviIrcServerDataBase::updateServerIp(const QString & server,unsigned short int port,const QString & ip)
{
	for(KviIrcNetwork * net = m_pNetworkList->first();net;net = m_pNetworkList->next())
	{
		KviIrcServer * srv = net->findServer(server,port);
		if(srv)
		{
			srv->m_szIp = ip;
			return;
		}
	}
}

//void KviIrcServerDataBase::disableCaching(const char * server)
//{
//	for(KviIrcNetwork * net = m_pNetworkList->first();net;net = m_pNetworkList->next())
//	{
//		KviIrcServer * srv = net->findServer(server);
//		if(srv)
//		{
//			srv->setCacheIp(false);
//		}
//	}
//}

//void KviIrcServerDataBase::iWantThisServerToBeTheCurrent(const char * server,unsigned short int uPort,bool bPortIsValid,bool bIpV6,bool bSSL,const char * linkFilter)
void KviIrcServerDataBase::makeCurrentServer(KviIrcServerDefinition * d)
{
	for(KviIrcNetwork * net = m_pNetworkList->first();net;net = m_pNetworkList->next())
	{
		KviIrcServer * srv = net->findServer(d->szServer,d->uPort);
		if(srv)
		{
			if((d->bIpV6 == srv->isIpV6()) && (d->bSSL == srv->useSSL()))
			{
				bool bThisOne = true;
				if(!d->szLinkFilter.isEmpty())
				{
					if(!KviQString::equalCI(d->szLinkFilter,srv->m_szLinkFilter))bThisOne = false;
				}
				if(bThisOne)
				{
					if(d->bPortIsValid)srv->setPort(d->uPort);
					if(!d->szNick.isEmpty())srv->m_szNick = d->szNick;
					if(!d->szPass.isEmpty())srv->m_szPass = d->szPass; // don't clear the pass!
					if(!d->szInitUMode.isEmpty())srv->m_szInitUMode = d->szInitUMode;
					setCurrentNetwork(net);
					net->setCurrentServer(srv);
					return;
				}
			}
		}
	}

	// no such server: is it a valid ip address or hostname ?
	bool bIsValidIpV4 = kvi_isValidStringIp(d->szServer.latin1());
#ifdef COMPILE_IPV6_SUPPORT
	bool bIsValidIpV6 = kvi_isValidStringIp_V6(d->szServer.latin1());
#else
	bool bIsValidIpV6 = false;
#endif

	if(!(bIsValidIpV4 || bIsValidIpV6))
	{
		// is it a valid hostname ? (must contain at least one dot)
		if(d->szServer.contains('.') < 1)
		{
			// assume it is a network name!
			KviIrcNetwork * n = findNetwork(d->szServer);
			if(n)
			{
				setCurrentNetwork(n);

				// find a round-robin server in that network

				for(KviIrcServer * s = n->m_pServerList->first();s;s = n->m_pServerList->next())
				{
					if(s->m_szDescription.contains("random",false) ||
						(s->m_szDescription.contains("round",false) && s->m_szDescription.contains("robin",false)))
					{
						n->setCurrentServer(s);
						return;
					}
				}

				// no explicit round robin... try some common names

				QString tryAlso1,tryAlso2,tryAlso3;

				KviQString::sprintf(tryAlso1,"irc.%Q.org",&(d->szServer));
				KviQString::sprintf(tryAlso2,"irc.%Q.net",&(d->szServer));
				KviQString::sprintf(tryAlso3,"irc.%Q.com",&(d->szServer));

				for(KviIrcServer * s = n->m_pServerList->first();s;s = n->m_pServerList->next())
				{
					if(KviQString::equalCI(s->m_szHostname,tryAlso1) ||
						KviQString::equalCI(s->m_szHostname,tryAlso2) ||
						KviQString::equalCI(s->m_szHostname,tryAlso3))
					{
						n->setCurrentServer(s);
						return;
					}
				}

				// a random one in this network
				return;
			} // else unreasonable..but :D
		}
	}

	// a valid hostname or ip address , not found in list : add it and make it current

	KviIrcNetwork * n = findNetwork(__tr2qs("orphan_servers"));
	if(!n)
	{
		n = new KviIrcNetwork(__tr2qs("orphan_servers"));
		insertNetwork(n);
	}
	KviIrcServer * s = new KviIrcServer();
	s->m_szHostname = d->szServer;
	if(bIsValidIpV4)
	{
		s->m_szIp = d->szServer;
		s->setCacheIp(true);
#ifdef COMPILE_IPV6_SUPPORT
	} else {
		if(bIsValidIpV6)
		{
			s->m_szIp = d->szServer;
			s->setCacheIp(true);
			d->bIpV6 = true;
		}
	}
#else
	}
#endif
	s->m_uPort = d->bPortIsValid ? d->uPort : 6667;
	s->setLinkFilter(d->szLinkFilter);
	s->m_szPass= d->szPass;
	s->m_szNick= d->szNick;
    s->m_szInitUMode = d->szInitUMode;
	s->setIpV6(d->bIpV6);
	s->setUseSSL(d->bSSL);
	n->insertServer(s);
	setCurrentNetwork(n);
	n->setCurrentServer(s);
}

void KviIrcServerDataBase::setCurrentNetwork(KviIrcNetwork *n)
{
	if(m_pNetworkList->findRef(n) != -1)m_pCurrentNetwork = n;
}


void KviIrcServerDataBase::load(const QString & filename)
{
	clear();
	KviConfig cfg(filename,KviConfig::Read);

	KviConfigIterator it(*(cfg.dict()));
	while(it.current())
	{
		if(it.current()->count() > 0)
		{
			KviIrcNetwork * n = new KviIrcNetwork(it.currentKey());
			insertNetwork(n);
			cfg.setGroup(it.currentKey());
			int nServers = cfg.readIntEntry("NServers",0);
			if(cfg.readBoolEntry("Current",false))m_pCurrentNetwork = n;
			for(int i=0;i < nServers;i++)
			{
				KviIrcServer *s = new KviIrcServer();
				n->m_pServerList->append(s);
				QString tmp;
				KviQString::sprintf(tmp,"%d_Current",i);
				if(cfg.readBoolEntry(tmp,false))n->m_pCurrentServer = s;
				KviQString::sprintf(tmp,"%d_Hostname",i);
				s->m_szHostname = cfg.readQStringEntry(tmp,"irc.unknown.net");
				KviQString::sprintf(tmp,"%d_Ip",i);
				s->m_szIp = cfg.readQStringEntry(tmp,"");
				KviQString::sprintf(tmp,"%d_Description",i);
				s->m_szDescription = cfg.readQStringEntry(tmp,"");
				KviQString::sprintf(tmp,"%d_User",i);
				s->m_szUser = cfg.readQStringEntry(tmp,"");
				KviQString::sprintf(tmp,"%d_Pass",i);
				s->m_szPass = cfg.readQStringEntry(tmp,"");
				KviQString::sprintf(tmp,"%d_Nick",i);
				s->m_szNick = cfg.readQStringEntry(tmp,"");
				KviQString::sprintf(tmp,"%d_RealName",i);
				s->m_szRealName = cfg.readQStringEntry(tmp,"");
				KviQString::sprintf(tmp,"%d_InitUmode",i);
				s->m_szInitUMode = cfg.readQStringEntry(tmp,"");
				KviQString::sprintf(tmp,"%d_AutoJoinChannels",i);
				QStringList l = cfg.readStringListEntry(tmp,QStringList());
				if(l.count() > 0)s->setAutoJoinChannelList(new QStringList(l));
				KviQString::sprintf(tmp,"%d_AutoConnect",i);
				bool bTmp = cfg.readBoolEntry(tmp,false);
				s->setAutoConnect(bTmp);
				if(bTmp)
				{
					if(!m_pAutoConnectOnStartupServers)
					{
						m_pAutoConnectOnStartupServers = new KviPtrList<KviIrcServer>;
						m_pAutoConnectOnStartupServers->setAutoDelete(false);
					}
					m_pAutoConnectOnStartupServers->append(s);
				}
				KviQString::sprintf(tmp,"%d_OnConnectCommand",i);
				s->m_szOnConnectCommand = cfg.readEntry(tmp,"");
				KviQString::sprintf(tmp,"%d_OnLoginCommand",i);
				s->m_szOnLoginCommand = cfg.readEntry(tmp,"");
				KviQString::sprintf(tmp,"%d_LinkFilter",i);
				s->m_szLinkFilter = cfg.readEntry(tmp,"");
				KviQString::sprintf(tmp,"%d_Port",i);
				s->m_uPort = cfg.readUShortEntry(tmp,6667);
				KviQString::sprintf(tmp,"%d_IpV6",i);
				bTmp = cfg.readBoolEntry(tmp,false);
				s->setIpV6(bTmp);
				KviQString::sprintf(tmp,"%d_CacheIp",i);
				bTmp = cfg.readBoolEntry(tmp,true);
				s->setCacheIp(bTmp);
				KviQString::sprintf(tmp,"%d_SSL",i);
				bTmp = cfg.readBoolEntry(tmp,false);
				s->setUseSSL(bTmp);
			}
			if(!n->m_pCurrentServer)n->m_pCurrentServer = n->m_pServerList->first();
		}
		++it;
	}
	if(!m_pCurrentNetwork)m_pCurrentNetwork = m_pNetworkList->first();
}

void KviIrcServerDataBase::save(const QString &filename)
{
	KviConfig cfg(filename,KviConfig::Write);

	cfg.clear(); // clear any old entry

	for(KviIrcNetwork * n = m_pNetworkList->first();n;n = m_pNetworkList->next())
	{
		cfg.setGroup(n->m_szName);
		cfg.writeEntry("NServers",n->m_pServerList->count());
		if(n == m_pCurrentNetwork)cfg.writeEntry("Current",true);

		int i=0;
		for(KviIrcServer *s = n->m_pServerList->first();s;s = n->m_pServerList->next())
		{
			QString tmp;
			KviQString::sprintf(tmp,"%d_Hostname",i);
			cfg.writeEntry(tmp,s->m_szHostname);
			if(!s->m_szIp.isEmpty())
			{
				KviQString::sprintf(tmp,"%d_Ip",i);
				cfg.writeEntry(tmp,s->m_szIp);
			}
			if(!s->m_szDescription.isEmpty())
			{
				KviQString::sprintf(tmp,"%d_Description",i);
				cfg.writeEntry(tmp,s->m_szDescription);
			}
			if(!s->m_szUser.isEmpty())
			{
				KviQString::sprintf(tmp,"%d_User",i);
				cfg.writeEntry(tmp,s->m_szUser);
			}
			if(!s->m_szPass.isEmpty())
			{
				KviQString::sprintf(tmp,"%d_Pass",i);
				cfg.writeEntry(tmp,s->m_szPass);
			}
			if(!s->m_szNick.isEmpty())
			{
				KviQString::sprintf(tmp,"%d_Nick",i);
				cfg.writeEntry(tmp,s->m_szNick);
			}
			if(!s->m_szRealName.isEmpty())
			{
				KviQString::sprintf(tmp,"%d_RealName",i);
				cfg.writeEntry(tmp,s->m_szRealName);
			}
			if(!s->m_szInitUMode.isEmpty())
			{
				KviQString::sprintf(tmp,"%d_InitUMode",i);
				cfg.writeEntry(tmp,s->m_szInitUMode);
			}
			if(s->autoJoinChannelList())
			{
				KviQString::sprintf(tmp,"%d_AutoJoinChannels",i);
				cfg.writeEntry(tmp,*(s->autoJoinChannelList()));
			}
			if(s->autoConnect()) // otherwise it defaults to false anyway
			{
				KviQString::sprintf(tmp,"%d_AutoConnect",i);
				cfg.writeEntry(tmp,s->autoConnect());
			}
			if(!s->m_szOnConnectCommand.isEmpty())
			{
				KviQString::sprintf(tmp,"%d_OnConnectCommand",i);
				cfg.writeEntry(tmp,s->m_szOnConnectCommand);
			}
			if(!s->m_szOnLoginCommand.isEmpty())
			{
				KviQString::sprintf(tmp,"%d_OnLoginCommand",i);
				cfg.writeEntry(tmp,s->m_szOnLoginCommand);
			}
			if(!s->m_szLinkFilter.isEmpty())
			{
				KviQString::sprintf(tmp,"%d_LinkFilter",i);
				cfg.writeEntry(tmp,s->m_szLinkFilter);
			}
			if(s->m_uPort != 6667)
			{
				KviQString::sprintf(tmp,"%d_Port",i);
				cfg.writeEntry(tmp,s->m_uPort);
			}
			if(s->isIpV6())
			{
				KviQString::sprintf(tmp,"%d_IpV6",i);
				cfg.writeEntry(tmp,s->isIpV6());
			}
			if(!(s->cacheIp()))
			{
				KviQString::sprintf(tmp,"%d_CacheIp",i);
				cfg.writeEntry(tmp,s->cacheIp());
			}
			if(s->useSSL())
			{
				KviQString::sprintf(tmp,"%d_SSL",i);
				cfg.writeEntry(tmp,s->useSSL());
			}
			if(s == n->m_pCurrentServer)
			{
				KviQString::sprintf(tmp,"%d_Current",i);
				cfg.writeEntry(tmp,true);
			}
			i++;
		}
	}
}

