#define __KVIRC__
#define _KVI_LAGMETER_CPP_

//=============================================================================
//
//   File : kvi_lagmeter.cpp
//   Creation date : Fri Oct 18 13:31:36 CEST 2002 by Juanjo lvarez
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#include "kvi_lagmeter.h"
#include "kvi_options.h"
#include "kvi_event.h"
#include "kvi_parameterlist.h"

KviLagMeter::KviLagMeter(KviConsole * par)
{
    m_iLagTime = 0;
    m_iInterval = KVI_OPTION_UINT(KviOption_uintLagCheckingInterval) * 1000; 
    m_iLagAlarmTime = KVI_OPTION_UINT(KviOption_uintLagAlarmTime);
    m_bPingSent = false;
    m_bEnabled = false;
    m_bOnAlarm = false;
    m_pConsole = par;
    m_intervalTimer = new QTimer();
    m_tmpStr = new KviStr();
    connect(m_intervalTimer, SIGNAL(timeout()), this, SLOT(issuePing()));
}

KviLagMeter::~KviLagMeter()
{
    m_intervalTimer->stop();
    delete m_intervalTimer;
    delete m_tmpStr;
}

void KviLagMeter::enable(void)
{
    if(m_pConsole->state() != KviConsole::Connected)return;
    m_intervalTimer->start(m_iInterval);
    m_bEnabled = true;
}

void KviLagMeter::disable(void)
{
    m_intervalTimer->stop();
    m_bEnabled = false;
}

void KviLagMeter::issuePing(void)
{
    if(!m_bEnabled || (m_pConsole->state() != KviConsole::Connected))return;
    m_intervalTimer->stop();

    //FIXME: Better would be to add a ping command...
    KviStr command(KviStr::Format,"raw -q ping %s",m_pConsole->currentServerName());
    g_pUserParser->parseCommandBuffer(command.ptr(),m_pConsole);
    m_pingTime = QTime::currentTime();
    m_bPingSent = true;
}

void KviLagMeter::pongReceived(void)
{
    if(!m_bEnabled || !(m_pConsole->state() == KviConsole::Connected))return;
    m_iLagTime = m_pingTime.msecsTo(QTime::currentTime());
    
    if( (m_iLagTime > m_iLagAlarmTime) && (!m_bOnAlarm))
    {
        TRIGGER_EVENT_2PARAM_RETVALUE(KviEvent_OnLagAlarmTimeUp,m_pConsole,m_pConsole->currentServerName(),m_tmpStr->sprintf("%d",m_iLagTime).ptr());
        m_bOnAlarm = true;
        emit OnLagAlarmTimeUp();
    } else if ( (m_iLagTime < m_iLagAlarmTime) && m_bOnAlarm) 
    {
        TRIGGER_EVENT_2PARAM_RETVALUE(KviEvent_OnLagAlarmTimeDown,m_pConsole,m_pConsole->currentServerName(),m_tmpStr->sprintf("%d",m_iLagTime).ptr());
        m_bOnAlarm = false;
        emit OnLagAlarmTimeDown();
    }

    emit newLag(m_iLagTime);
    TRIGGER_EVENT_2PARAM_RETVALUE(KviEvent_OnLagCheck,m_pConsole,m_pConsole->currentServerName(),m_tmpStr->sprintf("%d",m_iLagTime).ptr());
            
    if(m_bEnabled)m_intervalTimer->start(m_iInterval);
}
 
#include "kvi_lagmeter.moc"
