//=============================================================================
//
//   File : kvi_kvs_eventmanager.cpp
//   Created on Thu Aug 17 2000 13:59:12 by Szymon Stefanek
//
//   This file is part of the KVIrc IRC client distribution
//   Copyright (C) 2000-2004 Szymon Stefanek <pragma at kvirc dot net>
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __KVIRC__

#include "kvi_kvs_eventmanager.h"
#include "kvi_config.h"
#include "kvi_kvs_script.h"
#include "kvi_kvs_variant.h"
#include "kvi_options.h"
#include "kvi_locale.h"
#include "kvi_out.h"
#include "kvi_module.h"


KviKvsEventManager * KviKvsEventManager::m_pInstance = 0;

KviKvsEventManager::KviKvsEventManager()
{
	m_pInstance = this;
	for(int i=0;i<KVI_KVS_NUM_RAW_EVENTS;i++)
		m_rawEventTable[i] = 0;
}

KviKvsEventManager::~KviKvsEventManager()
{
	clear();
}

void KviKvsEventManager::init()
{
	if(KviKvsEventManager::instance())
	{
		debug("WARNING: Trying to create KviKvsEventManager twice!");
		return;
	}
	(void) new KviKvsEventManager();
}

void KviKvsEventManager::done()
{
	if(!KviKvsEventManager::instance())
	{
		debug("WARNING: Trying to destroy the KviKvsEventManager twice!");
		return;
	}
	delete KviKvsEventManager::instance();
}

bool KviKvsEventManager::addAppHandler(unsigned int uEvIdx,KviKvsEventHandler * h)
{
	if(uEvIdx >= KVI_KVS_NUM_APP_EVENTS)return false;
	m_appEventTable[uEvIdx].addHandler(h);
	return true;
}

bool KviKvsEventManager::addRawHandler(unsigned int uRawIdx,KviKvsEventHandler * h)
{
	if(uRawIdx >= KVI_KVS_NUM_RAW_EVENTS)return false;
	if(!m_rawEventTable[uRawIdx])
	{
		m_rawEventTable[uRawIdx] = new KviPtrList<KviKvsEventHandler>();
		m_rawEventTable[uRawIdx]->setAutoDelete(true);
	}
	m_rawEventTable[uRawIdx]->append(h);
	return true;
}

void KviKvsEventManager::removeModuleAppHandler(unsigned int uEvIdx,KviKvsModuleInterface *i)
{
	if(uEvIdx >= KVI_KVS_NUM_APP_EVENTS)return;
	KviKvsEventHandler * h;
	for(h = m_appEventTable[uEvIdx].handlers()->first();h;h = m_appEventTable[uEvIdx].handlers()->next())
	{
		if(h->type() == KviKvsEventHandler::Module)
		{
			if(((KviKvsModuleEventHandler *)h)->moduleInterface() == i)
			{
				m_appEventTable[uEvIdx].removeHandler(h);
				return;
			}
		}
	}
}

void KviKvsEventManager::removeAllModuleAppHandlers(KviKvsModuleInterface *pIface)
{
	KviKvsEventHandler * h;
	for(unsigned int i =0;i< KVI_KVS_NUM_APP_EVENTS;i++)
	{
		if(!m_appEventTable[i].handlers())continue;
	
		KviPtrList<KviKvsEventHandler> l;
		l.setAutoDelete(false);
		for(h = m_appEventTable[i].handlers()->first();h;h = m_appEventTable[i].handlers()->next())
		{
			if(h->type() == KviKvsEventHandler::Module)
			{
				if(((KviKvsModuleEventHandler *)h)->moduleInterface() == pIface)
				{
					l.append(h);
				}
			}
		}
		for(h = l.first();h;h = l.next())m_appEventTable[i].removeHandler(h);
	}
}

void KviKvsEventManager::removeAllModuleRawHandlers(KviKvsModuleInterface *pIface)
{
	KviKvsEventHandler * h;
	for(unsigned int i =0;i< KVI_KVS_NUM_RAW_EVENTS;i++)
	{
		if(!m_rawEventTable[i])continue;

		KviPtrList<KviKvsEventHandler> l;
		l.setAutoDelete(false);
		for(h = m_rawEventTable[i]->first();h;h = m_rawEventTable[i]->next())
		{
			if(h->type() == KviKvsEventHandler::Module)
			{
				if(((KviKvsModuleEventHandler *)h)->moduleInterface() == pIface)
				{
					l.append(h);
				}
			}
		}
		for(h = l.first();h;h = l.next())m_rawEventTable[i]->removeRef(h);
		if(m_rawEventTable[i]->isEmpty())
		{
			delete m_rawEventTable[i];
			m_rawEventTable[i] = 0;
		}
	}
}



void KviKvsEventManager::removeModuleRawHandler(unsigned int uRawIdx,KviKvsModuleInterface *i)
{
	if(uRawIdx >= KVI_KVS_NUM_RAW_EVENTS)return;
	if(!m_rawEventTable[uRawIdx])return;
	KviKvsEventHandler * h;
	for(h = m_rawEventTable[uRawIdx]->first();h;h = m_rawEventTable[uRawIdx]->next())
	{
		if(h->type() == KviKvsEventHandler::Module)
		{
			if(((KviKvsModuleEventHandler *)h)->moduleInterface() == i)
			{
				m_rawEventTable[uRawIdx]->removeRef(h);
				if(m_rawEventTable[uRawIdx]->isEmpty())
				{
					delete m_rawEventTable[uRawIdx];
					m_rawEventTable[uRawIdx] = 0;
				}
				return;
			}
		}
	}
}


void KviKvsEventManager::clearAppScriptHandlers()
{
	for(int i=0;i< KVI_KVS_NUM_APP_EVENTS;i++)
	{
		m_appEventTable[i].clearScriptHandlers();
	}
}

void KviKvsEventManager::clearRawScriptHandlers()
{
	for(int i=0;i< KVI_KVS_NUM_RAW_EVENTS;i++)
	{
		if(m_rawEventTable[i])
		{
			KviPtrList<KviKvsEventHandler> dl;
			dl.setAutoDelete(false);
			KviKvsEventHandler * e;
			for(e = m_rawEventTable[i]->first();e;e = m_rawEventTable[i]->next())
			{
				if(e->type() == KviKvsEventHandler::Script)dl.append(e);
			}
			
			for(e = dl.first();e;e = dl.next())
			{
				m_rawEventTable[i]->removeRef(e);
			}

			if(m_rawEventTable[i]->isEmpty())
			{
				delete m_rawEventTable[i];
				m_rawEventTable[i] = 0;
			}
		}
	}
}

void KviKvsEventManager::clearRawEvents()
{
	for(int i=0;i<KVI_KVS_NUM_RAW_EVENTS;i++)
	{
		if(m_rawEventTable[i])delete m_rawEventTable[i];
		m_rawEventTable[i] = 0;
	}
}

void KviKvsEventManager::clearAppEvents()
{
	for(int i=0;i<KVI_KVS_NUM_APP_EVENTS;i++)
	{
		m_appEventTable[i].clear();
	}
}

void KviKvsEventManager::clear()
{
	clearRawEvents();
	clearAppEvents();
}

bool KviKvsEventManager::triggerHandlers(KviPtrList<KviKvsEventHandler> * pHandlers,KviWindow *pWnd,KviKvsVariantList *pParams)
{
	if(!pHandlers)return false;

	bool bGotHalt = false;
	for(KviKvsEventHandler * h = pHandlers->first();h;h = pHandlers->next())
	{
		switch(h->type())
		{
			case KviKvsEventHandler::Script:
			{
				KviKvsScript * s = ((KviKvsScriptEventHandler *)h)->script();
				KviKvsScript copy(*s);
				KviKvsVariant retVal;
				if(!copy.run(pWnd,pParams,&retVal))
				{
					// error! disable the handler if it's broken
					if(KVI_OPTION_BOOL(KviOption_boolDisableBrokenEventHandlers))
					{
						((KviKvsScriptEventHandler *)h)->setEnabled(false);
						pWnd->output(KVI_OUT_PARSERERROR,__tr2qs("Event handler %Q is broken: disabling"),&(s->name()));
					}
				}
				if(!bGotHalt)bGotHalt = !retVal.isNothing();
			}
			break;
			case KviKvsEventHandler::Module:
			{
				KviModule * m = (KviModule *)((KviKvsModuleEventHandler *)h)->moduleInterface();
				KviKvsModuleEventHandlerRoutine * proc = ((KviKvsModuleEventHandler *)h)->handlerRoutine();
				KviKvsModuleEventCall call(m,pParams);
				if(!(*proc)(&call))bGotHalt = true;
			}
			break;
		}
	}
	return bGotHalt;
}


void KviKvsEventManager::loadRawEvents(const QString &szFileName)
{
	KviConfig cfg(szFileName,KviConfig::Read);
	clearRawScriptHandlers();
	int i;

	for(i=0;i<KVI_KVS_NUM_RAW_EVENTS;i++)
	{
		QString tmp;
		KviQString::sprintf(tmp,"RAW%d",i);
		if(cfg.hasGroup(tmp))
		{
			cfg.setGroup(tmp);
			unsigned int nHandlers = cfg.readUIntEntry("NHandlers",0);
			if(nHandlers)
			{
				m_rawEventTable[i] = new KviPtrList<KviKvsEventHandler>();
				m_rawEventTable[i]->setAutoDelete(true);
				for(unsigned int index = 0;index < nHandlers;index++)
				{
					KviQString::sprintf(tmp,"Name%u",index);
					QString szName = cfg.readQStringEntry(tmp,"unnamed");
					KviQString::sprintf(tmp,"Buffer%u",index);
					QString * pszCode = new QString();
					*pszCode = cfg.readQStringEntry(tmp,"");
					KviQString::sprintf(tmp,"RawEvent%u::%Q",index,&szName);
					KviKvsScriptEventHandler * s = new KviKvsScriptEventHandler(szName,tmp,pszCode);
					KviQString::sprintf(tmp,"Enabled%u",index);
					s->setEnabled(cfg.readBoolEntry(tmp,false));
					m_rawEventTable[i]->append(s);
				}
			}
		}
	}
}

void KviKvsEventManager::saveRawEvents(const QString &szFileName)
{
	KviConfig cfg(szFileName,KviConfig::Write);
	cfg.clear();
	int i;

	for(i=0;i<KVI_KVS_NUM_RAW_EVENTS;i++)
	{
		if(m_rawEventTable[i])
		{
			QString tmp;
			KviQString::sprintf(tmp,"RAW%d",i);
			cfg.setGroup(tmp);

			int index = 0;
			for(KviKvsEventHandler * s = m_rawEventTable[i]->first();s; s = m_rawEventTable[i]->next())
			{
				if(s->type() == KviKvsEventHandler::Script)
				{
					KviQString::sprintf(tmp,"Name%d",index);
					cfg.writeEntry(tmp,((KviKvsScriptEventHandler *)s)->name());
					KviQString::sprintf(tmp,"Buffer%d",index);
					cfg.writeEntry(tmp,((KviKvsScriptEventHandler *)s)->code());
					KviQString::sprintf(tmp,"Enabled%d",index);
					cfg.writeEntry(tmp,((KviKvsScriptEventHandler *)s)->isEnabled());
					index++;
				}
			}
			cfg.writeEntry("NHandlers",index);
		}
	}

}

void KviKvsEventManager::loadAppEvents(const QString &szFileName)
{
	KviConfig cfg(szFileName,KviConfig::Read);
	clearAppScriptHandlers();

	int i;
	for(i=0;i<KVI_KVS_NUM_APP_EVENTS;i++)
	{
		if(cfg.hasGroup(m_appEventTable[i].name()))
		{
			cfg.setGroup(m_appEventTable[i].name());
			unsigned int nHandlers = cfg.readUIntEntry("NHandlers",0);
			if(nHandlers)
			{
				for(unsigned int index = 0;index < nHandlers;index++)
				{
					QString tmp;
					KviQString::sprintf(tmp,"Name%u",index);
					QString szName = cfg.readQStringEntry(tmp,"unnamed");
					KviQString::sprintf(tmp,"Buffer%u",index);
					QString * pszCode = new QString();
					*pszCode = cfg.readQStringEntry(tmp,"");
					KviQString::sprintf(tmp,"Enabled%u",index);
					bool bEnabled = cfg.readBoolEntry(tmp,false);
					QString szCntx;
					KviQString::sprintf(szCntx,"%Q::%Q",&(m_appEventTable[i].name()),&szName);
					KviKvsScriptEventHandler *s = new KviKvsScriptEventHandler(szName,szCntx,pszCode,bEnabled);
					m_appEventTable[i].addHandler(s);
				}
			}
		}
	}

}

void KviKvsEventManager::saveAppEvents(const QString &szFileName)
{
	KviConfig cfg(szFileName,KviConfig::Write);
	cfg.clear();
	int i;

	for(i=0;i<KVI_KVS_NUM_APP_EVENTS;i++)
	{
		if(m_appEventTable[i].hasHandlers())
		{
			cfg.setGroup(m_appEventTable[i].name());
			int index = 0;
			for(KviKvsEventHandler* s = m_appEventTable[i].handlers()->first();s; s = m_appEventTable[i].handlers()->next())
			{
				if(s->type() == KviKvsEventHandler::Script)
				{
					QString tmp;
					KviQString::sprintf(tmp,"Name%d",index);
					cfg.writeEntry(tmp,((KviKvsScriptEventHandler *)s)->name());
					KviQString::sprintf(tmp,"Buffer%d",index);
					cfg.writeEntry(tmp,((KviKvsScriptEventHandler *)s)->code());
					KviQString::sprintf(tmp,"Enabled%d",index);
					cfg.writeEntry(tmp,((KviKvsScriptEventHandler *)s)->isEnabled());
					index++;
				}
			}
			cfg.writeEntry("NHandlers",index);
		}
	}
}

