//=============================================================================
//
//   File : kvi_kvs_treenode_switchlist.cpp
//   Created on Tue 07 Oct 2003 02:06:53 by Szymon Stefanek
//
//   This file is part of the KVIrc IRC client distribution
//   Copyright (C) 2003 Szymon Stefanek <pragma at kvirc dot net>
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __KVIRC__

#include "kvi_kvs_treenode_switchlist.h"
#include "kvi_kvs_runtimecontext.h"

KviKvsTreeNodeSwitchList::KviKvsTreeNodeSwitchList(const QChar * pLocation)
: KviKvsTreeNode(pLocation)
{
	m_pSwitchDict = new QIntDict<KviKvsTreeNodeData>(11);
	m_pSwitchDict->setAutoDelete(true);
}

KviKvsTreeNodeSwitchList::~KviKvsTreeNodeSwitchList()
{
	delete m_pSwitchDict;
}



void KviKvsTreeNodeSwitchList::dump(const char * prefix)
{
	debug("%s SwitchList",prefix);
	QIntDictIterator<KviKvsTreeNodeData> it(*m_pSwitchDict);
	while(it.current())
	{
		QString tmp = prefix;
		tmp.append("  Sw(");
		QChar c((unsigned short)it.currentKey());
		tmp.append(c);
		tmp.append("): ");
		it.current()->dump(tmp.latin1());
		++it;
	}
}

void KviKvsTreeNodeSwitchList::addSwitch(int key,KviKvsTreeNodeData * p)
{
	m_pSwitchDict->replace(key,p);
	p->setParent(this);
}


bool KviKvsTreeNodeSwitchList::evaluate(KviKvsRunTimeContext * c,KviKvsSwitchList * pSwList)
{
	pSwList->clear();
	QIntDictIterator<KviKvsTreeNodeData> it(*m_pSwitchDict);
	while(KviKvsTreeNodeData * d = it.current())
	{
		KviKvsVariant * v = new KviKvsVariant();
		if(!d->evaluateReadOnly(c,v))
		{
			delete v; 
			return false;
		}
		pSwList->add(it.currentKey(),v);
		++it;
	}
	return true;
}

