//====================================================================================
//
//   File : kvi_sp_literal.cpp
//   Creation date : Thu Aug 3 2000 01:29:12 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2004 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//====================================================================================

#define __KVIRC__

#include "kvi_sparser.h"
#include "kvi_window.h"
#include "kvi_out.h"
#include "kvi_locale.h"
#include "kvi_irclink.h"
#include "kvi_options.h"
#include "kvi_ircmask.h"
#include "kvi_channel.h"
#include "kvi_topicw.h"
#include "kvi_frame.h"
#include "kvi_mirccntrl.h"
#include "kvi_query.h"
#include "kvi_userlistview.h"
#include "kvi_antispam.h"
#include "kvi_nickserv.h"
#include "kvi_uparser.h"
#include "kvi_event.h"
#include "kvi_parameterlist.h"
#include "kvi_ircuserdb.h"
#include "kvi_app.h"
#include "kvi_regusersdb.h"
#include "kvi_debug.h"
#include "kvi_time.h"
#include "kvi_useraction.h"

#include "kvi_settings.h"

#ifdef COMPILE_CRYPT_SUPPORT
	#include "kvi_crypt.h"
	#include "kvi_cryptcontroller.h"
#endif

//#include "kvi_regusersdb.h"
//#include "kvi_iconmanager.h"
#include <qdatetime.h>

// kvi_app.cpp
extern KviRegisteredUserDataBase * g_pRegisteredUserDataBase;

extern KviNickServDataBase * g_pNickServDataBase;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// PING
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void KviServerParser::parseLiteralPing(KviIrcMessage *msg)
{
	// PING
	// <optional_prefix> PING :<argument>

	msg->console()->link()->sendFmtData("PONG %s",msg->allParams());

	if(TRIGGER_EVENT_2PARAM_RETVALUE(KviEvent_OnPing,msg->console(),msg->safePrefix(),msg->allParams()))
		msg->setHaltOutput();

	if((!msg->haltOutput()) && KVI_OPTION_BOOL(KviOption_boolShowPingPong))
	{
		msg->console()->output(KVI_OUT_SERVERPING,
			__tr2wc("Received ping from \r!s\r%s\r (PING %s), replied pong"),msg->safePrefix(),msg->allParams());
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PONG
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void KviServerParser::parseLiteralPong(KviIrcMessage *msg)
{
	if(TRIGGER_EVENT_2PARAM_RETVALUE(KviEvent_OnPong,msg->console(),msg->safePrefix(),msg->allParams()))
		msg->setHaltOutput();
		
	if(msg->console()->lagMeter()->isEnabled()) 
	{
		msg->console()->lagMeter()->pongReceived();
		// this message should be hidden ONLY if we're sure that it was internally generated
	}

	if((!msg->haltOutput()) && KVI_OPTION_BOOL(KviOption_boolShowPingPong))
	{
		msg->console()->output(KVI_OUT_SERVERPING,
			__tr2wc("Received pong from \r!s\r%s\r (PONG %s)"),msg->safePrefix(),msg->allParams());
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// JOIN
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void KviServerParser::parseLiteralJoin(KviIrcMessage *msg)
{
	// JOIN
	// :<joiner_mask> JOIN :<channel>

	KviIrcMask joiner(msg->prefix());

	KviStr * channel = msg->trailingString();

	if(!channel)
	{
		// This is broken....
		UNRECOGNIZED_MESSAGE(msg,__tr("Missing channel parameter in join message"));
		return;
	}

	// check for extended join syntax.
	// it is used in splits only (AFAIK)
	// nick!user@host JOIN :#channel\x07[o|v]
	const char * pExt = channel->ptr();
	char  chExtMode = 0;
	while(*pExt && (*pExt != 0x07))pExt++;
	if(*pExt)
	{
		++pExt;
		if(*pExt)
		{
			chExtMode = (*pExt);
			channel->cutRight(2); // assuming that we're at the end (we should be)
		} else {
			channel->cutRight(1); // senseless 0x07 in channel name (not allowed I guess...)
		}
	}

	// Now lookup the channel
	KviConsole * console = msg->console();
	KviChannel * chan = console->findChannel(channel->ptr());

	bool bIsMe = IS_ME(msg,joiner.nick());

	if(!chan)
	{
		// This must be me...(or desync)
		if(bIsMe)
		{
			console->setUserInfoFromServer(joiner.username(),joiner.host());
			chan = console->createChannel(channel->ptr()); // New channel
		} else {
			// Someone is joining an inexsisting channel!!!
			UNRECOGNIZED_MESSAGE(msg,__tr("Received a join message for an unknown channel, possible desync"));
			return;
		}

		// FIXME: #warning "STATUS BAR TEXT"

		// FIXME: #warning "IF VERBOSE SAY THAT WE'RE REQUESTING MODES & BAN LIST" (Synching channel)

		int iFlags = 0;
		switch(chExtMode)
		{
			case 'o': iFlags = KVI_USERFLAG_OP;     break;
			case 'v': iFlags = KVI_USERFLAG_VOICE;  break;
			case 'h': iFlags = KVI_USERFLAG_HALFOP; break;
			case 'u': iFlags = KVI_USERFLAG_USEROP; break;
			// we ignore +a and +q actually
		}

		KviUserListEntry * it = chan->join(joiner.nick(),joiner.username(),joiner.host(),iFlags);
		if(iFlags)chan->updateCaption();

// FIXME: #warning "Trigger also OnMeVoice and OnMeOp here ?"
		if(!(it->globalData()->avatar()))
		{
			KviAvatar * av = console->defaultAvatarFromOptions();
			if(av)
			{
				it->globalData()->setAvatar(av);
				console->avatarChanged(av,joiner.nick(),joiner.user(),joiner.host(),0);
			}
		}

		if(TRIGGER_EVENT_RETVALUE(KviEvent_OnMeJoin,chan))msg->setHaltOutput();

		// the channel requests must be sent AFTER we have created and accessed the chan
		// since it MAY happen that a sendFmtData() call fails by detecting a disconnect
		// and thus destroys the channel window!
		
		// If this problem persists in other parts of the KVIrc core then
		// we should disable disconnection detection during the parsing of a single
		// message in KviIrcLink. See the comment in KviIrcLink::processData() for more info.

		if(!console->link()->sendFmtData("MODE %s",channel->ptr()))return; // disconnected

		if(console->connectionInfo() && console->connectionInfo()->bServerSupportsModeIe)
		{
			if(!KVI_OPTION_BOOL(KviOption_boolDisableBanExceptionListRequestOnJoin))
			{
				if(!console->link()->sendFmtData("MODE %s e",channel->ptr()))return; // disconnected
				chan->setSentBanExceptionListRequest();
			}
			if(!KVI_OPTION_BOOL(KviOption_boolDisableInviteListRequestOnJoin))
			{
				if(!console->link()->sendFmtData("MODE %s I",channel->ptr()))return; // disconnected
				chan->setSentInviteListRequest();
			}
		}
		// MODE %s b MUST BE THE LAST AUTOMATIC CHANNEL QUERY
		// so we get RPL_ENDOFBANLIST as the last reply
		// and we know that the channel is in sync

		if(!KVI_OPTION_BOOL(KviOption_boolDisableWhoRequestOnJoin))
		{
			if(!console->link()->sendFmtData("WHO %s",channel->ptr()))return; // disconnected
			chan->setSentWhoRequest();
		}
		if(!KVI_OPTION_BOOL(KviOption_boolDisableBanListRequestOnJoin))
		{
			if(!console->link()->sendFmtData("MODE %s b",channel->ptr()))return; // disconnected
			chan->setSentBanListRequest();
		}

	} else {
		// This must be someone else...(or desync)
		int iFlags = 0;
		switch(chExtMode)
		{
			case 'o': iFlags = KVI_USERFLAG_OP;     break;
			case 'v': iFlags = KVI_USERFLAG_VOICE;  break;
			case 'h': iFlags = KVI_USERFLAG_HALFOP; break;
			case 'u': iFlags = KVI_USERFLAG_USEROP; break;
			// we ignore +a and +q actually
		}

		KviUserListEntry * it = chan->join(joiner.nick(),joiner.username(),joiner.host(),iFlags);

		// FIXME: #warning "Trigger also OnVoice and OnOp here ?"
		// Note: checkDefaultAvatar() makes a KviRegisteredUser lookup
		//       if later it is needed, make it return a pointer
		if(!(it->globalData()->avatar()))console->checkDefaultAvatar(it->globalData(),
			joiner.nick(),joiner.user(),joiner.host());

		if(TRIGGER_EVENT_3PARAM_RETVALUE(KviEvent_OnJoin,chan,joiner.nick(),joiner.username(),joiner.host()))
			msg->setHaltOutput();

// FIXME: #warning "WE COULD OPTIONALLY REQUEST A /WHO FOR THE USERS JOINING THAT WE DON'T KNOW THE HOST OF"
	}

	// Now say it to the world
	if(!msg->haltOutput())
	{
// FIXME: #warning "CHECK IF MESSAGES GO TO CONSOLE OR NOT"

		if(chExtMode != 0)
		{
			chan->output(KVI_OUT_JOIN,
				__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has joined \r!c\r%s\r [implicit +%c umode change]"),
				joiner.nick(),joiner.username(),joiner.host(),channel->ptr(),chExtMode);

		} else {
			chan->output(KVI_OUT_JOIN,
				__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has joined \r!c\r%s\r"),
				joiner.nick(),joiner.username(),joiner.host(),channel->ptr());
		}
	}

	if(!bIsMe && KVI_OPTION_BOOL(KviOption_boolEnableQueryTracing))
	{
		KviStr szChans;
		int iChans = console->getCommonChannels(joiner.nick(),szChans);
		for(KviQuery *q=console->queryList()->first();q;q=console->queryList()->next())
		{
			if(q->haveTarget(joiner.nick()))
			{
				q->output(KVI_OUT_QUERYTRACE,
					__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has just joined \r!c\r%s\r"),joiner.nick(),joiner.username(),
					joiner.host(),msg->safeParam(0));
				q->notifyCommonChannels(joiner.nick(),joiner.username(),joiner.host(),iChans,szChans);
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// PART
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void KviServerParser::parseLiteralPart(KviIrcMessage *msg)
{
	// PART
	// :<source_mask> PART <channel> :<part message>

	KviIrcMask source(msg->prefix());
	// Now lookup the channel
	KviConsole * console = msg->console();
	KviChannel * chan = console->findChannel(msg->safeParam(0));

	if(!chan)
	{
		UNRECOGNIZED_MESSAGE(msg,__tr("Received a part message for an unknown channel, possible desync"));
		return;
	}

	// always decode with the textEncoding of the channel

	KviWStr partMsg(KviWStr::Decode,(msg->paramCount() > 1) ? msg->safeTrailingString() : KviStr::emptyString(),chan->textEncoding());

	if(IS_ME(msg,source.nick()))
	{
		// FIXME: the events should be triggered as wide string text ?
		if(TRIGGER_EVENT_1PARAM_RETVALUE(KviEvent_OnMePart,chan,msg->safeTrailing()))
			msg->setHaltOutput();

		// It's me!
		chan->frame()->closeWindow(chan); // <-- deleted path

		if(!msg->haltOutput())
		{
			if(KVI_OPTION_BOOL(KviOption_boolShowOwnParts))
			{
				console->output(KVI_OUT_PART,
					(partMsg.hasData()) ? __tr2wc("You have left channel \r!c\r%s\r: %W") :
					__tr2wc("You have left channel \r!c\r%s\r"),msg->safeParam(0),&partMsg);

			}
		}

	} else {
		// Someone else

		if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnPart,chan,source.nick(),
				source.user(),source.host(),msg->safeTrailing()))
			msg->setHaltOutput();

		chan->part(source.nick());

		if(!msg->haltOutput())
		{
			chan->output(KVI_OUT_PART,
				(partMsg.hasData()) ? __tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has left \r!c\r%s\r: %W") :
				__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has left \r!c\r%s\r"),source.nick(),source.username(),
				source.host(),msg->safeParam(0),&partMsg);
		}

		if(KVI_OPTION_BOOL(KviOption_boolEnableQueryTracing))
		{
			KviStr szChans;
			int iChans = console->getCommonChannels(source.nick(),szChans);
			for(KviQuery *q=console->queryList()->first();q;q=console->queryList()->next())
			{
				if(q->haveTarget(source.nick()))
				{
					q->output(KVI_OUT_QUERYTRACE,
						(partMsg.hasData()) ? __tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has just left \r!c\r%s\r: %W") :
						__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has just left \r!c\r%s\r"),source.nick(),source.username(),
						source.host(),msg->safeParam(0),&partMsg);
					q->notifyCommonChannels(source.nick(),source.username(),source.host(),iChans,szChans);
				}
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// QUIT
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void KviServerParser::parseLiteralQuit(KviIrcMessage *msg)
{
	// QUIT
	// :<source_mask> QUIT :<quit message>

	KviIrcMask source(msg->prefix());
	KviConsole * console = msg->console();

	const char * aux = msg->safeTrailing();
	bool bWasSplit = false;
	//determine if signoff string matches "%.% %.%", and only one space (from eggdrop code)
	char *p = (char *)strchr(aux, ' ');
	if (p && (p == (char *)strrchr(aux,' '))){
		// one space detected. go ahead
		char *z1, *z2;
		*p = 0;
		z1 = (char *)strchr(p + 1, '.');
		z2 = (char *)strchr(aux, '.');
		if (z1 && z2 && (*(z1 + 1) > 47) && (z1 - 1 != p) && (z2 + 1 != p) && (z2 != aux) && console->connectionInfo())
		{
			// server split, or else it looked like it anyway
			*p=' ';
			bWasSplit = true;

			time_t curTime = kvi_unixTime();
			int diff = ((unsigned int)curTime) - ((unsigned int)console->connectionInfo()->lastNetsplitTime);
			bool bDuplicate = false;
			if(diff < 6)
			{
				if(kvi_strEqualCI(console->connectionInfo()->szLastNetsplit.ptr(),aux))
				{
					bDuplicate = true;
				}
			}
			console->connectionInfo()->szLastNetsplit = aux;
			console->connectionInfo()->lastNetsplitTime = curTime;
			if(!bDuplicate && !msg->haltOutput())
			{
// FIXME: #warning "OnNetsplit ?"
				console->output(KVI_OUT_SPLIT,__tr2wc("Netsplit detected: %s"),aux);
			}
		} else *p = ' ';
	}

// FIXME: #warning "Add a netsplit parameter ?"

	if(g_pEventManager->hasEventHandlers(KviEvent_OnQuit))
	{
		// compute the channel list
		KviStr chanlist;
		
		for(KviChannel *daChan=console->channelList()->first();daChan;daChan=console->channelList()->next())
		{
			if(daChan->isOn(source.nick()))
			{
				if(chanlist.isEmpty())chanlist = daChan->name();
				else {
					chanlist.append(',');
					chanlist.append(daChan->name());
				}
			}
		}

		if(g_pUserParser->triggerEvent(KviEvent_OnQuit,console,
					new KviParameterList(new KviStr(source.nick()),new KviStr(source.user()),
					new KviStr(source.host()),new KviStr(msg->safeTrailing()),new KviStr(chanlist))))
				msg->setHaltOutput();
	}


	for(KviChannel *c=console->channelList()->first();c;c=console->channelList()->next())
	{
		if(c->part(source.nick()))
		{
			if(bWasSplit)
			{
				if(!msg->haltOutput())c->output(KVI_OUT_QUITSPLIT,
					__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has quit IRC: NETSPLIT %s"),source.nick(),source.username(),
					source.host(),msg->safeTrailing());
			} else {
				// decode with the right text encoding
				KviWStr quitMsg(KviWStr::Decode,msg->safeTrailingString(),c->textEncoding());

				if(!msg->haltOutput())c->output(KVI_OUT_QUIT,
					__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has quit IRC: %W"),source.nick(),source.username(),
					source.host(),&quitMsg);
			}
		}
	}

	for(KviQuery *q=console->queryList()->first();q;q=console->queryList()->next())
	{
		if(q->haveTarget(source.nick()))
		{
			if(!msg->haltOutput())
			{
				if(bWasSplit)
				{
					q->output(KVI_OUT_QUITSPLIT,
						__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has quit IRC: NETSPLIT %s"),source.nick(),source.username(),
						source.host(),msg->safeTrailing());
				} else {
					KviWStr quitMsg(KviWStr::Decode,msg->safeTrailingString(),q->textEncoding());

					q->output(KVI_OUT_QUIT,
						__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has quit IRC: %W"),source.nick(),source.username(),
						source.host(),&quitMsg);
				}

				if(q->targetCount() > 1)
				{
					q->removeTarget(source.nick());
					q->output(KVI_OUT_SYSTEMWARNING,__tr2wc("\r!n\r%s\r, a target of this query, has just quit IRC and " \
							"has been removed from the target list. " \
							"Use \r![!dbl]addtarget %s[!txt]Double click here to re-add %s to the target list" \
							"\r/addtarget %s\r to add him again"),
						source.nick(),source.nick(),source.nick(),source.nick());
				}
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// KICK
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void KviServerParser::parseLiteralKick(KviIrcMessage *msg)
{
	// KICK
	// :<source_mask> KICK <channel> <nick> :<kick message>
	KviIrcMask source(msg->prefix());

//	KviStr nickBuffer;
//	KviStr hostBuffer;
//	formatNickLink(source.nick(),nickBuffer);
//	formatHostLink(source.host(),hostBuffer);

	KviStr szChan = msg->safeParam(0);
	KviStr szNick = msg->safeParam(1);

	KviConsole * console = msg->console();
	KviChannel * chan = console->findChannel(szChan.ptr());

	if(!chan){
		// Ooops , desync with the server.
		UNRECOGNIZED_MESSAGE(msg,__tr("Received a kick message for an unknown channel, possible desync"));
		return;
	}

	KviWStr szKickMsg(KviWStr::Decode,msg->safeTrailingString(),chan->textEncoding());

	if(IS_ME(msg,szNick.ptr()))
	{
		// ops...I have been kicked

		if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnMeKick,chan,
				source.nick(),source.user(),source.host(),msg->safeTrailing()))
			msg->setHaltOutput();

		KviStr szPass = chan->channelKey();

		if(KVI_OPTION_BOOL(KviOption_boolKeepChannelOpenOnKick))
		{
			chan->userAction(&source,KVI_USERACTION_KICK);

			chan->part(szNick.ptr());
			chan->setDeadChan();

			if(!msg->haltOutput())
			{
// FIXME: #warning "OPTION FOR THIS TO GO TO THE CONSOLE!"
				chan->output(KVI_OUT_MEKICK,
					__tr2wc("You have been kicked from \r!c\r%S\r by \r!n\r%s\r [%s@\r!h\r%s\r]: %W"),
					&szChan,source.nick(),source.username(),source.host(),&szKickMsg);
			}
		} else {
			chan->frame()->closeWindow(chan); // <-- deleted path

			if(!msg->haltOutput())
			{
// FIXME: #warning "This could go also to the active window!"
				console->output(KVI_OUT_MEKICK,
					__tr2wc("You have been kicked from \r!c\r%S\r by \r!n\r%s\r [%s@\r!h\r%s\r]: %W"),
					&szChan,source.nick(),source.username(),source.host(),&szKickMsg);
			}
		}
		if(KVI_OPTION_BOOL(KviOption_boolRejoinChannelOnKick))
		{
			if(KVI_OPTION_BOOL(KviOption_boolBeVerbose))
			{
// FIXME: #warning "This should go where the 'You have been kicked...' message was printed"
				console->output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("Attempting to rejoin \r!c\r%S\r..."),&szChan);
			}
			if(szPass.hasData())console->link()->sendFmtData("JOIN %s %s",szChan.ptr(),szPass.ptr());
			else console->link()->sendFmtData("JOIN %s",szChan.ptr());
		}
	} else {
		// ok...someone else has been kicked

		if(TRIGGER_EVENT_5PARAM_RETVALUE(KviEvent_OnKick,chan,
				source.nick(),source.user(),source.host(),szNick.ptr(),msg->safeTrailing()))
			msg->setHaltOutput();

		KviIrcUserEntry * e = console->userDataBase()->find(szNick.ptr());

		KviStr szHost;
		KviStr szUser;

		if(e)
		{
			szHost = e->host();
			szUser = e->user();
		} else {
			szHost = "*";
			szUser = "*";
		}


		chan->userAction(&source,KVI_USERACTION_KICK);

		chan->part(szNick.ptr());



		if(!msg->haltOutput())
		{
// FIXME: #warning "OPTION FOR THIS TO GO TO THE CONSOLE!"
			chan->output(KVI_OUT_KICK,
				__tr2wc("\r!n\r%S\r [%S@\r!h\r%S\r] has been kicked from \r!c\r%S\r by \r!n\r%s\r [%s@\r!h\r%s\r]: %W"),
				&szNick,&szUser,&szHost,&szChan,
				source.nick(),source.username(),source.host(),&szKickMsg);
		}

		if(KVI_OPTION_BOOL(KviOption_boolEnableQueryTracing))
		{
			KviStr szChans;
			int iChans = console->getCommonChannels(szNick.ptr(),szChans);
			for(KviQuery *q=console->queryList()->first();q;q=console->queryList()->next())
			{
				if(q->haveTarget(szNick.ptr()))
				{
					q->output(KVI_OUT_QUERYTRACE,
						__tr2wc("\r!n\r%S\r [%S@\r!h\r%S\r] has just been kicked from \r!c\r%S\r by \r!n\r%s\r [%s@\r!h\r%s\r]: %W"),
						&szNick,&szUser,&szHost,&szChan,
						source.nick(),source.username(),source.host(),&szKickMsg);
					q->notifyCommonChannels(szNick.ptr(),szUser.ptr(),szHost.ptr(),iChans,szChans);
				}
			}
		}
	}
}

#ifdef COMPILE_CRYPT_SUPPORT

// FIXME: 	#warning "SIMPLIFY THIS STUFF"

	#define OUTPUT_PRIVMSGORNOTICE(_cons,_target,_type,_type2,_talker,_txt) \
		if(KviCryptSessionInfo * cinf = _target->cryptSessionInfo()) \
		{ \
			if(cinf->bDoDecrypt) \
			{ \
				if(cinf->pEngine->isCryptographicEngine() && (*_txt == KVI_TEXT_CRYPT)) \
				{ \
					KviStr decryptedStuff; \
					if(cinf->pEngine->decrypt(_txt + 1,decryptedStuff)) \
					{ \
						/*_cons->outputPrivmsg(_target,_type,_talker.nick(),_talker.username(),_talker.host(),_txt + 1);*/ \
						_cons->outputPrivmsg(_target,_type2,_talker.nick(),_talker.username(),_talker.host(),decryptedStuff.ptr()); \
					} else { \
						_target->output(KVI_OUT_SYSTEMERROR, \
							__tr2wc("The following message appears to be encrypted, but the crypto engine failed to decode it: %s"), \
							cinf->pEngine->lastError()); \
						_cons->outputPrivmsg(_target,_type,_talker.nick(),_talker.username(),_talker.host(),_txt + 1); \
					} \
				} else { \
					if(cinf->pEngine->isCryptographicEngine()) \
					{ \
						_cons->outputPrivmsg(_target,_type,_talker.nick(),_talker.username(),_talker.host(),_txt); \
					} else { \
						KviStr decryptedStuff; \
						if(cinf->pEngine->decrypt(_txt,decryptedStuff)) \
						{ \
							/*_cons->outputPrivmsg(_target,_type,_talker.nick(),_talker.username(),_talker.host(),_txt + 1);*/ \
							_cons->outputPrivmsg(_target,_type,_talker.nick(),_talker.username(),_talker.host(),decryptedStuff.ptr()); \
						} else { \
							_target->output(KVI_OUT_SYSTEMERROR, \
								__tr2wc("The following message appears to be encrypted, but the crypto engine failed to decode it: %s"), \
								cinf->pEngine->lastError()); \
							_cons->outputPrivmsg(_target,_type,_talker.nick(),_talker.username(),_talker.host(),_txt); \
						} \
					} \
				} \
			} else _cons->outputPrivmsg(_target,_type,_talker.nick(),_talker.username(),_talker.host(),_txt); \
		} else _cons->outputPrivmsg(_target,_type,_talker.nick(),_talker.username(),_talker.host(),_txt);

#else

	#define OUTPUT_PRIVMSGORNOTICE(_cons,_target,_type,_type2,_talker,_txt) \
		_cons->outputPrivmsg(_target,_type,_talker.nick(),_talker.username(),_talker.host(),_txt);

#endif


void KviServerParser::parseLiteralPrivmsg(KviIrcMessage *msg)
{
	// PRIVMSG
	// :source PRIVMSG <target> :<message>

	KviIrcMask talker(msg->safePrefix());
	KviConsole * console = msg->console();
    KviRegisteredUser * u = g_pRegisteredUserDataBase->findMatchingUser(talker.nick(),talker.user(),talker.host());
	// Ignore it?
    if(u)
    {
        if(KVI_OPTION_BOOL(KviOption_boolEnableIgnoreOnPrivMsg))
        {
            if(u->getBoolProperty("ignore"))
            {
                if(KVI_OPTION_BOOL(KviOption_boolVerboseIgnore))
                    console->output(KVI_OUT_IGNORE,__tr2wc("Ignoring PRIVMSG from %s: %s"),msg->safePrefix(),msg->safeTrailing());
                return;
            }
        }
    }

    //Highlight it?
    
// FIXME: 	#warning "Stats for PRIVMSG traffic on a channel"

// FIXME: 	#warning "Highlighting ?"
// FIXME: 	#warning "DEDICATED CTCP WINDOW ?"
	KviStr * pTrailing = msg->trailingString();
	if(pTrailing)
	{
		if(*(pTrailing->ptr()) == 0x01){
			if(pTrailing->len() > 1)
			{
				if(pTrailing->lastCharIs(0x01))pTrailing->cutRight(1);
				pTrailing->cutLeft(1);
				KviCtcpMessage ctcp;
				ctcp.msg = msg;
				ctcp.pData = pTrailing->ptr();
				ctcp.pSource = &talker;
				ctcp.pTarget = msg->safeParam(0);
				ctcp.bIgnored = false;
				ctcp.bIsFlood = false;
				ctcp.bUnknown = false;
				parseCtcpRequest(&ctcp);
				return;
			}
		}
	}

	// Normal PRIVMSG
	if(IS_ME(msg,msg->safeParam(0)))
	{
		// FIXME: 	#warning "PROCESS MULTIMEDIA FILE REQUESTS"

		//		if(g_pOptions->m_bListenToMultimediaFileRequests)
		//		{
		//			if(*aux == '!')
		//			{
		//				const char *tmp = aux;
		//				tmp++;
		//				if(kvi_strEqualCI(tmp,m_pFrm->m_global.szCurrentNick.ptr()))
		//				{
		//					// A multimedia file request ?
		//					tmp += m_pFrm->m_global.szCurrentNick.len();
		//					if(((*tmp) == ' ') || ((*tmp) == '\t'))
		//					{
		//						while(((*tmp) == ' ') || ((*tmp) == '\t'))tmp++;
		//						if(*tmp)
		//						{
		//							KviStr file = tmp;
		//							KviStr filePath;
		//							m_pFrm->findMultimediaFileOffert(filePath,file);
		//							if(filePath.hasData())
		//							{
		//								m_pFrm->activeWindow()->output(KVI_OUT_INTERNAL,__tr("%s requests previously offered file %s: sending (%s)"),talker.nick(),file.ptr(),filePath.ptr());
		//								KviStr cmd(KviStr::Format,"DCC SEND %s %s",talker.nick(),filePath.ptr());
		//								m_pFrm->m_pUserParser->parseUserCommand(cmd,m_pConsole);
		//								return;
	
		//							} else {
		//								m_pFrm->activeWindow()->output(KVI_OUT_INTERNAL,__tr("%s requests file %s: no such file was offered , ignoring"),talker.nick(),file.ptr());
		//								return;
		//							}
		//						}
		//					}
		//				}
		//			}
		//		}
				
		// A query request
		// do we have a matching window ?
		KviQuery * query = console->findQuery(talker.nick());
		
		if(!query)
		{
			// New query requested. Check if we really should create it or not

			// first of all the anti spam , if desired.
			// the antispam blocks anything else
			// Eventually we could trigger a special event to notify the user of the
			// spam message...
			if(KVI_OPTION_BOOL(KviOption_boolUseAntiSpamOnPrivmsg))
			{
				KviStr * theMsg = msg->trailingString();
				if(theMsg)
				{
					KviStr spamWord;
					if(kvi_mayBeSpam(theMsg,spamWord))
					{
						// FIXME: OnSpam ?

						if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolSilentAntiSpam)))
						{
							console->output(KVI_OUT_SPAM,__tr2wc("Spam privmsg from \r!n\r%s\r [%s@\r!h\r%s\r]: %s (matching spamword \"%s\")"),
								talker.nick(),talker.user(),talker.host(),msg->safeTrailing(),spamWord.ptr());
						}
						return;
					}
				}
			}

			// this is not a spam, or at least it hasn't been recognized as spam
			
			// user option ? (this should again override any script)
			// if the scripters want really to force the query creation they can do
			// it manually or they can set the option to true at KVIrc startup
			if(KVI_OPTION_BOOL(KviOption_boolCreateQueryOnPrivmsg))
			{
				// We still want to create it
				// Give the scripter a chance to filter it out again
				if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnQueryWindowRequest,console,talker.nick(),talker.user(),talker.host(),msg->safeTrailing()))
				{
					// check if the scripter hasn't created it
					query = console->findQuery(talker.nick());
				} else {
					// no query yet, create it!
					// this will trigger OnQueryWindowCreated
					query = console->createQuery(talker.nick());
					// and this will trigger OnQueryTargetAdded
					query->addTarget(talker.nick(),talker.username(),talker.host());
				}
			}
		}

		// ok, now we either have a query or not
		if(query)
		{
			// ok, we have the query. Trigger the user action anyway
			query->userAction(&talker,KVI_USERACTION_PRIVMSG);
			// trigger the script event and eventually kill the output
			if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnQueryMessage,query,talker.nick(),talker.user(),talker.host(),msg->safeTrailing())) 
				msg->setHaltOutput();
			// spit out the message text
			if(!msg->haltOutput())
			{
				OUTPUT_PRIVMSGORNOTICE(console,query,KVI_OUT_QUERYPRIVMSG,KVI_OUT_QUERYPRIVMSGCRYPTED,talker,msg->safeTrailing())
			}
		} else {
			// no query creation
			// trigger the query message event in the console
			if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnQueryMessage,console,talker.nick(),talker.user(),talker.host(),msg->safeTrailing()))
				msg->setHaltOutput();
			// spit the message text out
			if(!msg->haltOutput())
			{
				KviWindow * pOut = KVI_OPTION_BOOL(KviOption_boolExternalMessagesToActiveWindow) ?
					console->activeWindow() : (KviWindow *)(console);

				if(KviConnectionInfo * pConnection = console->connectionInfo())
				{
					KviWindow * aWin = console->activeWindow();
					if((aWin->type() == KVI_WINDOW_TYPE_CHANNEL) && ((KviChannel *)aWin)->isOn(talker.nick()))
						pOut = aWin;
					else {
						for(KviChannel * c = pConnection->pChannelList->first();c;c = pConnection->pChannelList->next())
							if(c->isOn(talker.nick()))
							{
								pOut = (KviWindow *) c;
								break;
							}
					}
				}

				KviWStr szMessage(KviWStr::Decode,msg->safeTrailingString(),pOut->textEncoding());
				pOut->output(KVI_OUT_QUERYPRIVMSG,"[PRIVMSG \r!n\r%s\r]: %W",talker.nick(),&szMessage);
			}
		}
	} else {
		// Channel PRIVMSG
		KviChannel * chan = console->findChannel(msg->safeParam(0));

		if(!chan)
		{
			if(!msg->haltOutput())
			{
				KviWindow * pOut = KVI_OPTION_BOOL(KviOption_boolOperatorMessagesToActiveWindow) ? 
					console->activeWindow() : (KviWindow *)(console);
				KviStr broad(KviStr::Format,"[>> %s] %s",msg->safeParam(0),msg->safeTrailing());
				OUTPUT_PRIVMSGORNOTICE(console,pOut,KVI_OUT_BROADCASTPRIVMSG,KVI_OUT_BROADCASTPRIVMSG,talker,broad.ptr())
			}
		} else {
			chan->userAction(&talker,KVI_USERACTION_PRIVMSG);

			if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnChannelMessage,chan,talker.nick(),talker.user(),talker.host(),msg->safeTrailing()))
				msg->setHaltOutput();

			if(!msg->haltOutput())
			{
				OUTPUT_PRIVMSGORNOTICE(console,chan,KVI_OUT_CHANPRIVMSG,KVI_OUT_CHANPRIVMSGCRYPTED,talker,msg->safeTrailing())
			}
		}
	}
}


void KviServerParser::parseLiteralNotice(KviIrcMessage *msg)
{
	// NOTICE
	// :source NOTICE <target> :<message>

	KviIrcMask talker(msg->safePrefix());
	KviConsole * console = msg->console();

	if( *(talker.host()) == '*' )
	{
		if( *(talker.user()) == '*')
		{
			KviStr szServ = talker.nick();
			if(szServ.contains('.'))
			{
				// server notice
// FIXME: 		#warning "Dedicated window for server notices ?"
				if(TRIGGER_EVENT_2PARAM_RETVALUE(KviEvent_OnServerNotice, console, talker.nick(), msg->safeTrailing()))msg->setHaltOutput();
				if(!msg->haltOutput())
				{
					KviWindow * pOut = KVI_OPTION_BOOL(KviOption_boolServerNoticesToActiveWindow) ?
						console->activeWindow() : (KviWindow *)(console);
					KviWStr szMessage(KviWStr::Decode,msg->safeTrailingString(),console->textEncoding());
					pOut->output(KVI_OUT_SERVERNOTICE,"[\r!s\r%s\r]: %W",talker.nick(),&szMessage);
				}
				return;
			}
		}
	}

	// Ignore it?
	if(KVI_OPTION_BOOL(KviOption_boolEnableIgnoreOnNotice))
	{
		KviRegisteredUser * u = g_pRegisteredUserDataBase->findMatchingUser(talker.nick(),talker.user(),talker.host());
		if(u)
		{
			if(u->getBoolProperty("ignore"))
			{
				if(KVI_OPTION_BOOL(KviOption_boolVerboseIgnore))
					console->output(KVI_OUT_IGNORE,__tr2wc("Ignoring PRIVMSG from %s (%s)"), msg->safePrefix(), msg->safeTrailing());
				return;
			}
		}
	}

// FIXME: 	#warning "Highlighting ?"
// FIXME: 	#warning "DEDICATED CTCP WINDOW ?"

	KviStr * pTrailing = msg->trailingString();
	if(pTrailing)
	{
		if(*(pTrailing->ptr()) == 0x01){
			if(pTrailing->len() > 1)
			{
				if(pTrailing->lastCharIs(0x01))pTrailing->cutRight(1);
				pTrailing->cutLeft(1);
				KviCtcpMessage ctcp;
				ctcp.msg = msg;
				ctcp.pData = pTrailing->ptr();
				ctcp.pSource = &talker;
				ctcp.pTarget = msg->safeParam(0);
				ctcp.bIgnored = false;
				ctcp.bIsFlood = false;
				ctcp.bUnknown = false;
				parseCtcpReply(&ctcp);
				return;
			}
		}
	}


	// Normal NOTICE
	if(IS_ME(msg,msg->safeParam(0)))
	{

// FIXME: 	#warning "The NickServ and ChanServ handling should be optional!"

		if(kvi_strEqualCI(talker.nick(),"NickServ"))
		{
			if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnNickServNotice,console,talker.nick(),talker.user(),talker.host(),msg->safeTrailing()))
				msg->setHaltOutput();

			// nickname service... does it ask for identification ?
			if(!msg->haltOutput())
			{
				KviWindow * pOut = KVI_OPTION_BOOL(KviOption_boolServicesNoticesToActiveWindow) ?
					console->activeWindow() : (KviWindow *)(console);
				KviWStr szMessage(KviWStr::Decode,msg->safeTrailingString(),console->textEncoding());
				pOut->output(KVI_OUT_NICKSERV,"\r!n\r%s\r [%s@\r!h\r%s\r]: %W",talker.nick(),talker.user(),talker.host(),&szMessage);
			}

			if(KVI_OPTION_BOOL(KviOption_boolEnableAutomaticNickServIdentification))
			{
				KviStr szData = msg->safeTrailing();

				int msgIdx = szData.findFirstIdx("/msg",false);
				if(msgIdx > 0)
				{
					szData.cutLeft(msgIdx);
					msgIdx = 0;
				} else {
					if(msgIdx == -1)
					{
						// irc.dynamix.com uses /NickServ
						msgIdx = szData.findFirstIdx("/NickServ",false);
						if(msgIdx > 0)
						{
							szData.cutLeft(msgIdx);
							msgIdx = 0;
						}
					}
				}
				int nickServIdx = szData.findFirstIdx("NickServ",false);
				int identifyIdx = szData.findFirstIdx("identify",false);
				int passwordIdx = szData.findFirstIdx("password",false);

				if((identifyIdx != -1) && (nickServIdx != -1) && (msgIdx != -1) && (passwordIdx != -1)
					&& (msgIdx < nickServIdx) && (nickServIdx < identifyIdx) && (identifyIdx < passwordIdx))
				{
					// yep..this is almost surely an identification request
					KviStr szIdentifyCommand = g_pNickServDataBase->findIdentifyCommand(console->currentNickName(),console->currentServerName(),&talker);
					if(szIdentifyCommand.hasData())
					{
						// execute it!
						console->output(KVI_OUT_SYSTEMMESSAGE,__tr2wc("NickServ requests authentication, executing scheduled command"));
						if(!g_pUserParser->parseCommandBuffer(szIdentifyCommand.ptr(),console))
						{
							console->output(KVI_OUT_SYSTEMERROR,__tr2wc("The scheduled NickServ identification command appears to be broken, please change the setting"));
						}
					}
				}
			}

			return;
		}

		if(kvi_strEqualCI(talker.nick(),"ChanServ"))
		{
			if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnChanServNotice,console,talker.nick(),talker.user(),talker.host(),msg->safeTrailing()))
				msg->setHaltOutput();
			if(!msg->haltOutput())
			{
				KviWindow * pOut = KVI_OPTION_BOOL(KviOption_boolServicesNoticesToActiveWindow) ?
					console->activeWindow() : (KviWindow *)(console);
				KviWStr szMessage(KviWStr::Decode,msg->safeTrailingString(),console->textEncoding());
				pOut->output(KVI_OUT_CHANSERV,"\r!n\r%s\r [%s@\r!h\r%s\r]: %W",talker.nick(),talker.user(),talker.host(),&szMessage);
			}
			return;
		}

// FIXME: 	#warning "PROCESS MULTIMEDIA FILE REQUESTS"

	//		if(g_pOptions->m_bListenToMultimediaFileRequests)
	//		{
	//			if(*aux == '!')
	//			{
	//				const char *tmp = aux;
	//				tmp++;
	//				if(kvi_strEqualCI(tmp,m_pFrm->m_global.szCurrentNick.ptr()))
	//				{
	//					// A multimedia file request ?
	//					tmp += m_pFrm->m_global.szCurrentNick.len();
	//					if(((*tmp) == ' ') || ((*tmp) == '\t'))
	//					{
	//						while(((*tmp) == ' ') || ((*tmp) == '\t'))tmp++;
	//						if(*tmp)
	//						{
	//							KviStr file = tmp;
	//							KviStr filePath;
	//							m_pFrm->findMultimediaFileOffert(filePath,file);
	//							if(filePath.hasData())
	//							{
	//								m_pFrm->activeWindow()->output(KVI_OUT_INTERNAL,__tr("%s requests previously offered file %s: sending (%s)"),talker.nick(),file.ptr(),filePath.ptr());
	//								KviStr cmd(KviStr::Format,"DCC SEND %s %s",talker.nick(),filePath.ptr());
	//								m_pFrm->m_pUserParser->parseUserCommand(cmd,m_pConsole);
	//								return;

	//							} else {
	//								m_pFrm->activeWindow()->output(KVI_OUT_INTERNAL,__tr("%s requests file %s: no such file was offered , ignoring"),talker.nick(),file.ptr());
	//								return;
	//							}
	//						}
	//					}
	//				}
	//			}
	//		}

		// A query request
		// do we have a matching window ?
		KviQuery * query = console->findQuery(talker.nick());
		
		if(!query)
		{
			// New query requested. Check if we really should create it or not

			// first of all the anti spam , if desired.
			// the antispam blocks anything else
			// Eventually we could trigger a special event to notify the user of the
			// spam message...
			if(KVI_OPTION_BOOL(KviOption_boolUseAntiSpamOnNotice))
			{
				KviStr * theMsg = msg->trailingString();
				if(theMsg)
				{
					KviStr spamWord;
					if(kvi_mayBeSpam(theMsg,spamWord))
					{
						// FIXME: OnSpam ?

						if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolSilentAntiSpam)))
						{
							console->output(KVI_OUT_SPAM,__tr2wc("Spam notice from \r!n\r%s\r [%s@\r!h\r%s\r]: %s (matching spamword \"%s\")"),
								talker.nick(),talker.user(),talker.host(),msg->safeTrailing(),spamWord.ptr());
						}
						return;
					}
				}
			}

			// this is not a spam, or at least it hasn't been recognized as spam
			
			// user option ? (this should again override any script)
			// if the scripters want really to force the query creation they can do
			// it manually or they can set the option to true at KVIrc startup
			if(KVI_OPTION_BOOL(KviOption_boolCreateQueryOnNotice))
			{
				// We still want to create it
				// Give the scripter a chance to filter it out again
				if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnQueryWindowRequest,console,talker.nick(),talker.user(),talker.host(),msg->safeTrailing()))
				{
					// check if the scripter hasn't created it
					query = console->findQuery(talker.nick());
				} else {
					// no query yet, create it!
					// this will trigger OnQueryWindowCreated
					query = console->createQuery(talker.nick());
					// and this will trigger OnQueryTargetAdded
					query->addTarget(talker.nick(),talker.username(),talker.host());
				}
			}
		}

		// ok, now we either have a query or not
		if(query)
		{
			// ok, we have the query. Trigger the user action anyway
			query->userAction(&talker,KVI_USERACTION_NOTICE);
			// trigger the script event and eventually kill the output
			if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnQueryNotice,query,talker.nick(),talker.user(),talker.host(),msg->safeTrailing())) 
				msg->setHaltOutput();
			// spit out the message text
			if(!msg->haltOutput())
			{
				OUTPUT_PRIVMSGORNOTICE(console,query,KVI_OUT_QUERYNOTICE,KVI_OUT_QUERYNOTICECRYPTED,talker,msg->safeTrailing())
			}
		} else {
			// no query creation
			// trigger the query message event in the console
			if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnQueryNotice,console,talker.nick(),talker.user(),talker.host(),msg->safeTrailing()))
				msg->setHaltOutput();
			// spit the message text out
			if(!msg->haltOutput())
			{
				KviWindow * pOut = KVI_OPTION_BOOL(KviOption_boolExternalMessagesToActiveWindow) ?
					console->activeWindow() : (KviWindow *)(console);

				if(KviConnectionInfo * pConnection = console->connectionInfo())
				{
					KviWindow * aWin = console->activeWindow();
					if((aWin->type() == KVI_WINDOW_TYPE_CHANNEL) && ((KviChannel *)aWin)->isOn(talker.nick()))
						pOut = aWin;
					else {
						for(KviChannel * c = pConnection->pChannelList->first();c;c = pConnection->pChannelList->next())
							if(c->isOn(talker.nick()))
							{
								pOut = (KviWindow *) c;
								break;
							}
					}
				}

				KviWStr szMessage(KviWStr::Decode,msg->safeTrailingString(),pOut->textEncoding());
				pOut->output(KVI_OUT_QUERYNOTICE,"[NOTICE \r!n\r%s\r]: %W",talker.nick(),&szMessage);
			}
		}
		return;
	}


	// Channel PRIVMSG
	char flaggedNotice = 0;
	const char * aux = msg->safeParam(0);
	if(*aux == '@' || *aux == '+')
	{
		flaggedNotice = *aux;
		++aux;
	}
	KviChannel * chan = console->findChannel(aux);

	if(!chan)
	{
		if(!msg->haltOutput())
		{
			KviWindow * pOut = KVI_OPTION_BOOL(KviOption_boolOperatorMessagesToActiveWindow) ? 
				console->activeWindow() : (KviWindow *)(console);
			KviStr broad(KviStr::Format,"[>> %s] %s",msg->safeParam(0),msg->safeTrailing());
			OUTPUT_PRIVMSGORNOTICE(console,pOut,KVI_OUT_BROADCASTNOTICE,KVI_OUT_BROADCASTNOTICE,talker,broad.ptr())
			return;
		}
	} else {
		if(TRIGGER_EVENT_3PARAM_RETVALUE(KviEvent_OnChannelNotice,chan,talker.nick(),msg->safeTrailing(),msg->safeParam(0)))msg->setHaltOutput();
		chan->userAction(&talker,KVI_USERACTION_NOTICE);

		if(!msg->haltOutput())
		{
			KviStr broad;
			if(flaggedNotice)
				broad = KviStr(KviStr::Format,"[>> %c\r!c\r%s\r] %s",flaggedNotice,aux,msg->safeTrailing());
			else
				broad = msg->safeTrailing();

			OUTPUT_PRIVMSGORNOTICE(console,chan,KVI_OUT_CHANNELNOTICE,KVI_OUT_CHANNELNOTICECRYPTED,talker,broad.ptr())
		}
	}
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// TOPIC
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void KviServerParser::parseLiteralTopic(KviIrcMessage *msg)
{
	// TOPIC
	// :<source_mask> TOPIC <channel> :<topic>

	KviIrcMask source(msg->prefix());
	// Now lookup the channel
	KviChannel * chan = msg->console()->findChannel(msg->safeParam(0));

	if(!chan)
	{
		UNRECOGNIZED_MESSAGE(msg,__tr2wc("Received a topic message for an unknown channel, possible desync"));
		return;
	}

	if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnTopic,chan,
		source.nick(),source.user(),source.host(),msg->safeTrailing()))
			msg->setHaltOutput();

	KviWStr szTopic(KviWStr::Decode,msg->safeTrailingString(),chan->textEncoding());

	chan->topicWidget()->setTopic(szTopic.qstring());
	chan->topicWidget()->setTopicSetBy(source.nick());
	KviStr tmp = QDateTime::currentDateTime().toString();
	chan->topicWidget()->setTopicSetAt(tmp.ptr());

	chan->userAction(&source,KVI_USERACTION_TOPIC);

	if(!msg->haltOutput())
	{
		chan->output(KVI_OUT_TOPIC,
			__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] has changed topic to \"%W%c\""),
			source.nick(),source.username(),source.host(),
			&szTopic,KVI_TEXT_RESET);
	}
}

void KviServerParser::parseLiteralNick(KviIrcMessage *msg)
{
	// NICK
	// :source NICK <newnick>
	KviIrcMask source(msg->safePrefix());
	KviConsole * console = msg->console();

	bool bIsMe = IS_ME(msg,source.nick());

	if(bIsMe)
	{
		// We have changed our nick
		console->nickChange(msg->safeTrailing());

		if(TRIGGER_EVENT_2PARAM_RETVALUE(KviEvent_OnMeNickChange,console,source.nick(),msg->safeTrailing()))
				msg->setHaltOutput();
	} else {
		if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnNickChange,console,source.nick(),
				source.user(),source.host(),msg->safeTrailing()))
				msg->setHaltOutput();
	}

	for(KviChannel * c = console->channelList()->first();c;c = console->channelList()->next())
	{
		if(c->nickChange(source.nick(),msg->safeTrailing()))
		{
			if(!msg->haltOutput())
				c->output(KVI_OUT_NICK,__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] is now known as \r!n\r%s\r"),
					source.nick(),source.username(),source.host(),msg->safeTrailing());
		}
		if(bIsMe)c->updateCaption();
	}

// FIXME: #warning "NEW NICK MIGHT BE REGISTERED AND HAVE AN AVATAR!"

	for(KviQuery * q = console->queryList()->first();q;q = console->queryList()->next())
	{
		// some target may have changed the nick
		if(q->nickChange(source.nick(),msg->safeTrailing()))
		{
			if(!msg->haltOutput())
				q->output(KVI_OUT_NICK,__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] is now known as \r!n\r%s\r"),
					source.nick(),source.username(),source.host(),msg->safeTrailing());
		}
		q->updateCaption();
	}

// FIXME: #warning "UPDATE ALL THE OTHER CONNECTION RELATED WINDOW CAPTIONS WHEN bIsMe!!"
}

void KviServerParser::parseLiteralInvite(KviIrcMessage *msg)
{
	// INVITE
	// :source INVITE <target> <channel>

	KviIrcMask source(msg->safePrefix());

	KviStr target(msg->safeParam(0));
	KviStr channel(msg->safeParam(1));

	if(IS_ME(msg,target.ptr()))
	{
		if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnInvite,msg->console(),source.nick(),source.username(),source.host(),channel.ptr()))
			msg->setHaltOutput();
		
		if(!msg->haltOutput())
		{
			KviWindow * pOut = KVI_OPTION_BOOL(KviOption_boolInvitesToActiveWindow) ?
				msg->console()->activeWindow() : (KviWindow *)(msg->console());
			pOut->output(KVI_OUT_INVITE,__tr2wc("\r!n\r%s\r [%s@\r!h\r%s\r] invites you to channel \r!c\r%s\r (%w)"),
				source.nick(),source.user(),source.host(),
				channel.ptr(),KVI_OPTION_BOOL(KviOption_boolAutoJoinOnInvite) ? __tr2wc("autojoining") : __tr2wc("double-click the channel name to join"));
		}

		if(KVI_OPTION_BOOL(KviOption_boolAutoJoinOnInvite))
		{
			msg->console()->link()->sendFmtData("JOIN %s",channel.ptr());
		}

	} else {
		UNRECOGNIZED_MESSAGE(msg,__tr("Received an invite message directed to another nick, possible desync"));
	}
}

void KviServerParser::parseLiteralWallops(KviIrcMessage *msg)
{
	// WALLOPS
	// :source WALLOPS :msg

	KviIrcMask source(msg->safePrefix());

//#warning " FIXME: Event for WALLOPS"
//	if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnInvite,pOut,source.nick(),source.username(),source.host(),channel.ptr()))
//			msg->setHaltOutput();

	if(!msg->haltOutput())
	{
		KviWindow * pOut = KVI_OPTION_BOOL(KviOption_boolOperatorMessagesToActiveWindow) ?
			msg->console()->activeWindow() : (KviWindow *)(msg->console());
		KviWStr szMessage(KviWStr::Decode,msg->safeTrailingString(),pOut->textEncoding());
		pOut->output(KVI_OUT_WALLOPS,__tr2wc("WALLOPS from \r!n\r%s\r [%s@\r!h\r%s\r]: %W"),
			source.nick(),source.username(),source.host(),&szMessage);
	}
}

void KviServerParser::parseUserMode(KviIrcMessage *msg,const char * modeflptr)
{
	// changed my user mode
	bool bSet = true;
	while(*modeflptr)
	{
		switch(*modeflptr)
		{
			case '+': bSet = true; break;
			case '-': bSet = false; break;
			default:
				if(msg->console()->setUserMode(*modeflptr,bSet))
				{
					// There was a mode change
					if(g_pEventManager->hasEventHandlers(KviEvent_OnUserMode))
					{
						KviStr * modefl = new KviStr(KviStr::Format,"%c%c",bSet ? '+' : '-',*modeflptr);
						if(g_pUserParser->triggerEvent(KviEvent_OnUserMode,msg->console(),new KviParameterList(modefl)))
							msg->setHaltOutput();
					}
				}
			break;
		}
		++modeflptr;
	}
}



void KviServerParser::parseLiteralMode(KviIrcMessage *msg)
{
	// NICK
	// :source MODE target <params>
	//    :source MODE <me> +|-modeflag
	//    :source MODE <channel> +-modeflags [parameters]

	KviIrcMask source(msg->safePrefix());

//	if(!source.hasHost())
//	{
//		// This is a server or a channel service
//		KviStr snick = source.nick();
//		if(snick.contains('.'))source.setHost(source.nick()); // this is a server
//	}

	KviStr     target(msg->safeParam(0));
	KviStr     modefl(msg->safeParam(1));

	if(IS_ME(msg,target.ptr()))
	{
		parseUserMode(msg,modefl.ptr());
		if(!msg->haltOutput())
			msg->console()->output(KVI_OUT_MODE,__tr2wc("You have set user mode %s"),modefl.ptr());
	} else {
		// a channel mode
		KviChannel * chan = msg->console()->findChannel(target.ptr());

		if(!chan){
			// Ooops , desync with the server.
			UNRECOGNIZED_MESSAGE(msg,__tr("Received a mode change for an unknown channel, possible desync"));
			return;
		}

		chan->userAction(&source,KVI_USERACTION_CHANMODE);

		parseChannelMode(source,chan,modefl,msg,2);
	}
}

void KviServerParser::parseChannelMode(KviIrcMask &source,KviChannel * chan,KviStr &modefl,KviIrcMessage *msg,int curParam)
{
	// FIXME: freenode has two ugly incompatible extensions:
	// mode e: that is NOT viewable (???)
	// mode q that stands for "quiet-ban"
	//    mode #chan +q mask
	// adds mask to the banlist with the prefix %
	// and doesn't allow the users matching the mask to talk to the channel

	bool bSet = true;
	const char * aux = modefl.ptr();
	KviStr aParam;

	KviStr nickBuffer;
	KviStr hostBuffer;
	if(source.hasHost())
	{
		formatNickLink(source.nick(),nickBuffer);
		formatHostLink(source.host(),hostBuffer);
	} else {
		nickBuffer = source.nick();
		if(nickBuffer.contains('.'))
		{
			// This looks a lot like a server!
			formatServerLink(source.nick(),nickBuffer);
		} else {
			// Probably a service....whois should work
			formatNickLink(source.nick(),nickBuffer);
		}
		hostBuffer = source.host();
	}
	KviStr auxBuffer;
	KviIrcMask * auxMask;
	KviConsole * console = msg->console();

	int curParamSave = curParam;
	bool bIsMe;
	//FIXME: Use PREFIX in 005 numeric instead of bServerSupportsModeIe - get rid of it altogether
	bool bModeIe = console->connectionInfo() ? console->connectionInfo()->bServerSupportsModeIe : false;


	while(*aux)
	{
		switch(*aux)
		{
			case '+':
				bSet = true;
			break;
			case '-':
				bSet = false;
			break;

			case 'o':
				aParam = msg->safeParam(curParam++);
				chan->op(aParam.ptr(),bSet);
				bIsMe = IS_ME(msg,aParam.ptr());
				if(bIsMe)
				{
					if(TRIGGER_EVENT_3PARAM_RETVALUE(bSet ? KviEvent_OnMeOp : KviEvent_OnMeDeOp,chan,source.nick(),source.username(),source.host()))msg->setHaltOutput();
					chan->updateCaption();
				} else {
					if(TRIGGER_EVENT_4PARAM_RETVALUE(bSet ? KviEvent_OnOp : KviEvent_OnDeOp,chan,source.nick(),source.username(),source.host(),aParam.ptr()))msg->setHaltOutput();
				}
				if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges)))
				{
					chan->output(bSet ? (bIsMe ? KVI_OUT_MEOP : KVI_OUT_OP) : (bIsMe ? KVI_OUT_MEDEOP : KVI_OUT_DEOP),
						__tr2wc("%s [%s@%s] has set mode %co %s"),
						nickBuffer.ptr(),source.username(),hostBuffer.ptr(),
						bSet ? '+' : '-',formatNickLink(aParam.ptr(),auxBuffer));
				}
			break;
			case 'h':
				aParam = msg->safeParam(curParam++);
				chan->halfop(aParam.ptr(),bSet);
				bIsMe = IS_ME(msg,aParam.ptr());
				if(bIsMe)
				{
					if(TRIGGER_EVENT_3PARAM_RETVALUE(bSet ? KviEvent_OnMeHalfOp : KviEvent_OnMeDeHalfOp,chan,source.nick(),source.username(),source.host()))msg->setHaltOutput();
					chan->updateCaption();
				} else {
					if(TRIGGER_EVENT_4PARAM_RETVALUE(bSet ? KviEvent_OnHalfOp : KviEvent_OnDeHalfOp,chan,source.nick(),source.username(),source.host(),aParam.ptr()))msg->setHaltOutput();
				}
				if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges)))
				{
					chan->output(bSet ? (bIsMe ? KVI_OUT_MEHALFOP : KVI_OUT_HALFOP) : (bIsMe ? KVI_OUT_MEDEHALFOP : KVI_OUT_HALFDEOP),
						__tr2wc("%s [%s@%s] has set mode %ch %s"),
						nickBuffer.ptr(),source.username(),hostBuffer.ptr(),
						bSet ? '+' : '-',formatNickLink(aParam.ptr(),auxBuffer));
				}
			break;
			case 'v':
				aParam = msg->safeParam(curParam++);
				chan->voice(aParam.ptr(),bSet);
				bIsMe = IS_ME(msg,aParam.ptr());
				if(bIsMe)
				{
					if(TRIGGER_EVENT_3PARAM_RETVALUE(bSet ? KviEvent_OnMeVoice : KviEvent_OnMeDeVoice,chan,source.nick(),source.username(),source.host()))msg->setHaltOutput();
					chan->updateCaption();
				} else {
					if(TRIGGER_EVENT_4PARAM_RETVALUE(bSet ? KviEvent_OnVoice : KviEvent_OnDeVoice,chan,source.nick(),source.username(),source.host(),aParam.ptr()))msg->setHaltOutput();
				}
				if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges)))
				{
					chan->output(bSet ? (bIsMe ? KVI_OUT_MEVOICE : KVI_OUT_VOICE) : (bIsMe ? KVI_OUT_MEDEVOICE : KVI_OUT_DEVOICE),
						__tr2wc("%s [%s@%s] has set mode %cv %s"),
						nickBuffer.ptr(),source.username(),hostBuffer.ptr(),
						bSet ? '+' : '-',formatNickLink(aParam.ptr(),auxBuffer));
				}
			break;
			case 'u':
				// FIXME: Handle it ONLY if the server supports the +u mode?
				aParam = msg->safeParam(curParam++);
				chan->userop(aParam.ptr(),bSet);
				bIsMe = IS_ME(msg,aParam.ptr());
				if(bIsMe)
				{
					if(TRIGGER_EVENT_3PARAM_RETVALUE(bSet ? KviEvent_OnMeUserOp : KviEvent_OnMeDeUserOp,chan,source.nick(),source.username(),source.host()))msg->setHaltOutput();
					chan->updateCaption();
				} else {
					if(TRIGGER_EVENT_4PARAM_RETVALUE(bSet ? KviEvent_OnUserOp : KviEvent_OnDeUserOp,chan,source.nick(),source.username(),source.host(),aParam.ptr()))msg->setHaltOutput();
				}
				if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges)))
				{
					chan->output(bSet ? (bIsMe ? KVI_OUT_MEUSEROP : KVI_OUT_USEROP) : (bIsMe ? KVI_OUT_MEDEUSEROP : KVI_OUT_USERDEOP),
						__tr2wc("%s [%s@%s] has set mode %cu %s"),
						nickBuffer.ptr(),source.username(),hostBuffer.ptr(),
						bSet ? '+' : '-',formatNickLink(aParam.ptr(),auxBuffer));
				}
			break;
			case 'a':
				// FIXME : Handle it ONLY if the server supports the +a mode ?
				aParam = msg->safeParam(curParam++);
				chan->setChanAdmin(aParam.ptr(),bSet);
				bIsMe = IS_ME(msg,aParam.ptr());
				if(bIsMe)
				{
					if(TRIGGER_EVENT_3PARAM_RETVALUE(bSet ? KviEvent_OnMeChanAdmin : KviEvent_OnMeDeChanAdmin,chan,source.nick(),source.username(),source.host()))msg->setHaltOutput();
					chan->updateCaption();
				} else {
					if(TRIGGER_EVENT_4PARAM_RETVALUE(bSet ? KviEvent_OnChanAdmin : KviEvent_OnDeChanAdmin,chan,source.nick(),source.username(),source.host(),aParam.ptr()))msg->setHaltOutput();
				}
				if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges)))
				{
					chan->output(bSet ? (bIsMe ? KVI_OUT_MECHANADMIN : KVI_OUT_CHANADMIN) : (bIsMe ? KVI_OUT_MEDECHANADMIN : KVI_OUT_DECHANADMIN),
						__tr2wc("%s [%s@%s] has set mode %ca %s"),
						nickBuffer.ptr(),source.username(),hostBuffer.ptr(),
						bSet ? '+' : '-',formatNickLink(aParam.ptr(),auxBuffer));
				}
			break;
			case 'b':
				aParam = msg->safeParam(curParam++);
				chan->setMask(*aux,aParam.ptr(),bSet,msg->prefix(),QDateTime::currentDateTime().toString().ascii());
				auxMask = new KviIrcMask(aParam.ptr());
				bIsMe = auxMask->matchesFixed(console->currentNickName(),console->currentUserName(),console->currentLocalHostName());
				delete auxMask;

				if(bIsMe)
				{
					if(TRIGGER_EVENT_4PARAM_RETVALUE(bSet ? KviEvent_OnMeBan : KviEvent_OnMeUnban,chan,source.nick(),source.username(),source.host(),aParam.ptr()))msg->setHaltOutput();
				} else {
					if(TRIGGER_EVENT_4PARAM_RETVALUE(bSet ? KviEvent_OnBan : KviEvent_OnUnban,chan,source.nick(),source.username(),source.host(),aParam.ptr()))msg->setHaltOutput();
				}

				if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges)))
				{
					chan->output(bSet ? (bIsMe ? KVI_OUT_MEBAN : KVI_OUT_BAN) : (bIsMe ? KVI_OUT_MEUNBAN : KVI_OUT_UNBAN),
						__tr2wc("%s [%s@%s] has set mode %cb \r!m%cb\r%s\r"),
						nickBuffer.ptr(),source.username(),hostBuffer.ptr(),
						bSet ? '+' : '-',bSet ? '-' : '+',aParam.ptr());
				}
			break;
			case 'k':
				if(bSet)aParam = msg->safeParam(curParam++);
				else aParam = "";
				chan->setChannelKey(aParam.ptr());

			if(bSet) {
				if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnKeySet,chan,source.nick(),source.username(),source.host(),aParam.ptr()))
					msg->setHaltOutput();
			} else {
				if(TRIGGER_EVENT_3PARAM_RETVALUE(KviEvent_OnKeyUnset,chan,source.nick(),source.username(),source.host()))
					msg->setHaltOutput();
			}
				
				if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges)))
				{
					if(bSet)chan->output(KVI_OUT_KEY,
						__tr2wc("%s [%s@%s] has set channel key to \"\r!m-k\r%s\r\""),
						nickBuffer.ptr(),source.username(),hostBuffer.ptr(),aParam.ptr());
					else chan->output(KVI_OUT_KEY,
						__tr2wc("%s [%s@%s] has unset the channel key"),
						nickBuffer.ptr(),source.username(),hostBuffer.ptr());
				}
			break;
			case 'l':
				if(bSet)aParam = msg->safeParam(curParam++);
				else aParam = "";
				chan->setChannelLimit(aParam.ptr());
				
				if(bSet) {
					if(TRIGGER_EVENT_4PARAM_RETVALUE(KviEvent_OnLimitSet,chan,source.nick(),source.username(),source.host(),aParam.ptr()))
						msg->setHaltOutput();
				} else {
					if(TRIGGER_EVENT_3PARAM_RETVALUE(KviEvent_OnLimitUnset,chan,source.nick(),source.username(),source.host()))
						msg->setHaltOutput();
				}
				
				if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges)))
				{
					if(bSet)chan->output(KVI_OUT_LIMIT,
						__tr2wc("%s [%s@%s] has set channel \r!m-l\rlimit to %s\r"),
						nickBuffer.ptr(),source.username(),hostBuffer.ptr(),aParam.ptr());
					else chan->output(KVI_OUT_LIMIT,
						__tr2wc("%s [%s@%s] has unset the channel limit"),
						nickBuffer.ptr(),source.username(),hostBuffer.ptr());
				}
			break;
			case 'I':
				if(bModeIe) //FIXME: Use PREFIX in 005 numeric instead of bServerSupportsModeIe
				{
					aParam = msg->safeParam(curParam++);
					chan->setMask(*aux,aParam.ptr(),bSet,msg->prefix(),QDateTime::currentDateTime().toString().ascii());
					auxMask = new KviIrcMask(aParam.ptr());
					bIsMe = auxMask->matchesFixed(console->currentNickName(),console->currentUserName(),console->currentLocalHostName());
					delete auxMask;
	
					if(bIsMe)
					{
						if(TRIGGER_EVENT_4PARAM_RETVALUE(bSet ? KviEvent_OnMeInviteException : KviEvent_OnMeInviteExceptionRemove,chan,source.nick(),source.username(),source.host(),aParam.ptr()))msg->setHaltOutput();
					} else {
						if(TRIGGER_EVENT_4PARAM_RETVALUE(bSet ? KviEvent_OnInviteException : KviEvent_OnInviteExceptionRemove,chan,source.nick(),source.username(),source.host(),aParam.ptr()))msg->setHaltOutput();
					}
	
					if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges)))
					{
						chan->output(bSet ? (bIsMe ? KVI_OUT_MEINVITEEXCEPT : KVI_OUT_INVITEEXCEPT) : (bIsMe ? KVI_OUT_MEINVITEUNEXCEPT : KVI_OUT_INVITEUNEXCEPT),
							__tr2wc("%s [%s@%s] has set mode %cI \r!m%cI\r%s\r"),
							nickBuffer.ptr(),source.username(),hostBuffer.ptr(),
							bSet ? '+' : '-',bSet ? '-' : '+',aParam.ptr());
					}
					break;
				}
			case 'e':
				if(bModeIe)
				{
					aParam = msg->safeParam(curParam++);
					chan->setMask(*aux,aParam.ptr(),bSet,msg->prefix(),QDateTime::currentDateTime().toString().ascii());
					auxMask = new KviIrcMask(aParam.ptr());
					bIsMe = auxMask->matchesFixed(console->currentNickName(),console->currentUserName(),console->currentLocalHostName());
					delete auxMask;

					if(bIsMe)
					{
						if(TRIGGER_EVENT_4PARAM_RETVALUE(bSet ? KviEvent_OnMeBanException : KviEvent_OnMeBanExceptionRemove,chan,source.nick(),source.username(),source.host(),aParam.ptr()))msg->setHaltOutput();
					} else {
						if(TRIGGER_EVENT_4PARAM_RETVALUE(bSet ? KviEvent_OnBanException : KviEvent_OnBanExceptionRemove,chan,source.nick(),source.username(),source.host(),aParam.ptr()))msg->setHaltOutput();
					}

					if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges)))
					{
						chan->output(bSet ? (bIsMe ? KVI_OUT_MEBANEXCEPT : KVI_OUT_BANEXCEPT) : (bIsMe ? KVI_OUT_MEBANUNEXCEPT : KVI_OUT_BANUNEXCEPT),
							__tr2wc("%s [%s@%s] has set mode %ce \r!m%ce\r%s\r"),
							nickBuffer.ptr(),source.username(),hostBuffer.ptr(),
							bSet ? '+' : '-',bSet ? '-' : '+',aParam.ptr());
					}
					break;
				}
			default:
				chan->setChannelMode(*aux,bSet);
				if(!(msg->haltOutput() || KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges)))
				{
					chan->output(KVI_OUT_CHANMODE,
						__tr2wc("%s [%s@%s] has set channel \r!m%c%c\rmode %c%c\r"),
						nickBuffer.ptr(),source.username(),hostBuffer.ptr(),
						bSet ? '-' : '+',*aux,bSet ? '+' : '-',*aux);
				}
			break;
		}
		++aux;
	}

	if(KVI_OPTION_BOOL(KviOption_boolShowCompactModeChanges) && (!msg->haltOutput()) &&
			(!kvi_strEqualCS(modefl.ptr(),"+")))
	{
		KviStr param;
		KviStr params;
		param = msg->safeParam(curParamSave++);
		while(param.hasData())
		{
			if(params.hasData())params.append(' ');
			params.append(param);
			param = msg->safeParam(curParamSave++);
		}

		if(params.hasData())
		{
			chan->output(KVI_OUT_CHANMODE,__tr2wc("%s [%s@%s] has set mode %s %s"),
				nickBuffer.ptr(),source.username(),hostBuffer.ptr(),modefl.ptr(),
				params.ptr());
		} else {
			chan->output(KVI_OUT_CHANMODE,__tr2wc("%s [%s@%s] has set channel mode %s"),
				nickBuffer.ptr(),source.username(),hostBuffer.ptr(),modefl.ptr());
		}
	}
}
