//=============================================================================
//
//   File : kvi_frame.cpp
//   Creation date : Sun Jun 18 2000 17:59:02 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __KVIRC__
#define _KVI_DEBUG_CHECK_RANGE_

#include "kvi_debug.h"
#include "kvi_app.h"
#include "kvi_settings.h"
#include "kvi_frame.h"
#include "kvi_options.h"
#include "kvi_menubar.h"
#include "kvi_mdimanager.h"
#include "kvi_mdichild.h"
#include "kvi_iconmanager.h"
#include "kvi_window.h"
#include "kvi_taskbar.h"
#include "kvi_console.h"
#include "kvi_config.h"
#include "kvi_internalcmd.h"
#include "kvi_uparser.h"
#include "kvi_console.h"
#include "kvi_debug.h"
#include "kvi_irctoolbar.h"
#include "kvi_confignames.h"
#include "kvi_event.h"
#include "kvi_parameterlist.h"
#include "kvi_module.h"
#include "kvi_mextoolbar.h"
#include "kvi_locale.h"
#include "kvi_irccontext.h"

#ifdef COMPILE_SCRIPTTOOLBAR
	#include "kvi_scripttoolbar.h"
	#include "kvi_scripttoolbar_manager.h"
	#include "kvi_scripttoolbar_definition.h"
#endif

#include <qsplitter.h>
#include <qvariant.h>
#include <qlineedit.h>
#include <qmessagebox.h>
#include <qcheckbox.h>
#include <qaccel.h>

#include <qlayout.h>

#if QT_VERSION >= 300
	#include <qfile.h>
	#include <qtextstream.h>
	#include <qdockarea.h>
#endif

#include <time.h>

#ifdef COMPILE_PSEUDO_TRANSPARENCY
	#include <qpixmap.h>
	// kvi_app.h
	extern QPixmap * g_pShadedParentGlobalDesktopBackground;
	extern QPixmap * g_pShadedChildGlobalDesktopBackground;
// FIXME: #warning "When a toolbar is moved , MdiManager is resized but does not update the MdiChild backgrounds"
#endif

// FIXME: #warning "LAG-O-METER : With PINGS to server , disconnect and reconnect on lag too big"

// Declared and managed by KviApp (kvi_app.cpp)
extern KviConfig * g_pWinPropertiesConfig;

KviFrame::KviFrame(unsigned int id,const char * name)
: KviTalMainWindow(0,name)
{
	setIcon(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_KVIRC)));

	m_pWinList  = new KviPtrList<KviWindow>;
	m_pWinList->setAutoDelete(false);

	m_pModuleExtensionToolBarList = new KviPtrList<KviMexToolBar>;
	m_pModuleExtensionToolBarList->setAutoDelete(false);

	m_pActiveWindow = 0;

	m_pDockExtension = 0;

	m_pSplitter = new QSplitter(QSplitter::Horizontal,this,"main_splitter");
	setCentralWidget(m_pSplitter);

	setUsesBigPixmaps(KVI_OPTION_BOOL(KviOption_boolUseBigIcons));


	m_pMdi      = new KviMdiManager(m_pSplitter,this,"mdi_manager");


	connect(m_pMdi,SIGNAL(enteredSdiMode()),this,SLOT(enteredSdiMode()));
	connect(m_pMdi,SIGNAL(leftSdiMode()),this,SLOT(leftSdiMode()));

	m_pMenuBar   = new KviMenuBar(this,"main_menu_bar"); // This theoretically had to exists before KviMdiManager (that uses enterSdiMode)

	m_pTaskBar = 0;

	createTaskBar();

	m_pIrcContextToolBar = new KviIrcContextToolBar(this);

	if((KVI_OPTION_RECT(KviOption_rectFrameGeometry).width() < 100) || (KVI_OPTION_RECT(KviOption_rectFrameGeometry).height() < 100))
	{
		// Try to find some reasonable defaults
		KVI_OPTION_RECT(KviOption_rectFrameGeometry) = QRect(10,10,g_pApp->desktop()->width() - 200,g_pApp->desktop()->height() - 150);
	}

	resize(KVI_OPTION_RECT(KviOption_rectFrameGeometry).width(),
		KVI_OPTION_RECT(KviOption_rectFrameGeometry).height());
	move(KVI_OPTION_RECT(KviOption_rectFrameGeometry).x(),
		KVI_OPTION_RECT(KviOption_rectFrameGeometry).y());

	applyOptions();

	installAccelerators(this);

	layout()->setResizeMode(QLayout::FreeResize);

}

KviFrame::~KviFrame()
{
	KVI_OPTION_RECT(KviOption_rectFrameGeometry) = QRect(pos().x(),pos().y(),
			size().width(),size().height());

	KVI_OPTION_BOOL(KviOption_boolUseBigIcons) = usesBigPixmaps();

	KVI_OPTION_BOOL(KviOption_boolMdiManagerInSdiMode) = m_pMdi->isInSDIMode();

#ifdef COMPILE_SCRIPTTOOLBAR
	// "touch" the visibility state for the script toolbars
	g_pScriptToolBarManager->storeVisibilityState();
#endif // COMPILE_SCRIPTTOOLBAR

	saveToolBarPositions();

	saveModuleExtensionToolBars();
	// Explicitly kill all the module extension toolbars: qt has NOT to delete them: we must call their "die" method
	while(KviMexToolBar * t = m_pModuleExtensionToolBarList->first())t->die();
	delete m_pModuleExtensionToolBarList;

	KVI_OPTION_BOOL(KviOption_boolShowDockExtension) = m_pDockExtension;

	if(m_pDockExtension)
	{
		m_pDockExtension->die();
		m_pDockExtension = 0;
	}

	while(m_pWinList->first())closeWindow(m_pWinList->first());
	delete m_pWinList;
}



void KviFrame::registerModuleExtensionToolBar(KviMexToolBar * t)
{
	m_pModuleExtensionToolBarList->append(t);
}

void KviFrame::unregisterModuleExtensionToolBar(KviMexToolBar * t)
{
	m_pModuleExtensionToolBarList->removeRef(t);
}

void KviFrame::restoreModuleExtensionToolBars()
{
	for(QStringList::Iterator it = KVI_OPTION_STRINGLIST(KviOption_stringlistModuleExtensionToolbars).begin();it != KVI_OPTION_STRINGLIST(KviOption_stringlistModuleExtensionToolbars).end();++it)
	{
		KviStr szEntry = *it;
		int idx = szEntry.findFirstIdx(':');
		if(idx != -1)
		{
			KviStr szMod = szEntry.left(idx);
			szEntry.cutLeft(idx + 1);
			//debug("HAVE TO LOAD ENTRY %s",szEntry.ptr());
			g_pModuleExtensionManager->allocateExtension("toolbar",szEntry.ptr(),firstConsole(),0,0,szMod.ptr());
		}
	}
}

void KviFrame::saveModuleExtensionToolBars()
{
	KVI_OPTION_STRINGLIST(KviOption_stringlistModuleExtensionToolbars).clear();

	for(KviMexToolBar * t = m_pModuleExtensionToolBarList->first();t;t = m_pModuleExtensionToolBarList->next())
	{
		QString s = t->descriptor()->module()->name();
		s += ":";
		s += t->descriptor()->name().ptr();

		//debug("FOUND TOOLBAR %s",t->descriptor()->name().ptr());

		KVI_OPTION_STRINGLIST(KviOption_stringlistModuleExtensionToolbars).append(s);
	}
}

KviMexToolBar * KviFrame::moduleExtensionToolBar(int extensionId)
{
	for(KviMexToolBar * t = m_pModuleExtensionToolBarList->first();t;t = m_pModuleExtensionToolBarList->next())
	{
		if(extensionId == t->descriptor()->id())return t;
	}
	return 0;
}

/*
	@doc: keyboard
	@type:
		generic
	@title:
		Keyboard shortcuts
	@keyterms:
		Keyboard shortcuts
	@short:
		The list of the common keyboard shortcuts
	@body:
		[b]Ctrl+LeftArrow[/b]: Previous window[br]
		[b]Ctrl+RightArrow[/b]: Next window[br]
		[b]Alt+LeftArrow[/b]: Previous window in the same IRC context[/b]
		[b]Alt+RightArrow[/b]: Next window in the same IRC context[/b]
		[b]Ctrl+UpArrow[/b]: Maximize current window[br]
		[b]Ctrl+DownArrow[/b] or [b]ESC[/b]: Minimize current window[br]
		[b]Ctrl+&lt;digit&gt;[/b], [b]F1-F12[/b], [b]Shift+(F1-F12)[/b]: Script accelerators (see [event:onaccelkeypressed]OnAccelKeyPressed[/event])[br]
		[b]Tab in the first word of input[/b]: Completes nicknames in the current channel or query[br]
		[b]Tab after a leading /[/b]: Completes commands[br]
		[b]Tab after a / in the middle of input[/b]: Completes directories[br]
		[b]Tab after a $[/b]: Completes function names[br]
		[b]Shift+Tab after the first word of input[/b]: completes masks in the current channel or query[br]
		[b]Ctrl+B[/b]: Inserts the 'bold' mIRC text control character[br]
		[b]Ctrl+K[/b]: Inserts the 'color' mIRC text control character[br]
		[b]Ctrl+R[/b]: Inserts the 'reverse' mIRC text control character[br]
		[b]Ctrl+U[/b]: Inserts the 'underline' mIRC text control character[br]
		[b]Ctrl+O[/b]: Inserts the 'reset' mIRC text control character[br]
		[b]Ctrl+P[/b]: Inserts the 'non-crypt' (plain text) KVIrc control character used to disable encryption of the current text line[br]
		[b]Ctrl+C[/b]: Copies the selected text to clipboard[br]
		[b]Ctrl+X[/b]: Cuts the selected text[br]
		[b]Ctrl+V[/b]: Pastes the clipboard contents (same as middle mouse click)[br]
		[b]Ctrl+I[/b]: Inserts the 'icon' control code and pops up the icon list box
		[b]CursorUp[/b]: Moves backward in the command history[br]
		[b]CursorDown[/b]: Moves forward in the command history[br]
		[b]Ctrl+PageUp[/b]: Opens the history popup[br]
		[b]CursorRight[/b]: Moves the cursor to the right[br]
		[b]CursorLeft[/b]: Moves the cursor to the left :)[br]
		[b]Shift+CursorLeft[/b]: Moves the selection to the left[br]
		[b]Shift+RightCursor[/b]: Moves the selection to the right[br]
		[b]PageUp[/b]: Scrolls the output window up one page[br]
		[b]PageDown[/b]: Scrolls the output window down one page[b]
		[b]Shift+PageUp[/b]: Scrolls the output window up one line[br]
		[b]Shift+PageDown[/b]: Scrolls the output window down one line[b]
		[b]Alt+lt;numeric_sequence&gt;[/b]: Inserts the character by ASCII/Unicode code[br]
		[b]Ctrl+Backspace[/b]: Shows or hides the multiline editor[br]
		[b]Ctrl+End[/b]: Closes the current window[br]
		<example>
			Alt+32: Inserts ASCII/Unicode character 32: ' ' (a space)
			Alt+00032: Same as above :)
			Alt+13: Inserts the Carriage Return (CR) control character
			Alt+77: Inserts ASCII/Unicode character 77: 'M'
			Alt+23566: Inserts Unicode character 23566 (an ideogram)
		</example>
*/

QAccel * KviFrame::installAccelerators(QWidget * wnd)
{
	QAccel *ac = new QAccel(wnd ? (QWidget *)wnd : (QWidget *)this);

	static int accel_table[] = {
		Key_Left + CTRL ,    // prev window
		Key_Right + CTRL ,   // next window
		Key_Up + CTRL ,      // maximize window
		Key_Down + CTRL ,    // minimize window
		Key_Escape ,         // minimize window
		Key_Left + ALT ,     // prev window in context
		Key_Right + ALT ,    // next window in context
		Key_End + CTRL ,     // close current window
		Key_1 + CTRL ,       // script accels...
		Key_2 + CTRL ,
		Key_3 + CTRL ,
		Key_4 + CTRL ,
		Key_5 + CTRL ,
		Key_6 + CTRL ,
		Key_7 + CTRL ,
		Key_8 + CTRL ,
		Key_9 + CTRL ,
		Key_0 + CTRL ,
		Key_F1 ,
		Key_F2 ,
		Key_F3 ,
		Key_F4 ,
		Key_F5 ,
		Key_F6 ,
		Key_F7 ,
		Key_F8 ,
		Key_F9 ,
		Key_F10 ,
		Key_F11 ,
		Key_F12 ,
		Key_F1 + SHIFT ,
		Key_F2 + SHIFT ,
		Key_F3 + SHIFT ,
		Key_F4 + SHIFT ,
		Key_F5 + SHIFT ,
		Key_F6 + SHIFT ,
		Key_F7 + SHIFT ,
		Key_F8 + SHIFT ,
		Key_F9 + SHIFT ,
		Key_F10 + SHIFT ,
		Key_F11 + SHIFT ,
		Key_F12 + SHIFT ,
		0
	};

	int i=0;
	int keys;
	while((keys = accel_table[i]))
	{
		ac->insertItem(keys);
		i++;
	}

	connect(ac,SIGNAL(activated(int)),this,SLOT(accelActivated(int)));
	return ac;
}

void KviFrame::accelActivated(int id)
{
	QAccel * acc = (QAccel *)sender();

	int keys = (int)(acc->key(id));

//	KviStr keysstr = (QString)(acc->key(id));
//
//	debug("Accel keys: %s",keysstr.ptr());

	switch(keys)
	{
		case (Key_Left+CTRL):  switchToPrevWindow(); break;
		case (Key_Right+CTRL): switchToNextWindow(); break;
		case (Key_Up+CTRL):    maximizeWindow(); break;
		case (Key_Escape):
		case (Key_Down+CTRL):  minimizeWindow(); break;
		case (Key_Left+ALT):   switchToPrevWindowInContext(); break;
		case (Key_Right+ALT):  switchToNextWindowInContext(); break;
		case (Key_End+CTRL):   if(activeWindow())activeWindow()->close(); break;
		default:
		{
//			debug("Unhandled accel keys %d",keys);
#if QT_VERSION >= 300
			KviStr * keysstr = new KviStr((QString)(acc->key(id)));
#else
			KviStr * keysstr = new KviStr(QAccel::keyToString(acc->key(id)));
#endif
			TRIGGER_EVENT_1PARAM_NONEW(KviEvent_OnAccelKeyPressed,activeWindow(),keysstr);
		}
		break;
	};
}



void KviFrame::executeInternalCommand(int index)
{
	g_pUserParser->parseCommandBuffer(kvi_getInternalCommandBuffer(index),firstConsole());
}


void KviFrame::saveWindowProperties(KviWindow * wnd,const char * szSection)
{
	g_pWinPropertiesConfig->setGroup(szSection);
	g_pWinPropertiesConfig->writeEntry("EntryTimestamp",(unsigned int)time(0));

	// Allow max 30 window properties to be floating around
	while(g_pWinPropertiesConfig->groupsCount() > 30)
	{
		// Kill the oldest group
		KviConfigIterator it(*(g_pWinPropertiesConfig->dict()));
		KviStr minKey;
		unsigned int minVal = time(0);
		while(it.current() && minVal)
		{
			QString * pVal = it.current()->find("EntryTimestamp");
			if(pVal)
			{
				bool bOk;
				unsigned int uVal = pVal->toUInt(&bOk);
				if(bOk)
				{
					if(uVal < minVal)
					{
						minVal = uVal;
						minKey = it.currentKey();
					}
				} else {
					minVal = 0;
					minKey = it.currentKey();
				}
			} else {
				minVal = 0;
				minKey = it.currentKey();
			}
			++it;
		}

		if(minKey.hasData())g_pWinPropertiesConfig->clearGroup(minKey.ptr());
		else debug("Oops...no minimum key found!");
	}

	// The following line should NOT be needed...but just to be sure...
	g_pWinPropertiesConfig->setGroup(szSection);

	g_pWinPropertiesConfig->writeEntry("IsDocked",wnd->mdiParent());

//	KviWindow * top = m_pActiveWindow;
//	if(!top)top = wnd;
//	g_pWinPropertiesConfig->writeEntry("IsMaximized",top->isMaximized());
	
	g_pWinPropertiesConfig->writeEntry("WinRect",wnd->externalGeometry());

	wnd->saveProperties(g_pWinPropertiesConfig);
}


void KviFrame::closeWindow(KviWindow *wnd)
{
	wnd->triggerDestructionEvents();

	if(KVI_OPTION_BOOL(KviOption_boolWindowsRememberProperties)) // && (wnd->type() == KVI_WINDOW_TYPE_CHANNEL))
	{
		KviStr group;
		wnd->getConfigGroupName(group);
		// not uses default settings : store it always
		saveWindowProperties(wnd,group.ptr());
	}

	if(wnd == m_pActiveWindow)
	{
		m_pActiveWindow = 0;
		if(wnd == g_pApp->m_pActiveWindow)g_pApp->m_pActiveWindow = 0;
	}

	m_pWinList->removeRef(wnd);

	if(wnd->mdiParent())m_pMdi->destroyChild(wnd->mdiParent(),true);
	else delete wnd;
}



void KviFrame::addWindow(KviWindow *wnd,bool bShow)
{
	m_pWinList->append(wnd);
	wnd->createTaskBarItem(); // create the window taskbar item AFTER it has been constructed

	KviStr group;
	wnd->getConfigGroupName(group);

	bool bGroupSettings = false;

	if(g_pWinPropertiesConfig->hasGroup(group.ptr()))
		g_pWinPropertiesConfig->setGroup(group.ptr());
	else {
		bGroupSettings = true;
		if(g_pWinPropertiesConfig->hasGroup(wnd->typeString()))
			g_pWinPropertiesConfig->setGroup(wnd->typeString());
		else {
			g_pWinPropertiesConfig->setGroup("no_settings_group");
			wnd->loadProperties(g_pWinPropertiesConfig); // load it anyway (will set defaults if windows don't remember properties)
			goto default_docking; // no settings stored
		}
	}

	{
		wnd->loadProperties(g_pWinPropertiesConfig); // load it anyway (will set defaults if windows don't remember properties)
	
		if(KVI_OPTION_BOOL(KviOption_boolWindowsRememberProperties))
		{
			bool bDocked    = g_pWinPropertiesConfig->readBoolEntry("IsDocked",true);
			//bool bMaximized = g_pWinPropertiesConfig->readBoolEntry("IsMaximized",false);
			bool bMaximized;
			
			if(KVI_OPTION_BOOL(KviOption_boolMdiManagerInSdiMode))
			{
				bMaximized = true;
				//KVI_OPTION_BOOL(KviOption_boolMdiManagerInSdiMode) = false;
			} else bMaximized = false;
			
			QRect rect      = g_pWinPropertiesConfig->readRectEntry("WinRect",QRect(10,10,500,380));

			if(bDocked)
			{
				// when group settings are used , we always cascade the windows
				// this means that windows that have no specialized config group name
				// are always cascaded : this is true for consoles , queries (and other windows) but not channels (and some other windows)
				KviMdiChild * lpC = dockWindow(wnd,false,bGroupSettings,&rect);
				lpC->setRestoredGeometry(rect);
				wnd->triggerCreationEvents();
				if(bShow)
				{
					m_pMdi->showAndActivate(lpC);
					if(bMaximized)wnd->maximize();
					// Handle the special case of this top level widget not being the active one.
					// In this situation the child will not get the focusInEvent
					// and thus will not call out childWindowActivated() method
					if(!isActiveWindow())childWindowActivated(wnd);
				}
			} else {
				wnd->setGeometry(rect);
				wnd->triggerCreationEvents();
				if(bShow)
				{
					wnd->show();
					if(bMaximized)wnd->maximize();
				}
				wnd->youAreUndocked();
				if(bShow)
				{
					wnd->raise();
					wnd->setFocus();
				}
			}
			return;
		}
	}
	
default_docking:
	KviMdiChild * lpC = dockWindow(wnd,false,true); //cascade it
	wnd->triggerCreationEvents();
	if(bShow)
	{
		m_pMdi->showAndActivate(lpC);
		// Handle the special case of this top level widget not being the active one.
		// In this situation the child will not get the focusInEvent
		// and thus will not call out childWindowActivated() method
		if(!isActiveWindow())childWindowActivated(wnd);
	}
}

KviMdiChild * KviFrame::dockWindow(KviWindow *wnd,bool bShow,bool bCascade,QRect *setGeom)
{
	if(wnd->mdiParent())return wnd->mdiParent();
	KviStr tmp(KviStr::Format,"mdi_child_%s",wnd->name());
	KviMdiChild * lpC = new KviMdiChild(m_pMdi,tmp.ptr());
	lpC->setClient(wnd);
	wnd->youAreDocked();
	m_pMdi->manageChild(lpC,bCascade,setGeom);
	if(bShow)m_pMdi->showAndActivate(lpC);
	return lpC;
}

void KviFrame::undockWindow(KviWindow *wnd)
{
	if(!(wnd->mdiParent()))return;
	KviMdiChild * lpC = wnd->mdiParent();
	lpC->unsetClient();
	m_pMdi->destroyChild(lpC,false);
	wnd->youAreUndocked();
	wnd->raise();
	wnd->setFocus();
}


void KviFrame::newConsole()
{
	createNewConsole();
}

KviConsole * KviFrame::createNewConsole(bool bFirstInFrame)
{
	KviConsole * c = new KviConsole(this,bFirstInFrame ? KVI_CONSOLE_FLAG_FIRSTINFRAME : 0);
	addWindow(c);

	if(bFirstInFrame)
	{
		restoreModuleExtensionToolBars();
#ifdef COMPILE_SCRIPTTOOLBAR
		g_pScriptToolBarManager->createVisible(this);
#endif
		restoreToolBarPositions();
	}

	return c;
}

unsigned int KviFrame::consoleCount()
{
	unsigned int count = 0;
	for(KviWindow * wnd = m_pWinList->first();wnd;wnd = m_pWinList->next())
	{
		if(wnd->type() == KVI_WINDOW_TYPE_CONSOLE)count++;
	}
	return count;
}

KviConsole * KviFrame::firstConsole()
{
	for(KviWindow * wnd = m_pWinList->first();wnd;wnd = m_pWinList->next())
	{
		if(wnd->type() == KVI_WINDOW_TYPE_CONSOLE)return (KviConsole *)wnd;
	}
	//KviWindow * x;
	//x = 0;
	//x->name();
	//debug("%d",x->type());
	__range_valid(false);
	return 0; //newer here!
}

KviConsole * KviFrame::firstNotConnectedConsole()
{
	for(KviWindow * wnd = m_pWinList->first();wnd;wnd = m_pWinList->next())
	{
		if(wnd->type() == KVI_WINDOW_TYPE_CONSOLE)
		{
			if(!((KviConsole *)wnd)->connectionInProgress())
				return (KviConsole *)wnd;
		}
	}
	return 0;
}

KviWindow * KviFrame::activeWindow()
{
	return m_pActiveWindow ? m_pActiveWindow : m_pWinList->first();
}

void KviFrame::childWindowCloseRequest(KviWindow *wnd)
{
//	switch(wnd->type())
//	{
//		case KVI_WINDOW_TYPE_CONSOLE:
//			if(wnd->isNotConnected())closeWindow(wnd);
//			else {
//				// Attempt to disconnect first
//				
//			}
//		break;
//		case KVI_WINDOW_TYPE_CHANNEL:
//			
//		break;
//		default: closeWindow(wnd); break;
//	}
	closeWindow(wnd);
}

//KviIrcToolBar * KviFrame::getIrcToolBar()
//{
//	if(KVI_OPTION_BOOL(KviOption_boolSingleIrcToolBar))
//	{
//		__range_valid(m_pIrcToolBar);
//		return m_pIrcToolBar;
//	} else {
//		KviIrcToolBar * tb = new KviIrcToolBar(this,"main_irc_toolbar");
//		tb->show();
//		return tb;
//	}
//}
//
//KviIrcToolBarAppletContainer * KviFrame::frameAppletContainer()
//{
//	if(m_pIrcToolBar)return m_pIrcToolBar->frameWideContainer();
//	else return firstConsole()->icController()->parentToolBar()->frameWideContainer();
//}
//
//KviIrcToolBar * KviFrame::findFriendToolBar(KviIrcToolBar * bar)
//{
//	QObjectList * l = queryList("KviIrcToolBar");
//	if(!l)return 0;
//	QObjectListIt it(*l);
//	KviIrcToolBar * fr = 0;
//	while(it.current() && (!fr))
//	{
//		if(((KviIrcToolBar *)it.current()) != bar)fr = (KviIrcToolBar *)it.current(); // got a friend tool bar!
//		++it;
//	}
//	delete l;
//	return fr;
//}
//
//void KviFrame::updateIrcToolBars()
//{
//	if(KVI_OPTION_BOOL(KviOption_boolSingleIrcToolBar))
//	{
//		if(m_pIrcToolBar == 0)
//		{
//			// All the irc toolbars must reparent their applets to one
//			//__range_invalid(m_pIrcToolBar);
//	
//			QObjectList * l = queryList("KviIrcToolBar");
//			if(!l)return; // ops ?
//	
//			m_pIrcToolBar = new KviIrcToolBar(this,"main_irc_tool_bar");
//	
//			QObjectListIt it(*l);
//			while(it.current())
//			{
//				((KviIrcToolBar *)it.current())->reparentChildren(m_pIrcToolBar);
//				delete it.current();
//				++it;
//			}
//			delete l;
//	
//			m_pIrcToolBar->show();
//	
//			KviWindow * oldAct = m_pActiveWindow;
//			m_pActiveWindow = 0;
//			childWindowGainFocus(oldAct);
//		}
//
//	} else {
//		// The single irc toolbar must split her controllers thru multiple toolbars
//
//		if(m_pIrcToolBar)
//		{
//			// n-1 controllers get reparented
//	
//			bool bFirst = true;
//	
//			for(KviWindow * wnd = m_pWinList->first();wnd;wnd = m_pWinList->next())
//			{
//				if(wnd->type() == KVI_WINDOW_TYPE_CONSOLE)
//				{
//					if(!bFirst)
//					{
//						KviIrcToolBar * tb = getIrcToolBar();
//						tb->removeControllerNoDelete(((KviConsole *)wnd)->icController());
//						((KviConsole *)wnd)->icController()->reparent(tb->iccStack(),QPoint(0,0));
//						tb->addController(((KviConsole *)wnd)->icController());
//					} else {
//						bFirst = false; // skip the first one
//					}
//				}
//			}
//	
//			m_pIrcToolBar = 0;
//		}
//	}
//}

void KviFrame::setActiveWindow(KviWindow *wnd)
{
	if(wnd->isMinimized())wnd->restore();
	if(wnd->mdiParent())wnd->setFocus();
	else wnd->delayedAutoRaise();
}

void KviFrame::childWindowActivated(KviWindow *wnd)
{
	g_pApp->m_pActiveWindow = wnd;

	if(m_pActiveWindow != wnd)
	{
		m_pActiveWindow = wnd;
		if(wnd->isMaximized() && wnd->mdiParent())updateCaption();
		m_pIrcContextToolBar->activeWindowChanged();
		m_pTaskBar->setActiveItem(wnd->taskBarItem());
		TRIGGER_EVENT(KviEvent_OnWindowActivated,wnd);
	}
}

void KviFrame::windowActivationChange(bool bOldActive)
{
	// if we have just been activated by the WM
	// then update the active window task bar item
	// It will then reset its highlight state
	// and hopefully make the dock widget work correctly
	// in this case.
	// This will also trigger the OnWindowActivated event :)
	if(!bOldActive && isActiveWindow())
	{
		if(m_pActiveWindow)
		{
			KviWindow * pTmp = m_pActiveWindow;
			m_pActiveWindow = 0;
			childWindowActivated(pTmp);
		}
	}
}

void KviFrame::enteredSdiMode()
{
	updateCaption();
}

void KviFrame::leftSdiMode()
{
	updateCaption();
}

void KviFrame::updateCaption()
{
	KviWindow *w = activeWindow();
	if(w)
	{
		if(w->isMaximized() && w->mdiParent())
		{
			QString tmp = w->plainTextCaption();
			tmp += QChar(' ');
			tmp += KVI_OPTION_STRING(KviOption_stringFrameCaption).ptr();
			setCaption(tmp);
			return;
		}
	}
	setCaption(KVI_OPTION_STRING(KviOption_stringFrameCaption).ptr());
}


void KviFrame::closeEvent(QCloseEvent *e)
{
	e->ignore();
	
	if(KVI_OPTION_BOOL(KviOption_boolConfirmCloseWhenThereAreConnections))
	{
		// check for running connections
		
		bool bGotRunningConnection = false;
		for(KviWindow * w = m_pWinList->first();w;w = m_pWinList->next())
		{
			if(w->type() == KVI_WINDOW_TYPE_CONSOLE)
			{
				if(((KviConsole *)w)->connectionInProgress())
				{
					bGotRunningConnection = true;
					break;
				}
			}
		}
		
		if(bGotRunningConnection)
		{
			QString txt = "<p>";
			txt += __tr2qs("There are active connections, are you sure you wish to ");
			if(g_pApp->frameCount() == 1)txt += __tr2qs("quit KVIrc?");
			else txt += __tr2qs("close this window?");
			txt += "</p>";
	
			switch(QMessageBox::warning(this,__tr2qs("Confirmation - KVIrc"),txt,__tr2qs("&Yes"),__tr2qs("&Always"),__tr2qs("&No"),2,2))
			{
				case 0:
					// ok to close
				break;
				case 1:
					// ok to close but don't ask again
					KVI_OPTION_BOOL(KviOption_boolConfirmCloseWhenThereAreConnections) = false;
				break;
				case 2:
					return;
				break;
			}
		}
	}
		
	g_pApp->destroyFrame(this);
}

void KviFrame::resizeEvent(QResizeEvent *e)
{
	KVI_OPTION_RECT(KviOption_rectFrameGeometry) = QRect(pos().x(),pos().y(),
			size().width(),size().height());
	KviTalMainWindow::resizeEvent(e);
}

void KviFrame::updatePseudoTransparency()
{
#ifdef COMPILE_PSEUDO_TRANSPARENCY
	if(g_pShadedParentGlobalDesktopBackground)m_pMdi->viewport()->update();

	if(g_pShadedChildGlobalDesktopBackground)
	{
		for(KviWindow * wnd = m_pWinList->first();wnd;wnd = m_pWinList->next())wnd->updateBackgrounds();
		m_pIrcContextToolBar->updatePseudoTransparency();
		m_pTaskBar->updatePseudoTransparency();
	}
#endif
}

void KviFrame::moveEvent(QMoveEvent *e)
{
	KVI_OPTION_RECT(KviOption_rectFrameGeometry) = QRect(pos().x(),pos().y(),
			size().width(),size().height());
#ifdef COMPILE_PSEUDO_TRANSPARENCY
	updatePseudoTransparency();
#endif
	KviTalMainWindow::moveEvent(e);
}

void KviFrame::applyOptions()
{
	m_pMdi->update();
	for(KviWindow * wnd = m_pWinList->first();wnd;wnd = m_pWinList->next())wnd->applyOptions();
	updateCaption();
	
	m_pIrcContextToolBar->applyOptions();
	m_pTaskBar->applyOptions();
}

void KviFrame::fillToolBarsPopup(QPopupMenu * p)
{
	p->clear();

	disconnect(p,SIGNAL(activated(int)),this,SLOT(toolbarsPopupSelected(int))); // just to be sure
	connect(p,SIGNAL(activated(int)),this,SLOT(toolbarsPopupSelected(int)));

	int id = p->insertItem(__tr2qs("Show IRC Context Toolbar"),this,SLOT(toggleIrcContextToolBar()));
	p->setItemChecked(id,m_pIrcContextToolBar->isVisible());

	KviModuleExtensionDescriptorList * l = g_pModuleExtensionManager->getExtensionList("toolbar");
	if(l)
	{
		for(KviModuleExtensionDescriptor * d = l->first();d;d = l->next())
		{
			QString label(__tr2qs("Show "));
			label.append(d->visibleName().ptr());
			if(d->icon())id = p->insertItem(*(d->icon()),label);
			else id = p->insertItem(label);
			p->setItemChecked(id,moduleExtensionToolBar(d->id()));
			p->setItemParameter(id,d->id());
		}
	}

#ifdef COMPILE_SCRIPTTOOLBAR
	QDictIterator<KviScriptToolBarDefinition> it(*(g_pScriptToolBarManager->toolBarDefinitions()));
	if(it.current())
	{
		p->insertSeparator();
		while(KviScriptToolBarDefinition * def = it.current())
		{
			QString label(__tr2qs("Show "));
			label.append(def->label());
			id = p->insertItem(label);
			p->setItemChecked(id,def->existingInstance(this));
			p->setItemParameter(id,def->id());
			++it;
		}
	}
#endif // COMPILE_SCRIPTTOOLBAR
}


void KviFrame::toolbarsPopupSelected(int id)
{
	const QObject * o = sender();
	if(!o)return;
	if(!o->inherits("QPopupMenu"))return;
	const QPopupMenu * p = (const QPopupMenu *)o;
	int idext = p->itemParameter(id);

#ifdef COMPILE_SCRIPTTOOLBAR
	KviScriptToolBarDefinition * def = g_pScriptToolBarManager->find(idext);
	if(def)
	{
		if(KviScriptToolBar * pBar = def->existingInstance(this))
		{
			delete pBar; // kill it
		} else {
			def->instance(this);
		}
		return;
	}
#endif // COMPILE_SCRIPTTOOLBAR

	if(KviMexToolBar * t = moduleExtensionToolBar(idext))
	{
		t->die();
	} else {
		g_pModuleExtensionManager->allocateExtension("toolbar",idext,firstConsole());
	}
}


void KviFrame::toggleIrcContextToolBar()
{
	if(m_pIrcContextToolBar->isVisible())m_pIrcContextToolBar->hide();
	else m_pIrcContextToolBar->show();
}

bool KviFrame::focusNextPrevChild(bool next)
{
	//debug("FOCUS NEXT PREV CHILD");
	QWidget * w = focusWidget();
	if(w)
	{
		if(w->focusPolicy() == QWidget::StrongFocus)return false;
		//QVariant v = w->property("KviProperty_FocusOwner");
		//if(v.isValid())return false; // Do NOT change the focus widget!
		
		if(w->parent())
		{
			QVariant v = w->parent()->property("KviProperty_ChildFocusOwner");
			if(v.isValid())return false; // Do NOT change the focus widget!
		}
	}
	// try to focus the widget on top of the Mdi
	if(m_pMdi->topChild())
	{
		m_pMdi->focusTopChild();
		return false;
	}
	return KviTalMainWindow::focusNextPrevChild(next);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Toolbar positioning stuff
////////////////////////////////////////////////////////////////////////////////////////////////////

#if QT_VERSION < 300

static QMainWindow::ToolBarDock g_dockSites[7]= {
	QMainWindow::Top ,
	QMainWindow::Left ,
	QMainWindow::Right ,
	QMainWindow::Bottom ,
	QMainWindow::TornOff ,
	QMainWindow::Unmanaged ,
	QMainWindow::Minimized
};

#endif

void KviFrame::saveToolBarPositions()
{
	KviStr szTemp;
	g_pApp->getLocalKvircDirectory(szTemp,KviApp::Config,KVI_CONFIGFILE_TOOLBARS);

#if QT_VERSION < 300

	KviConfig cfg(szTemp.ptr(),KviConfig::Write);

	int idx = 0;

	for(int i=0;i<7;i++)
	{
		KviPtrList<QToolBar> l = toolBars(g_dockSites[i]);
	
		for(QToolBar * b = l.first();b;b = l.next())
		{
			if(b->parent() == this)
			{
				cfg.setGroup("Main");
				KviStr tmp(KviStr::Format,"Toolbar%d",idx);
				KviStr lab = b->label();
				cfg.writeEntry(tmp.ptr(),lab.ptr());
				saveToolBarState(b,&cfg);
			}
			idx++;
		}
	}

	cfg.setGroup("Main");
	cfg.writeEntry("NumToolbars",idx);
#else
	QFile f(szTemp.ptr());
	if(f.open(IO_WriteOnly | IO_Truncate))
	{
		QTextStream ts(&f);
		ts << *this;
		f.close();
	}
#endif
}


void KviFrame::restoreToolBarPositions()
{
	KviStr szTemp;
	g_pApp->getLocalKvircDirectory(szTemp,KviApp::Config,KVI_CONFIGFILE_TOOLBARS);

#if QT_VERSION < 300
	KviConfig cfg(szTemp.ptr(),KviConfig::Read);
	cfg.setGroup("Main");

	int nToolbars = cfg.readIntEntry("NumToolbars",0);

	if(nToolbars > 0)
	{
		for(int i=0;i<nToolbars;i++)
		{
			cfg.setGroup("Main");
			KviStr tmp(KviStr::Format,"Toolbar%d",i);
			KviStr label = cfg.readEntry(tmp.ptr(),"");
			if(label.hasData())
			{
				QObjectList * l = queryList("QToolBar",0,false,false);
				if(l)
				{
					QObjectListIt it(*l);
					while(it.current())
					{
						if(it.current()->parent() == this)
						{
							KviStr tbLab = ((QToolBar *)it.current())->label();
							if(kvi_strEqualCI(tbLab.ptr(),label.ptr()))
							{
								restoreToolBarState((QToolBar *)it.current(),&cfg,QMainWindow::Left,0,false,0);
								break;
							}
						}
						++it;
					}
					delete l;
				}
			}
		}
	}
#else
	QFile f(szTemp.ptr());
	if(f.open(IO_ReadOnly))
	{
		QTextStream ts(&f);
		ts >> *this;
		f.close();
	} else {
		// reasonable defaults
		QString def = "\n\n\nIrcToolBar,\n[IrcToolBar,0,0,-1,-1,1]\nTaskbar,\n[Taskbar,-18,1,150,-1,1]\n\n\n\n";
		QTextStream ts(def,IO_ReadOnly);
		ts >> *this;
	}

	if(m_pTaskBar->inherits("KviTreeTaskBar"))
	{
		QDockArea * a = m_pTaskBar->area();
		if((a == topDock()) || (a == bottomDock()))
		{
			// nope.... need to move it
			a->removeDockWindow(m_pTaskBar,true,false);
			leftDock()->moveDockWindow(m_pTaskBar);
		}
	} else if(m_pTaskBar->inherits("KviClassicTaskBar"))
	{
		QDockArea * a = m_pTaskBar->area();
		if((a == leftDock()) || (a == rightDock()))
		{
			// nope.... need to move it
			a->removeDockWindow(m_pTaskBar,true,false);
			bottomDock()->moveDockWindow(m_pTaskBar);
			bottomDock()->lineUp(true);
		}
	}


#endif
}

#if QT_VERSION < 300

void KviFrame::saveToolBarState(QToolBar * t,KviConfig * cfg)
{

	QMainWindow::ToolBarDock dock;
	int index;
	bool bNl;
	int extra;

	if(getLocation(t,dock,index,bNl,extra))
	{
		KviStr tmp =t->label();
		cfg->setGroup(tmp.ptr());

		cfg->writeEntry("Visible",t->isVisible());
		cfg->writeEntry("Dock",(int)dock);
		cfg->writeEntry("Index",index);
		cfg->writeEntry("Newline",bNl);
		cfg->writeEntry("ExtraSpace",extra);
	}
}

#endif

#if QT_VERSION < 300

void KviFrame::restoreToolBarState(QToolBar * t,KviConfig * cfg,QMainWindow::ToolBarDock defDock,int defIndex,bool bDefNewline,int defExtra)
{
	QMainWindow::ToolBarDock dock;
	int index , extra;
	bool bVisible , bNl;

	KviStr tmp = t->label().ascii();

	cfg->setGroup(tmp.ptr());

	bVisible   = cfg->readBoolEntry("Visible",true);
	dock       = (QMainWindow::ToolBarDock)cfg->readIntEntry("Dock",defDock);
	index      = cfg->readIntEntry("Index",defIndex);
	bNl        = cfg->readBoolEntry("Newline",bDefNewline);
	extra      = cfg->readIntEntry("ExtraSpace",defExtra);

	if(bVisible)t->show();
	else t->hide();

	moveToolBar(t,dock,bNl,index,extra);
}

#endif

void KviFrame::createTaskBar()
{
	if(KVI_OPTION_BOOL(KviOption_boolUseTreeWindowListTaskBar))
	{
		m_pTaskBar = new KviTreeTaskBar(this);
		setDockEnabled(m_pTaskBar,Qt::DockTop,false);
		setDockEnabled(m_pTaskBar,Qt::DockBottom,false);
	} else {
		m_pTaskBar = new KviClassicTaskBar(this);
		setDockEnabled(m_pTaskBar,Qt::DockTop,true);
		setDockEnabled(m_pTaskBar,Qt::DockBottom,true);
	}
	setDockEnabled(m_pTaskBar,Qt::DockLeft,true);
	setDockEnabled(m_pTaskBar,Qt::DockRight,true);
}

void KviFrame::recreateTaskBar()
{
	QString szOldClass = m_pTaskBar->className();

	saveToolBarPositions();
	KviWindow * w;
	for(w = m_pWinList->first();w;w = m_pWinList->next())
	{
		w->destroyTaskBarItem();
	}
	removeDockWindow(m_pTaskBar);
	delete m_pTaskBar;
	createTaskBar();
	for(w = m_pWinList->first();w;w = m_pWinList->next())
	{
		w->createTaskBarItem();
	}
	restoreToolBarPositions();
	
	/*
	QString szNewClass = m_pTaskBar->className();
	if(szOldClass != szNewClass)
	{
		// the class changed...
		// make sure that the tree task bar is in the left or right dock
		// and the classic one is in the top or bottom on
		
		Qt::Dock dock;
		int index;
		bool nl;
		int eo;
		getLocation(m_pTaskBar,dock,index,nl,eo);

		if(KVI_OPTION_BOOL(KviOption_boolUseTreeWindowListTaskBar))
		{
			if((dock == Qt::Bottom) || (dock == Qt::Top))
				moveDockWindow(m_pTaskBar,Qt::Left);
		} else {
			if((dock == Qt::Left) || (dock == Qt::Right))
				moveDockWindow(m_pTaskBar,Qt::Bottom);
		}
	}
	*/

	if(m_pActiveWindow)m_pTaskBar->setActiveItem(m_pActiveWindow->taskBarItem());
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Some accelerators
////////////////////////////////////////////////////////////////////////////////////////////////////

void KviFrame::maximizeWindow(void)
{
	if(m_pActiveWindow)
	{
		if(m_pActiveWindow->isMaximized())
		{
			m_pActiveWindow->restore();
		} else {
			m_pActiveWindow->maximize();
		}
	}
}

void KviFrame::minimizeWindow(void)
{
	if(m_pActiveWindow)m_pActiveWindow->minimize();
}

void KviFrame::switchToPrevWindow(void)
{
	m_pTaskBar->switchWindow(false,false);
}

void KviFrame::switchToNextWindow(void)
{
	m_pTaskBar->switchWindow(true,false);
}

void KviFrame::switchToPrevWindowInContext(void)
{
	m_pTaskBar->switchWindow(false,true);
}

void KviFrame::switchToNextWindowInContext(void)
{
	m_pTaskBar->switchWindow(true,true);
}

#include "kvi_frame.moc"
