//=============================================================================
//
//   File : kvi_irctoolbar.cpp
//   Creation date : Thu Oct 29 2000 14:13:13 CEST by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2000-2001 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __KVIRC__
#include "kvi_irctoolbar.h"
#include "kvi_console.h"
#include "kvi_frame.h"
#include "kvi_locale.h"
#include "kvi_iconmanager.h"
#include "kvi_settings.h"
#include "kvi_options.h"
#include "kvi_frame.h"
#include "kvi_app.h"
#include "kvi_dynamictooltip.h"
#include "kvi_ircurl.h"
#include "kvi_internalcmd.h"
#include "kvi_uparser.h"
#include "kvi_list.h"

//#include <qobjcoll.h>
#include <qstyle.h>
#include <qpainter.h>
#include <qpopupmenu.h>
#include <qlayout.h>


#ifdef COMPILE_PSEUDO_TRANSPARENCY
	extern QPixmap * g_pShadedChildGlobalDesktopBackground;
#endif

static QPixmap                                  * g_pIccMemBuffer               = 0;
static KviPtrList<KviToolBarGraphicalApplet>    * g_pToolBarGraphicalAppletList = 0;

KviToolBarGraphicalApplet::KviToolBarGraphicalApplet(QWidget * par)
: QWidget(par)
{
	if(!g_pToolBarGraphicalAppletList)
	{
		g_pToolBarGraphicalAppletList = new KviPtrList<KviToolBarGraphicalApplet>();
		g_pToolBarGraphicalAppletList->setAutoDelete(false);
		g_pIccMemBuffer = new QPixmap(1,1);
	}
	
	g_pToolBarGraphicalAppletList->append(this);
	setBackgroundMode(QWidget::NoBackground);

	setMouseTracking(true);
	m_bResizeMode = false;

	m_sizeHint = QSize(180,32);
	m_bSizeLoaded = false;
}

unsigned int KviToolBarGraphicalApplet::loadAppletWidth()
{
	if(KVI_OPTION_UINT(KviOption_uintIrcContextAppletWidth) < 32)
		KVI_OPTION_UINT(KviOption_uintIrcContextAppletWidth) = 32;
	return KVI_OPTION_UINT(KviOption_uintIrcContextAppletWidth);
}

void KviToolBarGraphicalApplet::saveAppletWidth(unsigned int uWidth)
{
	KVI_OPTION_UINT(KviOption_uintIrcContextAppletWidth) = uWidth;
}

void KviToolBarGraphicalApplet::setupSizeHint()
{
	m_sizeHint = QSize(loadAppletWidth(),22);
	m_bSizeLoaded = true;
}

QSize KviToolBarGraphicalApplet::sizeHint() const
{
	// forget constness :(
	KviToolBarGraphicalApplet * that = (KviToolBarGraphicalApplet *)this;
	if(!m_bSizeLoaded)that->setupSizeHint();
	return m_sizeHint;
}


void KviToolBarGraphicalApplet::mouseMoveEvent(QMouseEvent * e)
{
	if(e->state() & LeftButton)
	{
		//debug("A)");
		if(m_bResizeMode)
		{
			int w = e->pos().x();
			if(w < 32)w = 32;
			if(w > 480)w = 480;
			//debug("B) : %d",w);
			m_sizeHint = QSize(w,22);
			resize(w,height());
			g_pApp->postEvent(parentWidget(),new QEvent(QEvent::LayoutHint));
			//((QToolBar *)parentWidget())->boxLayout()->invalidate();
			//parentWidget()->updateGeometry();
			//((QToolBar *)parentWidget())->boxLayout()->invalidate();
			//if(parentWidget()->layout())parentWidget()->layout()->invalidate();
			//parentWidget()->updateGeometry();
		}
	} else {
		if(e->pos().x() > width() - 4)
			setCursor(sizeHorCursor);
		else
			setCursor(arrowCursor);
	}
}

void KviToolBarGraphicalApplet::mousePressEvent(QMouseEvent * e)
{
	if(e->button() & LeftButton)
	{
		m_bResizeMode = (e->pos().x() > (width() - 4));
	}
}

void KviToolBarGraphicalApplet::mouseReleaseEvent(QMouseEvent * e)
{
	m_bResizeMode = false;
	setCursor(arrowCursor);
}


KviToolBarGraphicalApplet::~KviToolBarGraphicalApplet()
{
	saveAppletWidth(m_sizeHint.width());
	g_pToolBarGraphicalAppletList->removeRef(this);
	if(g_pToolBarGraphicalAppletList->isEmpty())
	{
		delete g_pToolBarGraphicalAppletList;
		g_pToolBarGraphicalAppletList = 0;
		delete g_pIccMemBuffer;
		g_pIccMemBuffer = 0;
	} else {
		// resize the mem buffer to match the maximum width / height of the applets
		resizeMemBuffer();
	}
}

void KviToolBarGraphicalApplet::resizeMemBuffer()
{
	unsigned int uMaxW = 0;
	unsigned int uMaxH = 0;
	for(KviToolBarGraphicalApplet * a = g_pToolBarGraphicalAppletList->first();a;a = g_pToolBarGraphicalAppletList->next())
	{
		if(uMaxW < a->width())uMaxW = a->width();
		if(uMaxH < a->height())uMaxH = a->height();
	}
	g_pIccMemBuffer->resize(uMaxW,uMaxH);
}

void KviToolBarGraphicalApplet::paintEvent(QPaintEvent *e)
{
	if(!isVisible())return;

	QPainter pa(g_pIccMemBuffer);

#ifdef COMPILE_PSEUDO_TRANSPARENCY
	if(g_pShadedChildGlobalDesktopBackground)
	{
		QPoint pnt = mapToGlobal(QPoint(0,0));
		pa.drawTiledPixmap(e->rect().left(),e->rect().top(),e->rect().width(),e->rect().height(),*g_pShadedChildGlobalDesktopBackground,pnt.x(),pnt.y());
	} else {
#endif
		if(KVI_OPTION_PIXMAP(KviOption_pixmapIrcToolBarAppletBackground).pixmap())
		{
			QPoint pnt = mapToGlobal(QPoint(0,0));
			pa.drawTiledPixmap(e->rect().left(),e->rect().top(),e->rect().width(),e->rect().height(),*(KVI_OPTION_PIXMAP(KviOption_pixmapIrcToolBarAppletBackground).pixmap()),pnt.x(),pnt.y());
		} else {
			pa.fillRect(e->rect().left(),e->rect().top(),e->rect().width(),e->rect().height(),KVI_OPTION_COLOR(KviOption_colorIrcToolBarAppletBackground));
		}
#ifdef COMPILE_PSEUDO_TRANSPARENCY
	}
#endif

	drawContents(&pa);

	//Need to draw the sunken rect around the view now...
	pa.setPen(colorGroup().dark());
	pa.drawLine(0,0,width(),0);
	pa.drawLine(0,0,0,width());
	pa.setPen(colorGroup().light());
	pa.drawLine(1,height() - 1,width() - 1,height() - 1);
	pa.drawLine(width() - 1,1,width() - 1,height());

	bitBlt(this,e->rect().left(),e->rect().top(),g_pIccMemBuffer,e->rect().left(),e->rect().top(),e->rect().width(),e->rect().height(),Qt::CopyROP);
}

void KviToolBarGraphicalApplet::drawContents(QPainter *)
{
	// nothing here
}

void KviToolBarGraphicalApplet::resizeEvent(QResizeEvent *e)
{
	unsigned int uBufferW = g_pIccMemBuffer->width();
	unsigned int uBufferH = g_pIccMemBuffer->height();
	unsigned int uW = width();
	unsigned int uH = height();

	if((uBufferW != uW) || (uBufferH != uH))
	{
		if((uBufferW < uW) && (uBufferH < uH))g_pIccMemBuffer->resize(uW,uH);
		else resizeMemBuffer();
	}
}


//////////////////////////////////////////////////////////////////////////////
//
// KviIrcContextDisplay
//
//      Main applet of all irc contexts
//      Displays the server connection status, server name
//      nickname, user mode and the graphical indication of the context
//
//////////////////////////////////////////////////////////////////////////////

KviIrcContextDisplay::KviIrcContextDisplay(QWidget * par,KviFrame * f)
: KviToolBarGraphicalApplet(par)
{
    m_bIsLagMeterConnected = false;
	m_pFrm = f;
	KviDynamicToolTip * tip = new KviDynamicToolTip(this);
	connect(tip,SIGNAL(tipRequest(KviDynamicToolTip *,const QPoint &)),this,SLOT(tipRequest(KviDynamicToolTip *,const QPoint &)));
}


KviIrcContextDisplay::~KviIrcContextDisplay()
{
}

void KviIrcContextDisplay::tipRequest(KviDynamicToolTip * tip,const QPoint &)
{
	QString txt;

	KviConsole * c = m_pFrm->activeWindow()->console();

	static QString b = "<b>";
	static QString nb = "</b>";
	static QString br = "<br>";

	if(c)
	{
		KviConnectionInfo * i = c->connectionInfo();


		txt = b;

		if(!i)
		{
			txt += __tr2qs("No connection");
			txt += nb;
			txt += br;
		} else {
			KviStr nickAndMode = c->currentNickName();
			if(!c->userModeEmpty())nickAndMode.append(KviStr::Format," (+%s)",c->currentUserMode());

			txt += c->currentServerName();
			txt += nb;
			txt += br;
			txt += nickAndMode.ptr();
			txt += br;
		}

		QString szNum;
		szNum.setNum(c->ircContextId());

		QString szIrcContext = QChar('(');
		szIrcContext += __tr2qs("IRC Context");
		szIrcContext += QChar(' ');
		szIrcContext += szNum;
		szIrcContext += QChar(')');
		txt += szIrcContext;

        if( (KVI_OPTION_BOOL(KviOption_boolUseLagMeterEngine) && (c->lagMeter()->isEnabled()) && 
            (KVI_OPTION_BOOL(KviOption_boolShowLagOnContextDisplay)) && m_bIsLagMeterConnected) )
        {
            txt += br;
            txt.sprintf("%sLag: %.2f", txt.ascii(), c->lagMeter()->getLag()/1000.0);

        }
	} else {
		txt += __tr2qs("No IRC context");
		txt += nb;
	}

	tip->tip(rect(),txt);
}


/*
QSize KviIrcContextDisplay::sizeHint() const
{
	return QSize(160,22);
}
*/

#define KVI_APPLETIRCCONTEXTINDICATORWIDTH 12

void KviIrcContextDisplay::drawContents(QPainter * p)
{
	// The context indicator
	KviWindow * wnd = m_pFrm->activeWindow();
	KviConsole * c = wnd ? wnd->console() : 0;

	if(c)
	{
		KviConnectionInfo * i = c->connectionInfo();
		QString serv,nick;
        QString tmp;
		if(!i)
		{
			serv = __tr2qs("Not connected");
            if(m_bIsLagMeterConnected)
            {
                disconnect(c->lagMeter(),SIGNAL(newLag(int)),this,SLOT(update()));
                m_bIsLagMeterConnected = false;
            }
		} else {
			if(c->isConnected())
			{
				nick = c->currentNickName();
				if(!c->userModeEmpty())
				{
					static QString spp(" (+");
					nick += spp;
					nick += c->currentUserMode();
					if(c->userIsAway())
					{
						nick += QChar(' ');
						nick += __tr2qs("away");
					}
					nick += QChar(')');
				} else {
					if(c->userIsAway())
					{
						static QString ugly(" (");
						nick += ugly;
						nick += __tr2qs("away");
						nick += QChar(')');
					}
				}
				serv = c->currentServerName();

                if((KVI_OPTION_BOOL(KviOption_boolUseLagMeterEngine)) && (c->lagMeter()->isEnabled()) && (KVI_OPTION_BOOL(KviOption_boolShowLagOnContextDisplay)))
                {
                    if(!m_bIsLagMeterConnected)
                    {
                        //I don't really know if connecting several times the same signal to the
                        //same slot is inocuous but is not elegant, so this switch
                        connect(c->lagMeter(),SIGNAL(newLag(int)),this,SLOT(update()));
                        m_bIsLagMeterConnected = true;
                    }
                    tmp.sprintf(" Lag: %.2f",c->lagMeter()->getLag()/1000.0);
                    nick += tmp.ascii();
                }
			} else {
				serv = __tr2qs("In progress...");
			}
		}
	
		//p->setFont(KVI_OPTION_FONT(KviOption_fontIrcToolBarApplet));
		p->setPen(KVI_OPTION_COLOR(KviOption_colorIrcToolBarAppletForegroundHighContrastActive1));
	
		p->setClipRect(KVI_APPLETIRCCONTEXTINDICATORWIDTH + 2,2,width() - (KVI_APPLETIRCCONTEXTINDICATORWIDTH + 4),height() - 4);

		if(height() < 30)
		{
			static QString xxx(" [");
			serv += xxx;
			serv += nick;
			serv += QChar(']');
			p->drawText(KVI_APPLETIRCCONTEXTINDICATORWIDTH + 4,16,serv,serv.length());
		} else {
			p->drawText(KVI_APPLETIRCCONTEXTINDICATORWIDTH + 4,16,serv,serv.length());
			p->drawText(KVI_APPLETIRCCONTEXTINDICATORWIDTH + 4,30,nick,nick.length());
		}
	
		p->setClipping(false);

		QColor base = colorGroup().background();
		QColor cntx = KVI_OPTION_ICCOLOR(c->ircContextId() % KVI_NUM_ICCOLOR_OPTIONS);
		base.setRgb((base.red() + cntx.red()) >> 1,(base.green() + cntx.green()) >> 1,
			(base.blue() + cntx.blue()) >> 1);
	
		p->fillRect(2,2, KVI_APPLETIRCCONTEXTINDICATORWIDTH - 2,height() - 4,base);
	}

	p->setPen(KVI_OPTION_COLOR(KviOption_colorIrcToolBarAppletForegroundMidContrast));
	p->drawLine(1,1,width() - 1,1);
	p->drawLine(1,1,1,height() - 1);
	p->drawLine(2,height() - 2,width() - 1,height() - 2);
	p->drawLine(width() - 2,1,width() - 2,height());
	p->drawLine(KVI_APPLETIRCCONTEXTINDICATORWIDTH,2,KVI_APPLETIRCCONTEXTINDICATORWIDTH,height() - 2);

}

//
//void KviIrcContextDisplay::mousePressEvent(QMouseEvent *e)
//{
//	if(!console())return;
//	console()->frame()->setActiveWindow(console());
//}



KviIrcContextToolBar::KviIrcContextToolBar(KviFrame * pFrm)
: KviToolBar(__tr2qs("IRC Context Toolbar"),pFrm,QT_DOCK_TOP,false,"irccontexttoolbar")
{
	m_pFrm = pFrm;
	m_pLastConsole = 0;

	m_pDisplay = new KviIrcContextDisplay(this,pFrm);

	m_pConnectButton = new QToolButton(
			*(g_pIconManager->getBigIcon(KVI_BIGICON_DISCONNECTED)),
			__tr2qs("Connect"),
			__tr2qs("Connect"),
			this,
			SLOT(connectButtonClicked()),
			this);

	m_pToolsButton = new QToolButton(
			*(g_pIconManager->getBigIcon(KVI_BIGICON_TOOLS)),
			__tr2qs("IRC Tools"),
			__tr2qs("IRC Tools"),
			this,
			SLOT(toolsButtonClicked()),
			this);

	m_pToolsPopup = new QPopupMenu(this);
	connect(m_pToolsPopup,SIGNAL(aboutToShow()),this,SLOT(setupToolsPopup()));
	m_pToolsButton->setPopup(m_pToolsPopup);
	m_pToolsButton->setPopupDelay(1);

	m_pActionsButton = new QToolButton(
			*(g_pIconManager->getBigIcon(KVI_BIGICON_ACTIONS)),
			__tr2qs("Actions"),
			__tr2qs("Actions"),
			this,
			SLOT(actionsButtonClicked()),
			this);

	m_pActionsPopup = new QPopupMenu(this);
	connect(m_pActionsPopup,SIGNAL(aboutToShow()),this,SLOT(setupActionsPopup()));
	m_pActionsButton->setPopup(m_pActionsPopup);
	m_pActionsButton->setPopupDelay(1);

	// Defer creation
	m_pConnectToServerPopup = 0;
	m_pChangeNicknamePopup = 0;
	m_pJoinChannelPopup = 0;
	m_pUserModePopup = 0;

	connect(g_pApp,SIGNAL(reloadImages()),this,SLOT(reloadImages()));
}

void KviIrcContextToolBar::applyOptions()
{
	m_pDisplay->update();
}

KviIrcContextToolBar::~KviIrcContextToolBar()
{
}

void KviIrcContextToolBar::setupToolsPopup()
{
	m_pToolsPopup->clear();

	int id = m_pToolsPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_SPY)),
								__tr2qs("Show &Socket Spy"),this,SLOT(executeInternalCommand(int)));
	m_pToolsPopup->setItemParameter(id,KVI_INTERNALCOMMAND_SOCKETSPY_OPEN);
	m_pToolsPopup->insertSeparator();
	id = m_pToolsPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_LINKS)),
								__tr2qs("Get Network &Links"),this,SLOT(executeInternalCommand(int)));
	m_pToolsPopup->setItemParameter(id,KVI_INTERNALCOMMAND_LINKS_OPEN);
	id = m_pToolsPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_LIST)),
								__tr2qs("Get &Channel List"),this,SLOT(executeInternalCommand(int)));
	m_pToolsPopup->setItemParameter(id,KVI_INTERNALCOMMAND_LIST_OPEN);
}

void KviIrcContextToolBar::reloadImages()
{
	m_pToolsButton->setPixmap(*(g_pIconManager->getBigIcon(KVI_BIGICON_TOOLS)));
	m_pActionsButton->setPixmap(*(g_pIconManager->getBigIcon(KVI_BIGICON_ACTIONS)));

	if(g_pApp->windowExists(m_pLastConsole))setPixmapsForConsole(m_pLastConsole);
}

void KviIrcContextToolBar::activeWindowChanged()
{
	KviConsole * c = m_pFrm->activeWindow()->console();

	if(c != m_pLastConsole)
	{
		m_pConnectButton->setEnabled(c);
		m_pToolsButton->setEnabled(c);
		m_pActionsButton->setEnabled(c);
		m_pDisplay->update();
		if(c)setPixmapsForConsole(c);
		m_pLastConsole = c;
	}
}

void KviIrcContextToolBar::setPixmapsForConsole(KviConsole * c)
{
	switch(c->state())
	{
		case KviConsole::NotConnected:
			m_pConnectButton->setPixmap(*(g_pIconManager->getBigIcon(KVI_BIGICON_DISCONNECTED)));
			m_pConnectButton->setTextLabel(__tr2qs("Connect"),true);
		break;
		case KviConsole::Connecting:
			m_pConnectButton->setPixmap(*(g_pIconManager->getBigIcon(KVI_BIGICON_CONNECTING)));
			m_pConnectButton->setTextLabel(__tr2qs("Abort connection"),true);
		break;
		case KviConsole::Connected:
			m_pConnectButton->setPixmap(*(g_pIconManager->getBigIcon(KVI_BIGICON_CONNECTED)));
			m_pConnectButton->setTextLabel(__tr2qs("Disconnect"),true);
		break;
	}
}

void KviIrcContextToolBar::connectStateChanged(KviConsole * c)
{
	if(c != m_pLastConsole)return; // not displaying that console
	m_pDisplay->update();
	setPixmapsForConsole(c);
}

void KviIrcContextToolBar::updateStatusDisplay(KviConsole * c)
{
	if(c != m_pLastConsole)return; // not displaying that console
	m_pDisplay->update();
}

void KviIrcContextToolBar::updatePseudoTransparency()
{
	m_pDisplay->update();
}


void KviIrcContextToolBar::executeInternalCommand(int id)
{
	KviConsole * c = m_pFrm->activeWindow()->console();
	if(!c)return;
	c->executeInternalCommand(id);
}

void KviIrcContextToolBar::connectButtonClicked()
{
	KviConsole * c = m_pFrm->activeWindow()->console();
	if(c)c->connectButtonClicked();
}

void KviIrcContextToolBar::toolsButtonClicked()
{
}

void KviIrcContextToolBar::actionsButtonClicked()
{
}

void KviIrcContextToolBar::setupActionsPopup()
{
	KviConsole * c = m_pFrm->activeWindow()->console();
	if(!c)return;

	m_pActionsPopup->clear();
	int id;

	id = m_pActionsPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_AWAY)),__tr2qs("&Away Mode"),this,SLOT(awayClicked()));
	if(!c->isConnected())m_pActionsPopup->setItemEnabled(id,false);
	else m_pActionsPopup->setItemChecked(id,c->userIsAway());

	if(!m_pUserModePopup)
	{
		m_pUserModePopup = new QPopupMenu(m_pActionsPopup);
		connect(m_pUserModePopup,SIGNAL(aboutToShow()),this,SLOT(setupUserModePopup()));
		connect(m_pUserModePopup,SIGNAL(activated(int)),this,SLOT(userModeSelected(int)));
	}
	id = m_pActionsPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_MODE)),__tr2qs("&User Mode"),m_pUserModePopup);
	if(!c->isConnected())m_pActionsPopup->setItemEnabled(id,false);
	m_pActionsPopup->insertSeparator();


	if(!m_pJoinChannelPopup)
	{
		m_pJoinChannelPopup = new QPopupMenu(m_pActionsPopup);
		connect(m_pJoinChannelPopup,SIGNAL(aboutToShow()),this,SLOT(setupJoinChannelPopup()));
		connect(m_pJoinChannelPopup,SIGNAL(activated(int)),this,SLOT(joinChannelSelected(int)));
	}
	id = m_pActionsPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_CHANNEL)),__tr2qs("&Join Channel"),m_pJoinChannelPopup);
	if(!c->isConnected())m_pActionsPopup->setItemEnabled(id,false);

	if(!m_pChangeNicknamePopup)
	{
		m_pChangeNicknamePopup = new QPopupMenu(m_pActionsPopup);
		connect(m_pChangeNicknamePopup,SIGNAL(aboutToShow()),this,SLOT(setupChangeNicknamePopup()));
		connect(m_pChangeNicknamePopup,SIGNAL(activated(int)),this,SLOT(changeNicknameSelected(int)));
	}
	id = m_pActionsPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_NICK)),__tr2qs("Change &Nickname"),m_pChangeNicknamePopup);
	if(!c->isConnected())m_pActionsPopup->setItemEnabled(id,false);

	m_pActionsPopup->insertSeparator();

	if(!m_pConnectToServerPopup)
	{
		m_pConnectToServerPopup = new QPopupMenu(m_pActionsPopup);
		connect(m_pConnectToServerPopup,SIGNAL(aboutToShow()),this,SLOT(setupConnectToServerPopup()));
		connect(m_pConnectToServerPopup,SIGNAL(activated(int)),this,SLOT(connectToServerSelected(int)));
	}
	m_pActionsPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_SERVER)),__tr2qs("&Connect To"),m_pConnectToServerPopup);

}




void KviIrcContextToolBar::setupConnectToServerPopup()
{
	KviConsole * c = g_pApp->activeWindow()->console();
	if(!c)return;

	g_pApp->fillRecentServersPopup(m_pConnectToServerPopup);
	m_pConnectToServerPopup->insertSeparator();
	m_pConnectToServerPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_SERVER)),__tr2qs("&Other..."));
}

void KviIrcContextToolBar::setupChangeNicknamePopup()
{
	KviConsole * c = g_pApp->activeWindow()->console();
	if(!c)return;

	g_pApp->fillRecentNicknamesPopup(m_pChangeNicknamePopup,c);
}

void KviIrcContextToolBar::setupJoinChannelPopup()
{
	KviConsole * c = g_pApp->activeWindow()->console();
	if(!c)return;

	g_pApp->fillRecentChannelsPopup(m_pJoinChannelPopup,c);
	m_pJoinChannelPopup->insertSeparator();
	m_pJoinChannelPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_CHANNEL)),__tr2qs("&Other..."));
}

void KviIrcContextToolBar::setupUserModePopup()
{
	KviConsole * c = g_pApp->activeWindow()->console();
	if(!c)return;


	m_pUserModePopup->clear();
	int id;
	id = m_pUserModePopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_WALLOPS)),__tr2qs("Wallops (+w)"));
	m_pUserModePopup->setItemChecked(id,c->hasUserMode('w'));
	id = m_pUserModePopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_SERVERNOTICE)),__tr2qs("Server Notices (+s)"));
	m_pUserModePopup->setItemChecked(id,c->hasUserMode('s'));
	id = m_pUserModePopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_INVISIBLE)),__tr2qs("Invisible (+i)"));
	m_pUserModePopup->setItemChecked(id,c->hasUserMode('i'));
}

void KviIrcContextToolBar::awayClicked()
{
	KviConsole * c = g_pApp->activeWindow()->console();
	if(!c)return;

	if(!c->isConnected())return;
	KviStr command = "if($away)back; else away";
	g_pUserParser->parseCommandBuffer(command.ptr(),c);
}

void KviIrcContextToolBar::userModeSelected(int id)
{
	KviConsole * c = g_pApp->activeWindow()->console();
	if(!c)return;

	KviStr text = m_pUserModePopup->text(id);
	if(!c->isConnected())return;
	if(text.hasData())
	{
		text.cutFromFirst(' ');
		KviStr command = "mode ";
		command.append(c->currentNickName());
		command.append(c->hasUserMode(*(text.ptr())) ? " -" : " +");
		command.append(*(text.ptr()));
		g_pUserParser->parseCommandBuffer(command.ptr(),c);
	}
}

void KviIrcContextToolBar::joinChannelSelected(int id)
{
	KviConsole * c = g_pApp->activeWindow()->console();
	if(!c)return;

	QString text = m_pJoinChannelPopup->text(id);
	if(!text.isEmpty())
	{
		KviStr szText;
		if(__tr2qs("&Other...") == text)szText = "channelsjoin.open";
		else {
			szText = text;
			szText.prepend("join ");
		}
		g_pUserParser->parseCommandBuffer(szText.ptr(),c);
	}
}

void KviIrcContextToolBar::connectToServerSelected(int id)
{
	KviConsole * c = g_pApp->activeWindow()->console();
	if(!c)return;

	QString text = m_pConnectToServerPopup->text(id);
	if(!text.isEmpty())
	{
		if(__tr2qs("&Other...") == text)
		{
			g_pUserParser->parseCommandBuffer("options.edit KviServerOptionsWidget",c);
		} else {
			KviStr szCommand;
			KviStr szText = text;
			if(parseIrcUrl(szText.ptr(),szCommand,KVI_IRCURL_CONTEXT_THIS))
			{
				g_pUserParser->parseCommandBuffer(szCommand.ptr(),c);
			}
		}
	}
}

void KviIrcContextToolBar::changeNicknameSelected(int id)
{
	KviConsole * c = g_pApp->activeWindow()->console();
	if(!c)return;

	KviStr text = m_pChangeNicknamePopup->text(id);
	if(text.hasData())
	{
		text.prepend("nick ");
		g_pUserParser->parseCommandBuffer(text.ptr(),c);
	}
}


#ifdef Bool
	#undef Bool
#endif

#include "kvi_irctoolbar.moc"
