//
//   File : kvi_query.cpp
//   Creation date : Tue Aug 7 2000 14:23:22 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
#define __KVIRC__
#include "kvi_query.h"
#include "kvi_console.h"
#include "kvi_iconmanager.h"
#include "kvi_ircview.h"
#include "kvi_input.h"
#include "kvi_options.h"
#include "kvi_locale.h"
#include "kvi_settings.h"
#include "kvi_irclink.h"
#include "kvi_out.h"
#include "kvi_taskbar.h"
#include "kvi_config.h"
#include "kvi_event.h"
#include "kvi_uparser.h"
#include "kvi_mirccntrl.h"
#include "kvi_settings.h"
#include "kvi_themedlabel.h"
#include "kvi_useraction.h"
#include "kvi_parameterlist.h"

#ifdef COMPILE_CRYPT_SUPPORT
	#include "kvi_crypt.h"
	#include "kvi_cryptcontroller.h"
#endif

#include <qpixmap.h>
#include <qsplitter.h>
#include <qhbox.h>
#include <qtoolbutton.h>


KviQuery::KviQuery(KviFrame * lpFrm,KviConsole * lpConsole,const char * name)
: KviWindow(KVI_WINDOW_TYPE_QUERY,lpFrm,name,lpConsole)
{
	m_pConsole->registerQuery(this);
	m_iFlags = 0;

	//m_pTopSplitter = new QSplitter(QSplitter::Horizontal,this,"top_splitter");
	m_pButtonBox = new QHBox(this,"button_box");
	KviThemedLabel * l = new KviThemedLabel(m_pButtonBox,"dummy_label");
	m_pButtonBox->setStretchFactor(l,1);
	QToolTip::add(l,__tr2qs("What's this?"));

	// The button box on the right
	//QHBox * box = new QHBox(m_pTopSplitter,"button_box");

	m_pListViewButton = createToolButton(m_pButtonBox,"list_view_button",KVI_SMALLICON_HIDELISTVIEW,KVI_SMALLICON_SHOWLISTVIEW,__tr2qs("Show User List"),true);
	connect(m_pListViewButton,SIGNAL(clicked()),this,SLOT(toggleListView()));

#ifdef COMPILE_CRYPT_SUPPORT
	createCryptControllerButton(m_pButtonBox);
#endif

	createTextEncodingButton(m_pButtonBox);

	m_pSplitter = new QSplitter(QSplitter::Horizontal,this,"main_splitter");
	m_pIrcView = new KviIrcView(m_pSplitter,lpFrm,this);
	connect(m_pIrcView,SIGNAL(rightClicked()),this,SLOT(textViewRightClicked()));
	m_pUserListView = new KviUserListView(m_pSplitter,m_pConsole->userDataBase(),this,7,__tr("Query Targets"),"user_list_view");

	m_pInput   = new KviInput(this,m_pUserListView);

	if(KVI_OPTION_BOOL(KviOption_boolAutoLogQueries))m_pIrcView->startLogging(0);
// FIXME: #warning "Maybe tell the user all that we know about the remote end(s)....channels..."

	updateTargets();
}

KviQuery::~KviQuery()
{
	removeAllTargets();
	m_pConsole->unregisterQuery(this);
}

void KviQuery::triggerCreationEvents()
{
	TRIGGER_EVENT(KviEvent_OnQueryWindowCreated,this);
}

void KviQuery::getBaseLogFileName(KviStr &buffer)
{
	if (console()->connection())
		buffer.sprintf("%s.%s",name(),console()->currentNetworkName());
	else
		buffer.sprintf("%s.%u",name(),console()->ircContextId());
}

void KviQuery::textViewRightClicked()
{
	TRIGGER_EVENT(KviEvent_OnQueryPopupRequest,this);
}

void KviQuery::saveProperties(KviConfig *cfg)
{
	KviWindow::saveProperties(cfg);
	cfg->writeEntry("Splitter",m_pSplitter->sizes());
	cfg->writeEntry("UserListViewVisible",m_pUserListView->isVisible());
}

void KviQuery::loadProperties(KviConfig *cfg)
{
	int w = width();
	KviWindow::loadProperties(cfg);
	QValueList<int> def;
	def.append((w * 80) / 100);
	def.append((w * 20) / 100);
	m_pSplitter->setSizes(cfg->readIntListEntry("Splitter",def));
	showListView(cfg->readBoolEntry("UserListViewVisible",false));
}

void KviQuery::userAction(const char * nick,unsigned int uActionType)
{
	int iTemperature = kvi_getUserActionTemperature(uActionType);
	m_pUserListView->userAction(nick,iTemperature);
	if(KVI_OPTION_BOOL(KviOption_boolFlashQueryWindowOnNewMessages))
	{
		demandAttention();
	}
}

void KviQuery::userAction(KviIrcMask *user,unsigned int uActionType)
{
	int iTemperature = kvi_getUserActionTemperature(uActionType);
	m_pUserListView->userAction(user,iTemperature);
	if(KVI_OPTION_BOOL(KviOption_boolFlashQueryWindowOnNewMessages))
	{
		demandAttention();
	}
}


KviUserListEntry * KviQuery::addTarget(const char *nick,const char * user,const char * host)
{
	KviUserListEntry * e = m_pUserListView->join(nick,user,host);
	if((!e->globalData()->avatar()) && user)
		m_pConsole->checkDefaultAvatar(e->globalData(),nick,user,host);
	updateTargets();

	if(KVI_OPTION_BOOL(KviOption_boolEnableQueryTracing))
	{
		KviStr szChans;
		int iChans = m_pConsole->getCommonChannels(nick,szChans);
		notifyCommonChannels(nick,user,host,iChans,szChans);
	}

	TRIGGER_EVENT_3PARAM(KviEvent_OnQueryTargetAdded,this,nick,user,host);

	return e;
}

static const char * star = "*";

void KviQuery::notifyCommonChannels(const char * nick,const char * user,const char * host,int iChans,const char * chans)
{

	if(iChans > 0)
	{
		KviWStr szChannels(KviWStr::Decode,chans,textEncoding());
		output(KVI_OUT_QUERYTRACE,__tr2wc("Common channels for \r!n\r%s\r [%s@\r!h\r%s\r]: %W"),nick,user ? user : star,host ? host : star,&szChannels);
	} else {
		output(KVI_OUT_QUERYTRACE,__tr2wc("No common channels for \r!n\r%s\r [%s@\r!h\r%s\r]"),nick,user ? user : star,host ? host : star);
	}
}

bool KviQuery::removeTarget(const char * nick)
{
	// FIXME: OnQueryTargetRemoved ?
	bool bRet = m_pUserListView->part(nick);
	updateTargets();
	return bRet;
}

void KviQuery::removeAllTargets()
{
	m_pUserListView->partAll();
	updateTargets();
}

void KviQuery::updateTargets()
{
	m_szTargets = "";

	if(targetCount() > 0)
	{
		QAsciiDictIterator<KviUserListEntry> it(*(m_pUserListView->entryDict()));
		while(it.current())
		{
			if(m_szTargets.hasData())m_szTargets.append(',');
			m_szTargets.append(it.currentKey());
			++it;
		}
	}

	if(m_szTargets.hasData())setName(m_szTargets.ptr());
	else setName(__tr("[No targets]"));

	updateCaption();
}

void KviQuery::fillCaptionBuffers()
{
	static QString begin("<nobr><font color=\"");
	static QString boldbegin("\"><b>");
	static QString endofbold("</b></font> <font color=\"");
	static QString endoffont("\">");
	static QString end("</font></nobr>");

	if(!console()->connection())
	{
		QString dead = __tr2qs("[Dead query]");

		m_szPlainTextCaption = name();
		m_szPlainTextCaption += " : ";
		m_szPlainTextCaption += dead;

		m_szHtmlActiveCaption = begin;
		m_szHtmlActiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextActive).name();
		m_szHtmlActiveCaption += boldbegin;
		m_szHtmlActiveCaption += name();
		m_szHtmlActiveCaption += endofbold;
		m_szHtmlActiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextActive2).name();
		m_szHtmlActiveCaption += endoffont;
		m_szHtmlActiveCaption += dead;
		m_szHtmlActiveCaption += end;

		m_szHtmlInactiveCaption = begin;
		m_szHtmlInactiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextInactive).name();
		m_szHtmlInactiveCaption += boldbegin;
		m_szHtmlInactiveCaption += name();
		m_szHtmlInactiveCaption += endofbold;
		m_szHtmlInactiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextInactive2).name();
		m_szHtmlInactiveCaption += endoffont;
		m_szHtmlInactiveCaption += dead;
		m_szHtmlInactiveCaption += end;

		return;
	}

	QString szNickOnServer = QChar('[');
	szNickOnServer += m_pConsole->currentNickName();
	szNickOnServer += __tr2qs(" on ");
	szNickOnServer += m_pConsole->currentServerName();
	szNickOnServer += QChar(']');

	m_szPlainTextCaption = name();
	m_szPlainTextCaption += QChar(' ');
	m_szPlainTextCaption += szNickOnServer;

	m_szHtmlActiveCaption = begin;
	m_szHtmlActiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextActive).name();
	m_szHtmlActiveCaption += boldbegin;
	m_szHtmlActiveCaption += name();
	m_szHtmlActiveCaption += endofbold;
	m_szHtmlActiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextActive2).name();
	m_szHtmlActiveCaption += endoffont;
	m_szHtmlActiveCaption += szNickOnServer;
	m_szHtmlActiveCaption += end;

	m_szHtmlInactiveCaption = begin;
	m_szHtmlInactiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextInactive).name();
	m_szHtmlInactiveCaption += boldbegin;
	m_szHtmlInactiveCaption += name();
	m_szHtmlInactiveCaption += endofbold;
	m_szHtmlInactiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextInactive2).name();
	m_szHtmlInactiveCaption += endoffont;
	m_szHtmlInactiveCaption += szNickOnServer;
	m_szHtmlInactiveCaption += end;
}

bool KviQuery::nickChange(const char * oldNick,const char * newNick)
{
	bool bRet = m_pUserListView->nickChange(oldNick,newNick);
	updateTargets();
	return bRet;
}

void KviQuery::showListView(bool bShow)
{
	if(!bShow)
	{
		m_pUserListView->hide();
		if(m_pListViewButton->isOn())m_pListViewButton->setOn(false);
	} else {
		m_pUserListView->show();
		if(!(m_pListViewButton->isOn()))m_pListViewButton->setOn(true);
	}
}

void KviQuery::toggleListView()
{
	showListView(!m_pUserListView->isVisible());
}

void KviQuery::setDeadQuery()
{
	m_iFlags |= KVI_QUERY_FLAG_DEAD;

	m_pUserListView->enableUpdates(false);
	m_pUserListView->partAll();
	m_pUserListView->enableUpdates(true);
	m_pUserListView->setUserDataBase(0);
	// this should be moved to KviIrcContext!
	m_pConsole->setQueryAsDead(this);

	setType(KVI_WINDOW_TYPE_DEADQUERY);

	updateIcon();
	updateCaption();
};

void KviQuery::setAliveQuery()
{
	m_iFlags = 0;
	m_pUserListView->setUserDataBase(console()->connection()->userDataBase());
	setType(KVI_WINDOW_TYPE_QUERY);
	m_pConsole->setQueryAsAlive(this);
	// Update log file name
	if(m_pIrcView->isLogging())m_pIrcView->startLogging(0);
	updateIcon();
	updateCaption();
}

void KviQuery::applyOptions()
{
	m_pUserListView->applyOptions();
	// this applies options for IrcView and Input and forces the window to relayout
	KviWindow::applyOptions();
}

QPixmap * KviQuery::myIconPtr()
{
	return g_pIconManager->getSmallIcon(isDeadQuery() ? KVI_SMALLICON_DEADQUERY : KVI_SMALLICON_QUERY);
}

void KviQuery::resizeEvent(QResizeEvent *e)
{
	int hght = m_pInput->heightHint();
	int hght2 = m_pButtonBox->sizeHint().height();
	m_pButtonBox->setGeometry(0,0,width(),hght2);
	m_pSplitter->setGeometry(0,hght2,width(),height() - (hght + hght2));
	m_pInput->setGeometry(0,height() - hght,width(),hght);
}

QSize KviQuery::sizeHint() const
{
	QSize ret(m_pSplitter->sizeHint().width(),m_pIrcView->sizeHint().height() + m_pInput->heightHint());
	return ret;
}

void KviQuery::ownMessage(const char * buffer)
{
	if(targetCount() == 0)
	{
		outputNoFmt(KVI_OUT_SYSTEMWARNING,__tr2wc("This query has no active targets, no message sent"));
	} else {

#ifdef COMPILE_CRYPT_SUPPORT
		if(cryptSessionInfo())
		{
			if(cryptSessionInfo()->bDoEncrypt)
			{
				if(*buffer != KVI_TEXT_CRYPT)
				{
					KviStr encrypted;
					cryptSessionInfo()->pEngine->setMaxEncryptLen(500 - kvi_strLen(name()));
					if(cryptSessionInfo()->pEngine->encrypt(buffer,encrypted))
					{
						if(cryptSessionInfo()->pEngine->isCryptographicEngine())
						{
							m_pConsole->link()->sendFmtData("PRIVMSG %s :%c%s",name(),KVI_TEXT_CRYPT,encrypted.ptr());
							m_pConsole->outputPrivmsg(this,KVI_OUT_OWNPRIVMSGCRYPTED,0,0,0,buffer);
					} else {
							m_pConsole->link()->sendFmtData("PRIVMSG %s :%s",name(),encrypted.ptr());
							m_pConsole->outputPrivmsg(this,KVI_OUT_OWNPRIVMSG,0,0,0,encrypted.ptr());
						}
						//m_pConsole->outputPrivmsg(this,KVI_OUT_OWNPRIVMSG,0,0,0,encrypted.ptr());
						m_pUserListView->userAction(m_pConsole->currentNickName(),KVI_USERACTION_PRIVMSG);
						return;
					} else {
						output(KVI_OUT_SYSTEMERROR,
							__tr2wc("The crypto engine was unable to encrypt the current message (%s): %s, no data sent to the server"),
							buffer,cryptSessionInfo()->pEngine->lastError());
						return;
					}
				} else {
					buffer++; //eat the escape code
				}
			}
		}
#endif

		m_pConsole->link()->sendFmtData("PRIVMSG %s :%s",m_szTargets.ptr(),buffer);
		m_pConsole->outputPrivmsg(this,KVI_OUT_OWNPRIVMSG,0,0,0,buffer);
		m_pUserListView->userAction(m_pConsole->currentNickName(),KVI_USERACTION_PRIVMSG);
	}
}

void KviQuery::ownAction(const char * buffer)
{
	if(targetCount() == 0)
	{
		outputNoFmt(KVI_OUT_SYSTEMWARNING,__tr2wc("This query has no active targets, no message sent"));
	} else {
		KviWStr szWBuffer(KviWStr::Decode,buffer,textEncoding());
		m_pConsole->link()->sendFmtData("PRIVMSG %s :%cACTION %s%c",
			m_szTargets.ptr(),0x01,buffer,0x01);
		output(KVI_OUT_ACTION,"\r!n\r%s\r %W",m_pConsole->currentNickName(),&szWBuffer);
		m_pUserListView->userAction(m_pConsole->currentNickName(),KVI_USERACTION_ACTION);
	}
}

#include "kvi_query.moc"
