//
//   File : kvi_window.cpp
//   Creation date : Tue Jul 6 1999 14:52:11 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
#define __KVIRC__
#define KVI_WINDOW_MIN_WIDTH 100
#define KVI_WINDOW_MIN_HEIGHT 100

#define _KVI_WINDOW_CPP_

#define _KVI_DEBUG_CHECK_RANGE_

#include "kvi_debug.h"
#include "kvi_app.h"
#include "kvi_window.h"
#include "kvi_frame.h"
#include "kvi_taskbar.h"
#include "kvi_iconmanager.h"
#include "kvi_mdichild.h"
#include "kvi_locale.h"
#include "kvi_ircview.h"
#include "kvi_out.h"
#include "kvi_malloc.h"
#include "kvi_input.h"
#include "kvi_fileutils.h"
#include "kvi_options.h"
#include "kvi_wstring.h"
#include "kvi_config.h"
#include "kvi_irccontext.h"
#include "kvi_console.h"

#include <qpixmap.h>
#include <qpopupmenu.h>
#include <qcursor.h>
#include <qtimer.h>
#include <qsplitter.h>
#include <qmetaobject.h>
#include <qdatetime.h>
#include <qtextcodec.h>

#if QT_VERSION >= 0x030005
        #include <qobjectlist.h>
#else
        #include <qobjcoll.h>
#endif

#include <qvariant.h>
#include <qaccel.h>
#include <qtoolbutton.h>
#include <qtooltip.h>
#include <qmessagebox.h>

#ifdef COMPILE_CRYPT_SUPPORT
	#include "kvi_crypt.h"
	#include "kvi_cryptcontroller.h"
#endif

#ifdef COMPILE_KDE_SUPPORT
	#include <kwin.h>
	#include <kdeversion.h>
#endif

#ifdef COMPILE_ON_WINDOWS
	#include <windows.h>
#endif

static QPopupMenu * g_pMdiWindowSystemMainPopup = 0;
static QPopupMenu * g_pMdiWindowSystemTextEncodingPopup = 0;


unsigned int g_uUniqueWindowId = 1;

// FIXME: #warning "Mouse wheel suppport"

KviWindow::KviWindow(int type,KviFrame * lpFrm,const char * name,KviConsole * lpConsole)
		: QWidget(0,name)
{
	m_szId.setNum(g_uUniqueWindowId);
	g_uUniqueWindowId++;

	g_pApp->registerWindow(this);

	m_pFrm                  = lpFrm;
	m_iType                  = type;
	m_pFocusHandler         = 0;

	m_pIrcView              = 0;
	m_pInput                = 0;
	m_pSplitter             = 0;
	m_pButtonBox            = 0;
	m_pConsole              = lpConsole;
	m_pIrcContext           = lpConsole ? lpConsole->ircContext() : 0;
	m_pLastFocusedChild     = 0;
	setTextEncoding(KviApp::defaultTextEncoding());

#ifdef COMPILE_CRYPT_SUPPORT

	m_pCryptControllerButton = 0;
	m_pCryptController = 0;
	m_pCryptSessionInfo = 0;
#endif

	m_pAccel = 0;

	m_pTaskBarItem = 0;

	setMinimumSize(KVI_WINDOW_MIN_WIDTH,KVI_WINDOW_MIN_HEIGHT);
	setBackgroundMode(NoBackground);
	setFocusPolicy(StrongFocus);
	
	connect(g_pApp,SIGNAL(reloadImages()),this,SLOT(reloadImages()));
}

KviWindow::~KviWindow()
{
	destroyTaskBarItem();
	g_pApp->unregisterWindow(this);
	if(g_pApp->windowCount() == 0)
	{
		// this is the last window!
		if(g_pMdiWindowSystemMainPopup)
			delete g_pMdiWindowSystemMainPopup;
		if(g_pMdiWindowSystemTextEncodingPopup)
			delete g_pMdiWindowSystemTextEncodingPopup;
	}
#ifdef COMPILE_CRYPT_SUPPORT
	if(m_pCryptSessionInfo)
		KviCryptController::destroyCryptSessionInfo(&m_pCryptSessionInfo);
#endif
}

void KviWindow::reloadImages()
{
	updateIcon();
}

void KviWindow::demandAttention()
{
	if(mdiParent())
	{
		if(frame()->isActiveWindow())return;
#ifdef COMPILE_ON_WINDOWS
		FLASHWINFO fwi;
		fwi.cbSize = sizeof(fwi);
		fwi.hwnd = (HWND)(frame()->winId());
		fwi.dwFlags = FLASHW_TRAY | FLASHW_TIMERNOFG;
		fwi.uCount = 20;
		fwi.dwTimeout = 500;
		FlashWindowEx(&fwi);
#else
	#ifdef COMPILE_KDE_SUPPORT
		#if (KDE_VERSION_MAJOR >= 3) && (KDE_VERSION_MINOR >= 2)
			KWin::demandAttention(frame()->winId(),true);
		#endif
	#endif
#endif
	} else {
		if(isActiveWindow())return;
#ifdef COMPILE_ON_WINDOWS
		FLASHWINFO fwi;
		fwi.cbSize = sizeof(fwi);
		fwi.hwnd = (HWND)winId();
		fwi.dwFlags = FLASHW_TRAY | FLASHW_TIMERNOFG;
		fwi.uCount = 20;
		fwi.dwTimeout = 500;
		FlashWindowEx(&fwi);
#else
	#ifdef COMPILE_KDE_SUPPORT
		#if (KDE_VERSION_MAJOR >= 3) && (KDE_VERSION_MINOR >= 2)
			KWin::demandAttention(winId(),true);
		#endif
	#endif
#endif
	}
}

bool KviWindow::focusNextPrevChild(bool next)
{
	//debug("FOCUS NEXT PREV CHILD");
	QWidget * w = focusWidget();
	if(w)
	{
		if(w->focusPolicy() == QWidget::StrongFocus)return false;
		//QVariant v = w->property("KviProperty_FocusOwner");
		//if(v.isValid())return false; // Do NOT change the focus widget!
		
		if(w->parent())
		{
			QVariant v = w->parent()->property("KviProperty_ChildFocusOwner");
			if(v.isValid())return false; // Do NOT change the focus widget!
		}
	}

	return QWidget::focusNextPrevChild(next);
}

void KviWindow::setTextEncoding(unsigned int uTextEncoding)
{
	if(uTextEncoding >= KviWStr::numberOfEncodings())
		uTextEncoding = KviApp::defaultTextEncoding();
		
	m_uTextEncoding = uTextEncoding;
	m_pTextCodec = kvi_textEncodingCodec(uTextEncoding);
}

QCString KviWindow::encodeText(const QString &szText)
{
	if(!m_pTextCodec)return QCString();
	return m_pTextCodec->fromUnicode(szText);
}

QString KviWindow::decodeText(const char * szText)
{
	if(!m_pTextCodec)return QString();
	return m_pTextCodec->toUnicode(szText);
}


bool KviWindow::activityMeter(unsigned int *,unsigned int *)
{
	return false;
}


const char * KviWindow::m_typeTable[KVI_WINDOW_NUM_TYPES + 1]=
    {
        "console",
        "channel",
        "query",
        "help",
        "terminal",
        "editor",
        "dccchat",
        "dccsend",
        "socketspy",
        "links",
        "tool",
        "gnutella",
        "dirbrowser",
        "dcccanvas",
        "dccvoice",
        "list",
        "offer",
        "logview",
        "deadchannel",
        "deadquery",
        "scripteditor",
        "scriptobject",
        // <------ NEW TYPES GO HERE!
        "unknown"
    };

const char * KviWindow::typeString()
{
	if(m_iType < KVI_WINDOW_NUM_TYPES)
	{
		return m_typeTable[m_iType];
	}
	return m_typeTable[KVI_WINDOW_NUM_TYPES];
}

void KviWindow::setType(int iType)
{
	m_iType = iType;
}
/*
unsigned int KviWindow::getActivityValue()
{
	return KVI_ACTIVITY_NONE; // no activity
}
 
unsigned int KviWindow::getActivityType()
{
	return 0;
}
*/

void KviWindow::createTaskBarItem()
{
	if(m_pTaskBarItem)
		return;
	m_pTaskBarItem = m_pFrm->m_pTaskBar->addItem(this);
}

void KviWindow::destroyTaskBarItem()
{
	if(!m_pTaskBarItem)
		return;
	m_pFrm->m_pTaskBar->removeItem(m_pTaskBarItem);
	//	m_pTaskBarItem = 0; // actually the taskBarItem destructor sets it
}

QToolButton * KviWindow::createToolButton(QWidget * par,const char * nam,int pixon,int pixoff,const QString & tooltip,bool bOn)
{
	QToolButton * b = new QToolButton(par,nam);
	b->setToggleButton(true);
	b->setUsesBigPixmap(false);
#if QT_VERSION >= 300

	QIconSet is1;
	is1.setPixmap(*(g_pIconManager->getSmallIcon(pixon)),QIconSet::Small,QIconSet::Normal,QIconSet::On);
	is1.setPixmap(*(g_pIconManager->getSmallIcon(pixoff)),QIconSet::Small,QIconSet::Normal,QIconSet::Off);
	b->setIconSet(is1);
#else

	b->setOnIconSet(*(g_pIconManager->getSmallIcon(pixon)));
	b->setOffIconSet(*(g_pIconManager->getSmallIcon(pixoff)));
#endif

	QToolTip::add
		(b,tooltip);
	b->setOn(bOn);
	return b;
}

// This is always defined...
void KviWindow::createCryptControllerButton(QWidget * par)
{
#ifdef COMPILE_CRYPT_SUPPORT
	m_pCryptControllerButton = createToolButton(par,"crypt_controller_button",KVI_SMALLICON_UNLOCKEDOFF,KVI_SMALLICON_UNLOCKED,__tr2qs("Show Crypto Controller"),false);
	connect(m_pCryptControllerButton,SIGNAL(clicked()),this,SLOT(toggleCryptController()));
#endif // COMPILE_CRYPT_SUPPORT
}

void KviWindow::createTextEncodingButton(QWidget * par)
{
	m_pTextEncodingButton = createToolButton(par,"text_encoding_button",KVI_SMALLICON_TEXTENCODING,KVI_SMALLICON_TEXTENCODING,__tr2qs("Text Encoding"),false);
	connect(m_pTextEncodingButton,SIGNAL(clicked()),this,SLOT(textEncodingButtonClicked()));
}

void KviWindow::textEncodingButtonClicked()
{
	createSystemTextEncodingPopup();
	g_pMdiWindowSystemTextEncodingPopup->popup(m_pTextEncodingButton->mapToGlobal(QPoint(0,m_pTextEncodingButton->height())));
	m_pTextEncodingButton->setOn(false);
}

const QString & KviWindow::lastLineOfText()
{
	if(m_pIrcView)
		return m_pIrcView->lastLineOfText();
	return QString::null;
}

// The following three have to be here even if the crypt support is disabled...moc does not support conditional compilations
void KviWindow::toggleCryptController()
{
#ifdef COMPILE_CRYPT_SUPPORT
	if(m_pCryptController)
	{
		delete m_pCryptController;
		m_pCryptController = 0;
		if(!m_pCryptControllerButton)
			return;
		if(m_pCryptControllerButton->isOn())
			m_pCryptControllerButton->setOn(false);
	}
	else
	{
		if(m_pSplitter && m_pInput)
		{
			m_pCryptController = new KviCryptController(m_pSplitter,"crypt_controller",this,m_pCryptSessionInfo);
			connect(m_pCryptController,SIGNAL(done()),this,SLOT(cryptControllerFinished()));
			//setFocusHandlerNoClass(m_pInput,m_pCryptController,"QLineEdit"); //link it!
			m_pCryptController->show();
			if(!m_pCryptControllerButton)
				return;
			if(!(m_pCryptControllerButton->isOn()))
				m_pCryptControllerButton->setOn(true);
		}
	}
#endif // COMPILE_CRYPT_SUPPORT
}

void KviWindow::cryptControllerFinished()
{
#ifdef COMPILE_CRYPT_SUPPORT
	if(m_pCryptSessionInfo)
		KviCryptController::destroyCryptSessionInfo(&m_pCryptSessionInfo);
	m_pCryptSessionInfo = m_pCryptController->getNewSessionInfo();
	if(m_pCryptSessionInfo)
	{
		connect(m_pCryptSessionInfo->pEngine,SIGNAL(destroyed()),this,SLOT(cryptSessionInfoDestroyed()));
	}
	if(m_pCryptControllerButton)
	{
#if QT_VERSION >= 300
		QIconSet is;
		is.setPixmap(*(g_pIconManager->getSmallIcon(m_pCryptSessionInfo ? KVI_SMALLICON_LOCKEDOFF : KVI_SMALLICON_UNLOCKEDOFF)),QIconSet::Small,QIconSet::Normal,QIconSet::Off);
		is.setPixmap(*(g_pIconManager->getSmallIcon(m_pCryptSessionInfo ? KVI_SMALLICON_LOCKED : KVI_SMALLICON_UNLOCKED)),QIconSet::Small,QIconSet::Normal,QIconSet::On);
		m_pCryptControllerButton->setIconSet(is);
#else

		m_pCryptControllerButton->setOnIconSet(
			*(g_pIconManager->getSmallIcon(m_pCryptSessionInfo ? KVI_SMALLICON_LOCKEDOFF : KVI_SMALLICON_UNLOCKEDOFF)));
		m_pCryptControllerButton->setOffIconSet(
			*(g_pIconManager->getSmallIcon(m_pCryptSessionInfo ? KVI_SMALLICON_LOCKED : KVI_SMALLICON_UNLOCKED)));
#endif

		if(m_pCryptControllerButton->isOn())
			m_pCryptControllerButton->setOn(false);
	}
	delete m_pCryptController;
	m_pCryptController = 0;
#endif
}

void KviWindow::cryptSessionInfoDestroyed()
{
#ifdef COMPILE_CRYPT_SUPPORT
	output(KVI_OUT_SYSTEMERROR,__tr2wc("Ops...I've accidentally lost the crypting engine..."));
	m_pCryptSessionInfo->pEngine = 0;
	delete m_pCryptSessionInfo;
	m_pCryptSessionInfo = 0;
#endif
}



void KviWindow::setProgress(int progress)
{
	m_pTaskBarItem->setProgress(progress);
}

void KviWindow::listWindowTypes()
{
	outputNoFmt(KVI_OUT_SYSTEMMESSAGE,__tr2wc("List of window types available in this release of KVIrc:"));
	for(int i=0;i< KVI_WINDOW_NUM_TYPES;i++)
		outputNoFmt(KVI_OUT_SYSTEMMESSAGE,m_typeTable[i]);
}

void KviWindow::getConfigGroupName(KviStr &buf)
{
	buf = typeString();
}

void KviWindow::getDefaultLogFileName(KviStr &buffer)
{
	// FIXME: #warning "Make it configurable ?"
	QDate dt(QDate::currentDate());
	KviStr date(KviStr::Format,"%d.",dt.year());
	if(dt.month() < 10)
		date.append('0');
	date.append(KviStr::Format,"%d.",dt.month());
	if(dt.day() < 10)
		date.append('0');
	date.append(KviStr::Format,"%d",dt.day());
	KviStr base;
	kvi_encodeFileName(base);
	getBaseLogFileName(base);
	base.replaceAll(' ',"_");
	base.replaceAll(KVI_PATH_SEPARATOR_CHAR,"-");
	base.toLower();
	KviStr tmp(KviStr::Format,"%s_%s_%s.log",typeString(),base.ptr(),date.ptr());
	g_pApp->getLocalKvircDirectory(buffer,KviApp::Log,tmp.ptr());
}

void KviWindow::getBaseLogFileName(KviStr &buffer)
{
	buffer = name();
}

void KviWindow::saveProperties(KviConfig *cfg)
{
	cfg->writeEntry("TextEncoding",m_uTextEncoding);
}

void KviWindow::loadProperties(KviConfig *cfg)
{
	setTextEncoding(cfg->readUIntEntry("TextEncoding",KviApp::defaultTextEncoding()));
}

QPixmap * KviWindow::myIconPtr()
{
	return g_pIconManager->getSmallIcon(KVI_SMALLICON_DEFAULTICON);
}

void KviWindow::getTaskBarTipText(QString &buffer)
{
	buffer = m_szPlainTextCaption;
}

void KviWindow::setFixedCaption(const QString &szCaption)
{
	m_szPlainTextCaption = szCaption;
}

void KviWindow::fillCaptionBuffers()
{
	QString szCaption = m_szPlainTextCaption;
	if(szCaption.isEmpty())
		szCaption = name();

	fillSingleColorCaptionBuffers(szCaption);
}

void KviWindow::fillSingleColorCaptionBuffers(const QString &szName)
{
	static QString p1("<nobr><font color=\"");
	static QString p2("\"><b>");
	static QString p3("</b></font></nobr>");

	m_szPlainTextCaption = szName;

	m_szHtmlActiveCaption = p1;
	m_szHtmlActiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextActive).name();
	m_szHtmlActiveCaption += p2;
	m_szHtmlActiveCaption += szName;
	m_szHtmlActiveCaption += p3;

	m_szHtmlInactiveCaption = p1;
	m_szHtmlInactiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextInactive).name();
	m_szHtmlInactiveCaption += p2;
	m_szHtmlInactiveCaption += szName;
	m_szHtmlInactiveCaption += p3;
}

void KviWindow::updateCaption()
{
	fillCaptionBuffers();
	if(mdiParent())
		mdiParent()->setCaption(plainTextCaption(),htmlActiveCaption(),htmlInactiveCaption());
	else
		setCaption(plainTextCaption());
	if(m_pTaskBarItem)m_pTaskBarItem->captionChanged();
	if(mdiParent() && isMaximized() && (m_pFrm->activeWindow() == this))
		m_pFrm->updateCaption();
}

void KviWindow::createSystemTextEncodingPopup()
{
	if(!g_pMdiWindowSystemTextEncodingPopup)
		g_pMdiWindowSystemTextEncodingPopup = new QPopupMenu();
	else
	{
		g_pMdiWindowSystemTextEncodingPopup->clear();
		g_pMdiWindowSystemTextEncodingPopup->disconnect();
	}

	for(unsigned int u = 0;u < KviWStr::numberOfEncodings();u++)
	{
		KviStr tmp(KviStr::Format,"%s (%s)",KviWStr::encodingName(u),KviWStr::encodingNameAlias(u));
		int id = g_pMdiWindowSystemTextEncodingPopup->insertItem(tmp.ptr());
		if(u == m_uTextEncoding)
			g_pMdiWindowSystemTextEncodingPopup->setItemChecked(id,true);
		g_pMdiWindowSystemTextEncodingPopup->setItemParameter(id,u);
	}

	connect(g_pMdiWindowSystemTextEncodingPopup,SIGNAL(activated(int)),this,SLOT(systemTextEncodingPopupActivated(int)));

}

void KviWindow::systemPopupRequest(const QPoint &pnt)
{
	if(!g_pMdiWindowSystemMainPopup)
		g_pMdiWindowSystemMainPopup = new QPopupMenu();
	else
	{
		g_pMdiWindowSystemMainPopup->clear();
		g_pMdiWindowSystemMainPopup->disconnect();
	}

	createSystemTextEncodingPopup();

	if(mdiParent())
		g_pMdiWindowSystemMainPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_UNDOCK)),
		                                        __tr2qs("&Undock"),this,SLOT(undock()));
	else
		g_pMdiWindowSystemMainPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_DOCK)),
		                                        __tr2qs("&Dock"),this,SLOT(dock()));

	g_pMdiWindowSystemMainPopup->insertSeparator();

	int id = g_pMdiWindowSystemMainPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_MINIMIZE)),
		                                        __tr2qs("Mi&nimize"),this,SLOT(minimize()));
	g_pMdiWindowSystemMainPopup->setItemEnabled(id,!isMinimized());
	id = g_pMdiWindowSystemMainPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_MAXIMIZE)),
		                                        __tr2qs("Ma&ximize"),this,SLOT(maximize()));
	g_pMdiWindowSystemMainPopup->setItemEnabled(id,!isMaximized());
	id = g_pMdiWindowSystemMainPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_RESTORE)),
		                                        __tr2qs("&Restore"),this,SLOT(restore()));
	g_pMdiWindowSystemMainPopup->setItemEnabled(id,isMinimized()||isMaximized());

	g_pMdiWindowSystemMainPopup->insertSeparator();

	g_pMdiWindowSystemMainPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_CLOSE)),
	                                        __tr2qs("Close"),this,SLOT(close()));

	g_pMdiWindowSystemMainPopup->insertSeparator();

	g_pMdiWindowSystemMainPopup->insertItem(__tr2qs("Text &Encoding"),g_pMdiWindowSystemTextEncodingPopup);


	g_pMdiWindowSystemMainPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_XY)),
	                                        __tr2qs("Sa&ve Window Properties"),this,SLOT(savePropertiesAsDefault()));

	fillContextPopup(g_pMdiWindowSystemMainPopup);

	g_pMdiWindowSystemMainPopup->popup(pnt);
}

void KviWindow::systemTextEncodingPopupActivated(int id)
{
	if(!g_pMdiWindowSystemTextEncodingPopup)
		return;
	setTextEncoding(g_pMdiWindowSystemTextEncodingPopup->itemParameter(id));
}

void KviWindow::savePropertiesAsDefault()
{
	KviStr group;
	getConfigGroupName(group);

	if(!kvi_strEqualCI(group.ptr(),typeString()))
	{
		// save also the settings for THIS specialized window
		m_pFrm->saveWindowProperties(this,group.ptr());
	}

	m_pFrm->saveWindowProperties(this,typeString());
}

void KviWindow::contextPopup()
{
	systemPopupRequest(QCursor::pos());
}

void KviWindow::fillContextPopup(QPopupMenu *)
{
	// nothing here
}

void KviWindow::undock()
{
	m_pFrm->undockWindow(this);
}

void KviWindow::dock()
{
	m_pFrm->dockWindow(this);
}

void KviWindow::delayedAutoRaise()
{
	QTimer::singleShot(0,this,SLOT(autoRaise()));
}

void KviWindow::autoRaise()
{
	if(!mdiParent())
	{
		raise();
		setActiveWindow();
	}
	if(m_pFocusHandler)
		m_pFocusHandler->setFocus();
	else
		setFocus();
}

void KviWindow::delayedClose()
{
	QTimer::singleShot(0,this,SLOT(close()));
}

void KviWindow::closeEvent(QCloseEvent *e)
{
	e->ignore();
	m_pFrm->childWindowCloseRequest(this);
}

void KviWindow::updateIcon()
{
	if(parent())
	{
		((KviMdiChild *)parent())->setIcon(*myIconPtr());
	} else {
		setIcon(*myIconPtr());
	}
}

void KviWindow::youAreDocked()
{
	if(m_pAccel)
	{
		delete m_pAccel;
		m_pAccel = 0;
	}
	((KviMdiChild *)parent())->setIcon(*myIconPtr());
	updateCaption();
	connect(((KviMdiChild *)parent()),SIGNAL(systemPopupRequest(const QPoint &)),this,SLOT(systemPopupRequest(const QPoint &)));
}

void KviWindow::youAreUndocked()
{
	m_pAccel = m_pFrm->installAccelerators(this);
	setIcon(*myIconPtr());
	updateCaption();
}

#ifdef FocusIn
// Hack for X.h
#undef FocusIn
#endif

void KviWindow::activateSelf()
{
	if(mdiParent())
		mdiParent()->activate(false);
	m_pFrm->childWindowActivated(this);
	// this is now done by KviFrame in childWindowActivated
	//m_pFrm->m_pTaskBar->setActiveItem(m_pTaskBarItem);
}


void KviWindow::focusInEvent(QFocusEvent *)
{
	activateSelf();
	if(m_pFocusHandler)
	{
		if(m_pLastFocusedChild)
		{
			if(m_pLastFocusedChild->hasFocus() && m_pLastFocusedChild->isVisible())
				return; // the last focused child still has focus
		}
		m_pFocusHandler->setFocus();
	}
}

bool KviWindow::eventFilter(QObject *o,QEvent *e)
{
	switch(e->type())
	{
	case QEvent::FocusIn:
		//debug("FILTER FOCUS IN %d",o);
		m_pLastFocusedChild = (QWidget *)o;
		activateSelf();
		break;
	case QEvent::Enter:
		// this is a handler moved here from KviMdiChild::eventFilter
		if(QApplication::overrideCursor())
			QApplication::restoreOverrideCursor();
		break;
	case QEvent::MouseButtonPress:
		//debug("MOUSE BUTTON PRESS %d",o);
		if( (((QWidget *)o)->focusPolicy() == QWidget::NoFocus) ||
		        (((QWidget *)o)->focusPolicy() == QWidget::TabFocus))
		{
			// this will not focus our window
			// set the focus to the focus handler
			if(m_pLastFocusedChild)
			{
				//debug("ACCESSING LAST FOCUSED CHILD %d",m_pLastFocusedChild);
				if(m_pLastFocusedChild->hasFocus() && m_pLastFocusedChild->isVisible())
					return false;
			}

			if(m_pFocusHandler)
			{
				m_pFocusHandler->setFocus();
			} else {
				setFocus(); // we grab the focus (someone must do it , damn :D)
			}
		}
		break;
	case QEvent::ChildInserted:
		//debug("CHILD OF CHILD %d INSERTED",o);
		if(((QChildEvent *)e)->child()->isWidgetType())
			childInserted((QWidget *)((QChildEvent *)e)->child());
		break;
	case QEvent::ChildRemoved:
		//debug("CHILD OF CHILD %d REMOVED",o);
		if(((QChildEvent *)e)->child()->isWidgetType())
			childRemoved((QWidget *)((QChildEvent *)e)->child());
		break;
	}
	return false;
}


void KviWindow::childInserted(QWidget * o)
{
	//debug("CHILD INSERTED %d",o);
	o->removeEventFilter(this); // ensure that we don't filter twice
	o->installEventFilter(this); // we filter its events
	connect(o,SIGNAL(destroyed()),this,SLOT(childDestroyed()));

	if(o->inherits("KviInput"))
		m_pFocusHandler = o;
	else
	{
		if(!m_pFocusHandler && (o->focusPolicy() == QWidget::StrongFocus))
		{
			m_pFocusHandler = o;
		}
	}

	QObjectList *list = (QObjectList *)(o->children());
	if(list)
	{
		for(QObject * c = list->first();c;c = list->next())
		{
			if(c->isWidgetType())
				childInserted((QWidget *)c);
		}
	}
}

void KviWindow::childDestroyed()
{
	QWidget * s = (QWidget *)sender();
	childRemoved(s);
}

void KviWindow::childRemoved(QWidget * o)
{
	//debug("CHILD REMOVED %d",o);
	o->removeEventFilter(this);
	if(o == m_pFocusHandler)
		m_pFocusHandler = 0;
	if(o == m_pLastFocusedChild)
		m_pLastFocusedChild = 0;

	QObjectList *list = (QObjectList *)(o->children());
	if(list)
	{
		for(QObject * c = list->first();c;c = list->next())
		{
			if(c->isWidgetType())
				childRemoved((QWidget *)c);
		}
	} //else debug("The removed object has no children");

}

void KviWindow::childEvent(QChildEvent *e)
{
	if(e->child()->isWidgetType())
	{
		if(e->removed())
			childRemoved((QWidget *)(e->child()));
		else
			childInserted((QWidget *)(e->child()));
	}
	QWidget::childEvent(e);
}

void KviWindow::wheelEvent(QWheelEvent *e)
{
	/* NOTHING HERE FOR NOW (FIXME) */
}


void KviWindow::childrenTreeChanged(QWidget * widgetAdded)
{
	//	if(widgetAdded && m_pFocusHandler)setFocusHandler(m_pFocusHandler,widgetAdded);
	// FIXME: This might be useless
	QResizeEvent * e = new QResizeEvent(size(),size());
	resizeEvent(e);
	delete e;
}


void KviWindow::updateBackgrounds(QObject * obj)
{
	if(!obj)
		obj = this;
	QObjectList * list = (QObjectList *)(obj->children());
	if(list)
	{
		for(unsigned int i=0;i<list->count();i++)
		{
			QObject * child = list->at(i);
#if QT_VERSION >= 300
			// FIXME: check if this code can work with qt < 3.0.0 too
			if(child->metaObject()->findProperty("TransparencyCapable",true) != -1)
				((QWidget *)child)->update();
#else

			QVariant v = list->at(i)->property("TransparencyCapable");
			if(v.isValid())
				((QWidget *)child)->update();
#endif

			updateBackgrounds(child);
		}
	}
}

void KviWindow::moveEvent(QMoveEvent *e)
{
#ifdef COMPILE_PSEUDO_TRANSPARENCY
	updateBackgrounds();
#endif

	QWidget::moveEvent(e);
}

void KviWindow::minimize()
{
	if(mdiParent())
	{
		if(!isMinimized())
			mdiParent()->minimize();
	}
	else
		showMinimized();
}

void KviWindow::maximize()
{
	if(mdiParent())
	{
		if(!isMaximized())
			mdiParent()->maximize();
	}
	else
		showMaximized();
	autoRaise();
}

bool KviWindow::isMinimized()
{
	if(mdiParent())
		return (mdiParent()->state() == KviMdiChild::Minimized);
	else
		return QWidget::isMinimized();
}

bool KviWindow::isMaximized()
{
	if(mdiParent())
		return (mdiParent()->state() == KviMdiChild::Maximized);
	// Heh...how to check it ?
	// Empirical check
	int wdth = (g_pApp->desktop()->width() * 75) / 100;
	int hght = (g_pApp->desktop()->height() * 75) / 100;

	return ((x() <= 1)&&(y() <= 1)&&(width() >= wdth)&&(height() >= hght));
}

void KviWindow::restore()
{
	if(mdiParent())
	{
		if(isMinimized()||isMaximized())
			mdiParent()->restore();
	}
	else
		showNormal();
	autoRaise();
}

QRect KviWindow::externalGeometry()
{
	return mdiParent() ? mdiParent()->restoredGeometry() : frameGeometry();
}

void KviWindow::applyOptions()
{
	updateCaption();
	if(m_pIrcView)m_pIrcView->applyOptions();
	if(m_pInput)m_pInput->applyOptions();

	// trick: relayout
	resize(width() - 1,height() - 1);
	resize(width() + 1,height() + 1);
}

KviWindow * KviWindow::outputProxy()
{
	return 0;
}

void KviWindow::output(int msg_type,const char *format,...)
{
	kvi_wchar_t txt_ptr[512]; //It should be enough for all outputs...
	kvi_va_list list;
	kvi_va_start(list,format);
	if(kvi_wvsnprintcf(txt_ptr,512,format,list) < 0)
	{
		//Just in case...
		kvi_va_end(list);
		int len = 512;
		kvi_wchar_t *long_txt_ptr = 0;
		int result;
		do
		{
			len += 512;
			//first time long_txt_ptr == 0 so it is equivalent to malloc
			//At least the man page says that...
			long_txt_ptr = (kvi_wchar_t *)kvi_realloc((void *)long_txt_ptr,len * sizeof(kvi_wchar_t));
			kvi_va_start(list,format);
			result = kvi_wvsnprintcf(long_txt_ptr,len,format,list);
			kvi_va_end(list);
		}
		while(result < 0);
		if(m_pIrcView)
			m_pIrcView->appendText(msg_type,long_txt_ptr);
		else
		{
			// Redirect to the output proxy
			KviWindow *wnd = outputProxy();
			if(wnd)
				wnd->outputNoFmt(msg_type,long_txt_ptr);
		}
		kvi_free((void *)long_txt_ptr);
	}
	else
	{
		//Succesful vsnprintf
		kvi_va_end(list);
		if(m_pIrcView)
			m_pIrcView->appendText(msg_type,txt_ptr);
		else
		{
			// Redirect to the output proxy
			KviWindow *wnd = outputProxy();
			if(wnd)
				wnd->outputNoFmt(msg_type,txt_ptr);
		}
	}
	if(doHighlight(msg_type) && m_pTaskBarItem)
		m_pTaskBarItem->highlight(KVI_OPTION_MSGTYPE(msg_type).level());
}

void KviWindow::output(int msg_type,const QString &szFmt,...)
{
	kvi_va_list l;
	kvi_va_start_by_reference(l,szFmt);

	QString szBuf;
	KviQString::vsprintf(szBuf,szFmt,l);

	kvi_va_end(l);

	const QChar * pC = KviQString::nullTerminatedArray(szBuf);
	if(!pC)return;

	if(m_pIrcView)m_pIrcView->appendText(msg_type,(kvi_wchar_t *)pC);
	else {
		// Redirect to the output proxy
		KviWindow *wnd = outputProxy();
		if(wnd)wnd->outputNoFmt(msg_type,(kvi_wchar_t *)(szBuf.ucs2()));
	}

	if (doHighlight(msg_type) && m_pTaskBarItem)
		m_pTaskBarItem->highlight(KVI_OPTION_MSGTYPE(msg_type).level());
}

void KviWindow::output(int msg_type,const kvi_wchar_t *format,...)
{
	kvi_wchar_t txt_ptr[512]; //It should be enough for all outputs...
	kvi_va_list list;
	kvi_va_start(list,format);
	if(kvi_wvsnprintf(txt_ptr,512,format,list) < 0)
	{
		//Just in case...
		kvi_va_end(list);
		int len = 512;
		kvi_wchar_t *long_txt_ptr = 0;
		int result;
		do
		{
			len += 512;
			//first time long_txt_ptr == 0 so it is equivalent to malloc
			//At least the man page says that...
			long_txt_ptr = (kvi_wchar_t *)kvi_realloc((void *)long_txt_ptr,len * sizeof(kvi_wchar_t));
			kvi_va_start(list,format);
			result = kvi_wvsnprintf(long_txt_ptr,len,format,list);
			kvi_va_end(list);
		}
		while(result < 0);
		if(m_pIrcView)
			m_pIrcView->appendText(msg_type,long_txt_ptr);
		else
		{
			// Redirect to the output proxy
			KviWindow *wnd = outputProxy();
			if(wnd)
				wnd->outputNoFmt(msg_type,long_txt_ptr);
		}
		kvi_free((void *)long_txt_ptr);
	}
	else
	{
		//Succesful vsnprintf
		kvi_va_end(list);
		if(m_pIrcView)
			m_pIrcView->appendText(msg_type,txt_ptr);
		else
		{
			// Redirect to the output proxy
			KviWindow *wnd = outputProxy();
			if(wnd)
				wnd->outputNoFmt(msg_type,txt_ptr);
		}
	}

	if(doHighlight(msg_type) && m_pTaskBarItem)
		m_pTaskBarItem->highlight(KVI_OPTION_MSGTYPE(msg_type).level());
}

//================ outputNoFmt ================//

void KviWindow::outputNoFmt(int msg_type,const char * text)
{
	__range_valid(text);
	if(m_pIrcView)
		m_pIrcView->appendText(msg_type,text);
	else
	{
		// Redirect to the output proxy
		KviWindow *wnd = outputProxy();
		if(wnd)
			wnd->outputNoFmt(msg_type,text);
	}

	if (doHighlight(msg_type) && m_pTaskBarItem)
		m_pTaskBarItem->highlight(KVI_OPTION_MSGTYPE(msg_type).level());

	//  QMessageBox::information(0,"msg type2",QString().setNum(msg_type));
}

void KviWindow::outputNoFmt(int msg_type,const kvi_wchar_t * text)
{
	__range_valid(text);
	if(m_pIrcView)
		m_pIrcView->appendText(msg_type,text);
	else
	{
		// Redirect to the output proxy
		KviWindow *wnd = outputProxy();
		if(wnd)
			wnd->outputNoFmt(msg_type,text);
	}

	if (doHighlight(msg_type) && m_pTaskBarItem)
		m_pTaskBarItem->highlight(KVI_OPTION_MSGTYPE(msg_type).level());

	//   QMessageBox::information(0,"msg type3",QString().setNum(msg_type));
}

void KviWindow::outputNoFmt(int msg_type,const QString &szText)
{
	const QChar * pC = KviQString::nullTerminatedArray(szText);
	if(!pC)return;

	if(m_pIrcView)m_pIrcView->appendText(msg_type,(kvi_wchar_t *)pC);
	else {
		// Redirect to the output proxy
		KviWindow *wnd = outputProxy();
		if(wnd)wnd->outputNoFmt(msg_type,(kvi_wchar_t *)(szText.ucs2()));
	}

	if (doHighlight(msg_type) && m_pTaskBarItem)
		m_pTaskBarItem->highlight(KVI_OPTION_MSGTYPE(msg_type).level());
}

bool KviWindow::doHighlight(int msg_type)
{
	// normally we highlight
	bool highlight=true;

	// if this option is checked we dont highlight other than channel msg
	if (KVI_OPTION_BOOL(KviOption_boolHighlightOnlyNormalMsg))
	{
		if (msg_type != KVI_OUT_CHANPRIVMSG && msg_type != KVI_OUT_CHANPRIVMSGCRYPTED)
		{
			highlight=false;

			// look for exeptions
			// queries too
			if (KVI_OPTION_BOOL(KviOption_boolHighlightOnlyNormalMsgQueryToo) &&
			        ( msg_type == KVI_OUT_QUERYPRIVMSG
			          || msg_type == KVI_OUT_QUERYTRACE
			          || msg_type == KVI_OUT_QUERYPRIVMSGCRYPTED
			          || msg_type == KVI_OUT_QUERYNOTICE
			          || msg_type == KVI_OUT_QUERYNOTICECRYPTED ))
			{
				highlight=true;
			}

			// highlight msgs too (user definied words to highlight or nickname)
			if (KVI_OPTION_BOOL(KviOption_boolHighlightOnlyNormalMsgHighlightInChanToo) &&
			        msg_type == KVI_OUT_HIGHLIGHT)
			{
				highlight=true;
			}

		}
	}

	if(KVI_OPTION_BOOL(KviOption_boolHighlightOnlyAtCostumHighlightLevel) &&
			(KVI_OPTION_MSGTYPE(msg_type).level() >= KVI_OPTION_UINT(KviOption_uintMinHighlightLevel)))
	{
		highlight=true;
	}


	return highlight;
}

#include "kvi_window.moc"
