//=============================================================================
//
//   File : kvi_scripttoolbar_definition.cpp
//   Creation date : Fri Nov 08 2002 02:22:12 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __KVIRC__

#include "kvi_scripttoolbar_definition.h"
#include "kvi_scripttoolbar_manager.h"

#ifdef COMPILE_SCRIPTTOOLBAR

	#include "kvi_scripttoolbar.h"
	
	#include "kvi_frame.h"
	#include "kvi_config.h"
	#include "kvi_app.h"
	#include "kvi_cmdformatter.h"
	
	KviScriptToolBarDefinition::KviScriptToolBarDefinition(KviScriptToolBarManager * pManager)
	{
		m_iId = KviApp::getGloballyUniqueId();

		m_pManager = pManager;

		m_bVisible = true;

		m_pItems = new QDict<KviScriptToolBarItemDefinition>(17,false);
		m_pItems->setAutoDelete(true);
		
		m_pItemList = new KviPtrList<KviScriptToolBarItemDefinition>();
		m_pItemList->setAutoDelete(false);
		
		m_pToolBars = new KviPtrList<KviScriptToolBar>();
		m_pToolBars->setAutoDelete(false);
	}
	
	KviScriptToolBarDefinition::~KviScriptToolBarDefinition()
	{
		delete m_pItems;
		delete m_pItemList;
		while(!m_pToolBars->isEmpty())delete m_pToolBars->first();
		delete m_pToolBars;
	}
	
	void KviScriptToolBarDefinition::load(const char * szPrefix,KviConfig * cfg)
	{
		KviStr tmp;
		
		tmp.sprintf("%s_Name",szPrefix);
		m_szName = cfg->readQStringEntry(tmp.ptr(),"");
	
		tmp.sprintf("%s_Label",szPrefix);
		m_szLabel = cfg->readQStringEntry(tmp.ptr(),"");
	
		tmp.sprintf("%s_Visible",szPrefix);
		m_bVisible = cfg->readBoolEntry(tmp.ptr(),true);
	
		tmp.sprintf("%s_Count",szPrefix);
		unsigned int cnt = cfg->readUIntEntry(tmp.ptr(),0);
		for(unsigned int i=0;i<cnt;i++)
		{
			tmp.sprintf("%s_%d",szPrefix,i);
			KviScriptToolBarItemDefinition * d = new KviScriptToolBarItemDefinition();
			d->load(tmp.ptr(),cfg);
			if(d->name().isEmpty())delete d;
			else {
				KviScriptToolBarItemDefinition * old = m_pItems->find(d->name());
				if(old)m_pItemList->removeRef(old);
				m_pItems->replace(d->name(),d);
				m_pItemList->append(d);
			}
		}
	}
	
	void KviScriptToolBarDefinition::save(const char * szPrefix,KviConfig * cfg)
	{
		KviStr tmp;
		
		tmp.sprintf("%s_Name",szPrefix);
		cfg->writeEntry(tmp.ptr(),m_szName);
	
		tmp.sprintf("%s_Label",szPrefix);
		cfg->writeEntry(tmp.ptr(),m_szLabel);
		
		tmp.sprintf("%s_Visible",szPrefix);
		cfg->writeEntry(tmp.ptr(),m_bVisible); // gets visible if there are active instances
		
		tmp.sprintf("%s_Count",szPrefix);
		cfg->writeEntry(tmp.ptr(),m_pItems->count());
	
		unsigned int i = 0;
		for(KviScriptToolBarItemDefinition * d = m_pItemList->first();d;d = m_pItemList->next())
		{
			tmp.sprintf("%s_%d",szPrefix,i);
			d->save(tmp.ptr(),cfg);
			i++;
		}
	}

	void KviScriptToolBarDefinition::generateKvsDefinition(QString &szBuffer)
	{
		KviStr szCode;

		for(KviScriptToolBarItemDefinition * d = m_pItemList->first();d;d = m_pItemList->next())
		{
			QString szItem;
			d->generateKvsDefinition(szItem);
			szCode.append(szItem);
			szCode.append('\n');
		}

		KviCommandFormatter::blockFromBuffer(szCode);

		szBuffer = "toolbar.define(\"";
		szBuffer += m_szName;
		szBuffer += "\",\"";
		szBuffer += m_szLabel;
		szBuffer += "\")\n";
		QString bla(szCode);
		szBuffer += bla;
		szBuffer += "\n";
	}
	
	void KviScriptToolBarDefinition::setLabel(const QString &szLabel)
	{
		m_szLabel = szLabel;
		// update the existing bars!
		for(KviScriptToolBar * bar = m_pToolBars->first();bar;bar = m_pToolBars->next())
		{
			bar->setLabel(m_szLabel);
		}
	}
	
	void KviScriptToolBarDefinition::attach(KviScriptToolBar * pTb)
	{
		m_pToolBars->append(pTb);
	}
	
	void KviScriptToolBarDefinition::detach(KviScriptToolBar * pTb)
	{
		m_pToolBars->removeRef(pTb);
	}

	void KviScriptToolBarDefinition::storeVisibilityState()
	{
		// this is called as a "checkpoint" by KviFrame before closing down
		// and before saving settings
		m_bVisible = m_pToolBars->count() > 0; // will get initially visible if we have actual instances
	}


	KviScriptToolBar * KviScriptToolBarDefinition::existingInstance(KviFrame * pFrm)
	{
		for(KviScriptToolBar * pBar = m_pToolBars->first();pBar;pBar = m_pToolBars->next())
		{
			if(pBar->frame() == pFrm)return pBar;
		}
		return 0;
	}
	
	KviScriptToolBar * KviScriptToolBarDefinition::instance(KviFrame * pFrm)
	{
		KviScriptToolBar * pBar = existingInstance(pFrm);
		if(pBar)return pBar;
		return createInstance(pFrm);
	}
	
	KviScriptToolBar * KviScriptToolBarDefinition::createInstance(KviFrame * pFrm)
	{
		KviScriptToolBar * pBar = new KviScriptToolBar(pFrm,this,m_szLabel);

		for(KviScriptToolBarItemDefinition * d = m_pItemList->first();d;d = m_pItemList->next())
		{
			d->add(pBar);
		}

		pBar->show();
		return pBar;
	}

	KviScriptToolBarItemDefinition * KviScriptToolBarDefinition::getItemDefinition(const QString & szName,KviScriptToolBarItemDefinition::Type t)
	{
		KviScriptToolBarItemDefinition * d = findItemDefinition(szName);
		if(!d)
		{
			d = new KviScriptToolBarItemDefinition();
			d->setName(szName);
			d->setType(t);
			KviScriptToolBarItemDefinition * old = m_pItems->find(szName);
			if(old)m_pItemList->removeRef(old);
			m_pItems->replace(szName,d);
			m_pItemList->append(d);
			
			// add to all the existing toolbars
			for(KviScriptToolBar * pBar = m_pToolBars->first();pBar;pBar = m_pToolBars->next())
			{
				d->add(pBar);
			}
		}
		return d;
	}

	void KviScriptToolBarDefinition::clear()
	{
		m_pItems->clear();
		m_pItemList->clear();
		for(KviScriptToolBar * pBar = m_pToolBars->first();pBar;pBar = m_pToolBars->next())
		{
			pBar->clear();
		}
	}



	void KviScriptToolBarDefinition::copyFrom(KviScriptToolBarDefinition * d)
	{
		// clear the existing toolbars
		KviScriptToolBar * pBar;
		for(pBar = m_pToolBars->first();pBar;pBar = m_pToolBars->next())
		{
			pBar->clear();
		}

		m_pItems->clear();
		m_pItemList->clear();
		
		m_szLabel = d->label();
		m_bVisible = d->isVisible();

		for(KviScriptToolBarItemDefinition * i = d->m_pItemList->first();i;i = d->m_pItemList->next())
		{
			KviScriptToolBarItemDefinition * x = new KviScriptToolBarItemDefinition();
			x->copyFrom(i);
			m_pItems->replace(i->name(),x);
			m_pItemList->append(x);
		}
		
		// re-add
		for(KviScriptToolBarItemDefinition * def = m_pItemList->first();def;def = m_pItemList->next())
		{
			for(pBar = m_pToolBars->first();pBar;pBar = m_pToolBars->next())
			{
				def->add(pBar);
			}
		}
	}

	void KviScriptToolBarDefinition::rename(const QString &szNewName)
	{
		m_pManager->rename(this,szNewName);
	}

#endif // COMPILE_SCRIPTTOOLBAR
