//=============================================================================
//
//   File : kvi_scripttoolbar_itemdefinition.cpp
//   Creation date : Fri Nov 08 2002 02:22:12 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __KVIRC__

#include "kvi_scripttoolbar_itemdefinition.h"

#ifdef COMPILE_SCRIPTTOOLBAR

	#include "kvi_scripttoolbar_item.h"
	#include "kvi_scripttoolbar_button.h"
	#include "kvi_scripttoolbar.h"
	#include "kvi_config.h"
	#include "kvi_frame.h"
	#include "kvi_cmdformatter.h"
	
	
	KviScriptToolBarItemDefinition::KviScriptToolBarItemDefinition()
	{
		m_pItems = 0;
		m_bEnabled = true;
	}

	KviScriptToolBarItemDefinition::~KviScriptToolBarItemDefinition()
	{
		while(m_pItems)
		{
			KviScriptToolBarItem * it = m_pItems->first();
			if(it)delete it;
		}
	}

	void KviScriptToolBarItemDefinition::setIcon(const QString &szIcon)
	{
		if(m_type == Separator)return;
		
		m_szIcon = szIcon;
		if(m_pItems)
		{
			for(KviScriptToolBarItem * i = m_pItems->first();i;i = m_pItems->next())
			{
				i->updateIcon();
			}
		}
	}
	
	void KviScriptToolBarItemDefinition::setText(const QString &szText)
	{
		if(m_type == Separator)return;

		m_szText = szText;
		if(m_pItems)
		{
			for(KviScriptToolBarItem * i = m_pItems->first();i;i = m_pItems->next())
			{
				i->updateText();
			}
		}

	}
	
	void KviScriptToolBarItemDefinition::setEnabled(bool bEnabled,KviFrame * pFrm)
	{
		if(m_type == Separator)return;

		m_bEnabled = bEnabled;

		if(m_pItems)
		{
			if(pFrm)
			{
				for(KviScriptToolBarItem * i = m_pItems->first();i;i = m_pItems->next())
				{
					if(i->toolBar()->frame() == pFrm)i->setEnabledState(bEnabled);
				}
			} else {
				for(KviScriptToolBarItem * i = m_pItems->first();i;i = m_pItems->next())
				{
					i->setEnabledState(bEnabled);
				}
			}
		}
	}

	void KviScriptToolBarItemDefinition::attach(KviScriptToolBarItem * pButton)
	{
		if(!m_pItems)
		{
			m_pItems = new KviPtrList<KviScriptToolBarItem>();
			m_pItems->setAutoDelete(false);
		}
		m_pItems->append(pButton);
	}
	
	void KviScriptToolBarItemDefinition::detach(KviScriptToolBarItem * pButton)
	{
		if(m_pItems)
		{
			m_pItems->removeRef(pButton);
			if(m_pItems->isEmpty())
			{
				delete m_pItems;
				m_pItems = 0;
			}
		} else debug("KviScriptToolBarItemDefinition::detach(): the item list is invalid");
	}
	
	void KviScriptToolBarItemDefinition::add(KviScriptToolBar * pBar)
	{
		if(m_type == Separator)
		{
			pBar->addSeparator();
		} else {
			KviScriptToolBarButton * it = new KviScriptToolBarButton(pBar,this);
		}
	}
	
	void KviScriptToolBarItemDefinition::load(const char * szPrefix,KviConfig * cfg)
	{
		KviStr tmp;
		
		tmp.sprintf("%s_Name",szPrefix);
		m_szName = cfg->readQStringEntry(tmp.ptr(),"");
		tmp.sprintf("%s_Code",szPrefix);
		m_szCode = cfg->readQStringEntry(tmp.ptr(),"");
		tmp.sprintf("%s_Icon",szPrefix);
		m_szIcon = cfg->readQStringEntry(tmp.ptr(),"");
		tmp.sprintf("%s_Text",szPrefix);
		m_szText = cfg->readQStringEntry(tmp.ptr(),"");
		tmp.sprintf("%s_Enabled",szPrefix);
		m_bEnabled = cfg->readBoolEntry(tmp.ptr(),true);
		tmp.sprintf("%s_Type",szPrefix);
		tmp = cfg->readEntry(tmp.ptr(),"separator");
		if(kvi_strEqualCI(tmp.ptr(),"button"))
		{
			m_type = Button;
		} else {
			m_type = Separator;
		}
	}
	
	void KviScriptToolBarItemDefinition::save(const char * szPrefix,KviConfig * cfg)
	{
		KviStr tmp;
		
		tmp.sprintf("%s_Name",szPrefix);
		cfg->writeEntry(tmp.ptr(),m_szName);
		tmp.sprintf("%s_Code",szPrefix);
		cfg->writeEntry(tmp.ptr(),m_szCode);
		tmp.sprintf("%s_Icon",szPrefix);
		cfg->writeEntry(tmp.ptr(),m_szIcon);
		tmp.sprintf("%s_Text",szPrefix);
		cfg->writeEntry(tmp.ptr(),m_szText);
		tmp.sprintf("%s_Enabled",szPrefix);
		cfg->writeEntry(tmp.ptr(),m_bEnabled);
		KviStr szType;
		switch(m_type)
		{
			case Separator:
				szType = "separator";
			break;
			case Button:
				szType = "button";
			break;
			default:
				szType = "unknown";
			break;
		}
		tmp.sprintf("%s_Type",szPrefix);
		cfg->writeEntry(tmp.ptr(),szType);
	}

	void KviScriptToolBarItemDefinition::generateKvsDefinition(QString &szBuffer)
	{
		switch(m_type)
		{
			case Separator:
				szBuffer = "separator ";
				szBuffer += m_szName;
			break;
			case Button:
			{
				KviStr szTmp = m_szCode;
				KviCommandFormatter::blockFromBuffer(szTmp);
				szBuffer = "button(\"";
				szBuffer += m_szName;
				szBuffer += "\",\"";
				szBuffer += m_szIcon;
				szBuffer += "\",\"";
				szBuffer += m_szText;
				szBuffer += "\")\n";
				szBuffer += szTmp.ptr();
			}
			break;
			default:
				szBuffer = "";
			break;
		}
	}

	void KviScriptToolBarItemDefinition::copyFrom(KviScriptToolBarItemDefinition * d)
	{
		if(m_pItems)
		{
			delete m_pItems;
			m_pItems = 0;
		}
		m_szName = d->name();
		m_szCode = d->code();
		m_szIcon = d->icon();
		m_szText = d->text();
		m_type = d->type();
		m_bEnabled = d->enabled();
	}

#endif // COMPILE_SCRIPTTOOLBAR
