//=============================================================================
//
//   File : kvi_scripttoolbar_manager.cpp
//   Creation date : Fri Nov 08 2002 02:22:12 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __KVIRC__

#define _KVI_SCRIPTTOOLBAR_MANAGER_CPP_

#include "kvi_scripttoolbar_manager.h"

#ifdef COMPILE_SCRIPTTOOLBAR

	#include "kvi_scripttoolbar_definition.h"
	#include "kvi_config.h"
	
	// created and destroyed by kvi_app.cpp
	KVIRC_API KviScriptToolBarManager * g_pScriptToolBarManager = 0;
	
	
	KviScriptToolBarManager::KviScriptToolBarManager()
	{
		m_pBars = new QDict<KviScriptToolBarDefinition>(17,false);
		m_pBars->setAutoDelete(true);
	}
	
	KviScriptToolBarManager::~KviScriptToolBarManager()
	{
		delete m_pBars;
	}
	
	void KviScriptToolBarManager::createVisible(KviFrame * pFrm)
	{
		QDictIterator<KviScriptToolBarDefinition> it(*m_pBars);
		while(KviScriptToolBarDefinition * d = it.current())
		{
			if(d->isVisible())d->instance(pFrm);
			++it;
		}
	}

	void KviScriptToolBarManager::load(const char * szFileName)
	{
		KviConfig cfg(szFileName,KviConfig::Read);
	
		unsigned int cnt = cfg.readUIntEntry("Count",0);
	
		KviStr tmp;
	
		for(unsigned int i=0;i<cnt;i++)
		{
			KviScriptToolBarDefinition * d = new KviScriptToolBarDefinition(this);
			tmp.sprintf("%d",i);
			d->load(tmp.ptr(),&cfg);
			if(d->name().isEmpty())delete d;
			else m_pBars->replace(d->name(),d);
		}
	}
	
	void KviScriptToolBarManager::save(const char * szFileName)
	{
		KviConfig cfg(szFileName,KviConfig::Write);
		cfg.clear();
		
		unsigned int idx = 0;
		cfg.writeEntry("Count",m_pBars->count());
		
		KviStr tmp;
	
		QDictIterator<KviScriptToolBarDefinition> it(*m_pBars);
		while(KviScriptToolBarDefinition * d = it.current())
		{
			tmp.sprintf("%d",idx);
			d->save(tmp.ptr(),&cfg);
			idx++;
			++it;
		}
	}

	void KviScriptToolBarManager::rename(KviScriptToolBarDefinition * d,const QString &szNewName)
	{
		KviScriptToolBarDefinition * old = m_pBars->find(szNewName);
		if(old)
		{
			// collision
			if(old == d)return; // has already that name
			int idx = 1;
			QString tmp = szNewName;
			tmp += QString("%1").arg(idx);
			while(m_pBars->find(tmp))
			{
				idx++;
				tmp = szNewName;
				tmp += QString("%1").arg(idx);
			}
			rename(d,tmp);
			return;
		}
		// ok.. will work: there is no collision
		m_pBars->setAutoDelete(false);
		m_pBars->remove(d->name());
		d->setName(szNewName);
		m_pBars->replace(d->name(),d);
		m_pBars->setAutoDelete(true);
	}


	
	KviScriptToolBarDefinition * KviScriptToolBarManager::get(const QString & szName)
	{
		KviScriptToolBarDefinition * d = find(szName);
		if(!d)
		{
			d = new KviScriptToolBarDefinition(this);
			d->setName(szName);
			m_pBars->insert(szName,d);
		}
		return d;
	}
	
	KviScriptToolBarDefinition * KviScriptToolBarManager::find(int id)
	{
		QDictIterator<KviScriptToolBarDefinition> it(*m_pBars);
		while(KviScriptToolBarDefinition * d = it.current())
		{
			if(d->id() == id)return d;
			++it;
		}
		return 0;
	}
	
	KviScriptToolBarDefinition * KviScriptToolBarManager::getNew()
	{
		int idx = 2;
		QString szName = "toolbar1";
		while(find(szName))
		{
			szName = "toolbar";
			szName += QString("%1").arg(idx);
			idx++;
		}
//		KviScriptToolBarDefinition * def = get(szName);
//		def->setVisible(true); // we show it by default
//		return def;
		return get(szName);
	}

	void KviScriptToolBarManager::storeVisibilityState()
	{
		// this is called as a "checkpoint" by KviApp before closing down
		// and before saving settings
		QDictIterator<KviScriptToolBarDefinition> it(*m_pBars);
		while(KviScriptToolBarDefinition * d = it.current())
		{
			d->storeVisibilityState();
			++it;
		}
	}

	void KviScriptToolBarManager::remove(const QString & szName)
	{
		m_pBars->remove(szName);
	}

	void KviScriptToolBarManager::clear()
	{
		delete m_pBars;
		m_pBars = new QDict<KviScriptToolBarDefinition>(17,false);
		m_pBars->setAutoDelete(true);
	}
	
	void KviScriptToolBarManager::copyFrom(KviScriptToolBarManager * m)
	{
		QDictIterator<KviScriptToolBarDefinition> it(*(m->toolBarDefinitions()));
		
		KviPtrList<KviScriptToolBarDefinition> remaining;
		remaining.setAutoDelete(false);
		
		while(KviScriptToolBarDefinition * r = it.current())
		{
			KviScriptToolBarDefinition * theDef = get(r->name());
			remaining.append(theDef);
			theDef->copyFrom(r);
			++it;
		}
		
		// remove the dead toolbars
		KviPtrList<KviScriptToolBarDefinition> dead;
		dead.setAutoDelete(false);
		
		QDictIterator<KviScriptToolBarDefinition> it2(*m_pBars);
		while(KviScriptToolBarDefinition * existing = it2.current())
		{
			if(remaining.findRef(existing) == -1)dead.append(existing);
			++it2;
		}
		
		for(KviScriptToolBarDefinition * zombie = dead.first();zombie;zombie = dead.next())
		{
			m_pBars->remove(zombie->name());
		}
	}

#endif // COMPILE_SCRIPTTOOLBAR
