//
//   File : send.cpp
//   Creation date : Tue Sep 20 09 2000 15:14:14 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "send.h"
#include "broker.h"
#include "marshal.h"
#include "broker.h"

#ifdef COMPILE_ON_WINDOWS
	// Ugly Windoze compiler...
	#include "dialogs.h"
#endif

#define _KVI_DEBUG_CHECK_RANGE_
#include "kvi_debug.h"
#include "kvi_app.h"
#include "kvi_options.h"
#include "kvi_ircview.h"
#include "kvi_iconmanager.h"
#include "kvi_locale.h"
#include "kvi_error.h"
#include "kvi_out.h"
#include "kvi_netutils.h"
#include "kvi_console.h"
#include "kvi_frame.h"
#include "kvi_malloc.h"
#include "kvi_memmove.h"
#include "kvi_thread.h"
#include "kvi_irclink.h"
#include "kvi_uparser.h"
#include "kvi_mediatype.h"
#include "kvi_socket.h"
#include "kvi_event.h"
#include "kvi_parameterlist.h"

//#include <qsplitter.h>
#include <qevent.h>
//#include <qtooltip.h>
#include <qfile.h>
//#include <qtoolbutton.h>
#include <qpainter.h>

//#include <unistd.h> //close()

// FIXME: SSL Support here!
// FIXME: The events OnDCCConnect etc are in wrong places here...!

extern KviDccBroker * g_pDccBroker;

extern KVIRC_API KviMediaManager * g_pMediaManager; // kvi_app.cpp


static KviPtrList<KviDccFileTransfer> * g_pDccFileTransfers = 0;
static QPixmap * g_pDccFileTransferIcon = 0;


//#warning "The events that have a KviStr data pointer should become real classes, that take care of deleting the data pointer!"
//#warning "Otherwise, when left undispatched we will be leaking memory (event class destroyed but not the data ptr)"

KviDccRecvThread::KviDccRecvThread(QObject * par,kvi_socket_t fd,KviDccRecvThreadOptions * opt)
: KviDccThread(par,fd)
{
	m_pOpt                  = opt;
//	m_iProgress             = 0;
	m_iAverageSpeed         = 0;
	m_iInstantSpeed         = 0;
	m_iFilePosition         = 0;
	m_iStartTime            = kvi_unixTime();

	m_iTotalReceivedBytes   = 0;
	m_iInstantReceivedBytes = 0;
	m_pFile                 = 0;
}

KviDccRecvThread::~KviDccRecvThread()
{
	if(m_pOpt)delete m_pOpt;
	if(m_pFile)delete m_pFile;
}

bool KviDccRecvThread::sendAck(int filePos)
{
	int size = htonl(filePos);
	if(kvi_socket_send(m_fd,(void *)(&size),4) != 4)
	{
		postErrorEvent(KviError_acknowledgeError);
		return false;
	}
	return true;
}

void KviDccRecvThread::updateStats()
{
	int iCurTime = (int)kvi_unixTime();
	int iDiffTime = iCurTime - m_iLastInstantCheckTime;

	m_pMutex->lock();
	int iElapsedTime = (int)(iCurTime - m_iStartTime);
	if(iElapsedTime < 1)iElapsedTime = 1;
	m_iFilePosition = m_pFile->at();
//	if(m_pOpt->iTotalFileSize > 0)
//		m_iProgress = ((m_iFilePosition * 100) / m_pOpt->iTotalFileSize);
	m_iAverageSpeed = m_iTotalReceivedBytes / iElapsedTime;
	if(iDiffTime > 1)
	{
		m_iInstantSpeed = m_iInstantReceivedBytes / iDiffTime;
		m_iLastInstantCheckTime = iCurTime;
		m_iInstantReceivedBytes = 0;
	}
	m_pMutex->unlock();
}

void KviDccRecvThread::postMessageEvent(const char * m)
{
	KviThreadDataEvent<KviStr> * e = new KviThreadDataEvent<KviStr>(KVI_DCC_THREAD_EVENT_MESSAGE);
	e->setData(new KviStr(m));
	postEvent(parent(),e);
}


void KviDccRecvThread::run()
{
	m_pMutex->lock();
	m_iStartTime              = kvi_unixTime();
	m_pMutex->unlock();

	m_iLastInstantCheckTime = kvi_unixTime();

	int iProbableTerminationTime = 0;

	m_pFile = new QFile(m_pOpt->szFileName.ptr());

	if(m_pOpt->bResume)
	{
		if(!m_pFile->open(IO_WriteOnly | IO_Append))
		{
			postErrorEvent(KviError_cantOpenFileForAppending);
			goto exit_dcc;
		} // else pFile is already at end
	} else {
		if(!m_pFile->open(IO_WriteOnly))
		{
			postErrorEvent(KviError_cantOpenFileForWriting);
			goto exit_dcc;
		}
	}

	if(m_pOpt->bSendZeroAck && (!m_pOpt->bNoAcks))
	{
		if(!sendAck(m_pFile->at()))goto exit_dcc;
	}


	for(;;)
	{
		// Dequeue events
		while(KviThreadEvent * e = dequeueEvent())
		{
			if(e->id() == KVI_THREAD_EVENT_TERMINATE)
			{
				delete e;
				goto exit_dcc;
			} else {
				// Other events are senseless to us
				delete e;
			}
		}

		bool bCanRead;
		bool bDummy;

		if(kvi_select(m_fd,&bCanRead,&bDummy))
		{
			if(bCanRead)
			{
				// Read a data block
				char buffer[1024];
				int readLen = kvi_socket_recv(m_fd,buffer,1024);

				if(readLen > 0)
				{
					// Readed something useful...write back
					if((m_pOpt->iTotalFileSize > -1) && ((readLen + m_pFile->at()) > m_pOpt->iTotalFileSize))
					{
						postMessageEvent(__tr_no_lookup_ctx("WARNING: The peer is sending garbage data past the end of the file","dcc"));
						postMessageEvent(__tr_no_lookup_ctx("WARNING: Ignoring data past the declared end of file and closing the connection","dcc"));

						readLen = m_pOpt->iTotalFileSize - m_pFile->at();
						if(readLen > 0)
						{
							if(m_pFile->writeBlock(buffer,readLen) != readLen)
								postErrorEvent(KviError_fileIOError);
						}
						break;

					} else {
						if(m_pFile->writeBlock(buffer,readLen) != readLen)
						{
							postErrorEvent(KviError_fileIOError);
							break;
						}
					}

					// Update stats
					m_iTotalReceivedBytes += readLen;
					m_iInstantReceivedBytes += readLen;

					updateStats();
					// Now send the ack
					if(m_pOpt->bNoAcks)
					{
						// No acks...
						// Interrupt if the whole file has been received
						if(m_pOpt->iTotalFileSize > 0)
						{
							if(m_pFile->at() == m_pOpt->iTotalFileSize)
							{
								// Received the whole file...die
								KviThreadEvent * e = new KviThreadEvent(KVI_DCC_THREAD_EVENT_SUCCESS);
								postEvent(parent(),e);
								break;
							}
						}
					} else {
						// Must send the ack... the peer must close the connection
						if(!sendAck(m_pFile->at()))break;
					}
				} else {
					updateStats();
					// Read problem...

					if(readLen == 0)
					{
						// readed EOF..
						if((m_pFile->at() == m_pOpt->iTotalFileSize) || (m_pOpt->iTotalFileSize < 0))
						{
							// success if we got the whole file or if we don't know the file size (we trust the peer)
							KviThreadEvent * e = new KviThreadEvent(KVI_DCC_THREAD_EVENT_SUCCESS);
							postEvent(parent(),e);
							break;
						}
					}
					if(!handleInvalidSocketRead(readLen))break;
				}
			} else {
				// Can't read stuff
				updateStats();

				if(m_pFile->at() == m_pOpt->iTotalFileSize)
				{
					// Wait for the peer to close the connection
					if(iProbableTerminationTime == 0)
					{
						iProbableTerminationTime = (int)kvi_unixTime();
						m_pFile->flush();
						postMessageEvent(__tr_no_lookup_ctx("Data transfer terminated, waiting 30 seconds for the peer to close the connection...","dcc"));
						// FIXME: Close the file ?
					} else {
						int iDiff = (((int)kvi_unixTime()) - iProbableTerminationTime);
						if(iDiff > 30)
						{
							// success if we got the whole file or if we don't know the file size (we trust the peer)
							postMessageEvent(__tr_no_lookup_ctx("Data transfer was terminated 30 seconds ago, closing the connection","dcc"));
							KviThreadEvent * e = new KviThreadEvent(KVI_DCC_THREAD_EVENT_SUCCESS);
							postEvent(parent(),e);
							break;
						}
					}
				}
			}
			msleep(10);
		} else {
			updateStats();
			msleep(50);
		}
	}

exit_dcc:
	if(m_pFile)
	{
		m_pFile->close();
		delete m_pFile;
		m_pFile = 0;
	}
	kvi_socket_close(m_fd);
	m_fd = KVI_INVALID_SOCKET;
}

int KviDccRecvThread::elapsedTime()
{
	return (int)(kvi_unixTime() - m_iStartTime);
}

void KviDccRecvThread::initGetInfo()
{
	m_pMutex->lock();
}

void KviDccRecvThread::doneGetInfo()
{
	m_pMutex->unlock();
}

KviDccSendThread::KviDccSendThread(QObject * par,kvi_socket_t fd,KviDccSendThreadOptions * opt)
: KviDccThread(par,fd)
{
	m_pOpt = opt;
	// stats
	m_iStartTime     = kvi_unixTime();
//	m_iProgress      = 0;
	m_iAverageSpeed  = 0;
	m_iInstantSpeed  = 0;
	m_iFilePosition  = 0;
//	m_iAckedProgress = 0;
}

KviDccSendThread::~KviDccSendThread()
{
	if(m_pOpt)delete m_pOpt;
}

void KviDccSendThread::run()
{

	m_pMutex->lock();
	m_iStartTime              = kvi_unixTime();
	m_pMutex->unlock();

	int iLastInstantCheckTime = m_iStartTime;
	int iTotalSentBytes       = 0;
	int iInstantSentBytes     = 0;
	int iLastAck;

	if(m_pOpt->iPacketSize < 32)m_pOpt->iPacketSize = 32;
	char * buffer = (char *)kvi_malloc(m_pOpt->iPacketSize * sizeof(char));

	QFile * pFile = new QFile(m_pOpt->szFileName.ptr());

	if(!pFile->open(IO_ReadOnly))
	{
		postErrorEvent(KviError_cantOpenFileForReading);
		goto exit_dcc;
	}

	if(pFile->size() < 1)
	{
		postErrorEvent(KviError_cantSendAZeroSizeFile);
		goto exit_dcc;
	}

	if(m_pOpt->iStartPosition > 0)
	{
		// seek
		if(!(pFile->at(m_pOpt->iStartPosition)))
		{
			postErrorEvent(KviError_fileIOError);
			goto exit_dcc;
		}
	}

	iLastAck = m_pOpt->iStartPosition;
	for(;;)
	{
		// Dequeue events
		while(KviThreadEvent * e = dequeueEvent())
		{
			if(e->id() == KVI_THREAD_EVENT_TERMINATE)
			{
				delete e;
				goto exit_dcc;
			} else {
				// Other events are senseless to us
				delete e;
			}
		}

		bool bCanRead;
		bool bCanWrite;

		if(kvi_select(m_fd,&bCanRead,&bCanWrite))
		{
			if(bCanRead)
			{
				if(!m_pOpt->bNoAcks)
				{
					int iAck;
					// FIXME: this could fail : the network/kernel could deliver only a part of the packet (?)
					int readLen = kvi_socket_recv(m_fd,(void *)&iAck,4);
					if(readLen == 4)
					{
						int iNewAck = ntohl(iAck);
						if((iNewAck > pFile->at()) || (iNewAck < iLastAck))
						{
							// the peer is drunk or is trying to fool us
							postErrorEvent(KviError_acknowledgeError);
							break;
						}

						iLastAck = iNewAck;
					} else {
						if(readLen > 0)
						{
							postErrorEvent(KviError_acknowledgeError);
							break;
						} else {
							if(!handleInvalidSocketRead(readLen))break;
						}
					}

					// update stats
					m_pMutex->lock(); // is this really necessary ?
					m_iAckedBytes = iLastAck;
					m_pMutex->unlock(); 
	
					if(iLastAck >= (int)(pFile->size()))
					{
						KviThreadEvent * e = new KviThreadEvent(KVI_DCC_THREAD_EVENT_SUCCESS);
						postEvent(parent(),e);
						break;
					}
				} else {
					// No acknowledges
					if(m_pOpt->bIsTdcc)
					{
						// We expect the remote end to close the connection when the whole file has been sent
						if(pFile->atEnd())
						{
							int iAck;
							int readLen = kvi_socket_recv(m_fd,(void *)&iAck,4);
							if(readLen == 0)
							{
								// done...success
								KviThreadEvent * e = new KviThreadEvent(KVI_DCC_THREAD_EVENT_SUCCESS);
								postEvent(parent(),e);
								break;
							} else {
								if(readLen < 0)
								{
									if(!handleInvalidSocketRead(readLen))break;
								} else {
									KviThreadDataEvent<KviStr> * e = new KviThreadDataEvent<KviStr>(KVI_DCC_THREAD_EVENT_MESSAGE);
									e->setData(new KviStr(__tr_no_lookup_ctx("WARNING: Received data in a DCC TSEND, there should be no acknowledges","dcc")));
									postEvent(parent(),e);
								}
							}
						}
					}
				}
			}
			if(bCanWrite)
			{
				if(!pFile->atEnd())
				{
					if(m_pOpt->bFastSend || m_pOpt->bNoAcks || (iLastAck == pFile->at()))
					{
						int toRead = pFile->size() - pFile->at();
						if(toRead > m_pOpt->iPacketSize)toRead = m_pOpt->iPacketSize;
						int readed = pFile->readBlock(buffer,toRead);
						if(readed < toRead)
						{
							postErrorEvent(KviError_fileIOError);
							break;
						}
						int written = kvi_socket_send(m_fd,buffer,toRead);
						if(written < toRead)
						{
							if(written < 0)
							{
								// error ?
								if(!handleInvalidSocketRead(written))break;
							} else {
								pFile->at(pFile->at() - (toRead - written));
							}
						}
						iTotalSentBytes += written;
						iInstantSentBytes += written;
						int iCurTime = (int)kvi_unixTime();
						int iDiffTime = iCurTime - iLastInstantCheckTime;
						m_pMutex->lock();
						int iElapsedTime = (int)(iCurTime - m_iStartTime);
						if(iElapsedTime < 1)iElapsedTime = 1;
						m_iFilePosition = pFile->at();

						if(m_pOpt->bNoAcks)
						{
							// There are no acks : the avg speed is based on the sent bytes
							m_iAverageSpeed = iTotalSentBytes / iElapsedTime;
						} else {
							// acknowledges : we compute the avg speed based on the acks we receive
							m_iAverageSpeed = (iLastAck - m_pOpt->iStartPosition) / iElapsedTime;
						}

						if(iDiffTime > 1)
						{
							m_iInstantSpeed = iInstantSentBytes / iDiffTime;
							iLastInstantCheckTime = iCurTime;
							iInstantSentBytes = 0;
						}
						m_pMutex->unlock();
					}
				} else {
					if(m_pOpt->bNoAcks && !m_pOpt->bIsTdcc)
					{
						// at end of the file in a blind dcc send...
						// not in a tdcc: we can close the file...
						KviThreadEvent * e = new KviThreadEvent(KVI_DCC_THREAD_EVENT_SUCCESS);
						postEvent(parent(),e);
						break;
					}
				}
			}
		}
//#warning "iIdleStepLengthInMSec could increase if stalled or decrease of received data"
//#warning "obviously into a predefined range... (adaptive dcc)"
		msleep(m_pOpt->iIdleStepLengthInMSec);
	}

exit_dcc:
	kvi_free(buffer);
	pFile->close();
	delete pFile;
	pFile = 0;
	kvi_socket_close(m_fd);
	m_fd = KVI_INVALID_SOCKET;
}

int KviDccSendThread::elapsedTime()
{
	return (int)(kvi_unixTime() - m_iStartTime);
}

void KviDccSendThread::initGetInfo()
{
	m_pMutex->lock();
}

void KviDccSendThread::doneGetInfo()
{
	m_pMutex->unlock();
}


KviDccFileTransfer::KviDccFileTransfer(KviDccDescriptor * dcc)
: KviFileTransfer()
{
	init(); // ensure we're initialized
	g_pDccFileTransfers->append(this);

	m_szTransferIdString.sprintf(__tr_ctx("TRANSFER %d","dcc"),id());

	m_pDescriptor = dcc;


	m_pMarshal = new KviDccMarshal(this);

	connect(m_pMarshal,SIGNAL(error(int)),this,SLOT(handleMarshalError(int)));
	connect(m_pMarshal,SIGNAL(connected()),this,SLOT(connected()));
	connect(m_pMarshal,SIGNAL(inProgress()),this,SLOT(connectionInProgress()));
#ifdef COMPILE_SSL_SUPPORT
	connect(m_pMarshal,SIGNAL(startingSSLHandshake()),this,SLOT(startingSSLHandshake()));
	connect(m_pMarshal,SIGNAL(sslError(const char *)),this,SLOT(sslError(const char *)));
#endif

	m_szDccType = dcc->bIsTdcc ? (dcc->bRecvFile ? "TRECV" : "TSEND") : (dcc->bRecvFile ? "RECV" : "SEND");

	m_pSlaveRecvThread = 0;
	m_pSlaveSendThread = 0;

	m_szStatusString = __tr2qs_ctx("Setting up the connection","dcc");
	m_eGeneralStatus = Connecting;

	bool bOk;
	m_uTotalFileSize = dcc->bRecvFile ? dcc->szFileSize.toUInt(&bOk) :  dcc->szLocalFileSize.toUInt(&bOk);
	if(!bOk)m_uTotalFileSize = 0;


	startConnection();
}

KviDccFileTransfer::~KviDccFileTransfer()
{
	g_pDccFileTransfers->removeRef(this);

	if(m_pSlaveRecvThread)
	{
		m_pSlaveRecvThread->terminate();
		delete m_pSlaveRecvThread;
		m_pSlaveRecvThread = 0;
	}

	if(m_pSlaveSendThread)
	{
		m_pSlaveSendThread->terminate();
		delete m_pSlaveSendThread;
		m_pSlaveSendThread = 0;
	}

	KviThreadManager::killPendingEvents(this);

	delete m_pDescriptor;
	delete m_pMarshal;
}

void KviDccFileTransfer::startConnection()
{
	if(!(m_pDescriptor->bActive))
	{
		// PASSIVE CONNECTION
		m_szStatusString = __tr2qs_ctx("Attempting a passive DCC %1 connection","dcc").arg(m_szDccType.ptr());
		outputAndLog(m_szStatusString);

		int ret = m_pMarshal->dccListen(m_pDescriptor->szListenIp.ptr(),m_pDescriptor->szListenPort.ptr(),m_pDescriptor->bDoTimeout);
		if(ret != KviError_success)handleMarshalError(ret);

	} else {
		// ACTIVE CONNECTION
		m_szStatusString = __tr2qs_ctx("Attempting an active DCC %1 connection","dcc").arg(m_szDccType.ptr());
		outputAndLog(m_szStatusString);

		if(m_pDescriptor->bResume && m_pDescriptor->bRecvFile)
		{
			KviStr fName  = m_pDescriptor->szFileName.ptr();
			if(fName.contains(' '))
			{
				fName.prepend("\"");
				fName.append("\"");
			}
			m_pDescriptor->console()->link()->sendFmtData("PRIVMSG %s :%cDCC RESUME %s %s %s%c",
				m_pDescriptor->szNick.ptr(),0x01,fName.ptr(),m_pDescriptor->szPort.ptr(),
				m_pDescriptor->szLocalFileSize.ptr(),0x01);
			m_szStatusString = __tr2qs_ctx("Sent DCC RESUME request to %1, waiting for ACCEPT","dcc").arg(m_pDescriptor->szNick.ptr());
			outputAndLog(m_szStatusString);
		} else {
			int ret = m_pMarshal->dccConnect(m_pDescriptor->szIp.ptr(),m_pDescriptor->szPort.ptr(),m_pDescriptor->bDoTimeout);
			if(ret != KviError_success)handleMarshalError(ret);
		}
	}

	displayUpdate();
}


KviWindow * KviDccFileTransfer::dccMarshalOutputWindow()
{
	return transferWindow();
}

const char * KviDccFileTransfer::dccMarshalOutputContextString()
{
	return m_szTransferIdString;
}

void KviDccFileTransfer::die()
{
	delete this;
}

QString KviDccFileTransfer::localFileName()
{
	return QString(m_pDescriptor->szLocalFileName.ptr());
}


void KviDccFileTransfer::abort()
{
	if(m_pSlaveRecvThread)m_pSlaveRecvThread->terminate();
	if(m_pSlaveSendThread)m_pSlaveSendThread->terminate();
	if(m_pMarshal)m_pMarshal->abort();

	if(m_pDescriptor->bRecvFile)
			g_pApp->fileDownloadTerminated(false,m_pDescriptor->szFileName.ptr(),m_pDescriptor->szLocalFileName.ptr(),m_pDescriptor->szNick.ptr(),__tr_ctx("Aborted","dcc"));

	m_szStatusString = __tr2qs_ctx("Transfer failed: ","dcc");
	m_szStatusString += __tr2qs_ctx("Aborted","dcc");
	outputAndLog(KVI_OUT_DCCERROR,m_szStatusString);
	m_eGeneralStatus = Failure;
	displayUpdate();
}


void KviDccFileTransfer::fillContextPopup(QPopupMenu * m,int column)
{
	if(active())m->insertItem(__tr2qs_ctx("Abort","dcc"),this,SLOT(abort()));
}


bool KviDccFileTransfer::active()
{
	return ((m_eGeneralStatus == Connecting) || (m_eGeneralStatus == Transferring));
}

void KviDccFileTransfer::displayPaint(QPainter * p,int column,int width,int height)
{

	QString txt;
	bool bIsTerminated = ((m_eGeneralStatus == Success) || (m_eGeneralStatus == Failure));

	switch(column)
	{
		case COLUMN_TRANSFERTYPE:
		{

			int xoffset;
			int yoffset = 0;
			if(m_pDescriptor->bRecvFile)yoffset = 64;
			switch(m_eGeneralStatus)
			{
				case Connecting: xoffset = 0; break;
				case Transferring: xoffset = 48; break;
				case Success: xoffset = 96; break;
				case Failure: xoffset = 144; break;
			}
			p->drawPixmap(3,3,*g_pDccFileTransferIcon,xoffset,yoffset,48,64);
		}
		break;
		case COLUMN_FILEINFO:
		{

			QFontMetrics fm(p->font());

			QString szFrom = __tr2qs_ctx("From: ","dcc");
			QString szTo   = __tr2qs_ctx("To: ","dcc");

			int daW1 = fm.width(szFrom);
			int daW2 = fm.width(szTo);
			if(daW1 < daW2)daW1 = daW2;
			int iLineSpacing = fm.lineSpacing();

			int iY = 4;

			p->setPen(Qt::black);

			KviStr szRemote(KviStr::Format,"dcc://%s@%s:%s/%s",m_pDescriptor->szNick.ptr(),m_pDescriptor->szIp.ptr(),m_pDescriptor->szPort.ptr(),
					m_pDescriptor->szFileName.ptr());

			p->drawText(4 + daW1,iY,width - (8 + daW1),height - 8,Qt::AlignTop | Qt::AlignLeft,
					m_pDescriptor->bRecvFile ? szRemote.ptr() : m_pDescriptor->szLocalFileName.ptr());
			iY += iLineSpacing;

			p->drawText(4 + daW1,iY,width - (8 + daW1),height - 8,Qt::AlignTop | Qt::AlignLeft,
					m_pDescriptor->bRecvFile ? m_pDescriptor->szLocalFileName.ptr() : szRemote.ptr());
			iY += iLineSpacing;


			p->setPen(Qt::darkGray);

			p->drawText(4,4,width - 8,height - 8,Qt::AlignTop | Qt::AlignLeft,szFrom);
			p->drawText(4,4 + iLineSpacing,width - 8,height - 8,Qt::AlignTop | Qt::AlignLeft,szTo);


			p->setPen(QColor(180,180,200));

			iLineSpacing += 2;

			p->drawRect(4,height - (iLineSpacing + 4),width - 8,iLineSpacing);
			p->fillRect(5,height - (iLineSpacing + 3),width - 10,iLineSpacing - 2,bIsTerminated ? QColor(210,210,210) : QColor(190,190,240));

			p->setPen(Qt::black);

			p->drawText(7,height - (iLineSpacing + 4),width - 14,iLineSpacing,Qt::AlignVCenter | Qt::AlignLeft,m_szStatusString);

		}
		break;
		case COLUMN_PROGRESS:
		{

			QFontMetrics fm(p->font());

			int iW = width - 8;
			int iAvgSpeed = -1;
			int iInstantSpeed = -1;
			int iAckedBytes = -1;

			int iEta = -1;

			unsigned int uTransferred = 0;

			if(m_pDescriptor->bRecvFile)
			{
				if(m_pSlaveRecvThread)
				{
					m_pSlaveRecvThread->initGetInfo();
					iAvgSpeed = m_pSlaveRecvThread->averageSpeed();
					iInstantSpeed = m_pSlaveRecvThread->instantSpeed();
					uTransferred = m_pSlaveRecvThread->filePosition();
/*
					int el = m_pSlaveRecvThread->elapsedTime();
*/
					m_pSlaveRecvThread->doneGetInfo();
				}
			} else {
				if(m_pSlaveSendThread)
				{
					m_pSlaveSendThread->initGetInfo();
					iAvgSpeed = m_pSlaveSendThread->averageSpeed();
					iInstantSpeed = m_pSlaveSendThread->instantSpeed();
					uTransferred = m_pSlaveSendThread->filePosition();
					iAckedBytes = m_pSlaveSendThread->ackedBytes();

/*
					int el = m_pSlaveSendThread->elapsedTime();
*/
					m_pSlaveSendThread->doneGetInfo();
				}
			}

			p->setPen(bIsTerminated ? Qt::lightGray : QColor(210,210,240));
			p->drawRect(4,4,iW,12);

			iW -= 2;

			if(m_uTotalFileSize > 0)
			{
				if(iAvgSpeed > 0)
				{
					unsigned int uRemaining = m_uTotalFileSize - uTransferred;
					iEta = uRemaining / iAvgSpeed;
				}

				if(!m_pDescriptor->bNoAcks && (iAckedBytes > 0) && (iAckedBytes < uTransferred))
				{
					// we are sending a file and are getting acks

					double dPerc1 = (double)(((double)uTransferred) * 100.0) / (double)m_uTotalFileSize;
					unsigned int uPerc1 = (unsigned int)dPerc1;
					int iL1 = (int) ((((double)iW) * dPerc1) / 100.0);
					double dPerc2 = (double)(((double)iAckedBytes) * 100.0) / (double)m_uTotalFileSize;
					unsigned int uPerc2 = (unsigned int)dPerc2;
					int iL2 = (int) ((((double)iW) * dPerc2) / 100.0);
					int iW2 = iL1 - iL2;
					if(iW2 > 0)p->fillRect(5 + iL2,5,iW2,10,bIsTerminated ? QColor(150,130,110) : QColor(220,170,100));
					p->fillRect(5,5,iL2,10,bIsTerminated ? QColor(140,110,110) : QColor(200,100,100));

					txt = QString(__tr2qs_ctx("%1 of %2 bytes (%3%)","dcc")).arg(iAckedBytes).arg(m_uTotalFileSize).arg(dPerc2,0,'f',2);
				} else {
					// we are receiving a file or not sending acks
					double dPerc = (double)(((double)uTransferred) * 100.0) / (double)m_uTotalFileSize;
					unsigned int uPerc = (unsigned int)dPerc;
					int iL = (int) ((((double)iW) * dPerc) / 100.0);
					p->fillRect(5,5,iL,10,bIsTerminated ? QColor(140,110,110) : QColor(200,100,100));

					txt = QString(__tr2qs_ctx("%1 of %2 bytes (%3%)","dcc")).arg(uTransferred).arg(m_uTotalFileSize).arg(dPerc,0,'f',2);
				}

			} else {
				txt = QString(__tr2qs_ctx("%1 bytes","dcc")).arg(uTransferred);
			}

			p->setPen(Qt::black);

			p->drawText(4,19,width - 8,height - 8,Qt::AlignTop | Qt::AlignLeft,txt);

			int iLeftHalf = (iW - 2) / 2;
			int iRightHalf = iW - (iLeftHalf + 1);
			int iLineSpacing = fm.lineSpacing() + 2;

			if(iAvgSpeed >= 0)txt = QString(__tr2qs_ctx("Avg: %1 bytes/sec","dcc")).arg(iAvgSpeed);
			else txt = __tr2qs_ctx("Avg: ? bytes/sec","dcc");


			int iDaH = height - (iLineSpacing + 4);

			p->setPen(QColor(180,180,200));
			p->drawRect(4,iDaH,iLeftHalf,iLineSpacing);
			p->fillRect(5,iDaH + 1,iLeftHalf - 2,iLineSpacing - 2,bIsTerminated ? QColor(210,210,210) : QColor(190,190,240));
			p->setPen(bIsTerminated ? Qt::darkGray : Qt::black);
			p->drawText(6,iDaH,iLeftHalf - 4,iLineSpacing,Qt::AlignLeft | Qt::AlignVCenter,txt);

			unsigned int uD,uH,uM,uS;

			if(bIsTerminated)
			{
				/*
				kvi_secondsToDaysHoursMinsSecs(kvi_timeSpan(m_tTransferEndTime,m_tTransferStartTime),&uD,&uH,&uM,&uS);
				txt = "TOT: ";
				if(uD > 0)txt += QString(__tr2qs_ctx("%1d %2h %3m %4s","http")).arg(uD).arg(uH).arg(uM).arg(uS);
				else if(uH > 0)txt += QString(__tr2qs_ctx("%2h %3m %4s","http")).arg(uH).arg(uM).arg(uS);
				else txt += QString(__tr2qs_ctx("%3m %4s","http")).arg(uM).arg(uS);
				*/
			} else {
				if(iEta >= 0)
				{
					kvi_secondsToDaysHoursMinsSecs(iEta,&uD,&uH,&uM,&uS);
					txt = "ETA: ";
					if(uD > 0)txt += QString(__tr2qs_ctx("%1d %2h %3m %4s","dcc")).arg(uD).arg(uH).arg(uM).arg(uS);
					else if(uH > 0)txt += QString(__tr2qs_ctx("%2h %3m %4s","dcc")).arg(uH).arg(uM).arg(uS);
					else txt += QString(__tr2qs_ctx("%3m %4s","dcc")).arg(uM).arg(uS);
				} else {
					txt = "ETA: Unknown";
				}
			}

			p->setPen(QColor(180,180,200));
			p->drawRect(width - (4 + iRightHalf),iDaH,iRightHalf,iLineSpacing);
			p->fillRect(width - (3 + iRightHalf),iDaH + 1,iRightHalf - 2,iLineSpacing - 2,bIsTerminated ? QColor(210,210,210) : QColor(190,190,240));
			p->setPen(bIsTerminated ? Qt::darkGray : Qt::black);
			p->drawText(width - (2 + iRightHalf),iDaH,iRightHalf - 4,iLineSpacing,Qt::AlignLeft | Qt::AlignVCenter,txt);

/*

	m_pSlaveRecvThread->initGetInfo();


			QFontMetrics fm(p->font());

			unsigned int uTotal = m_pHttpRequest->totalSize();
			unsigned int uRecvd = m_pHttpRequest->receivedSize();


			int iW = width - 8;

			p->setPen(bIsTerminated ? Qt::lightGray : QColor(210,210,240));
			p->drawRect(4,4,iW,12);

			int iAvgSpeed = -1;
			int iEta = -1;

			if(m_tTransferStartTime > 0)
			{
				int tSpan = kvi_timeSpan(m_tTransferEndTime > 0 ? m_tTransferEndTime : kvi_unixTime(),m_tTransferStartTime);
				if(tSpan > 0)
				{
					//debug("SPAN: %d (%d - %d)",tSpan,m_tTransferEndTime > 0 ? m_tTransferEndTime : kvi_unixTime(),m_tTransferStartTime);
					iAvgSpeed = uRecvd / tSpan;
					if(!bIsTerminated && (uTotal >= uRecvd))
					{
						unsigned int uRemaining = uTotal - uRecvd;
						iEta = uRemaining / iAvgSpeed;
					}
				}
			}

			if(uTotal > 0)
			{
				double dPerc = (double)(((double)uRecvd) * 100.0) / (double)uTotal;
				unsigned int uPerc = (unsigned int)dPerc;
				iW -= 2;
				int iL = (int) ((((double)iW) * dPerc) / 100.0);
				//iR = iW - iL;
				p->fillRect(5,5,iL,10,bIsTerminated ? QColor(140,110,110) : QColor(200,100,100));

				txt = QString(__tr2qs_ctx("%1 of %2 bytes (%3 %)","http")).arg(uRecvd).arg(uTotal).arg(dPerc,0,'f',2);
			} else {
				txt = QString(__tr2qs_ctx("%1 bytes","http")).arg(m_pHttpRequest->receivedSize());
			}

			p->setPen(Qt::black);

			p->drawText(4,19,width - 8,height - 8,Qt::AlignTop | Qt::AlignLeft,txt);

			int iLeftHalf = (iW - 2) / 2;
			int iRightHalf = iW - (iLeftHalf + 1);
			int iLineSpacing = fm.lineSpacing() + 2;

			if(iAvgSpeed >= 0)txt = QString(__tr2qs_ctx("AVG: %1 bytes/sec","http")).arg(iAvgSpeed);
			else txt = __tr2qs_ctx("AVG: ? bytes/sec","http");


			int iDaH = height - (iLineSpacing + 4);

			p->setPen(QColor(180,180,200));
			p->drawRect(4,iDaH,iLeftHalf,iLineSpacing);
			p->fillRect(5,iDaH + 1,iLeftHalf - 2,iLineSpacing - 2,bIsTerminated ? QColor(210,210,210) : QColor(190,190,240));
			p->setPen(bIsTerminated ? Qt::darkGray : Qt::black);
			p->drawText(6,iDaH,iLeftHalf - 4,iLineSpacing,Qt::AlignLeft | Qt::AlignVCenter,txt);

			unsigned int uD,uH,uM,uS;

			if(bIsTerminated)
			{
				kvi_secondsToDaysHoursMinsSecs(kvi_timeSpan(m_tTransferEndTime,m_tTransferStartTime),&uD,&uH,&uM,&uS);
				txt = "TOT: ";
				if(uD > 0)txt += QString(__tr2qs_ctx("%1d %2h %3m %4s","http")).arg(uD).arg(uH).arg(uM).arg(uS);
				else if(uH > 0)txt += QString(__tr2qs_ctx("%2h %3m %4s","http")).arg(uH).arg(uM).arg(uS);
				else txt += QString(__tr2qs_ctx("%3m %4s","http")).arg(uM).arg(uS);
			} else {
				if(iEta >= 0)
				{
					kvi_secondsToDaysHoursMinsSecs(iEta,&uD,&uH,&uM,&uS);
					txt = "ETA: ";
					if(uD > 0)txt += QString(__tr2qs_ctx("%1d %2h %3m %4s","http")).arg(uD).arg(uH).arg(uM).arg(uS);
					else if(uH > 0)txt += QString(__tr2qs_ctx("%2h %3m %4s","http")).arg(uH).arg(uM).arg(uS);
					else txt += QString(__tr2qs_ctx("%3m %4s","http")).arg(uM).arg(uS);
				} else {
					txt = "ETA: ???";
				}
			}

			p->setPen(QColor(180,180,200));
			p->drawRect(width - (4 + iRightHalf),iDaH,iRightHalf,iLineSpacing);
			p->fillRect(width - (3 + iRightHalf),iDaH + 1,iRightHalf - 2,iLineSpacing - 2,bIsTerminated ? QColor(210,210,210) : QColor(190,190,240));
			p->setPen(bIsTerminated ? Qt::darkGray : Qt::black);
			p->drawText(width - (2 + iRightHalf),iDaH,iRightHalf - 4,iLineSpacing,Qt::AlignLeft | Qt::AlignVCenter,txt);

			p->setPen(Qt::lightGray);
			p->drawRect(1,1,width - 2,height - 2);
*/
		}
		break;
	}

}

int KviDccFileTransfer::displayHeight()
{
	return 70;
}

QString KviDccFileTransfer::tipText()
{

	QString s;

	s = QString("<table><tr><td bgcolor=\"#000000\"><font color=\"#FFFFFF\"><b>DCC %1 (ID %2)</b></font></td></tr>").arg(m_szDccType.ptr()).arg(id());

	s += "<tr><td bgcolor=\"#404040\"><font color=\"#FFFFFF\">";
	s += __tr2qs_ctx("Transfer Log","dcc");
	s += "</font></td></tr>";
	s += "<tr><td bgcolor=\"#C0C0C0\">";
	s += m_szTransferLog;
	s += "</td></tr>";
	s += "<table>";

	return s;
}

void KviDccFileTransfer::init()
{
	if(g_pDccFileTransfers)return;
	g_pDccFileTransfers = new KviPtrList<KviDccFileTransfer>;
	g_pDccFileTransfers->setAutoDelete(false);

	QPixmap * pix = g_pIconManager->getImage("kvi_dccfiletransfericons.png");
	if(pix)g_pDccFileTransferIcon = new QPixmap(*pix);
	else g_pDccFileTransferIcon = new QPixmap(192,128);
}

void KviDccFileTransfer::done()
{
	if(!g_pDccFileTransfers)return;
	while(KviDccFileTransfer * t = g_pDccFileTransfers->first())
		delete t;
	delete g_pDccFileTransfers;
	g_pDccFileTransfers = 0;
	delete g_pDccFileTransferIcon;
	g_pDccFileTransferIcon = 0;
}

unsigned int KviDccFileTransfer::transferCount()
{
	if(!g_pDccFileTransfers)return 0;
	return g_pDccFileTransfers->count();
}

unsigned int KviDccFileTransfer::runningTransfersCount()
{
	if(!g_pDccFileTransfers)return 0;
	unsigned int cnt = 0;
	for(KviDccFileTransfer * t = g_pDccFileTransfers->first();t;t = g_pDccFileTransfers->next())
	{
		if(t->active())cnt++;
	}
	return cnt;
}

bool KviDccFileTransfer::handleResumeAccepted(const char * filename,const char * port)
{
	if(!g_pDccFileTransfers)return false;

	for(KviDccFileTransfer * t = g_pDccFileTransfers->first();t;t = g_pDccFileTransfers->next())
	{
		if(t->resumeAccepted(filename,port))return true;
	}

	return false;
}

bool KviDccFileTransfer::handleResumeRequest(const char * filename,const char * port,unsigned int filePos)
{
	if(!g_pDccFileTransfers)return false;

	for(KviDccFileTransfer * t = g_pDccFileTransfers->first();t;t = g_pDccFileTransfers->next())
	{
		if(t->doResume(filename,port,filePos))return true;
	}

	return false;
}

void KviDccFileTransfer::outputAndLog(const QString &s)
{
	KviWindow * out = transferWindow();
	addToTransferLog(s);
	if(out)out->output(KVI_OUT_DCCMSG,"[%s]: %Q",m_szTransferIdString.ptr(),&s);
}

void KviDccFileTransfer::outputAndLog(int msgtype,const QString &s)
{
	KviWindow * out = transferWindow();
	addToTransferLog(s);
	if(out)out->output(msgtype,"[%s]: %Q",m_szTransferIdString.ptr(),&s);
}


void KviDccFileTransfer::addToTransferLog(const QString &s)
{
	QDateTime dt = QDateTime::currentDateTime();
	QString ts;
	ts.sprintf("[%4d.%2d.%2d %2d:%2d:%2d] ",dt.date().year(),dt.date().month(),dt.date().day(),dt.time().hour(),dt.time().minute(),dt.time().second());
	m_szTransferLog += ts+s;
	m_szTransferLog += "<br>";
}


void KviDccFileTransfer::connectionInProgress()
{
	KviWindow * out = transferWindow();

	if(m_pDescriptor->bActive)
	{
		// ACTIVE CONNECTION
//		if((kvi_strEqualCS(m_szDccType.ptr(), "RECV")) || (kvi_strEqualCS(m_szDccType.ptr(),"TRECV"))) 
//		{
//			// FIXME: that's not true!... we're NOT connected here
//			if(TRIGGER_EVENT_5PARAM_RETVALUE(KviEvent_OnDCCGetConnected,this,m_pDescriptor->szPort.ptr(),m_pDescriptor->szFileName.ptr(),m_pDescriptor->szNick.ptr(),m_pDescriptor->szUser.ptr(),m_pDescriptor->szHost.ptr()));
//		} else {
//			if(TRIGGER_EVENT_5PARAM_RETVALUE(KviEvent_OnDCCSendConnected,this,m_pDescriptor->szPort.ptr(),m_pDescriptor->szFileName.ptr(),m_pDescriptor->szNick.ptr(),m_pDescriptor->szUser.ptr(),m_pDescriptor->szHost.ptr()));
//		}
//
		m_szStatusString = __tr2qs_ctx("Contacting host %1 on port %2","dcc").arg(m_pDescriptor->szIp.ptr()).arg(m_pDescriptor->szPort.ptr());
		outputAndLog(m_szStatusString);
		displayUpdate();
		return;
	}

	// PASSIVE CONNECTION
	m_szStatusString = __tr2qs_ctx("Listening on interface %1 port %2","dcc").arg(m_pMarshal->localIp()).arg(m_pMarshal->localPort());
	outputAndLog(m_szStatusString);

	if(m_pDescriptor->bSendRequest)
	{
		KviStr ip;
		if(m_pDescriptor->szFakeIp.hasData())
		{
			ip = m_pDescriptor->szFakeIp;
		} else {
			ip = m_pDescriptor->szListenIp;

			if(KVI_OPTION_BOOL(KviOption_boolDccGuessIpFromServerWhenLocalIsUnroutable))
			{
				if(!kvi_isRoutableIpString(ip.ptr()))
				{
					// try to get the IP that the IRC server can see
					if(m_pDescriptor->console())
					{
						KviStr tmp = m_pDescriptor->console()->currentLocalHostIp();
						if(tmp.hasData())
						{
							ip = tmp;
							outputAndLog(__tr2qs_ctx("The local IP address is private, determining from IRC server: %1","dcc").arg(QString(ip.ptr())));
						} else {
							outputAndLog(__tr2qs_ctx("The local IP address is private, but unable to determine it from the IRC server","dcc"));
						}
					} else {
						outputAndLog(__tr2qs_ctx("The local IP address is private, but have no IRC server to determine it from","dcc"));
					}
				}
			}
		}
		
		KviStr port   = m_pDescriptor->szFakePort.hasData() ? m_pDescriptor->szFakePort.ptr() : m_pMarshal->localPort();
		//#warning "OPTION FOR SENDING 127.0.0.1 and so on (not an unsigned nuumber)"
		struct in_addr a;
		if(kvi_stringIpToBinaryIp(ip.ptr(),&a))ip.setNum(htonl(a.s_addr));

		KviStr fName  = m_pDescriptor->szFileName.ptr();
		fName.cutToLast('/'); // just to be sure...
		fName.replaceAll(' ',m_pDescriptor->bRecvFile ? "\\ " : "_"); // be cool :)

		// Zero port requests want DCC SEND as back-request
		KviStr szReq;
		if(m_pDescriptor->isZeroPortRequest())
		{
			szReq = "SEND";
			m_pDescriptor->console()->link()->sendFmtData("PRIVMSG %s :%cDCC %s %s %s %s %s %s%c",
					m_pDescriptor->szNick.ptr(),0x01,szReq.ptr(),
					fName.ptr(),ip.ptr(),port.ptr(),
					m_pDescriptor->szFileSize.ptr(),m_pDescriptor->zeroPortRequestTag(),0x01);
		} else {
			szReq = m_szDccType;
			m_pDescriptor->console()->link()->sendFmtData("PRIVMSG %s :%cDCC %s %s %s %s %s%c",
					m_pDescriptor->szNick.ptr(),0x01,szReq.ptr(),
					fName.ptr(),ip.ptr(),port.ptr(),
					m_pDescriptor->szLocalFileSize.ptr(),0x01);
		}
		outputAndLog(__tr2qs_ctx("Sent DCC %1 request to %2, waiting for remote client to connect...","dcc").arg(szReq.ptr()).arg(m_pDescriptor->szNick.ptr()));
	} else {
		outputAndLog(__tr2qs_ctx("DCC %1 request not sent, awaiting manual connection","dcc").arg(m_szDccType.ptr()));
	}
	displayUpdate();
}

void KviDccFileTransfer::startingSSLHandshake()
{
#ifdef COMPILE_SSL_SUPPORT
	outputAndLog(KVI_OUT_SSL,__tr2qs_ctx("Low-level transport connection established","dcc"));
	outputAndLog(KVI_OUT_SSL,__tr2qs_ctx("Starting Secure Socket Layer handshake","dcc"));
#endif
}

void KviDccFileTransfer::sslError(const char * msg)
{
#ifdef COMPILE_SSL_SUPPORT
	outputAndLog(KVI_OUT_DCCERROR,__tr2qs_ctx("[SSL ERROR]: %1","dcc").arg(msg));
#endif
}




bool KviDccFileTransfer::event(QEvent *e)
{
	if(e->type() == KVI_THREAD_EVENT)
	{
		switch(((KviThreadEvent *)e)->id())
		{
			case KVI_DCC_THREAD_EVENT_ERROR:
			{
				int * err = ((KviThreadDataEvent<int> *)e)->getData();
				QString szErrorString = kvi_getErrorString(*err);
				delete err;
				if(m_pDescriptor->bRecvFile)
					g_pApp->fileDownloadTerminated(false,m_pDescriptor->szFileName.ptr(),m_pDescriptor->szLocalFileName.ptr(),m_pDescriptor->szNick.ptr(),szErrorString.latin1());
				m_szStatusString = __tr2qs_ctx("Transfer failed: ","dcc");
				m_szStatusString += szErrorString;
				outputAndLog(KVI_OUT_DCCERROR,m_szStatusString);
				m_eGeneralStatus = Failure;
				displayUpdate();
				return true;
			}
			break;
			case KVI_DCC_THREAD_EVENT_SUCCESS:
//				if ((kvi_strEqualCS(m_szDccType.ptr(), "RECV")) || (kvi_strEqualCS(m_szDccType.ptr(),"TRECV"))) 
//				{
//					if(TRIGGER_EVENT_5PARAM_RETVALUE(KviEvent_OnDCCGetTransferComplete,this,m_pDescriptor->szPort.ptr(),m_pDescriptor->szFileName.ptr(),m_pDescriptor->szNick.ptr(),m_pDescriptor->szUser.ptr(),m_pDescriptor->szHost.ptr()));
//				} else {
//					if(TRIGGER_EVENT_5PARAM_RETVALUE(KviEvent_OnDCCSendTransferComplete,this,m_pDescriptor->szPort.ptr(),m_pDescriptor->szFileName.ptr(),m_pDescriptor->szNick.ptr(),m_pDescriptor->szUser.ptr(),m_pDescriptor->szHost.ptr()));
//				}

				if(KVI_OPTION_BOOL(KviOption_boolNotifyDccSendSuccessInConsole))
				{
					g_pApp->activeConsole()->output(KVI_OUT_DCCMSG,__tr2wc_ctx("DCC %s transfer with %s@%s:%s completed: \r![!dbl]play $0\r%s\r","dcc"),
						m_pDescriptor->bIsTdcc ? (m_pDescriptor->bRecvFile ? "TRECV" : "TSEND") : (m_pDescriptor->bRecvFile ? "RECV" : "SEND"),
						m_pDescriptor->szNick.ptr(),m_pDescriptor->szIp.ptr(),m_pDescriptor->szPort.ptr(),
						m_pDescriptor->szLocalFileName.ptr());
				}
				if(m_pDescriptor->bRecvFile)g_pApp->fileDownloadTerminated(true,m_pDescriptor->szFileName.ptr(),m_pDescriptor->szLocalFileName.ptr(),m_pDescriptor->szNick.ptr());
/*
				FIXME: Autoclear ?
				if(KVI_OPTION_BOOL(KviOption_boolAutoCloseDccSendOnSuccess))close();
*/
				m_szStatusString = __tr2qs_ctx("Transfer completed","dcc");
				outputAndLog(m_szStatusString);
				m_eGeneralStatus = Success;
				displayUpdate();
				return true;
			break;
			case KVI_DCC_THREAD_EVENT_MESSAGE:
			{
				KviStr * str = ((KviThreadDataEvent<KviStr> *)e)->getData();
				outputAndLog(QString(__tr_no_xgettext_ctx(str->ptr(),"dcc")));
				delete str;
				return true;
			}
			break;
			default:
				debug("Invalid event type %d received",((KviThreadEvent *)e)->id());
			break;
		}
	}
//#warning "Remove this!"
//	if(e->type() == QEvent::Close)debug("Close event received");
	return KviDccFileTransfer::event(e);
}

void KviDccFileTransfer::handleMarshalError(int err)
{
	KviStr szErr = kvi_getErrorString(err);
	m_eGeneralStatus = Failure;
	m_szStatusString = __tr2qs_ctx("Transfer failed: ","dcc");
	m_szStatusString += szErr.ptr();
	outputAndLog(m_szStatusString);
	displayUpdate();
}

void KviDccFileTransfer::connected()
{
	outputAndLog(__tr2qs_ctx("Connected to %1:%2","dcc").arg(m_pMarshal->remoteIp()).arg(m_pMarshal->remotePort()));
	outputAndLog(__tr2qs_ctx("Local end is %1:%2","dcc").arg(m_pMarshal->localIp()).arg(m_pMarshal->localPort()));

	if(!(m_pDescriptor->bActive))
	{
		m_pDescriptor->szIp   = m_pMarshal->remoteIp();
		m_pDescriptor->szPort = m_pMarshal->remotePort();
		m_pDescriptor->szHost = m_pMarshal->remoteIp();
	}

	if(m_pDescriptor->bRecvFile)
	{
		KviDccRecvThreadOptions * o = new KviDccRecvThreadOptions;
		o->szFileName     = m_pDescriptor->szLocalFileName.ptr();
		bool bOk;
		o->iTotalFileSize = m_pDescriptor->szFileSize.toInt(&bOk);
		if(!bOk)o->iTotalFileSize = -1;
		o->bResume        = m_pDescriptor->bResume;
		o->bIsTdcc        = m_pDescriptor->bIsTdcc;
		o->bSendZeroAck   = KVI_OPTION_BOOL(KviOption_boolSendZeroAckInDccRecv);
		o->bNoAcks         = m_pDescriptor->bNoAcks;
		m_pSlaveRecvThread = new KviDccRecvThread(this,m_pMarshal->releaseSocket(),o);
		m_pSlaveRecvThread->start();
	} else {
		KviDccSendThreadOptions * o = new KviDccSendThreadOptions;
		o->szFileName     = m_pDescriptor->szLocalFileName.ptr();
		o->bFastSend      = KVI_OPTION_BOOL(KviOption_boolUseFastDccSend);
		o->iIdleStepLengthInMSec = KVI_OPTION_UINT(KviOption_uintDccSendIdleStepInMSec);
		bool bOk;
		o->bIsTdcc        = m_pDescriptor->bIsTdcc;
		o->iStartPosition  = m_pDescriptor->szFileSize.toInt(&bOk);
		if(!bOk || (o->iStartPosition < 0))o->iStartPosition = 0;
		o->iPacketSize     = KVI_OPTION_UINT(KviOption_uintDccSendPacketSize);
		if(o->iPacketSize < 32)o->iPacketSize = 32;
		o->bNoAcks         = m_pDescriptor->bNoAcks;
		m_pSlaveSendThread = new KviDccSendThread(this,m_pMarshal->releaseSocket(),o);
		m_pSlaveSendThread->start();
	}

	m_szStatusString = __tr2qs_ctx("Transferring data","dcc");
	outputAndLog(m_szStatusString);
	m_eGeneralStatus = Transferring;
	displayUpdate();
}

bool KviDccFileTransfer::resumeAccepted(const char *filename,const char* port)
{
	if(kvi_strEqualCI(filename,m_pDescriptor->szFileName.ptr()) || KVI_OPTION_BOOL(KviOption_boolAcceptBrokenFileNameDccResumeRequests))
	{
		if(kvi_strEqualCI(port,m_pDescriptor->szPort.ptr()) &&
			(!m_pSlaveRecvThread) && m_pDescriptor->bResume && m_pDescriptor->bRecvFile)
		{
			outputAndLog(__tr2qs_ctx("RESUME accepted, transfer will begin at position %1","dcc").arg(m_pDescriptor->szLocalFileSize.ptr()));

			int ret = m_pMarshal->dccConnect(m_pDescriptor->szIp.ptr(),
							m_pDescriptor->szPort.ptr(),m_pDescriptor->bDoTimeout);

			if(ret != KviError_success)handleMarshalError(ret);
			else {
				m_szStatusString = __tr2qs_ctx("Contacting host %1 on port %2","dcc").arg(m_pDescriptor->szIp.ptr()).arg(m_pDescriptor->szPort.ptr());
				outputAndLog(m_szStatusString);
				displayUpdate();
			}
	
			return true;
		}
	}
	return false;
}

bool KviDccFileTransfer::doResume(const char * filename,const char * port,unsigned int filePos)
{
	if(kvi_strEqualCI(port,m_pMarshal->dccPort()) &&
		(!m_pSlaveRecvThread) && (!m_pDescriptor->bRecvFile))
	{
		if(kvi_strEqualCI(filename,m_pDescriptor->szFileName.ptr()) || KVI_OPTION_BOOL(KviOption_boolAcceptBrokenFileNameDccResumeRequests))
		{
			bool bOk;
			unsigned int iLocalFileSize = m_pDescriptor->szLocalFileSize.toUInt(&bOk);
			if(!bOk)
			{
				// ops...internal error
				outputAndLog(KVI_OUT_DCCERROR,__tr2qs_ctx("Internal error in RESUME request","dcc"));
				return false;
			}
			if(iLocalFileSize <= filePos)
			{
				outputAndLog(KVI_OUT_DCCERROR,__tr2qs_ctx("Invalid RESUME request: Position %1 is larger than file size","dcc").arg(filePos));
				return false;
			}

			outputAndLog(KVI_OUT_DCCERROR,__tr2qs_ctx("Accepting RESUME request, transfer will begin at position %1","dcc").arg(filePos));

			m_pDescriptor->szFileSize.setNum(filePos);
			m_pDescriptor->console()->link()->sendFmtData("PRIVMSG %s :%cDCC ACCEPT %s %s %u%c",
				m_pDescriptor->szNick.ptr(),0x01,filename,port,filePos,0x01);
			return true;
		}
	}
	return false;
}



#include "m_send.moc"
