#ifndef _SEND_H_
#define _SEND_H_
//
//   File : send.h
//   Creation date : Tue Sep 24 09 2000 15:06:12 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "kvi_window.h"
#include "kvi_string.h"

#include "descriptor.h"
#include "window.h"
#include "thread.h"

#include "kvi_sockettype.h"

#include "kvi_list.h"
#include <qlabel.h>
#include <qprogressbar.h>
#include <qpopupmenu.h>
#include <qhbox.h>
#include <qvbox.h>
#include <qfile.h>

#include "kvi_filetransfer.h"
#include "kvi_time.h"


typedef struct _KviDccSendThreadOptions
{
	KviStr szFileName;
	int    iStartPosition;
	int    iPacketSize;
	int    iIdleStepLengthInMSec;
	bool   bFastSend;
	bool   bNoAcks;
	bool   bIsTdcc;
} KviDccSendThreadOptions;


class KviDccSendThread : public KviDccThread
{
public:
	KviDccSendThread(QObject * par,kvi_socket_t fd,KviDccSendThreadOptions * opt);
	~KviDccSendThread();
private:
	KviDccSendThreadOptions * m_pOpt;
	// stats: SHARED!!!
	kvi_time_t  m_iStartTime;
	int         m_iAverageSpeed;
	int         m_iInstantSpeed;
	int         m_iFilePosition;
	int         m_iAckedBytes;
public:
	void initGetInfo();
	int averageSpeed(){ return m_iAverageSpeed; };
	int instantSpeed(){ return m_iInstantSpeed; };
	int elapsedTime();
	int filePosition(){ return m_iFilePosition; };
	void doneGetInfo();
	int ackedBytes(){ return m_iAckedBytes; };
protected:
	virtual void run();
};

typedef struct _KviDccRecvThreadOptions
{
	bool   bResume;
	KviStr szFileName;
	int    iTotalFileSize;
	bool   bSendZeroAck;
	bool   bNoAcks;
	bool   bIsTdcc;
} KviDccRecvThreadOptions;

class KviDccRecvThread : public KviDccThread
{
public:
	KviDccRecvThread(QObject * par,kvi_socket_t fd,KviDccRecvThreadOptions * opt);
	~KviDccRecvThread();
protected:
	KviDccRecvThreadOptions * m_pOpt;

	// stats: SHARED!
	kvi_time_t m_iStartTime;
	int        m_iAverageSpeed;
	int        m_iInstantSpeed;
	int        m_iFilePosition;

	// internal
	int        m_iTotalReceivedBytes;
	int        m_iInstantReceivedBytes;
	int        m_iLastInstantCheckTime;
	QFile    * m_pFile;
public:
	void initGetInfo();
	int averageSpeed(){ return m_iAverageSpeed; };
	int instantSpeed(){ return m_iInstantSpeed; };
	int elapsedTime();
	int filePosition(){ return m_iFilePosition; };
	void doneGetInfo();
protected:
	void postMessageEvent(const char * msg);
	void updateStats();
	bool sendAck(int filePos);
	virtual void run();
};


class KviDccMarshal;
class QPainter;
class QPopupMenu;

class KviDccFileTransfer : public KviFileTransfer, public KviDccMarshalOutputContext
{
	enum GeneralStatus { Connecting , Transferring , Success , Failure };
	Q_OBJECT
public:
	KviDccFileTransfer(KviDccDescriptor * dcc);
	~KviDccFileTransfer();
private:
	KviDccSendThread       * m_pSlaveSendThread;
	KviDccRecvThread       * m_pSlaveRecvThread;
	KviDccDescriptor       * m_pDescriptor;
	KviDccMarshal          * m_pMarshal;

	KviStr                   m_szTarget;
	KviStr                   m_szDccType;
	KviStr                   m_szTransferIdString;
	QPopupMenu             * m_pFilePopup;

	QString                  m_szStatusString;
	GeneralStatus            m_eGeneralStatus;

	QString                  m_szTransferLog; // html

	// cached stats
	unsigned int             m_uTotalFileSize; // total file size to transfer
public:
	bool resumeAccepted(const char * filename,const char * port);
	bool doResume(const char * filename,const char * port,unsigned int filePos);

	static void init();
	static void done();
	static unsigned int runningTransfersCount();
	static unsigned int transferCount();
	static bool handleResumeAccepted(const char * filename,const char * port);
	static bool handleResumeRequest(const char * filename,const char * port,unsigned int filePos);


	virtual bool event(QEvent *e);

	virtual KviWindow * dccMarshalOutputWindow();
	virtual const char * dccMarshalOutputContextString();


	virtual void displayPaint(QPainter * p,int column,int width,int height);
	virtual int displayHeight();
	virtual void fillContextPopup(QPopupMenu * m,int column);
	virtual bool active();
	virtual void die();
	virtual QString tipText();
	virtual QString localFileName();
protected:
	void startConnection();
	void addToTransferLog(const QString &s);
	void outputAndLog(const QString &s);
	void outputAndLog(int msgtype,const QString &s);
protected slots:
	void connectionInProgress();
	void sslError(const char * msg);
	void startingSSLHandshake();
	void handleMarshalError(int err);
	void connected();

	void abort();
/*
	void updateDccRecv();
	void updateDccSend();
*/
};

#endif //_SEND_H_
