//
//   File : class_window.cpp
//   Creation date : Tue Now 26 21:39:59 CEST 2002 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//



#include "class_window.h"

#include "kvi_error.h"
#include "kvi_debug.h"
#include "kvi_command.h"
#include "kvi_locale.h"
#include "kvi_window.h"
#include "kvi_frame.h"

static KviScriptObjectClass * g_pWindowClass = 0;

static KviScriptObject * windowClassCreateInstance(KviScriptObjectClass * c,
	KviScriptObject * p, const char * n)
{
	return new KviScriptWindowObject(c, p, n);
}

KviScriptWindowObject::KviScriptWindowObject(KviScriptObjectClass * c, KviScriptObject * p,
	const char *n) : KviScriptWidgetObject(c, p, n)
{
}

KviScriptWindowObject::~KviScriptWindowObject()
{
}

/*
	@doc:	window
	@title:
		window class
	@type:
		class
	@short:
		A MDI window object
	@inherits:
		[class]object[/class]
		[class]widget[/class]
	@description:
		This class rappresents a MDI window in a KVIrc frame.
		Regardless of the parent object passed, the window will be always
		a "physical" child of the frame window: you can't embed a window object
		inside a widget or another window. The parent object will just "own" this
		window object and eventually destroy it when it dies.
	@functions:
		!fn: $insertItem(<text>,<popup>)
		Returns an integer that identifies the inserted item.
*/

bool KviScriptWindowObject::init(KviCommand *c)
{
	KviWindow * w = new KviWindow(KVI_WINDOW_TYPE_SCRIPTOBJECT,c->window()->frame(),name());
	setObject(w);
	c->window()->frame()->addWindow(w,false);
	return true;
}

#define windowFuncReg(__nam, __func) \
	g_pWindowClass->registerFunctionHandler(__nam, \
	(KviScriptObjectFunctionHandlerProc)(KVI_PTR2MEMBER(KviScriptWindowObject::__func)), \
	0, true);

void KviScriptWindowObject::registerSelf()
{
	KviScriptObjectClass * base = g_pScriptObjectController-> \
		lookupClass("widget");
	__range_valid(base);

	g_pWindowClass = new KviScriptObjectClass(base, "window",
		windowClassCreateInstance, true);

	windowFuncReg("setIcon", functionSetIcon);
}

void KviScriptWindowObject::unregisterSelf()
{
	delete g_pWindowClass;
}


bool KviScriptWindowObject::functionSetIcon(KviCommand * c,KviParameterList * p,KviStr & b)
{
	ENTER_CONTEXT(c,"window::setIcon");

	KviStr * pIcon = p->safeFirst();

//	if(widget())((QWindow *)widget())->insertItem(pText->ptr());

	return c->leaveContext();
}



#include "m_class_window.moc"
