//
//   File : libkviobjects.cpp
//   Creation date : Wed Sep 09 2000 20:59:01 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "kvi_module.h"

#include "class_button.h"
#include "class_checkbox.h"
#include "class_urllabel.h"
#include "class_lineedit.h"
#include "class_multilineedit.h"
#include "class_combobox.h"
#include "class_widget.h"
#include "class_list.h"
#include "class_file.h"
#include "class_string.h"
#include "class_socket.h"
#include "class_layout.h"
#include "class_menubar.h"
#include "class_window.h"


static bool objects_module_init(KviModule * m)
{
	KviScriptListObject::registerSelf();
	KviScriptFileObject::registerSelf();
	KviScriptStringObject::registerSelf();
	KviScriptSocketObject::registerSelf();

	KviScriptWidgetObject::registerSelf();
	// add new widget-inherited objects here. NOT before the previous line ;)

	KviScriptButtonObject::registerSelf();
	KviScriptCheckBoxObject::registerSelf(); // must be after button !
	KviScriptLabelObject::registerSelf();
	KviScriptUrlLabelObject::registerSelf();
	KviScriptLineEditObject::registerSelf();
	KviScriptMLEditObject::registerSelf();
	KviScriptComboBoxObject::registerSelf();

	KviScriptMenuBarObject::registerSelf();

	KviScriptWindowObject::registerSelf();

	KviScriptLayoutObject::registerSelf();

	return true;
}

static bool objects_module_cleanup(KviModule *m)
{
	KviScriptLayoutObject::unregisterSelf();

	KviScriptWindowObject::unregisterSelf();

	KviScriptMenuBarObject::unregisterSelf();

	KviScriptCheckBoxObject::unregisterSelf();
	KviScriptButtonObject::unregisterSelf();

	KviScriptUrlLabelObject::unregisterSelf();
	KviScriptLabelObject::unregisterSelf();
	KviScriptLineEditObject::unregisterSelf();
	KviScriptMLEditObject::unregisterSelf();
	KviScriptComboBoxObject::unregisterSelf();
	
	KviScriptWidgetObject::unregisterSelf();

	KviScriptFileObject::unregisterSelf();
	KviScriptListObject::unregisterSelf();
	KviScriptStringObject::unregisterSelf();
	KviScriptSocketObject::unregisterSelf();

	m->unregisterMetaObject("KviScriptMenuBarObject");
	m->unregisterMetaObject("KviScriptWindowObject");
	m->unregisterMetaObject("KviScriptCheckBoxObject");
	m->unregisterMetaObject("KviScriptButtonObject");
	m->unregisterMetaObject("KviScriptLabelObject");
	m->unregisterMetaObject("KviScriptUrlLabelObject");
	m->unregisterMetaObject("KviScriptLineEditObject");
	m->unregisterMetaObject("KviScriptMLEditObject");
	m->unregisterMetaObject("KviScriptComboBoxObject");
	m->unregisterMetaObject("KviScriptWidgetObject");
	m->unregisterMetaObject("KviScriptFileObject");
	m->unregisterMetaObject("KviScriptListObject");
	m->unregisterMetaObject("KviScriptStringObject");
	m->unregisterMetaObject("KviScriptSocketObject");
	m->unregisterMetaObject("KviScriptLayoutObject");
	return true;
}

static bool objects_module_can_unload(KviModule *)
{
	// Actually this object remains locked in memory

	// FIXME: unlock when no more instances & derived classes!
	return false;
}

KviModuleInfo kvirc_module_info =
{
	"Objects",
	"1.0.0",
	"Copyright (C) 2000:\n" \
	"  Szymon Stefanek (pragma at kvirc dot net)\n" \
	"  Krzysztof Godlewski (kristoff at poczta dot wprost dot pl)",
	"Object classes for KVirc scripting language\n",
	objects_module_init,
	objects_module_can_unload,
	0,
	objects_module_cleanup
};
