//
//   File : optw_lag.cpp
//   Creation date : Wed Cct 16 10:45:54 CEST 2002 by Juanjo lvarez (juanjux@yahoo.es)
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//


#include "optw_lag.h"

#include <qlayout.h>
#include <qtooltip.h>

#include "kvi_options.h"
#include "kvi_locale.h"

//#warning "Info tips"


KviLagOptionsWidget::KviLagOptionsWidget(QWidget * parent)
: KviOptionsWidget(parent,"lag_options_widget")
{
	createLayout(3,1);

	KviBoolSelector * pUse = addBoolSelector(0,0,0,0, __tr2qs("Enable lag meter"), KviOption_boolUseLagMeterEngine);
    QGroupBox *g = addGroupBox(0,1,0,1,1,QGroupBox::Horizontal,__tr2qs("Configuration"),KVI_OPTION_BOOL(KviOption_boolUseLagMeterEngine));
    KviUIntSelector * pInterval = addUIntSelector(g, __tr2qs("Interval between lag checks:"), KviOption_uintLagCheckingInterval, 2, 10000, 15, KVI_OPTION_BOOL(KviOption_boolUseLagMeterEngine));
    pInterval->setSuffix(__tr2qs(" sec"));
#ifdef COMPILE_INFO_TIPS
    QToolTip::add(pInterval,__tr2qs("<center>Minimum value: <b>2 sec</b><br>Maximum value: <b>10000 sec</b></center>"));
#endif
    connect(pUse,SIGNAL(toggled(bool)),pInterval,SLOT(setEnabled(bool)));

    KviUIntSelector * pAlarm = addUIntSelector(g, __tr2qs("Trigger event if lag exceeds:"), KviOption_uintLagAlarmTime, 1000, 1000000, 30000, KVI_OPTION_BOOL(KviOption_boolUseLagMeterEngine));
    pAlarm->setSuffix(__tr2qs(" msec"));
#ifdef COMPILE_INFO_TIPS
        QToolTip::add(pAlarm,__tr2qs("<center>Minimum value: <b>1000 msec</b><br>Maximum value: <b>1000000 msec</b></center>"));
#endif
    connect(pUse,SIGNAL(toggled(bool)),pAlarm,SLOT(setEnabled(bool)));

    KviBoolSelector * pShow = addBoolSelector(g, __tr2qs("Show lag in IRC context display"), KviOption_boolShowLagOnContextDisplay, KVI_OPTION_BOOL(KviOption_boolUseLagMeterEngine));
    connect(pUse,SIGNAL(toggled(bool)),pShow,SLOT(setEnabled(bool)));

    connect(pUse,SIGNAL(toggled(bool)),g,SLOT(setEnabled(bool)));

	addRowSpacer(0,2,0,2);

#ifdef COMPILE_INFO_TIPS
    QToolTip::add(pUse,__tr2qs("<center>This enables the lag meter engine, which checks at regular intervals how much lag (latency) the server has.</center>"));
    QToolTip::add(pInterval,__tr2qs("<center>This specifies the time in seconds that the lag meter will wait between lag checks. Please note that setting this too low can actually cause lag.</center>"));
    QToolTip::add(pAlarm,__tr2qs("<center>This specifies the time in milliseconds that will trigger the <b>OnLagAlarmTimeUp</b> event that may be used in scripts for taking action when the lag is too high (e.g. reconnecting to the server). Once this level is reached, the next time the lag returns to a lower level, the event <b>OnLagAlarmTimeDown</b> will be triggered.</center>"));
    QToolTip::add(pShow,__tr2qs("<center>This makes the IRC context display applet show the current lag after the user's nickname (in seconds)</center>"));
#endif
}

KviLagOptionsWidget::~KviLagOptionsWidget()
{
}


#include "m_optw_lag.moc"
