//
//   File : optw_servers.cpp
//   Creation date : Wed Jun 12 2000 03:29:51 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2000 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "optw_servers.h"
#include "kvi_query.h"
#include "kvi_channel.h"
#include "kvi_locale.h"
#include "kvi_iconmanager.h"
#include "kvi_ircserverdb.h"
#include "kvi_ipeditor.h"
#include "kvi_netutils.h"
#include "kvi_settings.h"
#include "kvi_filedialog.h"
#include "kvi_config.h"
#include "kvi_scripteditor.h"
#include "kvi_options.h"
#include "kvi_uparser.h"
#include "kvi_app.h"
#include "kvi_console.h"
#include "kvi_ircurl.h"
#include "kvi_module.h"
#include "kvi_modulemanager.h"
#include "kvi_moduleextension.h"
#include "kvi_msgbox.h"
#include "kvi_mexserverimport.h"

#include <qtoolbutton.h>
#include <qlineedit.h>
#include <qcheckbox.h>
#include <qpopupmenu.h>
#include <qcursor.h>
#include <qtooltip.h>
#include <qvalidator.h>
#include <qmessagebox.h>
#include <qasciidict.h>
#include <qcombobox.h>
#include <qbuttongroup.h>
#include <qmessagebox.h>
#include <qinputdialog.h>

KviServerDetailsWidget::KviServerDetailsWidget(QWidget * par,KviIrcServer * s)
    : QDialog(par,"server_details",true)
{
	QGridLayout * g = new QGridLayout(this,4,4,10,4);

	setCaption(__tr2qs("Server Details"));
	setIcon(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_SERVER)));

	QString szTmp = "<font size=\"+1\"><b>irc";
	if(s->isIpV6())szTmp += "6";
	szTmp += "://";
	szTmp += s->m_szHostname;
	szTmp += ":";

  QString num;
  num.setNum(s->port());
  szTmp += num;
  szTmp += "</b></font>";

  QLabel * l = new QLabel(szTmp,this);
  l->setFrameStyle(QFrame::Raised | QFrame::StyledPanel);
  l->setAlignment(AlignCenter);
  l->setMargin(10);
  g->addMultiCellWidget(l,0,0,0,3);
  l->setMinimumWidth(200);


  l = new QLabel(__tr2qs("Description:"),this);
  g->addWidget(l,1,0);
  m_pDescEditor = new QLineEdit(this);
  g->addMultiCellWidget(m_pDescEditor,1,1,1,3);
  m_pDescEditor->setText(s->description());
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pDescEditor,__tr2qs("<center>This is a brief description of this server. " \
                                        "This field has no restrictions but many server lists use it to describe the server's physical location</center>"));
#endif


  // Identity tab

  QTabWidget * tw = new QTabWidget(this);
  g->addMultiCellWidget(tw,2,2,0,3);

  QWidget * tab = new QWidget(tw);
  QGridLayout * gl = new QGridLayout(tab,3,1,10,4);

  QGroupBox *gbox = new QGroupBox(2,QGroupBox::Horizontal,__tr2qs("Properties"),tab);
  gl->addWidget(gbox,0,0);

  l = new QLabel(__tr2qs("Username:"),gbox);
  m_pUserEditor = new QLineEdit(gbox);
  m_pUserEditor->setText(s->username());
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pUserEditor,__tr2qs("<center>This is the <b>username</b> that KVIrc will use to login to this server.\n" \
                                        "If this field is left empty (most common case), KVIrc will use the default username " \
                                        "specified in the \"Identity\" options tab.</center>"));
#endif

  l = new QLabel(__tr2qs("Password:"),gbox);
  m_pPassEditor = new QLineEdit(gbox);
  m_pPassEditor->setEchoMode(QLineEdit::Password);
  m_pPassEditor->setText(s->password());
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pPassEditor,__tr2qs("<center>If this server requires a password, put it in this field, otherwise leave it empty.</center>"));
#endif

  l = new QLabel(__tr2qs("Nickname:"),gbox);
  m_pNickEditor = new QLineEdit(gbox);
  QValidator * v = new QRegExpValidator(QRegExp("[^-0-9 ][^ ]*"),gbox);
  m_pNickEditor->setValidator(v);
  m_pNickEditor->setText(s->nickname());
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pNickEditor,__tr2qs("<center>You can specify a \"special\" <b>nickname</b> that will be used to log in to this server.<br>" \
                                        "If this field is left empty (most common case), the default nickname (specified in the \"Identity\" settings) will be used.</center>"));
#endif

  l = new QLabel(__tr2qs("Real name:"),gbox);
  m_pRealEditor = new QLineEdit(gbox);
  m_pRealEditor->setText(s->realName());
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pRealEditor,__tr2qs("<center>You can specify a \"special\" <b>real name</b> that will be used to login with this server.<br>" \
                                        "If you leave this field empty (most common case), the default \"real name\" (specified in the \"Identity\" settings) will be used.</center>"));
#endif


  gbox = new QGroupBox(1,QGroupBox::Horizontal,__tr2qs("User Mode"),tab);
  gl->addWidget(gbox,1,0);

  m_pUseDefaultInitUMode = new QCheckBox(__tr2qs("Use default user mode"),gbox);
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pUseDefaultInitUMode,__tr2qs("<center>If this is enabled, the global initial <b>user mode</b> (configured from"\
                  " the identity dialog) will be used. If disabled, you can configure an initial user mode for this server"));
#endif
  bool bHasUmode = !(s->initUMode().isEmpty());
  KviStr szDefUMode = KVI_OPTION_STRING(KviOption_stringDefaultUserMode);
  m_pUseDefaultInitUMode->setChecked(!bHasUmode);
  connect(m_pUseDefaultInitUMode,SIGNAL(toggled(bool)),this,SLOT(useDefaultInitUModeToggled(bool)));

  m_pIMode = new QCheckBox(__tr2qs("Invisible (+i)"),gbox);
  m_pIMode->setEnabled(bHasUmode);
  m_pIMode->setChecked(bHasUmode ? s->initUMode().contains('i',false) : szDefUMode.contains('i',false));

  m_pSMode = new QCheckBox(__tr2qs("Server notices (+s)"),gbox);
  m_pSMode->setEnabled(bHasUmode);
  m_pSMode->setChecked(bHasUmode ? s->initUMode().contains('s',false) : szDefUMode.contains('s',false));

  m_pWMode = new QCheckBox(__tr2qs("Wallops (+w)"),gbox);
  m_pWMode->setEnabled(bHasUmode);
  m_pWMode->setChecked(bHasUmode ? s->initUMode().contains('w',false) : szDefUMode.contains('w',false));


  l = new QLabel("",tab);
  gl->addWidget(l,2,0);

  gl->setRowStretch(2,1);
  //gl->setColStretch(1,1);


  tw->addTab(tab,__tr2qs("Identity"));



  // Connection tab

  tab = new QWidget(tw);
  gl = new QGridLayout(tab,6,1,10,4);


  l = new QLabel(__tr2qs("IP address:"),tab);
  gl->addWidget(l,0,0);
  m_pIpEditor = new KviIpEditor(tab,KviIpEditor::IpV4);
  gl->addWidget(m_pIpEditor,0,1);
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pIpEditor,__tr2qs("<center>This is the <b>IP address</b> of this server, it is here for caching purposes.<br>" \
                                      "If you leave this field empty, KVIrc will fill it in the first time it connects to the server. " \
                                      "If you enable the \"cache IP address\" option below, KVIrc will use it as a \"cached result\" " \
                                      "and avoid looking it up again.</center>"));
#endif
#ifdef COMPILE_IPV6_SUPPORT
  m_pIpEditor->setAddressType(s->isIpV6() ? KviIpEditor::IpV6 : KviIpEditor::IpV4);
#else
  m_pIpEditor->setAddressType(KviIpEditor::IpV4);
#endif

  if(!m_pIpEditor->setAddress(s->m_szIp))
  {
#ifdef COMPILE_IPV6_SUPPORT
    m_pIpEditor->setAddress(s->isIpV6() ? "0:0:0:0:0:0:0:0" : "0.0.0.0");
#else
    m_pIpEditor->setAddress("0.0.0.0");
#endif

  }

  m_pCacheIpCheck = new QCheckBox(__tr2qs("Cache IP address"),tab);
  gl->addMultiCellWidget(m_pCacheIpCheck,1,1,0,1);
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pCacheIpCheck,__tr2qs("<center>This check will enable <b>IP address caching</b> for this server:<br>" \
                                          "DNS lookups can be time-consuming and might be blocking on some platforms; " \
                                          "this option will cause KVIrc to look up the server hostname only once.</center>"));
#endif
  m_pCacheIpCheck->setChecked(s->cacheIp());


  m_pUseSSLCheck = new QCheckBox(__tr2qs("Use SSL protocol"),tab);
  gl->addMultiCellWidget(m_pUseSSLCheck,2,2,0,1);
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pUseSSLCheck,__tr2qs("<center>This check will cause the connection to use the <b>Secure Socket Layer</b> " \
                                         "encryption support. Obviously, this server must have support for this, too. :)</center>"));
#endif
#ifndef COMPILE_SSL_SUPPORT
  m_pUseSSLCheck->setEnabled(false);
#endif
  m_pUseSSLCheck->setChecked(s->useSSL());


  l = new QLabel(__tr2qs("Link filter:"),tab);
  gl->addWidget(l,3,0);
  m_pLinkFilterEditor = new QComboBox(true,tab);
  m_pLinkFilterEditor->setDuplicatesEnabled(false);
  gl->addWidget(m_pLinkFilterEditor,3,1);

  m_pLinkFilterEditor->insertItem("");

  g_pModuleManager->loadModulesByCaps("linkfilter");
  KviModuleExtensionDescriptorList * mexl = KviModuleExtensionManager::instance()->getExtensionList("linkfilter");


  if(mexl)
  {
    for(KviModuleExtensionDescriptor * d = mexl->first();d;d = mexl->next())
      m_pLinkFilterEditor->insertItem(d->name().ptr());
  }
  else
  {
    if(!s->linkFilter().isEmpty())
      m_pLinkFilterEditor->insertItem(s->linkFilter());
  }

  if(!s->linkFilter().isEmpty())
    m_pLinkFilterEditor->setCurrentText(s->linkFilter());
  else
    m_pLinkFilterEditor->setCurrentText("");


#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pLinkFilterEditor,__tr2qs("<center>This field specifies the name of a module that exports a link filter for this type of server.<br>" \
                  "For plain IRC connections, you don't need any link filters; this is used for incompatible protocols.</center>"));
#endif


	m_pUseAutoConnect = new QCheckBox(__tr2qs("Connect to this server at startup"),tab);
	m_pUseAutoConnect->setChecked(s->autoConnect());

	gl->addMultiCellWidget(m_pUseAutoConnect,4,4,0,1);
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pUseAutoConnect,__tr2qs("<center>This option will cause KVIrc to connect to the IRC server when it is started.</center>"));
#endif

  l = new QLabel("",tab);
  gl->addMultiCellWidget(l,5,5,0,1);

  gl->setRowStretch(5,1);
//  gl->setColStretch(1,1);
  tw->addTab(tab,__tr2qs("Connection"));

  // after login join channels
  tab = new QWidget(tw);

  gl = new QGridLayout(tab,1,1,10,4);

	if(s->autoJoinChannelList())
		m_lstChannels = *(s->autoJoinChannelList());
	m_pChannelListSelector = new KviStringListSelector(tab,
		__tr2qs("Channels to join automatically upon connect:"),&m_lstChannels,true);
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pChannelListSelector,__tr2qs("<center>Here you can set a list of channels to be joined automatically " \
		"after a connection to this server has been established. To add a channel, type its name in the " \
		"text input below and click \"<b>Add</b>\".</center>"));
#endif
	gl->addWidget(m_pChannelListSelector,0,0);

  tw->addTab(tab,__tr2qs("Join Channels"));




  // after connect execute

  tab = new QWidget(tw);
  gl = new QGridLayout(tab,1,1,10,4);


  m_pOnConnectEditor = KviScriptEditor::createInstance(tab);
  gl->addWidget(m_pOnConnectEditor,0,0);
  m_pOnConnectEditor->setText(s->onConnectCommand());
  m_pOnConnectEditor->setMinimumHeight(150);
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pOnConnectEditor,__tr2qs("<center>The following commands will be executed after a connection has been established.<br>" \
            "<b>Important:</b> Enter commands <b>without</b> a preceding slash (e.g. <tt>quote pass secret</tt> instead of <tt>/quote pass secret</tt>).<br>"\
            "KVIrc will first send the USER command, then eventually PASS and NICK and then execute this " \
            "command sequence.<br>This is particularly useful for IRC bouncers that require login commands.</center>"));
#endif


  tw->addTab(tab,__tr2qs("On Connect"));


  // after login execute

  tab = new QWidget(tw);
  gl = new QGridLayout(tab,1,1,10,4);

  m_pOnLoginEditor = KviScriptEditor::createInstance(tab);
  gl->addWidget(m_pOnLoginEditor,0,0);
  m_pOnLoginEditor->setText(s->onLoginCommand());

  m_pOnLoginEditor->setMinimumHeight(150);

#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pOnLoginEditor,__tr2qs("<center>The following commands will be executed after a successful login to this server.<br>" \
                                            "<b>Important:</b> Enter commands <b>without</b> a preceding slash (e.g. <tt>quote privatelog</tt> instead of <tt>/quote privatelog</tt>).<br>"
                                           "This is useful for automatically opening queries, setting variables, etc.</center>"));
#endif

	tw->addTab(tab,__tr2qs("On Login"));

	QPushButton * b = new QPushButton(__tr2qs("&OK"),this);
	b->setMinimumWidth(80);
	g->addWidget(b,3,2);
	b->setDefault(true);
	connect(b,SIGNAL(clicked()),this,SLOT(accept()));

	b = new QPushButton(__tr2qs("Cancel"),this);
	g->addWidget(b,3,3);
	b->setMinimumWidth(80);
	connect(b,SIGNAL(clicked()),this,SLOT(reject()));

	g->setRowStretch(2,1);
	g->setColStretch(1,1);

	//g->setResizeMode(QLayout::FreeResize);

	tw->setMinimumWidth(350);

}

KviServerDetailsWidget::~KviServerDetailsWidget()
{
	KviScriptEditor::destroyInstance(m_pOnConnectEditor);
	KviScriptEditor::destroyInstance(m_pOnLoginEditor);
}

void KviServerDetailsWidget::fillData(KviIrcServer * s)
{
	s->m_szUser = m_pUserEditor->text();
	s->m_szPass = m_pPassEditor->text();
	s->m_szNick = m_pNickEditor->text();
	s->m_szRealName = m_pRealEditor->text();
	s->m_szDescription = m_pDescEditor->text();
	s->m_szLinkFilter = m_pLinkFilterEditor->lineEdit()->text();
	s->m_szIp = "";

	s->setCacheIp(m_pCacheIpCheck->isChecked());
	s->setUseSSL(m_pUseSSLCheck->isChecked());

	KviStr szUMode;
	if(!m_pUseDefaultInitUMode->isChecked())
  {
    if(m_pIMode->isChecked())szUMode.append('i');
    if(m_pWMode->isChecked())szUMode.append('w');
    if(m_pSMode->isChecked())szUMode.append('s');
  }
  s->setInitUMode(szUMode.ptr());

  KviStr tmpAddr = m_pIpEditor->address();

  if(!m_pIpEditor->hasEmptyFields())
  {
#ifdef COMPILE_IPV6_SUPPORT
    if(s->isIpV6())
    {
      if((!kvi_strEqualCI(tmpAddr.ptr(),"0:0:0:0:0:0:0:0")) && kvi_isValidStringIp_V6(tmpAddr.ptr()))
      {
        s->m_szIp = tmpAddr;
      }
      else
      {
        s->setCacheIp(false);
        s->m_szIp = "";
      }
    }
    else
    {
#endif
      if((!kvi_strEqualCI(tmpAddr.ptr(),"0.0.0.0")) && kvi_isValidStringIp(tmpAddr.ptr()))
      {
        s->m_szIp = tmpAddr;
      }
      else
      {
        s->setCacheIp(false);
        s->m_szIp = "";
      }
#ifdef COMPILE_IPV6_SUPPORT

    }
#endif

  }
  else
  {
    s->setCacheIp(false);
    s->m_szIp = "";
  }

  QString tmp;
  m_pOnConnectEditor->getText(tmp);
  s->m_szOnConnectCommand = tmp;

  m_pOnLoginEditor->getText(tmp);
  s->m_szOnLoginCommand = tmp.data();

/*
  //crissi
  s->m_pChannelList->clear();
  int channelCount = channel_group->count();
  for (int channelNr=0;channelNr < channelCount;channelNr++)
  {
    if (((QCheckBox*)(channel_group->find(channelNr)))->isChecked())
    {
      s->m_pChannelList->append((channel_group->find(channelNr))->text());
    //  QMessageBox::information(0,"new channel",(channel_group->find(channelNr))->text());
    }


  }
*/

	m_pChannelListSelector->commit();
	s->setAutoJoinChannelList(new QStringList(m_lstChannels));

  s->setAutoConnect (m_pUseAutoConnect->isChecked());
}

void KviServerDetailsWidget::useDefaultInitUModeToggled(bool b)
{
  m_pIMode->setEnabled(!b);
  m_pWMode->setEnabled(!b);
  m_pSMode->setEnabled(!b);
}

/*
//void KviServerDetailsWidget::addNewChannel(QTabWidget *tw, QWidget *tab, KviIrcServer * s)
void KviServerDetailsWidget::addNewChannel()
{
  bool ok;
  QString channelName = QInputDialog::getText(
                   __tr("Add Channel - KVIrc"), __tr("Channel name:"), QLineEdit::Normal,
                   QString::null, &ok, this );
  if ( ok && !channelName.isEmpty() )
  {
    bool insert=true;
    for (int ii=0;ii< channel_group->count();ii++)
        if (((QCheckBox*)(channel_group->find(ii)))->text() == channelName)
            insert =false;
    if (insert == true) {
        server->m_pChannelList->append(channelName);
        QCheckBox *cb = new QCheckBox( channelName, this );
        cb->setChecked( true );
        channel_group->insert(cb);

        // FIXME checkbox is only shown after reopen dialog
        channel_group->adjustSize();
    }

  }

}
*/

// kvi_app.cpp
extern KVIRC_API KviIrcServerDataBase * g_pIrcServerDataBase;

KviServerOptionsListViewItem::KviServerOptionsListViewItem(QListView *parent,const QPixmap &pm,const QString & netname)
    : QListViewItem(parent,netname)
{
  setPixmap(0,pm);
  m_pServerData = 0;
}

KviServerOptionsListViewItem::KviServerOptionsListViewItem(QListViewItem *parent,const QPixmap &pm,const KviIrcServer *s)
    : QListViewItem(parent)
{
  setPixmap(0,pm);
  m_pServerData = new KviIrcServer(*s);
  setText(0,s->m_szHostname);
  setText(1,s->m_szDescription);
}

KviServerOptionsListViewItem::~KviServerOptionsListViewItem()
{
  if(m_pServerData)delete m_pServerData;
}









KviServerOptionsWidget::KviServerOptionsWidget(QWidget * parent)
    : KviOptionsWidget(parent,"server_options_widget")
{
  createLayout(4,2);

  //layout()->setSpacing(4);
  //layout()->setMargin(8);

  m_pContextPopup = new QPopupMenu(this);
  m_pImportPopup = new QPopupMenu(this);

  connect(m_pImportPopup,SIGNAL(aboutToShow()),this,SLOT(importPopupAboutToShow()));
  connect(m_pImportPopup,SIGNAL(activated(int)),this,SLOT(importPopupActivated(int)));


  m_pDetailsDialog = 0;
  m_pImportFilter = 0;

  m_pListView = new QListView(this);
  addWidgetToLayout(m_pListView,0,0,0,0);
  m_pListView->addColumn(__tr2qs("Server"));
  m_pListView->addColumn(__tr2qs("Description"));
  m_pListView->setRootIsDecorated(true);
  m_pListView->setAllColumnsShowFocus(true);
  m_pListView->setSelectionMode(QListView::Single);
  connect(m_pListView,SIGNAL(selectionChanged(QListViewItem *)),
          this,SLOT(listViewItemSelectionChanged(QListViewItem *)));
  connect(m_pListView,SIGNAL(rightButtonPressed(QListViewItem *,const QPoint &,int)),
          this,SLOT(listViewRightButtonPressed(QListViewItem *,const QPoint &,int)));
  connect(m_pListView,SIGNAL(doubleClicked(QListViewItem*, const QPoint&, int )),
        this,SLOT(detailsClicked()));

#ifdef COMPILE_INFO_TIPS
  QString tiptxt = __tr2qs("<center>This is the list of available IRC servers.<br>" \
                           "Right-click on the list to add or remove servers and perform other actions.<br>"\
                           "Double-click on a item for advanced options.</center>");
  QToolTip::add(m_pListView,tiptxt);
  QToolTip::add(m_pListView->viewport(),tiptxt);
#endif

  QVBox * vbox = new QVBox(this);
  addWidgetToLayout(vbox,1,0,1,0);

  m_pNewNetworkButton = new QToolButton(vbox);
  m_pNewNetworkButton->setPixmap(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_WORLD)));
  m_pNewNetworkButton->setAutoRaise(true);
  connect(m_pNewNetworkButton,SIGNAL(clicked()),this,SLOT(newNetwork()));
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pNewNetworkButton,__tr2qs("New Network"));
#endif

  m_pNewServerButton = new QToolButton(vbox);
  m_pNewServerButton->setPixmap(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_SERVER)));
  m_pNewServerButton->setAutoRaise(true);
  connect(m_pNewServerButton,SIGNAL(clicked()),this,SLOT(newServer()));
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pNewServerButton,__tr2qs("New Server"));
#endif

  m_pRemoveButton = new QToolButton(vbox);
  m_pRemoveButton->setPixmap(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_CUT)));
  m_pRemoveButton->setEnabled(false);
  m_pRemoveButton->setAutoRaise(true);
  connect(m_pRemoveButton,SIGNAL(clicked()),this,SLOT(removeCurrent()));
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pRemoveButton,__tr2qs("Remove Network/Server"));
#endif

  QFrame * f = new QFrame(vbox);
  f->setFrameStyle(QFrame::Sunken | QFrame::HLine);

  m_pCopyServerButton = new QToolButton(vbox);
  m_pCopyServerButton->setPixmap(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_COPY)));
  m_pCopyServerButton->setEnabled(false);
  m_pCopyServerButton->setAutoRaise(true);
  connect(m_pCopyServerButton,SIGNAL(clicked()),this,SLOT(copyServer()));
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pCopyServerButton,__tr2qs("Copy Server"));
#endif

  m_pPasteServerButton = new QToolButton(vbox);
  m_pPasteServerButton->setPixmap(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_PASTE)));
  m_pPasteServerButton->setEnabled(false);
  m_pPasteServerButton->setAutoRaise(true);
  connect(m_pPasteServerButton,SIGNAL(clicked()),this,SLOT(pasteServer()));
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pPasteServerButton,__tr2qs("Paste Server"));
#endif

  f = new QFrame(vbox);
  f->setFrameStyle(QFrame::Sunken | QFrame::HLine);

  m_pImportButton = new QToolButton(vbox);
  m_pImportButton->setPixmap(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_FOLDER)));
  m_pImportButton->setAutoRaise(true);
  m_pImportButton->setPopup(m_pImportPopup);
  m_pImportButton->setPopupDelay(1);
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pImportButton,__tr2qs("Import List"));
#endif

  QFrame * lll = new QFrame(vbox);
  vbox->setStretchFactor(lll,100);


  QGroupBox *gbox = addGroupBox(0,1,1,1,3,QGroupBox::Horizontal,__tr2qs("Configuration"));
  m_pSrvNetLabel = new QLabel(__tr2qs("Server:"),gbox);

  m_pSrvNetEdit = new QLineEdit(gbox);
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pSrvNetEdit,__tr2qs("<center>This is the <b>hostname</b> of the currently selected server.</center>"));
#endif


  m_pIpV6Check = new QCheckBox(__tr2qs("Use IPv6 protocol"),gbox);

#ifndef COMPILE_IPV6_SUPPORT
  m_pIpV6Check->setEnabled(false);
#endif

#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pIpV6Check,__tr2qs("<center>This check identifies IPv6 servers.<br>If enabled, KVIrc will attempt to use the IPv6 protocol " \
                                       "(thus your OS <b>must</b> have a working IPv6 stack and you <b>must</b> have an IPv6 connection).</center>"));
#endif


  m_pPortLabel = new QLabel(__tr2qs("Port:"),gbox);

  m_pPortEdit = new QLineEdit(gbox);
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pPortEdit,__tr2qs("<center>This is the default <b>port</b> that this server will be contacted on.<br>Usually <b>6667</b> is OK.</center>"));
#endif

  m_pDetailsButton = new QPushButton(__tr2qs("Advanced..."),gbox);
  connect(m_pDetailsButton,SIGNAL(clicked()),this,SLOT(detailsClicked()));
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pDetailsButton,__tr2qs("<center>Click here to edit advanced options for this server.</center>"));
#endif


  m_pConnectCurrent = new QPushButton(__tr2qs("Connect &Now"),this);
  addWidgetToLayout(m_pConnectCurrent,0,2,0,2);
  connect(m_pConnectCurrent,SIGNAL(clicked()),this,SLOT(connectCurrentClicked()));

#ifdef COMPILE_INFO_TIPS
  QToolTip::add(m_pConnectCurrent,__tr2qs("<center>Hit this button to connect to the currently selected server.</center>"));
#endif

  m_pRecentPopup = new QPopupMenu(this);
  connect(m_pRecentPopup,SIGNAL(aboutToShow()),this,SLOT(recentServersPopupAboutToShow()));
  connect(m_pRecentPopup,SIGNAL(activated(int)),this,SLOT(recentServersPopupClicked(int)));

  QToolButton * tb = new QToolButton(this);
  addWidgetToLayout(tb,1,2,1,2);
  tb->setPixmap(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_TIME)));
  tb->setPopup(m_pRecentPopup);
  tb->setPopupDelay(1);
#ifdef COMPILE_INFO_TIPS
  QToolTip::add(tb,__tr2qs("<center>This button shows a list of recently used servers. It allows you to quickly find them in the list.</center>"));
#endif


  KviBoolSelector * b = addBoolSelector(0,3,1,3,__tr2qs("Show this dialog at startup"),KviOption_boolShowServersConnectDialogOnStart);

#ifdef COMPILE_INFO_TIPS
  QToolTip::add(b,__tr2qs("<center>If this option is enabled, the Servers dialog will appear every time you start KVIrc</center>"));
#endif

  //	KviBoolSelector * c = addBoolSelector(0,6,2,6,__tr("Close after connection"),KviOption_boolCloseServerWidgetAfterConnect);

  m_pLastEditedItem = 0;

  m_pClipboard = 0;

  fillServerList();

  layout()->setRowStretch(0,1);
  //layout()->setColStretch(1,5);
  //layout()->setColStretch(2,2);
  layout()->setColStretch(0,1);
  setMinimumWidth(320);
}

KviServerOptionsWidget::~KviServerOptionsWidget()
{
  if(m_pImportFilter)
  {
    disconnect(m_pImportFilter,0,this,0);
    m_pImportFilter->die();
    m_pImportFilter = 0;
  }

  if(m_pClipboard)delete m_pClipboard;
  if(m_pDetailsDialog)delete m_pDetailsDialog;
}

void KviServerOptionsWidget::recentServersPopupAboutToShow()
{
  g_pApp->fillRecentServersPopup(m_pRecentPopup);
}

void KviServerOptionsWidget::recentServersPopupClicked(int id)
{
	KviStr data = m_pRecentPopup->text(id);
	data.cutToFirst('/');
	while(data.firstCharIs('/'))data.cutLeft(1);
	data.cutFromLast(':');
	// we should have a full server name now , with no port
	
	KviStr port = m_pRecentPopup->text(id);
	port.cutToLast(':');
	bool bOk;
	unsigned short uPort = port.toUShort(&bOk);
	// we have the port too

	QListViewItem * pFoundNet = 0;
	QListViewItem * pFoundSrv = 0;

	for(QListViewItem * net = m_pListView->firstChild();net;net = net->nextSibling())
	{
		for(QListViewItem * srv = net->firstChild();srv;srv = srv->nextSibling())
		{
			KviStr tmp = ((KviServerOptionsListViewItem *)srv)->m_pServerData->hostname();
			if(kvi_strEqualCI(tmp.ptr(),data.ptr()))
			{
				if(((KviServerOptionsListViewItem *)srv)->m_pServerData->port() == uPort)
				{
					net->setOpen(true);
					m_pListView->setCurrentItem(srv);
					m_pListView->ensureItemVisible(srv);
					return;
				} else {
					if(!pFoundNet)
					{
						// the port doesn't match.. check for
						// another entry with the correct port
						// but keep track of this entry in case we don't find it
						pFoundNet = srv;
						pFoundSrv = net;
					}
				}
			}
		}
	}
	
	// fallback to the server with the wrong port
	if(pFoundNet)
	{
		pFoundNet->setOpen(true);
		m_pListView->setCurrentItem(pFoundSrv);
		m_pListView->ensureItemVisible(pFoundSrv);
	}
}


void KviServerOptionsWidget::connectCurrentClicked()
{
  saveLastItem();
  commit();
  KviStr szServer = m_pSrvNetEdit->text();
  if (szServer.isEmpty())return;
  g_pUserParser->parseCommandBuffer("server -u",g_pApp->activeConsole());
}


void KviServerOptionsWidget::fillServerList()
{
  KviServerOptionsListViewItem * net;
  KviServerOptionsListViewItem * srv;

  KviPtrList<KviIrcNetwork> * l = g_pIrcServerDataBase->networkList();

  for(KviIrcNetwork * n = l->first();n;n = l->next())
  {
    net = new KviServerOptionsListViewItem(m_pListView,*(g_pIconManager->getSmallIcon(KVI_SMALLICON_WORLD)),n->name());
    KviPtrList<KviIrcServer> * sl = n->serverList();
    net->setOpen(n == g_pIrcServerDataBase->currentNetwork());
    //debug("Current server = %s",n->currentServer() ? n->currentServer()->m_szHostname.ptr() : "non");
    for(KviIrcServer * s = sl->first();s;s = sl->next())
    {
      srv = new KviServerOptionsListViewItem(net,*(g_pIconManager->getSmallIcon(KVI_SMALLICON_SERVER)),s);

      if((s == n->currentServer()) && (n == g_pIrcServerDataBase->currentNetwork()))
      {
        m_pListView->setSelected(srv,true);
        m_pListView->ensureItemVisible(srv);
      }
    }
  }
}

void KviServerOptionsWidget::listViewItemSelectionChanged(QListViewItem *it)
{
  //	m_pRecentListBox->clearSelection();
  saveLastItem();
  m_pLastEditedItem = (KviServerOptionsListViewItem *)it;

  if(m_pLastEditedItem)
  {
    m_pSrvNetLabel->setEnabled(true);
    m_pSrvNetEdit->setEnabled(true);
    m_pDetailsButton->setEnabled(m_pLastEditedItem->m_pServerData);
    m_pPortLabel->setEnabled(m_pLastEditedItem->m_pServerData);
    m_pPortEdit->setEnabled(m_pLastEditedItem->m_pServerData);
#ifdef COMPILE_IPV6_SUPPORT
    m_pIpV6Check->setEnabled(m_pLastEditedItem->m_pServerData);
#else
    m_pIpV6Check->setEnabled(false);
#endif
    m_pConnectCurrent->setEnabled(m_pLastEditedItem->m_pServerData);
    //m_pConnectNew->setEnabled(m_pLastEditedItem->m_pServerData);

    m_pRemoveButton->setEnabled(true);
    m_pCopyServerButton->setEnabled(m_pLastEditedItem->m_pServerData);

    if(m_pLastEditedItem->m_pServerData)
    {
      m_pSrvNetLabel->setText(__tr2qs("Server:"));
      m_pSrvNetEdit->setText(m_pLastEditedItem->m_pServerData->m_szHostname);
      KviStr tmp(KviStr::Format,"%u",m_pLastEditedItem->m_pServerData->m_uPort);
      m_pPortEdit->setText(tmp.ptr());
#ifdef COMPILE_IPV6_SUPPORT
      m_pIpV6Check->setChecked(m_pLastEditedItem->m_pServerData->isIpV6());
#else
      m_pIpV6Check->setChecked(false);
#endif

    }
    else
    {
      m_pSrvNetLabel->setText(__tr2qs("Network:"));
      m_pSrvNetEdit->setText(it->text(0));
      m_pPortEdit->setText("");
    }
  }
  else
  {
    m_pRemoveButton->setEnabled(false);
    m_pCopyServerButton->setEnabled(true);

    m_pConnectCurrent->setEnabled(false);
    //m_pConnectNew->setEnabled(false);
    m_pIpV6Check->setEnabled(false);
    m_pSrvNetLabel->setEnabled(false);
    m_pSrvNetEdit->setEnabled(false);
    m_pSrvNetEdit->setText(__tr2qs("No selection"));
    m_pDetailsButton->setEnabled(false);
    m_pPortLabel->setEnabled(false);
    m_pPortEdit->setEnabled(false);
    m_pPortEdit->setText("");
  }
}



void KviServerOptionsWidget::saveLastItem()
{
  if(m_pLastEditedItem)
  {
    if(m_pLastEditedItem->m_pServerData)
    {
      KviStr tmp = m_pSrvNetEdit->text();
      if(tmp.isEmpty())tmp = "irc.unknown.net";
      m_pLastEditedItem->setText(0,tmp.ptr());
      m_pLastEditedItem->m_pServerData->m_szHostname = tmp;
      m_pLastEditedItem->setText(1,m_pLastEditedItem->m_pServerData->m_szDescription);
      tmp = m_pPortEdit->text();
      bool bOk;
      unsigned short int uPort = tmp.toUShort(&bOk);
      if(!bOk)uPort = 6667;
      m_pLastEditedItem->m_pServerData->m_uPort = uPort;
#ifdef COMPILE_IPV6_SUPPORT
      m_pLastEditedItem->m_pServerData->setIpV6(m_pIpV6Check->isChecked());
#else
      m_pLastEditedItem->m_pServerData->setIpV6(false);
#endif

    }
    else
    {
      KviStr tmp = m_pSrvNetEdit->text();
      if(tmp.isEmpty())tmp = __tr2qs("UnknownNet");
      m_pLastEditedItem->setText(0,tmp.ptr());
    }
  }
}

void KviServerOptionsWidget::commit()
{
  //debug("COMMITTING");
  saveLastItem();
  g_pIrcServerDataBase->clear();
  KviServerOptionsListViewItem * it = (KviServerOptionsListViewItem *)m_pListView->firstChild();
  while(it)
  {
    KviStr tmp = it->text(0);
    if(tmp.hasData())
    {
      KviIrcNetwork * net = g_pIrcServerDataBase->findNetwork(tmp.ptr());
      if(!net)
      {
        net = new KviIrcNetwork(tmp.ptr());
        g_pIrcServerDataBase->insertNetwork(net);
      }
      if(it == m_pLastEditedItem)g_pIrcServerDataBase->setCurrentNetwork(net);

      KviServerOptionsListViewItem * ch = (KviServerOptionsListViewItem *)it->firstChild();
      KviIrcServer *srv;
      while(ch)
      {
        if(ch->m_pServerData)
        {
          if(!ch->m_pServerData->m_szHostname.isEmpty())
          {
            srv = net->findServer(ch->m_pServerData->m_szHostname,ch->m_pServerData->m_uPort);
            if(!srv)
            {
              srv = new KviIrcServer(*(ch->m_pServerData));
              net->insertServer(srv);
            }
            else *srv = *(ch->m_pServerData);
            if(ch == m_pLastEditedItem)
            {
              g_pIrcServerDataBase->setCurrentNetwork(net);
              net->setCurrentServer(srv);
            }
          }
        }
        ch = (KviServerOptionsListViewItem *)ch->nextSibling();
      }

      if(net->currentServer() == 0)net->setCurrentServer(net->serverList()->first());
    }
    it = (KviServerOptionsListViewItem *)it->nextSibling();
  }
  if(g_pIrcServerDataBase->currentNetwork() == 0)
  {
    g_pIrcServerDataBase->setCurrentNetwork(g_pIrcServerDataBase->networkList()->first());
  }

  KviOptionsWidget::commit();
}

void KviServerOptionsWidget::listViewRightButtonPressed(QListViewItem *it,const QPoint &pnt,int col)
{
  int id;
  bool bServer = (it && ((KviServerOptionsListViewItem *)it)->m_pServerData);
  m_pContextPopup->clear();
  m_pContextPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_WORLD)),__tr2qs("New Network"),this,SLOT(newNetwork()));
  id = m_pContextPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_CUT)),__tr2qs("Remove Network"),this,SLOT(removeCurrent()));
  m_pContextPopup->setItemEnabled(id,!bServer);
  m_pContextPopup->insertSeparator();
  id = m_pContextPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_SERVER)),__tr2qs("&New Server"),this,SLOT(newServer()));
  id = m_pContextPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_CUT)),__tr2qs("Re&move Server"),this,SLOT(removeCurrent()));
  m_pContextPopup->setItemEnabled(id,bServer);
  id = m_pContextPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_COPY)),__tr2qs("&Copy Server"),this,SLOT(copyServer()));
  m_pContextPopup->setItemEnabled(id,bServer);
  id = m_pContextPopup->insertItem(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_PASTE)),__tr2qs("&Paste Server"),this,SLOT(pasteServer()));
  m_pContextPopup->setItemEnabled(id,m_pClipboard);

  m_pContextPopup->insertSeparator();
  //	m_pContextPopup->insertItem(__c2q(__tr("Merge list from server.ini")),this,SLOT(importFromIni()));
  m_pContextPopup->insertItem(__tr2qs("Clear List"),this,SLOT(clearList()));
  m_pContextPopup->insertSeparator();
  m_pContextPopup->insertItem(__tr2qs("Import List"),m_pImportPopup);
  m_pContextPopup->popup(QCursor::pos());
}

void KviServerOptionsWidget::importPopupAboutToShow()
{
  g_pModuleManager->loadModulesByCaps("serverimport");
  m_pImportPopup->clear();

  KviModuleExtensionDescriptorList * l = KviModuleExtensionManager::instance()->getExtensionList("serverimport");

  if(!l)return;

  int id;

  for(KviModuleExtensionDescriptor * d = l->first();d;d = l->next())
  {
    if(d->icon())
      id = m_pImportPopup->insertItem(*(d->icon()),d->visibleName().ptr());
    else
      id = m_pImportPopup->insertItem(d->visibleName().ptr());
    m_pImportPopup->setItemParameter(id,d->id());
  }

}

void KviServerOptionsWidget::importPopupActivated(int id)
{
  // ensure that we have all the modules : they could have been unloaded while the popup was displayed
  g_pModuleManager->loadModulesByCaps("serverimport");

  KviModuleExtensionDescriptorList * l = KviModuleExtensionManager::instance()->getExtensionList("serverimport");

  if(!l)
  {
    // ops.. internal error: I thought to have a module capable of importing servers
    // but actually it's not the case.. something weird happened (in the best case
    // the user has just unloaded the module and removed it from disk ?)
    // FIXME: KviMessageBox::warning
    kvi_warningBox(__tr("Oops... something weird happened:<br>Can't find any module capable of importing servers."));
    return;
  }

  if(m_pImportFilter)
  {
    disconnect(m_pImportFilter,0,this,0);
    m_pImportFilter->die();
    m_pImportFilter = 0;
  }

  id = m_pImportPopup->itemParameter(id);

  m_pImportFilter = (KviMexServerImport *)KviModuleExtensionManager::instance()->allocateExtension("serverimport",id,0);

  if(!m_pImportFilter)
  {
    kvi_warningBox(__tr("Oops... something weird happened:<br>Can't find the module that was capable of this import action. :("));
    return;
  }

  connect(m_pImportFilter,SIGNAL(server(const KviIrcServer &,const char *)),this,SLOT(importServer(const KviIrcServer &,const char *)));
  connect(m_pImportFilter,SIGNAL(destroyed()),this,SLOT(importerDead()));

  m_pImportFilter->start();
}

void KviServerOptionsWidget::importerDead()
{
  m_pImportFilter = 0;
}

void KviServerOptionsWidget::importServer(const KviIrcServer &s,const char * network)
{
  KviServerOptionsListViewItem * net = findNetItem(network);
  if(!net)
  {
    net = new KviServerOptionsListViewItem(m_pListView,*(g_pIconManager->getSmallIcon(KVI_SMALLICON_WORLD)),network);
    net->setOpen(true);
  }

  for(KviServerOptionsListViewItem * srv = (KviServerOptionsListViewItem *)net->firstChild();srv;srv = (KviServerOptionsListViewItem *)srv->nextSibling())
  {
    if((srv)->m_pServerData->useSSL() == s.useSSL())
    {
      if(srv->m_pServerData->isIpV6() == s.isIpV6())
      {
        if(kvi_strEqualCI(srv->m_pServerData->hostname(),s.hostname()))
        {
          // update the port
          srv->m_pServerData->setPort(s.port());
          if(!s.ipAddress().isEmpty())srv->m_pServerData->setIpAddress(s.ipAddress());
          if(!s.password().isEmpty())srv->m_pServerData->setPassword(s.password());
          if(!s.nickname().isEmpty())srv->m_pServerData->setNickname(s.nickname());
          m_pListView->setCurrentItem(srv);
          m_pListView->ensureItemVisible(srv);
          return;
        }
      }
    }
  }

  // not found : add it!
  KviServerOptionsListViewItem * newServer = new KviServerOptionsListViewItem(net,*(g_pIconManager->getSmallIcon(KVI_SMALLICON_SERVER)),&s);
  m_pListView->setCurrentItem(newServer);
  m_pListView->ensureItemVisible(newServer);

}

void KviServerOptionsWidget::newNetwork()
{
  KviServerOptionsListViewItem * it = new KviServerOptionsListViewItem(m_pListView,*(g_pIconManager->getSmallIcon(KVI_SMALLICON_WORLD)),__tr2qs("New Network"));
  it->setOpen(true);
  m_pListView->setSelected(it,true);
  m_pListView->ensureItemVisible(it);
}

void KviServerOptionsWidget::newServer()
{
  if(m_pLastEditedItem)
  {
    KviServerOptionsListViewItem * net;
    if(m_pLastEditedItem->m_pServerData)
    {
      net = (KviServerOptionsListViewItem *)m_pLastEditedItem->parent();
      if(!net)return;
    }
    else net = m_pLastEditedItem;

    KviIrcServer tmp;
    tmp.m_szHostname = __tr2qs("irc.unknown.net");
    tmp.setCacheIp(true);

    KviServerOptionsListViewItem * it = new KviServerOptionsListViewItem(net,
                                        *(g_pIconManager->getSmallIcon(KVI_SMALLICON_SERVER)),&tmp);

    net->setOpen(true);

    m_pListView->setSelected(it,true);
    m_pListView->ensureItemVisible(it);
  }
}

void KviServerOptionsWidget::copyServer()
{
  if(m_pLastEditedItem)
  {
    if(m_pLastEditedItem->m_pServerData)
    {
      if(!m_pClipboard)m_pClipboard = new KviIrcServer();
      *m_pClipboard = *(m_pLastEditedItem->m_pServerData);
      m_pPasteServerButton->setEnabled(true);
    }
  }
}

void KviServerOptionsWidget::pasteServer()
{
  if(m_pClipboard)
  {
    if(m_pLastEditedItem)
    {
      KviServerOptionsListViewItem * net;
      if(m_pLastEditedItem->m_pServerData)
      {
        net = (KviServerOptionsListViewItem *)m_pLastEditedItem->parent();
        if(!net)return;
      }
      else net = m_pLastEditedItem;

      KviServerOptionsListViewItem * it = new KviServerOptionsListViewItem(net,
                                          *(g_pIconManager->getSmallIcon(KVI_SMALLICON_SERVER)),m_pClipboard);

      net->setOpen(true);

      m_pListView->setSelected(it,true);
      m_pListView->ensureItemVisible(it);
    }
  }
}

void KviServerOptionsWidget::removeCurrent()
{
  if(m_pLastEditedItem)
  {
    delete m_pLastEditedItem;
    m_pLastEditedItem = 0;
    QListViewItem * it = m_pListView->firstChild();
    if(it)
    {
      m_pListView->setSelected(it,true);
      m_pListView->ensureItemVisible(it);
    }
  }
}



KviServerOptionsListViewItem * KviServerOptionsWidget::findNetItem(const char * netname)
{
  KviServerOptionsListViewItem * it = (KviServerOptionsListViewItem *)m_pListView->firstChild();
  while(it)
  {
    KviStr szNetName = it->text(0);
    if(kvi_strEqualCI(szNetName.ptr(),netname))return it;
    it = (KviServerOptionsListViewItem *)it->nextSibling();
  }
  return 0;
}

void KviServerOptionsWidget::clearList()
{
  m_pListView->clear();
  m_pLastEditedItem = 0;
  listViewItemSelectionChanged(0);
}

void KviServerOptionsWidget::detailsClicked()
{
  if(!m_pLastEditedItem)return;
  if(!(m_pLastEditedItem->m_pServerData))return;

  saveLastItem();

  m_pDetailsDialog = new KviServerDetailsWidget(this,m_pLastEditedItem->m_pServerData);

  int retCode = m_pDetailsDialog->exec();
  if(retCode == QDialog::Accepted)
  {
    if(m_pLastEditedItem)
    {
      if(m_pLastEditedItem->m_pServerData)
      {
        m_pDetailsDialog->fillData(m_pLastEditedItem->m_pServerData);
        saveLastItem();
      }
    }
  }


  delete m_pDetailsDialog;
  m_pDetailsDialog = 0;
}

#include "m_optw_servers.moc"
