//
//   File : libkvitb_options.cpp
//   Creation date : Sun Oct 21 2001 13:29:12 CEST by Szymon Stefanek
//
//   Copyright (C) 2001 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "libkvitb_options.h"

#include "kvi_module.h"
#include "kvi_uparser.h"

#include "kvi_frame.h"
#include "kvi_window.h"
#include "kvi_locale.h"
#include "kvi_internalcmd.h"
#include "kvi_app.h"
#include "kvi_mdimanager.h"
#include "kvi_iconmanager.h"

#include <qtoolbutton.h>

static KviPtrList<KviOptionsToolBar> * g_pToolBarList = 0;

KviOptionsToolBar::KviOptionsToolBar(KviModuleExtensionDescriptor * d,KviFrame * par)
: KviMexToolBar(d,__tr2qs("Options Toolbar"),par)
{
	g_pToolBarList->append(this);


	m_pBtnServers = new QToolButton(*(g_pIconManager->getBigIcon("kvi_bigicon_servers.png")),__tr2qs("Configure Servers"),__tr2qs("Change server settings"),this,SLOT(showServerOptions()),this);
	m_pBtnIdentity = new QToolButton(*(g_pIconManager->getBigIcon("kvi_bigicon_identity.png")),__tr2qs("Configure Identity"),__tr2qs("Change identity settings"),this,SLOT(showIdentityOptions()),this);
	m_pBtnSettings = new QToolButton(*(g_pIconManager->getBigIcon("kvi_bigicon_settings.png")),__tr2qs("Configure KVIrc"),__tr2qs("Change KVIrc settings"),this,SLOT(showGlobalOptions()),this);
	
	connect(g_pApp,SIGNAL(reloadImages()),this,SLOT(reloadImages()));
}

KviOptionsToolBar::~KviOptionsToolBar()
{
	g_pToolBarList->removeRef(this);
}

void KviOptionsToolBar::reloadImages()
{
	m_pBtnServers->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_servers.png")));
	m_pBtnIdentity->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_identity.png")));
	m_pBtnSettings->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_settings.png")));
}

void KviOptionsToolBar::die()
{
	delete this;
}

void KviOptionsToolBar::showServerOptions()
{
	g_pUserParser->parseCommandBuffer("options.edit KviServerOptionsWidget",g_pApp->activeWindow());
}

void KviOptionsToolBar::showIdentityOptions()
{
	g_pUserParser->parseCommandBuffer("options.edit KviIdentityOptionsWidget",g_pApp->activeWindow());
}

void KviOptionsToolBar::showGlobalOptions()
{
	g_pUserParser->parseCommandBuffer(kvi_getInternalCommandBuffer(KVI_INTERNALCOMMAND_OPTIONS_DIALOG),g_pApp->activeWindow());
}


static KviModuleExtension * tb_options_alloc(KviModuleExtensionAllocStruct * s)
{
	if(!s->pWindow)return 0;
	return new KviOptionsToolBar(s->pDescriptor,s->pWindow->frame());
}


static bool tb_options_module_init(KviModule * m)
{
	g_pToolBarList = new KviPtrList<KviOptionsToolBar>;
	g_pToolBarList->setAutoDelete(false);

	KviModuleExtensionDescriptor * d = m->registerExtension("toolbar","options",__tr("Options Toolbar"),tb_options_alloc);

	if(d)d->setIcon(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_OPTIONS)));


	return true;
}

static bool tb_options_module_cleanup(KviModule *m)
{
	while(g_pToolBarList->first())delete g_pToolBarList->first();
	delete g_pToolBarList;

	m->unregisterMetaObject("KviOptionsToolBar");

	return true;
}

static bool tb_options_module_can_unload(KviModule *m)
{
	return (g_pToolBarList->isEmpty());
}

KVIMODULEEXPORTDATA KviModuleInfo kvirc_module_info=
{
	"List",                                                      // module name
	"1.0.0",                                                     // module version
	"Copyright (C) 2000-2001 Szymon Stefanek (pragma at kvirc dot net)", // author & (C)
	"Options toolbar extension",
	tb_options_module_init,
	tb_options_module_can_unload,
	0,
	tb_options_module_cleanup
};

#include "libkvitb_options.moc"
