//=====================================================================================
//
//   File : libkvitb_scripting.cpp
//   Creation date : Wed 10 Mar 01:30:54 2004 GMT by Szymon Stefanek
//
//   Copyright (C) 2004 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=====================================================================================

#include "libkvitb_scripting.h"

#include "kvi_module.h"
#include "kvi_uparser.h"

#include "kvi_frame.h"
#include "kvi_window.h"
#include "kvi_locale.h"
#include "kvi_app.h"
#include "kvi_iconmanager.h"
#include "kvi_internalcmd.h"

#include <qtoolbutton.h>

static KviPtrList<KviScriptingToolBar> * g_pToolBarList = 0;

KviScriptingToolBar::KviScriptingToolBar(KviModuleExtensionDescriptor * d,KviFrame * par)
: KviMexToolBar(d,__tr2qs("Scripting Toolbar"),par)
{
	g_pToolBarList->append(this);

	m_pBtnAliasEditor = new QToolButton(
		*(g_pIconManager->getBigIcon("kvi_bigicon_aliaseditor.png")),
		__tr2qs("Show alias editor"),__tr2qs("Show alias editor"),this,SLOT(showAliasEditor()),this);
	m_pBtnEventEditor = new QToolButton(
		*(g_pIconManager->getBigIcon("kvi_bigicon_eventeditor.png")),
		__tr2qs("Show event editor"),__tr2qs("Show event editor"),this,SLOT(showEventEditor()),this);
	m_pBtnPopupEditor = new QToolButton(
		*(g_pIconManager->getBigIcon("kvi_bigicon_popupeditor.png")),
		__tr2qs("Show popup editor"),__tr2qs("Show popup editor"),this,SLOT(showPopupEditor()),this);
	m_pBtnRawEditor = new QToolButton(
		*(g_pIconManager->getBigIcon("kvi_bigicon_raweditor.png")),
		__tr2qs("Show raw event editor"),__tr2qs("Show raw event editor"),this,SLOT(showRawEditor()),this);
	m_pBtnToolBarEditor = new QToolButton(
		*(g_pIconManager->getBigIcon("kvi_bigicon_toolbareditor.png")),
		__tr2qs("Show toolbar editor"),__tr2qs("Show toolbar editor"),this,SLOT(showToolBarEditor()),this);
	m_pBtnCodeTester = new QToolButton(
		*(g_pIconManager->getBigIcon("kvi_bigicon_codetester.png")),
		__tr2qs("New code tester"),__tr2qs("New code tester"),this,SLOT(newCodeTester()),this);

	connect(g_pApp,SIGNAL(reloadImages()),this,SLOT(reloadImages()));

}

KviScriptingToolBar::~KviScriptingToolBar()
{
	g_pToolBarList->removeRef(this);
}

void KviScriptingToolBar::reloadImages()
{
	m_pBtnAliasEditor->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_aliaseditor.png")));
	m_pBtnEventEditor->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_eventeditor.png")));
	m_pBtnPopupEditor->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_popupeditor.png")));
	m_pBtnRawEditor->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_raweditor.png")));
	m_pBtnToolBarEditor->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_toolbareditor.png")));
	m_pBtnCodeTester->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_codetester.png")));
}

void KviScriptingToolBar::die()
{
	delete this;
}

void KviScriptingToolBar::showAliasEditor()
{
	frame()->executeInternalCommand(KVI_INTERNALCOMMAND_ALIASEDITOR_OPEN);
}

void KviScriptingToolBar::showEventEditor()
{
	frame()->executeInternalCommand(KVI_INTERNALCOMMAND_EVENTEDITOR_OPEN);
}

void KviScriptingToolBar::showPopupEditor()
{
	frame()->executeInternalCommand(KVI_INTERNALCOMMAND_POPUPEDITOR_OPEN);
}

void KviScriptingToolBar::showRawEditor()
{
	frame()->executeInternalCommand(KVI_INTERNALCOMMAND_RAWEDITOR_OPEN);
}

void KviScriptingToolBar::showToolBarEditor()
{
	frame()->executeInternalCommand(KVI_INTERNALCOMMAND_TOOLBAREDITOR_OPEN);
}

void KviScriptingToolBar::newCodeTester()
{
	frame()->executeInternalCommand(KVI_INTERNALCOMMAND_CODETESTER_OPEN);
}

static KviModuleExtension * tb_scripting_alloc(KviModuleExtensionAllocStruct * s)
{
	if(!s->pWindow)return 0;
	return new KviScriptingToolBar(s->pDescriptor,s->pWindow->frame());
}


static bool tb_scripting_module_init(KviModule * m)
{
	g_pToolBarList = new KviPtrList<KviScriptingToolBar>;
	g_pToolBarList->setAutoDelete(false);

	KviModuleExtensionDescriptor * d = m->registerExtension("toolbar","scripting",__tr("Scripting Toolbar"),tb_scripting_alloc);
	if(d)d->setIcon(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_ALIAS)));

	return true;
}

static bool tb_scripting_module_cleanup(KviModule *m)
{
	while(g_pToolBarList->first())delete g_pToolBarList->first();
	delete g_pToolBarList;

	m->unregisterMetaObject("KviScriptingToolBar");

	return true;
}

static bool tb_scripting_module_can_unload(KviModule *m)
{
	return (g_pToolBarList->isEmpty());
}

KVIMODULEEXPORTDATA KviModuleInfo kvirc_module_info=
{
	"Scripting",                                            // module name
	"1.0.0",                                                // module version
	"Copyright (C) 2004 Szymon Stefanek (pragma at kvirc dot net)", // author & (C)
	"Scripting shortcuts toolbar extension",
	tb_scripting_module_init,
	tb_scripting_module_can_unload,
	0,
	tb_scripting_module_cleanup
};

#include "libkvitb_scripting.moc"
