//
//   File : libkvitb_winops.cpp
//   Creation date : Sun Oct 21 2001 13:29:12 CEST by Szymon Stefanek
//
//   Copyright (C) 2001 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "libkvitb_winops.h"

#include "kvi_module.h"
#include "kvi_uparser.h"

#include "kvi_frame.h"
#include "kvi_window.h"
#include "kvi_locale.h"
#include "kvi_app.h"
#include "kvi_mdimanager.h"
#include "kvi_iconmanager.h"

#include <qtoolbutton.h>

static KviPtrList<KviWindowOperationsToolBar> * g_pToolBarList = 0;

KviWindowOperationsToolBar::KviWindowOperationsToolBar(KviModuleExtensionDescriptor * d,KviFrame * par)
: KviMexToolBar(d,__tr2qs("Window Toolbar"),par)
{
	g_pToolBarList->append(this);

	m_pBtnCascade = new QToolButton(*(g_pIconManager->getBigIcon("kvi_bigicon_cascadewindows.png")),__tr2qs("Cascade"),__tr2qs("Cascade windows"),par->mdiManager(),SLOT(cascadeMaximized()),this);
	m_pBtnTile = new QToolButton(*(g_pIconManager->getBigIcon("kvi_bigicon_tilewindows.png")),__tr2qs("Tile"),__tr2qs("Tile windows"),par->mdiManager(),SLOT(tile()),this);
	m_pBtnMinimize = new QToolButton(*(g_pIconManager->getBigIcon("kvi_bigicon_minimizeallwindows.png")),__tr2qs("Minimize All"),__tr2qs("Minimize all windows"),par->mdiManager(),SLOT(minimizeAll()),this);

	connect(g_pApp,SIGNAL(reloadImages()),this,SLOT(reloadImages()));

}

KviWindowOperationsToolBar::~KviWindowOperationsToolBar()
{
	g_pToolBarList->removeRef(this);
}

void KviWindowOperationsToolBar::reloadImages()
{
	m_pBtnCascade->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_cascadewindows.png")));
	m_pBtnTile->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_tilewindows.png")));
	m_pBtnMinimize->setPixmap(*(g_pIconManager->getBigIcon("kvi_bigicon_minimizeallwindows.png")));

}

void KviWindowOperationsToolBar::die()
{
	delete this;
}


static KviModuleExtension * tb_winops_alloc(KviModuleExtensionAllocStruct * s)
{
	if(!s->pWindow)return 0;
	return new KviWindowOperationsToolBar(s->pDescriptor,s->pWindow->frame());
}


static bool tb_winops_module_init(KviModule * m)
{
	g_pToolBarList = new KviPtrList<KviWindowOperationsToolBar>;
	g_pToolBarList->setAutoDelete(false);

	KviModuleExtensionDescriptor * d = m->registerExtension("toolbar","winops",__tr("Window Toolbar"),tb_winops_alloc);
	if(d)d->setIcon(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_CASCADEWINDOWS)));

	return true;
}

static bool tb_winops_module_cleanup(KviModule *m)
{
	while(g_pToolBarList->first())delete g_pToolBarList->first();
	delete g_pToolBarList;

	m->unregisterMetaObject("KviWindowOperationsToolBar");

	return true;
}

static bool tb_winops_module_can_unload(KviModule *m)
{
	return (g_pToolBarList->isEmpty());
}

KVIMODULEEXPORTDATA KviModuleInfo kvirc_module_info=
{
	"List",                                             // module name
	"1.0.0",                                                // module version
	"Copyright (C) 2000-2001 Szymon Stefanek (pragma at kvirc dot net)", // author & (C)
	"Window operations toolbar extension",
	tb_winops_module_init,
	tb_winops_module_can_unload,
	0,
	tb_winops_module_cleanup
};

#include "libkvitb_winops.moc"
