//
//   File : libkvitoolbareditor.cpp
//   Creation date : Sun 10 Nov 2002 23:25:59 2002 GMT by Szymon Stefanek
//
//   This toolbar is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "kvi_module.h"
#include "kvi_uparser.h"
#include "kvi_locale.h"
#include "kvi_command.h"
#include "kvi_frame.h"

#ifdef COMPILE_SCRIPTTOOLBAR
	#include "toolbareditor.h"
	KviScriptToolBarEditorWindow * g_pToolBarEditorWindow = 0;
#endif // COMPILE_SCRIPTTOOLBAR


/*
	@doc: toolbareditor.open
	@type:
		command
	@title:
		toolbareditor.open
	@short:
		Shows the script toolbar editor
	@syntax:
		toolbareditor.open
	@description:
		Opens the script toolbar editor dialog.
		This command does nothing if the scripttoolbar support isn't compiled in the KVIrc executable
*/

bool toolbareditor_module_cmd_open(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"toolbareditor.open");

	KviStr dummy;
	if(!g_pUserParser->parseCmdFinalPart(c,dummy))return false;

#ifdef COMPILE_SCRIPTTOOLBAR

	if(!g_pToolBarEditorWindow)
	{
		g_pToolBarEditorWindow = new KviScriptToolBarEditorWindow(c->window()->frame());
		c->window()->frame()->addWindow(g_pToolBarEditorWindow);
	}

	g_pToolBarEditorWindow->setFocus();
	
#endif // COMPILE_SCRIPTTOOLBAR

    return c->leaveContext();
}


static bool toolbareditor_module_init(KviModule * m)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	m->registerCommand("open",toolbareditor_module_cmd_open);
	g_pToolBarEditorWindow = 0;
#endif // COMPILE_SCRIPTTOOLBAR
	return true;
}

static bool toolbareditor_module_can_unload(KviModule * m)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	return (g_pToolBarEditorWindow == 0);
#else // !COMPILE_SCRIPTTOOLBAR
	return true;
#endif // !COMPILE_SCRIPTTOOLBAR
}

static bool toolbareditor_module_cleanup(KviModule *m)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	if(g_pToolBarEditorWindow)delete g_pToolBarEditorWindow;
	m->unregisterMetaObject("KviScriptToolBarEditorWindow");
	m->unregisterMetaObject("KviScriptToolBarEditor");
	m->unregisterMetaObject("KviSingleScriptToolBarEditor");
#endif // COMPILE_SCRIPTTOOLBAR
	return true;
}

KVIMODULEEXPORTDATA KviModuleInfo kvirc_module_info=
{
	"ToolbarEditor",                                                 // module name
	"1.0.0",                                                // module version
	"Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)", // author & (C)
#ifdef COMPILE_SCRIPTTOOLBAR
	"Editor for the scriptable toolbars",
#else // !COMPILE_SCRIPTTOOLBAR
	"Editor for the scriptable toolbars (disabled at compiletime)",
#endif // !COMPILE_SCRIPTTOOLBAR
	toolbareditor_module_init,
	toolbareditor_module_can_unload,
	0,
	toolbareditor_module_cleanup
};
