//
//   File : toolbareditor.cpp
//   Creation date : Sun Nov 10 2002 23:50:18 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the linkss of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
#include "toolbareditor.h"

#include "kvi_iconmanager.h"
#include "kvi_options.h"
#include "kvi_locale.h"
#include "kvi_imagedialog.h"
#include "kvi_config.h"
#include "kvi_filedialog.h"
#include "kvi_fileutils.h"
#include "kvi_app.h"
#include "kvi_qstring.h"

#ifdef COMPILE_SCRIPTTOOLBAR
	#include "kvi_scripttoolbar_manager.h"
#endif // COMPILE_SCRIPTTOOLBAR

#include <qvbox.h>
#include <qasciidict.h>
#include <qlayout.h>
#include <qhbox.h>

#ifdef COMPILE_SCRIPTTOOLBAR
	extern KviScriptToolBarEditorWindow * g_pToolBarEditorWindow;
#endif // COMPILE_SCRIPTTOOLBAR

#ifdef COMPILE_SCRIPTTOOLBAR
	KviScriptToolBarEditorItem::KviScriptToolBarEditorItem()
	{
		m_pFrame = 0;
		m_pButton = 0;
	}

	KviScriptToolBarEditorItem::~KviScriptToolBarEditorItem()
	{
		if(m_pFrame)delete m_pFrame;
		if(m_pButton)delete m_pButton;
	}
#endif // COMPILE_SCRIPTTOOLBAR

KviSingleScriptToolBarEditor::KviSingleScriptToolBarEditor(QWidget * par)
: QWidget(par)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	m_pDef = 0;

	m_pItemList = 0;

	QGridLayout * g = new QGridLayout(this,8,3,2,2);

	QLabel * l = new QLabel(__tr2qs("Toolbar name:"),this);
	g->addWidget(l,0,0);

	m_pNameEdit = new QLineEdit(this);
	g->addMultiCellWidget(m_pNameEdit,0,0,1,1);

	m_pGenerateKvsButton = new QPushButton(__tr2qs("&Export To..."),this);
	connect(m_pGenerateKvsButton,SIGNAL(clicked()),this,SLOT(generateKvs()));
	g->addWidget(m_pGenerateKvsButton,0,2);

	l = new QLabel(__tr2qs("Toolbar label:"),this);
	g->addWidget(l,1,0);

	m_pLabelEdit = new QLineEdit(this);
	g->addMultiCellWidget(m_pLabelEdit,1,1,1,2);

	m_pToolBox = new QFrame(this);
	m_pToolBox->setMinimumHeight(32);
	m_pToolBox->setFrameStyle(QFrame::WinPanel | QFrame::Raised);
	g->addMultiCellWidget(m_pToolBox,2,2,0,2);


	QHBox * buttonBox = new QHBox(this);
	g->addMultiCellWidget(buttonBox,3,3,0,2);

	m_pAddItemButton = new QPushButton(__tr2qs("New &Button"),buttonBox);
	connect(m_pAddItemButton,SIGNAL(clicked()),this,SLOT(addButton()));
	m_pAddSeparatorButton = new QPushButton(__tr2qs("New &Separator"),buttonBox);
	connect(m_pAddSeparatorButton,SIGNAL(clicked()),this,SLOT(addSeparator()));
	m_pMoveLeftButton = new QPushButton(__tr2qs("Move Left"),buttonBox);
	connect(m_pMoveLeftButton,SIGNAL(clicked()),this,SLOT(moveLeft()));
	m_pMoveLeftButton->setEnabled(false);
	m_pMoveRightButton = new QPushButton(__tr2qs("Move Right"),buttonBox);
	connect(m_pMoveRightButton,SIGNAL(clicked()),this,SLOT(moveRight()));
	m_pMoveRightButton->setEnabled(false);
	m_pDeleteButton = new QPushButton(__tr2qs("Re&move"),buttonBox);
	connect(m_pDeleteButton,SIGNAL(clicked()),this,SLOT(deleteItem()));
	m_pDeleteButton->setEnabled(false);

	l = new QLabel(__tr2qs("Item name:"),this);
	g->addWidget(l,4,0);

	m_pItemNameEdit = new QLineEdit(this);
	g->addMultiCellWidget(m_pItemNameEdit,4,4,1,2);

	l = new QLabel(__tr2qs("Button text:"),this);
	g->addWidget(l,5,0);

	m_pItemTextEdit = new QLineEdit(this);
	g->addMultiCellWidget(m_pItemTextEdit,5,5,1,2);

	l = new QLabel(__tr2qs("Button icon:"),this);
	g->addWidget(l,6,0);

	m_pItemIconEdit = new QLineEdit(this);
	g->addWidget(m_pItemIconEdit,6,1);

	m_pItemIconButton = new QPushButton("...",this);
	connect(m_pItemIconButton,SIGNAL(clicked()),this,SLOT(chooseIcon()));
	g->addWidget(m_pItemIconButton,6,2);

	m_pItemCodeEdit = KviScriptEditor::createInstance(this);
	g->addMultiCellWidget(m_pItemCodeEdit,7,7,0,2);

	g->setRowStretch(7,1);

	m_pCurrentItem = 0;

	edit(0);
#endif // COMPILE_SCRIPTTOOLBAR
}

KviSingleScriptToolBarEditor::~KviSingleScriptToolBarEditor()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	KviScriptEditor::destroyInstance(m_pItemCodeEdit);
	if(m_pItemList)delete m_pItemList;
#endif // COMPILE_SCRIPTTOOLBAR
}

#ifdef COMPILE_SCRIPTTOOLBAR



void KviSingleScriptToolBarEditor::save()
{
	if(!m_pDef)return;

	// save the changes!
	if(m_pCurrentItem)
	{
		// save the item changes first
		saveCurrentItem();
	}

	QString szOldName = m_pDef->name();

	m_pDef->clear();
	m_pDef->setLabel(m_pLabelEdit->text());

	QString szNewName = m_pNameEdit->text();
	if(szOldName != szNewName)m_pDef->rename(szNewName);

	if(m_pItemList)
	{
		for(KviScriptToolBarEditorItem * i = m_pItemList->first();i;i = m_pItemList->next())
		{
			KviScriptToolBarItemDefinition * def;
			if(i->m_pButton)
			{
				def = m_pDef->getItemDefinition(i->m_szName,KviScriptToolBarItemDefinition::Button);
				def->setCode(i->m_szCode);
				def->setIcon(i->m_szIcon);
				def->setText(i->m_szText);
			} else {
				def = m_pDef->getItemDefinition(i->m_szName,KviScriptToolBarItemDefinition::Separator);
			}
		}
	}

	if(!KviQString::equalCI(szOldName,szNewName))emit barNameChanged(szOldName,szNewName);
}

void KviSingleScriptToolBarEditor::edit(KviScriptToolBarDefinition * pDef)
{
	if(m_pDef == pDef)return;

	save();

	if(m_pItemList)delete m_pItemList;
	m_pItemList = 0;

	m_pCurrentItem = 0;

	m_pDef = pDef;

	if(!m_pDef)
	{
		m_pAddItemButton->setEnabled(false);
		m_pAddSeparatorButton->setEnabled(false);
		m_pMoveLeftButton->setEnabled(false);
		m_pMoveRightButton->setEnabled(false);
		m_pDeleteButton->setEnabled(false);

		m_pItemCodeEdit->setEnabled(false);
		m_pItemIconEdit->setEnabled(false);
		m_pItemIconButton->setEnabled(false);
		m_pItemTextEdit->setEnabled(false);
		m_pItemNameEdit->setEnabled(false);
		m_pNameEdit->setEnabled(false);
		m_pLabelEdit->setEnabled(false);
		m_pGenerateKvsButton->setEnabled(false);
		return;
	}

	m_pAddItemButton->setEnabled(true);
	m_pAddSeparatorButton->setEnabled(true);
	m_pMoveLeftButton->setEnabled(false);
	m_pMoveRightButton->setEnabled(false);
	m_pDeleteButton->setEnabled(false);

	// fill it man :)
	m_pNameEdit->setText(m_pDef->name());
	m_pNameEdit->setEnabled(true);
	m_pGenerateKvsButton->setEnabled(true);
	m_pLabelEdit->setText(m_pDef->label());
	m_pLabelEdit->setEnabled(true);

	KviPtrList<KviScriptToolBarItemDefinition> * l = m_pDef->itemList();
	if(l)
	{
		m_pItemList = new KviPtrList<KviScriptToolBarEditorItem>;
		m_pItemList->setAutoDelete(true);

		for(KviScriptToolBarItemDefinition * d = l->first();d;d = l->next())
		{
			KviScriptToolBarEditorItem * edi = new KviScriptToolBarEditorItem();
			edi->m_szName = d->name();
			edi->m_szCode = d->code();
			edi->m_szIcon = d->icon();
			edi->m_szText = d->text();
			switch(d->type())
			{
				case KviScriptToolBarItemDefinition::Separator:
					edi->m_pFrame = new QFrame(m_pToolBox);
					edi->m_pFrame->setFrameStyle(QFrame::Raised | QFrame::VLine);
					edi->m_pFrame->installEventFilter(this);
				break;
				case KviScriptToolBarItemDefinition::Button:
					edi->m_pButton = new QToolButton(m_pToolBox);
					edi->m_pButton->setTextLabel(edi->m_szText);
					edi->m_pButton->setUsesBigPixmap(true);
					edi->m_pButton->setAutoRaise(true);
					edi->m_pButton->setToggleButton(true);
					edi->m_pButton->setOn(false);
					edi->m_pButton->installEventFilter(this);
					{
						QPixmap * pix = g_pIconManager->getImage(edi->m_szIcon);
						if(pix)edi->m_pButton->setIconSet(*pix);
					}
				break;
			}
			m_pItemList->append(edi);
		}
	}

	layoutItems();

	setCurrentItem(0);
}

void KviSingleScriptToolBarEditor::setCurrentItem(KviScriptToolBarEditorItem * it)
{
	if(m_pCurrentItem)saveCurrentItem();
	m_pCurrentItem = it;
	if(!m_pItemList)return;

	for(KviScriptToolBarEditorItem * i=m_pItemList->first();i;i = m_pItemList->next())
	{
		if(i != it)
		{
			if(i->m_pButton)i->m_pButton->setOn(false);
			else if(i->m_pFrame)i->m_pFrame->setFrameStyle(QFrame::Raised | QFrame::VLine);
		}
	}

	m_pItemNameEdit->setEnabled(it);
	if(it)m_pItemTextEdit->setEnabled(it->m_pButton);
	else m_pItemTextEdit->setEnabled(false);
	if(it)m_pItemCodeEdit->setEnabled(it->m_pButton);
	else m_pItemCodeEdit->setEnabled(false);
	if(it)m_pItemIconEdit->setEnabled(it->m_pButton);
	else m_pItemIconEdit->setEnabled(false);
	m_pItemIconButton->setEnabled(m_pItemIconEdit->isEnabled());
	if(it)m_pMoveLeftButton->setEnabled(it != m_pItemList->first());
	else m_pMoveLeftButton->setEnabled(false);
	if(it)m_pMoveRightButton->setEnabled(it != m_pItemList->last());
	else m_pMoveRightButton->setEnabled(false);
	m_pDeleteButton->setEnabled(it);

	if(!it)
	{
		m_pItemNameEdit->setText("");
		m_pItemTextEdit->setText("");
		m_pItemCodeEdit->setText("");
		m_pItemIconEdit->setText("");
		return;
	}

	if(it->m_pButton)it->m_pButton->setOn(true);
	else if(it->m_pFrame)it->m_pFrame->setFrameStyle(QFrame::Sunken | QFrame::VLine);

	m_pItemNameEdit->setText(it->m_szName);
	m_pItemTextEdit->setText(it->m_szText);
	m_pItemCodeEdit->setText(it->m_szCode);
	m_pItemIconEdit->setText(it->m_szIcon);
}

bool KviSingleScriptToolBarEditor::eventFilter(QObject * o,QEvent * e)
{
	if(!m_pItemList)return false;
	if(e->type() == QEvent::MouseButtonPress)
	{
		if(o->inherits("QFrame"))
		{
			QFrame * f = (QFrame *)o;
			KviScriptToolBarEditorItem * i = 0;
			for(i = m_pItemList->first();i;i = m_pItemList->next())
			{
				if(i->m_pFrame == f)break;
			}
			if(!i)return false;
			// set this item as current!
			setCurrentItem(i);

		} else if(o->inherits("QToolButton"))
		{
			QToolButton * b = (QToolButton *)o;
			KviScriptToolBarEditorItem * i = 0;
			for(i = m_pItemList->first();i;i = m_pItemList->next())
			{
				if(i->m_pButton == b)break;
			}
			if(!i)return false;
			// set this item as current!
			setCurrentItem(i);
		}
		return true;
	}

	return false;
}

bool KviSingleScriptToolBarEditor::itemNameInUse(const QString & szName,KviScriptToolBarEditorItem  * except)
{
	if(!m_pItemList)return false;
	
	for(KviScriptToolBarEditorItem  * i = m_pItemList->first();i;i = m_pItemList->next())
	{
		if(i != except)
		{
			if(KviQString::equalCI(szName,i->m_szName))return true;
		}
	}
	return false;
}

void KviSingleScriptToolBarEditor::findUniqueItemName(QString & szBuffer,KviScriptToolBarEditorItem * except,const QString & base)
{
	szBuffer = base;
	int idx = 1;
	while(itemNameInUse(szBuffer,except))
	{
		KviQString::sprintf(szBuffer,"%Q%d",&base,idx);
		idx++;
	}
}

void KviSingleScriptToolBarEditor::saveCurrentItem()
{
	if(!m_pDef)return;
	if(!m_pCurrentItem)return;
	if(!m_pItemList)return;

	QString szName = m_pItemNameEdit->text();

	if(m_pCurrentItem->m_pFrame)
	{
		findUniqueItemName(m_pCurrentItem->m_szName,m_pCurrentItem,szName.isEmpty() ? QString("separator") : szName);
		return;
	}

	findUniqueItemName(m_pCurrentItem->m_szName,m_pCurrentItem,szName.isEmpty() ? QString("button") : szName);

	QCString tmp;
	m_pItemCodeEdit->getText(tmp);
	m_pCurrentItem->m_szCode = tmp.data();
	m_pCurrentItem->m_szIcon = m_pItemIconEdit->text();
	m_pCurrentItem->m_szText = m_pItemTextEdit->text();

	m_pCurrentItem->m_pButton->setTextLabel(m_pItemTextEdit->text());

	QPixmap * pix = g_pIconManager->getImage(m_pCurrentItem->m_szIcon);
	if(pix)m_pCurrentItem->m_pButton->setIconSet(*pix);
	else m_pCurrentItem->m_pButton->setIconSet(QPixmap());
}

void KviSingleScriptToolBarEditor::layoutItems()
{
	if(!m_pItemList)return;

	// get the maximum height hint

	int maxHeight = 0;
	KviScriptToolBarEditorItem * i;

	for(i = m_pItemList->first();i;i = m_pItemList->next())
	{
		int theHeight = 0;
		if(i->m_pButton)theHeight = i->m_pButton->sizeHint().height();
		else if(i->m_pFrame)theHeight = i->m_pFrame->sizeHint().height();
		if(theHeight > maxHeight)maxHeight = theHeight;
	}

	if(maxHeight < 14)maxHeight = 14;

	m_pToolBox->setMinimumHeight(maxHeight + 6);

	int curX = 3;

	for(i = m_pItemList->first();i;i = m_pItemList->next())
	{
		if(i->m_pButton)
		{
			i->m_pButton->setGeometry(curX,3,maxHeight,maxHeight);
			if(!i->m_pButton->isVisible())i->m_pButton->show();
			curX += maxHeight + 3;
		} else {
			if(i->m_pFrame)
			{
				int theW = i->m_pFrame->sizeHint().width();
				i->m_pFrame->setGeometry(curX,3,theW,maxHeight);
				if(!i->m_pFrame->isVisible())i->m_pFrame->show();
				curX += theW + 3;
			}
		}
	}
}

#endif // COMPILE_SCRIPTTOOLBAR

void KviSingleScriptToolBarEditor::chooseIcon()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	if(!m_pDef)return;
	if(!m_pCurrentItem)return;
	KviImageDialog dlg(this);
	if(dlg.exec() == QDialog::Rejected)return;
	QString str = dlg.selectedImage();
	if(str.isNull())return;
	if(!m_pCurrentItem)return; //safety measure
	m_pItemIconEdit->setText(str);
	saveCurrentItem();
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviSingleScriptToolBarEditor::generateKvs()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	if(!m_pDef)return;
	saveCurrentItem();
	QString szBuffer;
	m_pDef->generateKvsDefinition(szBuffer);
	KviStr szName;
	// FIXME: __tr()
	if(KviFileDialog::askForSaveFileName(szName,__tr("Choose a Filename - KVIrc"),0,"*.kvs",false,true))
	{
		KviStr tmp = szBuffer;
		kvi_writeFile(szName.ptr(),tmp);
	}
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviSingleScriptToolBarEditor::addButton()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	if(!m_pDef)return;
	if(!m_pItemList)
	{
		m_pItemList = new KviPtrList<KviScriptToolBarEditorItem>;
		m_pItemList->setAutoDelete(true);
	}
	KviScriptToolBarEditorItem * edi = new KviScriptToolBarEditorItem();
	findUniqueItemName(edi->m_szName,edi,"button");
	edi->m_szCode = __tr2qs("# Insert button code here");
	edi->m_szText = "Button";

	edi->m_pButton = new QToolButton(m_pToolBox);
	edi->m_pButton->setTextLabel(edi->m_szText);
	edi->m_pButton->setUsesBigPixmap(true);
	edi->m_pButton->setAutoRaise(true);
	edi->m_pButton->setToggleButton(true);
	edi->m_pButton->setOn(false);
	edi->m_pButton->installEventFilter(this);

	m_pItemList->append(edi);
	layoutItems();
	setCurrentItem(edi);
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviSingleScriptToolBarEditor::addSeparator()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	if(!m_pDef)return;
	if(!m_pItemList)
	{
		m_pItemList = new KviPtrList<KviScriptToolBarEditorItem>;
		m_pItemList->setAutoDelete(true);
	}
	KviScriptToolBarEditorItem * edi = new KviScriptToolBarEditorItem();
	findUniqueItemName(edi->m_szName,edi,"separator");
	edi->m_pFrame = new QFrame(m_pToolBox);
	edi->m_pFrame->setFrameStyle(QFrame::Raised | QFrame::VLine);
	edi->m_pFrame->installEventFilter(this);
	m_pItemList->append(edi);
	layoutItems();
	setCurrentItem(edi);
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviSingleScriptToolBarEditor::moveLeft()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	if(!m_pCurrentItem)return;
	if(!m_pItemList)return;
	int idx = m_pItemList->findRef(m_pCurrentItem);
	if(idx == -1)return;
	if(idx == 0)return;
	m_pItemList->setAutoDelete(false);
	m_pItemList->removeRef(m_pCurrentItem);
	m_pItemList->setAutoDelete(true);
	m_pItemList->insert(idx - 1,m_pCurrentItem);
	m_pMoveLeftButton->setEnabled(m_pCurrentItem != m_pItemList->first());
	m_pMoveRightButton->setEnabled(m_pCurrentItem != m_pItemList->last());
	layoutItems();
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviSingleScriptToolBarEditor::moveRight()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	if(!m_pCurrentItem)return;
	if(!m_pItemList)return;
	int idx = m_pItemList->findRef(m_pCurrentItem);
	if(idx == -1)return;
	if(idx == (m_pItemList->count() - 1))return;
	m_pItemList->setAutoDelete(false);
	m_pItemList->removeRef(m_pCurrentItem);
	m_pItemList->setAutoDelete(true);
	idx++;
	m_pItemList->insert(idx,m_pCurrentItem);
	m_pMoveLeftButton->setEnabled(m_pCurrentItem != m_pItemList->first());
	m_pMoveRightButton->setEnabled(m_pCurrentItem != m_pItemList->last());
	layoutItems();
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviSingleScriptToolBarEditor::deleteItem()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	if(!m_pCurrentItem)return;
	if(!m_pItemList)return;
	m_pItemList->removeRef(m_pCurrentItem);
	m_pCurrentItem = 0;
	layoutItems();
	setCurrentItem(m_pItemList->first());
#endif // COMPILE_SCRIPTTOOLBAR
}




KviScriptToolBarEditor::KviScriptToolBarEditor(QWidget * par)
: QSplitter(QSplitter::Horizontal,par)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	QVBox * b = new QVBox(this);

	m_pListBox = new QListBox(b);
	connect(m_pListBox,SIGNAL(highlighted(int)),this,SLOT(itemSelected(int)));

	m_pAddButton = new QPushButton(__tr2qs("&New"),b);
	connect(m_pAddButton,SIGNAL(clicked()),this,SLOT(addClicked()));

	m_pRemoveButton = new QPushButton(__tr2qs("Re&move"),b);
	connect(m_pRemoveButton,SIGNAL(clicked()),this,SLOT(removeClicked()));
	m_pRemoveButton->setEnabled(false);
	
	m_pEditor = new KviSingleScriptToolBarEditor(this);
	connect(m_pEditor,SIGNAL(barNameChanged(const QString &,const QString &)),this,SLOT(barNameChanged(const QString &,const QString &)));

	b->setStretchFactor(m_pListBox,1);
	
	m_pManager = new KviScriptToolBarManager();
	m_pManager->copyFrom(g_pScriptToolBarManager);

	fillToolBarList();

	if(m_pListBox->count() > 0)
	{
		m_pListBox->setCurrentItem(0);
		itemSelected(0);
	}

#endif // COMPILE_SCRIPTTOOLBAR
}

KviScriptToolBarEditor::~KviScriptToolBarEditor()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	delete m_pManager;
#endif // COMPILE_SCRIPTTOOLBAR
}

#ifdef COMPILE_SCRIPTTOOLBAR
void KviScriptToolBarEditor::commit()
{
	// save the last bar edited
	m_pEditor->save();
	g_pScriptToolBarManager->copyFrom(m_pManager);
	g_pApp->saveToolBars();
}
#endif // COMPILE_SCRIPTTOOLBAR


void KviScriptToolBarEditor::addClicked()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	KviScriptToolBarDefinition * d = m_pManager->getNew();
	m_pListBox->insertItem(d->name());
	int cnt = m_pListBox->count();
	m_pListBox->setCurrentItem(cnt - 1);
	m_pEditor->edit(d);
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviScriptToolBarEditor::removeClicked()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	int idx = m_pListBox->currentItem();
	if((idx < 0) || (idx >= m_pListBox->count()))return;
	QString tmp = m_pListBox->text(idx);
	m_pEditor->edit(0); // clear the editor
	m_pManager->remove(tmp);
	m_pListBox->removeItem(idx);
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviScriptToolBarEditor::itemSelected(int idx)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	if((idx >= 0) && (idx < m_pListBox->count()))
	{
		QString tmp = m_pListBox->text(idx);
		editToolBar(tmp);
	}
#endif // COMPILE_SCRIPTTOOLBAR
}

#ifdef COMPILE_SCRIPTTOOLBAR
void KviScriptToolBarEditor::editToolBar(const QString &szName)
{
	KviScriptToolBarDefinition *d = m_pManager->find(szName);
	if(d)
	{
		m_szLastEditedToolBarName = szName;
		m_pEditor->edit(d);
		m_pRemoveButton->setEnabled(true);
		int idx1 = m_pListBox->currentItem();
		QListBoxItem * it = m_pListBox->findItem(szName,Qt::ExactMatch);
		if(it)
		{
			int idx2 = m_pListBox->index(it);
			if(idx1 != idx2)m_pListBox->setCurrentItem(idx2);
		}
	}
}
#endif // COMPILE_SCRIPTTOOLBAR

void KviScriptToolBarEditor::barNameChanged(const QString &szOldName,const QString &szNewName)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	QListBoxItem * it = m_pListBox->findItem(szOldName,Qt::ExactMatch);
	if(it)
	{
		int cur = m_pListBox->currentItem();
		int idx = m_pListBox->index(it);
		m_pListBox->changeItem(szNewName,idx);
		m_pListBox->setCurrentItem(cur); // it seems to loose the setting
		m_szLastEditedToolBarName = m_pListBox->text(cur);
	}
#endif // COMPILE_SCRIPTTOOLBAR
}

#ifdef COMPILE_SCRIPTTOOLBAR
void KviScriptToolBarEditor::fillToolBarList()
{
	m_pListBox->clear();
	QDictIterator<KviScriptToolBarDefinition> it(*(m_pManager->toolBarDefinitions()));
	while(KviScriptToolBarDefinition * d = it.current())
	{
		m_pListBox->insertItem(d->name());
		++it;
	}
}
#endif // COMPILE_SCRIPTTOOLBAR


KviScriptToolBarEditorWindow::KviScriptToolBarEditorWindow(KviFrame * lpFrm)
: KviWindow(KVI_WINDOW_TYPE_SCRIPTEDITOR,lpFrm,"toolbareditor",0)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	g_pToolBarEditorWindow = this;

	m_pEditor = new KviScriptToolBarEditor(this);

	m_pBase = new QWidget(this);
	QGridLayout * g = new QGridLayout(m_pBase,1,4,4,4);

	QPushButton * btn = new QPushButton(__tr2qs("&OK"),m_pBase);
	connect(btn,SIGNAL(clicked()),this,SLOT(okClicked()));
	btn->setIconSet(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_ACCEPT)));
	g->addWidget(btn,0,1);

	btn = new QPushButton(__tr2qs("&Apply"),m_pBase);
	connect(btn,SIGNAL(clicked()),this,SLOT(applyClicked()));
	btn->setIconSet(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_ACCEPT)));
	g->addWidget(btn,0,2);

	btn = new QPushButton(__tr2qs("Cancel"),m_pBase);
	connect(btn,SIGNAL(clicked()),this,SLOT(cancelClicked()));
	btn->setIconSet(*(g_pIconManager->getSmallIcon(KVI_SMALLICON_DISCARD)));
	g->addWidget(btn,0,3);

	g->setColStretch(0,1);
#endif // COMPILE_SCRIPTTOOLBAR
}

KviScriptToolBarEditorWindow::~KviScriptToolBarEditorWindow()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	g_pToolBarEditorWindow = 0;
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviScriptToolBarEditorWindow::okClicked()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	m_pEditor->commit();
	close();
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviScriptToolBarEditorWindow::applyClicked()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	m_pEditor->commit();
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviScriptToolBarEditorWindow::cancelClicked()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	close();
#endif // COMPILE_SCRIPTTOOLBAR
}


QPixmap * KviScriptToolBarEditorWindow::myIconPtr()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	return g_pIconManager->getSmallIcon(KVI_SMALLICON_TOOLBAR);
#else // !COMPILE_SCRIPTTOOLBAR
	return 0;
#endif // !COMPILE_SCRIPTTOOLBAR
}

void KviScriptToolBarEditorWindow::resizeEvent(QResizeEvent *e)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	int hght = m_pBase->sizeHint().height();
	m_pEditor->setGeometry(0,0,width(),height()- hght);
	m_pBase->setGeometry(0,height() - hght,width(),hght);
#endif // COMPILE_SCRIPTTOOLBAR
}

void KviScriptToolBarEditorWindow::fillCaptionBuffers()
{
#ifdef COMPILE_SCRIPTTOOLBAR
	m_szPlainTextCaption = __tr2qs("Toolbar Editor");

	static QString p1("<nobr><font color=\"");
	static QString p2("\"><b>");
	static QString p3("</b></font></nobr>");

	m_szHtmlActiveCaption = p1;
	m_szHtmlActiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextActive).name();
	m_szHtmlActiveCaption += p2;
	m_szHtmlActiveCaption += m_szPlainTextCaption;
	m_szHtmlActiveCaption += p3;

	m_szHtmlInactiveCaption = p1;
	m_szHtmlInactiveCaption += KVI_OPTION_COLOR(KviOption_colorCaptionTextInactive).name();
	m_szHtmlInactiveCaption += p2;
	m_szHtmlInactiveCaption += m_szPlainTextCaption;
	m_szHtmlInactiveCaption += p3;
#endif // COMPILE_SCRIPTTOOLBAR
}


void KviScriptToolBarEditorWindow::getConfigGroupName(KviStr &szName)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	szName = "toolbareditor";
#endif // COMPILE_SCRIPTTOOLBAR
}


void KviScriptToolBarEditorWindow::saveProperties(KviConfig *cfg)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	cfg->writeEntry("Sizes",m_pEditor->sizes());
	cfg->writeEntry("LastToolBar",m_pEditor->lastEditedToolBar());
	//debug("LAST EDITED=%s",m_pEditor->lastEditedToolBar().ptr());
#endif // COMPILE_SCRIPTTOOLBAR

}

void KviScriptToolBarEditorWindow::loadProperties(KviConfig *cfg)
{
#ifdef COMPILE_SCRIPTTOOLBAR
	QValueList<int> def;
	def.append(20);
	def.append(80);
	m_pEditor->setSizes(cfg->readIntListEntry("Sizes",def));
	QString tmp = cfg->readQStringEntry("LastToolBar","");
	m_pEditor->editToolBar(tmp);
	//debug("LAST EDITED WAS %s",tmp.ptr());
#endif // COMPILE_SCRIPTTOOLBAR
}

