#ifndef _TOOLBAREDITOR_H_
#define _TOOLBAREDITOR_H_
//
//   File : toolbareditor.h
//   Creation date : Sun Nov 10 2002 23:49:55 CEST by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the linkss of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "kvi_window.h"
#include "kvi_string.h"

#include <qsplitter.h>
#include <qlistbox.h>
#include <qpushbutton.h>
#include <qhbox.h>

#ifdef COMPILE_SCRIPTTOOLBAR
	#include <qlineedit.h>
	#include <qframe.h>
	#include <qtoolbutton.h>

	#include "kvi_scripttoolbar_definition.h"

	#include "kvi_list.h"
	#include "kvi_scripteditor.h"
#endif // COMPILE_SCRIPTTOOLBAR

#ifdef COMPILE_SCRIPTTOOLBAR
	class KviScriptToolBarManager;
#endif // COMPILE_SCRIPTTOOLBAR

#ifdef COMPILE_SCRIPTTOOLBAR
	class KviScriptToolBarEditorItem
	{
	public:
		KviScriptToolBarEditorItem();
		~KviScriptToolBarEditorItem();
	
		QFrame      * m_pFrame;
		QToolButton * m_pButton;
	
		QString       m_szName;
		QString       m_szCode;
		QString       m_szIcon;
		QString       m_szText;
	public:
		bool isButton(){ return (m_pButton != 0); };
		bool isSeparator(){ return (m_pFrame != 0); };
	};
#endif // COMPILE_SCRIPTTOOLBAR
	
class KviSingleScriptToolBarEditor : public QWidget
{
	Q_OBJECT
public:
	KviSingleScriptToolBarEditor(QWidget * par);
	~KviSingleScriptToolBarEditor();
#ifdef COMPILE_SCRIPTTOOLBAR
protected:
	KviScriptToolBarDefinition * m_pDef;
	QLineEdit                  * m_pNameEdit;
	QLineEdit                  * m_pLabelEdit;
	QFrame                     * m_pToolBox;
	QLineEdit                  * m_pItemNameEdit;
	QLineEdit                  * m_pItemTextEdit;
	QLineEdit                  * m_pItemIconEdit;
	KviScriptEditor            * m_pItemCodeEdit;
	QPushButton                * m_pGenerateKvsButton;
	QPushButton                * m_pItemIconButton;
	QPushButton                * m_pAddItemButton;
	QPushButton                * m_pAddSeparatorButton;
	QPushButton                * m_pMoveLeftButton;
	QPushButton                * m_pMoveRightButton;
	QPushButton                * m_pDeleteButton;
	KviPtrList<KviScriptToolBarEditorItem> * m_pItemList;
	KviScriptToolBarEditorItem * m_pCurrentItem;
public:
	void edit(KviScriptToolBarDefinition * pDef);
	void save();
protected:
	void layoutItems();
	bool eventFilter(QObject * o,QEvent * e);
	void setCurrentItem(KviScriptToolBarEditorItem * i);
	void saveCurrentItem();

	bool itemNameInUse(const QString & szName,KviScriptToolBarEditorItem  * except);
	void findUniqueItemName(QString & szBuffer,KviScriptToolBarEditorItem * except,const QString & base);

#endif // COMPILE_SCRIPTTOOLBAR
protected slots:
	void addButton();
	void addSeparator();
	void moveLeft();
	void moveRight();
	void deleteItem();
	void chooseIcon();
	void generateKvs();
signals:
	void barNameChanged(const QString &szOldName,const QString &szNewName);
};


class KviScriptToolBarEditor : public QSplitter
{
	friend class KviScriptToolBarEditorWindow;
	Q_OBJECT
public:
	KviScriptToolBarEditor(QWidget * par);
	~KviScriptToolBarEditor();
#ifdef COMPILE_SCRIPTTOOLBAR
protected:
	QListBox    * m_pListBox;
	QPushButton * m_pAddButton;
	QPushButton * m_pRemoveButton;
	KviSingleScriptToolBarEditor * m_pEditor;
	KviScriptToolBarManager * m_pManager;
	QString        m_szLastEditedToolBarName;
protected:
	const QString & lastEditedToolBar(){ return m_szLastEditedToolBarName; };
	virtual void commit();
	void fillToolBarList();
	void editToolBar(const QString &szName);
#endif // COMPILE_SCRIPTTOOLBAR
protected slots:
	void removeClicked();
	void addClicked();
	void itemSelected(int idx);
	void barNameChanged(const QString &szOldName,const QString &szNewName);
signals:
	void closeRequest();
};



class KviScriptToolBarEditorWindow : public KviWindow
{
	Q_OBJECT
public:
	KviScriptToolBarEditorWindow(KviFrame * lpFrm);
	~KviScriptToolBarEditorWindow();
protected:
	KviScriptToolBarEditor * m_pEditor;
	QWidget                * m_pBase;
protected:
	virtual QPixmap * myIconPtr();
	virtual void fillCaptionBuffers();
	virtual void resizeEvent(QResizeEvent *e);
	virtual void getConfigGroupName(KviStr &szName);
	virtual void saveProperties(KviConfig *);
	virtual void loadProperties(KviConfig *);
protected slots:
	void okClicked();
	void cancelClicked();
	void applyClicked();
};

#endif //_TOOLBAREDITOR_H_
